#region Using Statements
using System;
using System.Diagnostics;
using System.IO;
//using System.Xml.Serialization;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using matrix;
using utilities;
using DifferentiableFunction;
using Dynamics;
#endregion

namespace Dynamics
{
    /// <summary>
    /// Simple rigid body class for Microsoft XNA
    /// </summary>
    public class LinkXNA : Link
    {

        public LinkXNA(Function[,] transform, Function[] generalizedCoordinates, //Function[] generalizedForces,
            Function[,] centerOfMass, Function[,] inertiaMatrix, Function mass, string nodeName, Variable time)
            : base(transform, generalizedCoordinates, //generalizedForces,
            centerOfMass, inertiaMatrix, mass, nodeName, time)
        {
            translation = Vector3.Zero;
            rotation = Vector3.Zero;
            scale = Vector3.One;
        }

        //visualization
        public Model shape;         // Shape of a rigid body
        public Matrix matWorld;     // Body frame of the rigid body with respect to the world frame
        public Vector3 translation, rotation, scale; // Translation, rotation, and scale of the shape with respect to the body frame
        internal Matrix[] absoluteBoneTransforms;

        /// <summary>
        /// Set shape of the rigid body
        /// </summary>
        /// <param name="shape"></param>
        public void SetShape(Model shape)
        {
            this.shape = shape;
            absoluteBoneTransforms = new Matrix[shape.Bones.Count];
            shape.CopyAbsoluteBoneTransformsTo(absoluteBoneTransforms);
        }

        /// <summary>
        /// Draw the rigid body using basic effect
        /// </summary>
        /// <param name="viewMatrix"></param>
        /// <param name="projectionMatrix"></param>
        public void Draw(Matrix viewMatrix, Matrix projectionMatrix)
        {
            Matrix worldMatrix = Matrix.CreateScale(scale) *
                Matrix.CreateRotationX(rotation.X) *
                Matrix.CreateRotationY(rotation.Y) *
                Matrix.CreateRotationZ(rotation.Z) *
                Matrix.CreateTranslation(translation) *
                matWorld;
            foreach (ModelMesh mesh in shape.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.PreferPerPixelLighting = true;

                    effect.View = viewMatrix;
                    effect.Projection = projectionMatrix;
                    effect.World = absoluteBoneTransforms[mesh.ParentBone.Index] * worldMatrix;
                }
                mesh.Draw();
            }
        }
    }

    /// <summary>
    /// Simple rigid body system class for Micorsoft XNA
    /// </summary>
    public class MultiBodySystemXNA : MultiBodySystem
    {
        public MultiBodySystemXNA(Graph<Link> _systemGraph, Variable t, Function[,] explicitConst) : base(_systemGraph, t, explicitConst) { }

        public void PrepareToDraw()
        {
            LinkXNA[] links = (LinkXNA[])this.allNodes;
            for (int i = 0; i < W.Length; i++) { links[i].matWorld = makeXNAMatrix(W[i]); }
        }

        internal Matrix makeXNAMatrix(double[,] transform)
        {
            return new Matrix( //transpose
                (float)transform[0, 0], (float)transform[1, 0], (float)transform[2, 0], (float)transform[3, 0],
                (float)transform[0, 1], (float)transform[1, 1], (float)transform[2, 1], (float)transform[3, 1],
                (float)transform[0, 2], (float)transform[1, 2], (float)transform[2, 2], (float)transform[3, 2],
                (float)transform[0, 3], (float)transform[1, 3], (float)transform[2, 3], (float)transform[3, 3]);
        }

        /// <summary>
        /// Draw each rigid body
        /// </summary>
        /// <param name="viewMatrix"></param>
        /// <param name="projectionMatrix"></param>
        public void Draw(Matrix viewMatrix, Matrix projectionMatrix)
        {
            foreach (LinkXNA rb in this.allNodes)
            {
                if (rb.shape != null)
                {
                    rb.Draw(viewMatrix, projectionMatrix);
                }
            }

        }

    }

}


