using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

using matrix;
using utilities;
using DifferentiableFunction;
using Dynamics;


namespace ClosedLoop
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        #region Camera Model
        // The next set of variables are used to control the camera used in the sample. 
        // It is an arc ball camera, so it can rotate in a sphere around the target, and
        // zoom in and out.
        float cameraArc = 30.0f;
        float cameraRotation = -30.0f;
        float cameraDistance = 15.0f;
        float cameraMaxDistance = 20.0f;
        float cameraMinDistance = 0.2f;

        Matrix viewMatrix;
        Matrix projectionMatrix;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;
        #endregion

        MultiBodySystemXNA rbs;

        //target sphere
        Model target;
        Matrix targetMatrix;
        Matrix[] targetABT;

        //line connecting target and the root node
        Model line;
        Matrix lineMatrix;
        Matrix[] lineABT;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // define projection matrix
            float aspectRatio = graphics.GraphicsDevice.Viewport.Width /
                (float)graphics.GraphicsDevice.Viewport.Height;
            projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(45.0f), aspectRatio, .01f, 1000);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            spriteFont = Content.Load<SpriteFont>("SpriteFont1");

            // TODO: use this.Content to load your game content here
            rbs = buildClosedLoop(Content);

            //create jacobianEval
            createJacobianEval();
            
            //target
            target = Content.Load<Model>("unitSphereYellow");
            targetMatrix = Matrix.CreateScale(0.3f);
            targetABT = new Matrix[target.Bones.Count];
            target.CopyAbsoluteBoneTransformsTo(targetABT);

            //line
            line = Content.Load<Model>("unitCylinderRed");
            lineMatrix = Matrix.CreateScale(0);
            lineABT = new Matrix[line.Bones.Count];
            line.CopyAbsoluteBoneTransformsTo(lineABT);
        }
        
        
        #region Jacobian Functions

        RuntimeFunction jacobianEval;
        double[] jacobianInputs;
        bool[] jacobianInputIndices;
        int numConst = 3; // number of constraints. position X,Y,Z

        /// <summary>
        /// create RuntimeFunction jacobianEval
        /// </summary>
        void createJacobianEval()
        {
            int n = rbs.numberOfCoords;
            Function qdt = Function.derivative(Function.D(rbs.genCoords1D, rbs.time));
            Function qddt = Function.derivative(Function.D(rbs.genCoords1D, rbs.time, rbs.time));
            UnspecifiedFunction[] tempqdot = new UnspecifiedFunction[n];
            for (int i = 0; i < n; i++) { tempqdot[i] = (UnspecifiedFunction)qdt[i]; }

            //constraint equation
            Function[] g = VM.mult(rbs.allNodes[rbs.numberOfNodes - 1].Wi, new Function[] { 2, 0, 0, 1 });
            g = VM.minus(g, new Function[] { 2, 0, 0, 1 });
            Function gddt = Function.derivative(Function.D(VM.subArray(g, 0, numConst), rbs.time, rbs.time)); 
            
            //substituion map will be used to substitute qddt with 0
            Function[,] substitutionMap = new Function[n, 2];
            for (int i = 0; i < n; i++)
            {
                substitutionMap[i, 0] = qddt[i];
                substitutionMap[i, 1] = 0;
            }

            Function[,] J = new Function[numConst, n];
            Function[] Jdtqdt = new Function[numConst];

            //J is the coefficient of qddt in gddt
            //Jdtqdt is qddt when qddt = 0
            for (int i = 0; i < J.GetLength(0); i++)
            {
                Jdtqdt[i] = Function.substitute(gddt[i], substitutionMap); 
                for (int j = 0; j < J.GetLength(1); j++)
                {
                    J[i, j] = gddt[i].coefficient(qddt[j]); 
                }
            }

            //make 1d array
            Function J1d = new Function(VM.concatenate<Function>(VM.matrixToVector<Function>(J), Jdtqdt));
            
            //create jacobianInputIndices
            UnspecifiedFunction[] inputJacobian = VM.concatenate<UnspecifiedFunction>(rbs.genCoords1D, tempqdot);
            inputJacobian = MultiBodySystem.initializedInputUnspecifiedFunctions(J1d, inputJacobian, out jacobianInputIndices);
            jacobianInputs = new double[inputJacobian.Length];
            J1d.orderVariablesInDomain(inputJacobian);
            
            //compile
            jacobianEval = J1d.compile();
        }

        /// <summary>
        /// evaluate Jacobian matrices
        /// </summary>
        /// <param name="J"></param>
        /// <param name="Jdqd"></param>
        void evalJacobians(out double[,] J, out double[] Jdqd)
        {
            int n = rbs.numberOfCoords;
            double[] result = new double[numConst * rbs.numberOfCoords + numConst];

            updateJacobianInputs();
            jacobianEval.eval(result, jacobianInputs);

            J = VM.vectorToMatrix<double>(n, numConst, 0, result);
            Jdqd = VM.subArray<double>(result, numConst * n, numConst * n + numConst);
        }

        /// <summary>
        /// update jacobianInputs[] from current state of the system
        /// </summary>
        void updateJacobianInputs()
        {
            int i, j = 0, n = rbs.numberOfCoords;
            for (i = 0; i < n; ++i)
            {
                if (jacobianInputIndices[i]) jacobianInputs[j++] = rbs.q[i];
            }
            for (i = n; i < 2 * n; ++i)
            {
                if (jacobianInputIndices[i]) jacobianInputs[j++] = rbs.dq[i - n];
            }
        }

        #endregion

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            KeyboardInput();

            NextStep(gameTime);

            UpdateCamera();

            base.Update(gameTime);
        }

        double[] targetPos = new double[2];

        protected void KeyboardInput()
        {
            KeyboardState keyState = Keyboard.GetState();
            if (keyState.IsKeyDown(Keys.Up))
            {
                targetPos[1] -= 0.06;
            }
            if (keyState.IsKeyDown(Keys.Down))
            {
                targetPos[1] += 0.06;
            }
            if (keyState.IsKeyDown(Keys.Left))
            {
                targetPos[0] -= 0.06;
            }
            if (keyState.IsKeyDown(Keys.Right))
            {
                targetPos[0] += 0.06;
            }

            targetMatrix = Matrix.CreateScale(0.3f) * Matrix.CreateTranslation((float)targetPos[0],0,(float)targetPos[1]);
            
        }

        protected void NextStep(GameTime gameTime)
        {
            int n = rbs.numberOfCoords;
            //proceed to the next time step
            double timeStep = 0.005;
            int stepMult = Math.Max(1, (int)Math.Round(gameTime.ElapsedGameTime.TotalSeconds / timeStep));
            for (int l = 0; l < stepMult; l++)
            {
                //clear input forces
                rbs.setZeroAllForces();
                
                //add gravity
                double gravity = -9.8;
                for (int i = 0; i < rbs.numberOfNodes; ++i)
                {
                    rbs.addExternalForce(i, new double[] { 0, gravity * rbs.allNodes[i].mass(), 0 });
                }


                //add spring force between the target and the root
                double[] pos = new double[] { rbs.W[0][0, 3], rbs.W[0][1, 3], rbs.W[0][2, 3] }; //position of the root node
                pos[0] -= targetPos[0]; //x
                pos[2] -= targetPos[1]; //z
                double ks = 4;
                double[] f = VM.mult(pos, -ks);
                rbs.addExternalForce(0, f);

                //add damping force
                double Kd = 1; //derivative gain
                for (int i = 0; i < rbs.numberOfCoords; ++i)
                {
                    rbs.addGenForce(i, -Kd * rbs.dq[i]);
                }

                //evaluate matrices
                double[,] J;
                double[] Jdqd;
                rbs.evalDynamicsMatrices(out rbs.W);
                evalJacobians(out J, out Jdqd);

                //create matrix equation
                int len = n + numConst;
                double[,] A = new double[len, len];
                double[] b = new double[len];

                VM.copySubMatrix(A, rbs.Mv, 0, 0);
                VM.copySubMatrix(A, J, n, 0);
                VM.copySubMatrix(A, VM.transpose<double>(J), 0, n);

                VM.copySubArray(b, rbs.bv, 0);
                VM.copySubArray(b, VM.mult(Jdqd, -1), n);

                //solve matrix equation
                double[,] fA = VM.copy(A);
                double[] result = VM.JacobiSolve(fA, b, 10000);

                rbs.ddq = VM.subArray<double>(result, 0, n); //acceleration part
                double[] lambda = VM.subArray<double>(result, n, n + numConst); // Lagrange multipliers

                //sanity check : J dq = 0 ?
                //double[] res = VM.mult(J, rbs.dq); //should be zero

                //time integrate
                for (int i = 0; i < n; i++)
                {
                    rbs.q[i] += timeStep * rbs.dq[i];
                    rbs.dq[i] += timeStep * rbs.ddq[i];
                }
            }
        }

        protected override bool BeginDraw()
        {
            rbs.PrepareToDraw();

            //determine the configuration of line object
            Vector3 target = new Vector3((float)targetPos[0], 0, (float)targetPos[1]);
            Vector3 rootPos = new Vector3((float)rbs.W[0][0, 3], (float)rbs.W[0][1, 3], (float)rbs.W[0][2, 3]);
            Vector3 center = 0.5f * (target + rootPos);
            Vector3 x = target - rootPos;
            float length = x.Length();
            if (Math.Abs(length) > 1e-4f)
            {
                x = x / length;
                Vector3 z = Vector3.Cross(x, Vector3.UnitY);


                Matrix m = new Matrix(x.X, x.Y, x.Z, 0,
                    0, 1, 0, 0,
                    z.X, z.Y, z.Z, 0,
                    center.X, center.Y, center.Z, 1);

                lineMatrix = Matrix.CreateScale(0.05f, (float)length, 0.05f) * Matrix.CreateRotationZ(-(float)Math.PI / 2) * m;
            }
            else
            {
                lineMatrix = Matrix.CreateScale(0);
            }
            
            return base.BeginDraw();
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.RenderState.DepthBufferEnable = true;

            //draw ground
            DrawModel(target, targetMatrix, targetABT);

            //draw robot
            rbs.Draw(viewMatrix, projectionMatrix);

            //draw line
            DrawModel(line, lineMatrix, lineABT);

            //draw text
            spriteBatch.Begin();
            DrawText();
            spriteBatch.End();
            
            base.Draw(gameTime);
        }

        void DrawText()
        {
            spriteBatch.DrawString(spriteFont, "Press Arrows to drag one end of the link with the yellow sphere.", new Vector2(20, 45), Color.White);
        }

        /// <summary>
        /// Handles input for moving the camera.
        /// </summary>
        void UpdateCamera()
        {
            // clamp the camera distance so it doesn't get too close or too far away.
            cameraDistance = MathHelper.Clamp(cameraDistance, cameraMinDistance, cameraMaxDistance);

            Matrix unrotatedView = Matrix.CreateLookAt(
                new Vector3(0, 0, cameraDistance), new Vector3(3, 0, 0), Vector3.Up);

            viewMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(cameraRotation)) *
                          Matrix.CreateRotationX(MathHelper.ToRadians(cameraArc)) *
                          unrotatedView;
        }

        /// <summary>
        /// DrawModel is a helper function that takes a model, world matrix, and
        /// bone transforms. It does just what its name implies, and draws the model.
        /// </summary>
        /// <param name="model">the model to draw</param>
        /// <param name="worldTransform">where to draw the model</param>
        /// <param name="absoluteBoneTransforms">the model's bone transforms. this can
        /// be calculated using the function Model.CopyAbsoluteBoneTransformsTo</param>
        private void DrawModel(Model model, Matrix worldTransform, Matrix[] absoluteBoneTransforms)
        {
            // nothing tricky in here; this is the same model drawing code that we see
            // everywhere. we'll loop over all of the meshes in the model, set up their
            // effects, and then draw them.
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.PreferPerPixelLighting = true;

                    effect.View = viewMatrix;
                    effect.Projection = projectionMatrix;
                    effect.World = absoluteBoneTransforms[mesh.ParentBone.Index] * worldTransform;
                }
                mesh.Draw();
            }
        }

        public MultiBodySystemXNA buildClosedLoop(ContentManager Content)
        {
            Function.newContext();
            List<LinkXNA> alljoints = new List<LinkXNA>();
            Variable t = new Variable("t");
            Graph<Link> g = null;

            //root
            double mass_r = 1.0, len_r = 0.5;
            Function[,] com_r = { { 0 }, { 0 }, { 0 }, { 1 } };
            Function[,] inertia_r = MultiBodySystem.inertiaMatrixSphere(mass_r, len_r);
            UnspecifiedFunction[] q_r = new UnspecifiedFunction[2];//x,z translation
            for (int i = 0; i < 2; ++i) q_r[i] = UnspecifiedFunction.functionOf("qr" + i.ToString(), t);
            Function[,] A_r = VM.translate(q_r[0], 0, q_r[1]);

            LinkXNA root = new LinkXNA(A_r, q_r, com_r, inertia_r, mass_r, "root", t);
            root.SetShape(Content.Load<Model>("unitSphereBlue"));
            root.scale = new Vector3((float)len_r);
            alljoints.Add(root);

            //inertia of cylinder
            double mass = 1.0, len = 2, radius = 0.3;
            Function[,] com = { { 0.5 * len }, { 0 }, { 0 }, { 1 } };
            Function[,] inertia = MultiBodySystem.inertiaMatrixCylinder(mass, radius, len);
            int n = 4;
            UnspecifiedFunction[] q = new UnspecifiedFunction[n * 2];
            LinkXNA[] link = new LinkXNA[n];

            q[0] = UnspecifiedFunction.functionOf("q00", t);
            q[1] = UnspecifiedFunction.functionOf("q01", t);
            Function[,] A0 = VM.translate(0, -0.5 * 0, 0);
            A0 = VM.mult(A0, VM.rotZ(q[0], 0, 0, 0));
            A0 = VM.mult(A0, VM.rotY(q[1], 0, 0, 0));
            link[0] = new LinkXNA(A0, new Function[] { q[0], q[1] }, com, inertia, mass, "link0", t);
            link[0].SetShape(Content.Load<Model>("unitCylinderCham"));
            link[0].rotation = new Vector3(0, 0, (float)(Math.PI / 2));
            link[0].translation = new Vector3((float)(len / 2), 0, 0);
            link[0].scale = new Vector3((float)radius, (float)len, (float)radius);
            alljoints.Add(link[0]);
            root.c(link[0]);

            for (int i = 1; i < n; ++i)
            {
                q[2 * i] = UnspecifiedFunction.functionOf("q" + i.ToString() + "0", t);
                q[2 * i + 1] = UnspecifiedFunction.functionOf("q" + i.ToString() + "1", t);

                Function[,] A = VM.translate(len, 0, 0);
                A = VM.mult(A, VM.rotZ(q[2 * i], 0, 0, 0));
                A = VM.mult(A, VM.rotY(q[2 * i + 1], 0, 0, 0));

                link[i] = new LinkXNA(A, new Function[] { q[2 * i], q[2 * i + 1] }, com, inertia, mass, "link" + i.ToString(), t);
                link[i].SetShape(Content.Load<Model>("unitCylinderCham"));
                link[i].rotation = new Vector3(0, 0, (float)(Math.PI / 2));
                link[i].translation = new Vector3((float)(len / 2), 0, 0);
                link[i].scale = new Vector3((float)radius, (float)len, (float)radius);

                alljoints.Add(link[i]);
                link[i - 1].c(link[i]);
            }

            g = new Graph<Link>(root, alljoints.ToArray());

            MultiBodySystemXNA RBSTemp = new MultiBodySystemXNA(g, t, null);

            return RBSTemp;
        }


    }
}
