using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

using matrix;
using utilities;
using DifferentiableFunction;
using Dynamics;

namespace ExplicitConstraint
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        #region Camera Model
        // The next set of variables are used to control the camera used in the sample. 
        // It is an arc ball camera, so it can rotate in a sphere around the target, and
        // zoom in and out.
        float cameraArc = 30.0f;
        float cameraRotation = -30.0f;
        float cameraDistance = 15.0f;
        float cameraMaxDistance = 20.0f;
        float cameraMinDistance = 0.2f;

        Matrix viewMatrix;
        Matrix projectionMatrix;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;
        #endregion

        MultiBodySystemXNA rbs;

        Model ground;
        Matrix groundMatrix;
        Matrix[] groundABT;

        double[] qd = new double[] { 0, 0, 0 };

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // define projection matrix
            float aspectRatio = graphics.GraphicsDevice.Viewport.Width /
                (float)graphics.GraphicsDevice.Viewport.Height;
            projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(45.0f), aspectRatio, .01f, 1000);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            spriteFont = Content.Load<SpriteFont>("SpriteFont1");

            // TODO: use this.Content to load your game content here
            rbs = buildSliderCrank(Content);

            //ground
            ground = Content.Load<Model>("sliderGround");
            groundMatrix = Matrix.Identity;
            groundABT = new Matrix[ground.Bones.Count];
            ground.CopyAbsoluteBoneTransformsTo(groundABT);

        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            NextStep(gameTime);

            UpdateCamera();

            base.Update(gameTime);
        }

        protected void NextStep(GameTime gameTime)
        {
            //define applied forces
            rbs.setZeroAllForces();
            double gravity = -9.8;
            for (int i = 0; i < rbs.numberOfNodes; ++i)
            {
                rbs.addExternalForce(i, new double[] { 0, gravity * rbs.allNodes[i].mass(), 0 });
            }
            //PD control
            double Ks = 100;//proportional gain
            double Kd = 10; //derivative gain
            double qd = gameTime.TotalGameTime.TotalMilliseconds * 0.001;
            rbs.addGenForce(0, Ks* (qd - rbs.q[0]) - Kd * rbs.dq[0]);

            //proceed to the next time step
            double timeStep = 0.01;
            int stepMult = Math.Max(1, (int)Math.Round(gameTime.ElapsedGameTime.TotalSeconds / timeStep));
            for (int i = 0; i < stepMult; i++)
            {
                rbs.nextStepSymplecticEulerMethod(timeStep);
            }
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.RenderState.DepthBufferEnable = true;

            //draw ground
            DrawModel(ground, groundMatrix, groundABT);

            //draw robot
            rbs.PrepareToDraw();
            rbs.Draw(viewMatrix, projectionMatrix);

            base.Draw(gameTime);
        }

        /// <summary>
        /// Handles input for moving the camera.
        /// </summary>
        void UpdateCamera()
        {
            // clamp the camera distance so it doesn't get too close or too far away.
            cameraDistance = MathHelper.Clamp(cameraDistance, cameraMinDistance, cameraMaxDistance);

            Matrix unrotatedView = Matrix.CreateLookAt(
                new Vector3(0, 0, cameraDistance), new Vector3(3, 0, 0), Vector3.Up);

            viewMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(cameraRotation)) *
                          Matrix.CreateRotationX(MathHelper.ToRadians(cameraArc)) *
                          unrotatedView;
        }

        /// <summary>
        /// DrawModel is a helper function that takes a model, world matrix, and
        /// bone transforms. It does just what its name implies, and draws the model.
        /// </summary>
        /// <param name="model">the model to draw</param>
        /// <param name="worldTransform">where to draw the model</param>
        /// <param name="absoluteBoneTransforms">the model's bone transforms. this can
        /// be calculated using the function Model.CopyAbsoluteBoneTransformsTo</param>
        private void DrawModel(Model model, Matrix worldTransform, Matrix[] absoluteBoneTransforms)
        {
            // nothing tricky in here; this is the same model drawing code that we see
            // everywhere. we'll loop over all of the meshes in the model, set up their
            // effects, and then draw them.
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.PreferPerPixelLighting = true;

                    effect.View = viewMatrix;
                    effect.Projection = projectionMatrix;
                    effect.World = absoluteBoneTransforms[mesh.ParentBone.Index] * worldTransform;
                }
                mesh.Draw();
            }
        }

        public MultiBodySystemXNA buildSliderCrank(ContentManager Content)
        {
            Function.newContext();
            List<LinkXNA> alljoints = new List<LinkXNA>();
            Variable t = new Variable("t");
            Graph<Link> g = null;
            
            //link0 
            UnspecifiedFunction q0 = UnspecifiedFunction.functionOf("q0", t);
            float mass0 = 1.0f, lenX0 = 2, lenY0 = 0.3f, lenZ0 = 0.5f;
            Function[,] com0 = { { 0.5 * lenX0 }, { 0 }, { 0 }, { 1 } };
            Function[,] inertia0 = MultiBodySystem.inertiaMatrixBox(mass0, lenX0, lenY0, lenZ0);
            Function[,] A0 = VM.translate(0, 0, 0);
            A0 = VM.mult(A0, VM.rotY(q0, 0, 0, 0));
            LinkXNA rigidBody0 = new LinkXNA(A0, new Function[] { q0 }, com0, inertia0, mass0, "link0", t);
            alljoints.Add(rigidBody0);
            rigidBody0.SetShape(Content.Load<Model>("sliderLink0"));
            
            //link1
            UnspecifiedFunction q1 = UnspecifiedFunction.functionOf("q1", t);
            float mass1 = 2.0f, lenX1 = 4, lenY1 = 0.3f, lenZ1 = 0.5f;
            Function[,] com1 = { { 0.5 * lenX1 }, { 0 }, { 0 }, { 1 } };
            Function[,] inertia1 = MultiBodySystem.inertiaMatrixBox(mass1, lenX1, lenY1, lenZ1);
            Function[,] A1 = VM.translate(lenX0, 0, 0);
            A1 = VM.mult(A1, VM.rotY(q1, 0, 0, 0));
            LinkXNA rigidBody1 = new LinkXNA(A1, new Function[] { q1 }, com1, inertia1, mass1, "link1", t);
            alljoints.Add(rigidBody1);
            rigidBody0.c(rigidBody1);  
            rigidBody1.SetShape(Content.Load<Model>("sliderLink1"));
            
            //link2 
            UnspecifiedFunction q2 = UnspecifiedFunction.functionOf("q2", t);
            float mass2 = 0.2f, lenX2 = 1, lenY2 = 0.6f, lenZ2 = 0.5f;
            Function[,] com2 = { { 0 }, { 0 }, { 0 }, { 1 } };
            Function[,] inertia2 = MultiBodySystem.inertiaMatrixBox(mass2, lenX2, lenY2, lenZ2);
            Function[,] A2 = VM.translate(lenX1, 0, 0);
            A2 = VM.mult(A2, VM.rotY(q2, 0, 0, 0));
            LinkXNA rigidBody2 = new LinkXNA(A2, new Function[] { q2 }, com2, inertia2, mass2, "link2", t);
            alljoints.Add(rigidBody2);
            rigidBody1.c(rigidBody2);  
            rigidBody2.SetShape(Content.Load<Model>("sliderLink2"));
           
            g = new Graph<Link>(rigidBody0, alljoints.ToArray());

            //Function[,] explicitCon = new Function[,] { { q1, 2 * Math.PI - q0 - q2 }, { q2, Function.asin(lenX0 / lenX1 * Function.sin(q0)) } };
            Function[,] explicitCon = new Function[,] { { q1, 2 * Math.PI - q0 - q2 }, { q2, lenX0 / lenX1 * Function.sin(q0) } };

            MultiBodySystemXNA RBSTemp = new MultiBodySystemXNA(g, t, explicitCon);

            return RBSTemp;
        }


    }
}
