﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Diagnostics;
using DifferentiableFunction;
using matrix;
using utilities;
using System.IO;



namespace Dynamics
{

    public class Link : GraphNode
    {
        internal Function[,] A; //transformation from its parent
        internal UnspecifiedFunction[] q; //generalized coordinates
        internal Function[,] r; //position of center of mass, stored as 4x1 matrix, with 1 as the last element
        internal Function[,] I; //inertia matrix
        internal Function m; //mass of object
        internal string name; //user readable name
        internal Variable t; //time variable
        internal int nodeIndex;
        internal Function[] genForces; //generalized forces
        internal Function[,] fext; //externally applied force, assumed to pass through center of mass, stored as 4x1 matrix with 0 as the last element
        internal Function[,] tauext; //externally applied torque, in the local coordinate frame, stored as 4x1 matrix with 0 as the last element
        public Function[,] Wi; //transformation with respect to the world frame

        public Link() { }

        /// <summary>
        /// can have a maximum of six generalized coordinates per joint
        /// </summary>
        /// <param name="transform"></param>
        /// <param name="generalizedCoordinates"></param>
        /// <param name="centerOfMass"></param>
        /// <param name="inertiaMatrix"></param>
        /// <param name="mass"></param>
        /// <param name="nodeName"></param>
        /// <param name="time"></param>
        /// 
        public Link(Function[,] transform, Function[] generalizedCoordinates,
            Function[,] centerOfMass, Function[,] inertiaMatrix, Function mass, string nodeName, Variable time)
        {
            VM.checkSize<Function>(transform, 4, 4);
            A = VM.copy<Function>(transform);

            if (generalizedCoordinates.GetLength(0) > 6) { throw new Exception("too many generalized coordinates for one joint"); }

            //stupid thing I have to do for backward compatibility with visuals studio 2005 and XNA. Switch back to this code when XNA upgrades to Visual Studio 2008
            //q = (UnspecifiedFunction[])generalizedCoordinates.Clone();
            //beginning of stupid backward compatibility code
            q = new UnspecifiedFunction[generalizedCoordinates.Length];
            for (int i = 0; i < q.Length; i++) { q[i] = (UnspecifiedFunction)generalizedCoordinates[i]; }
            //end of stupid backward compatibility code

            genForces = new Function[generalizedCoordinates.GetLength(0)];
            for (int i = 0; i < generalizedCoordinates.GetLength(0); ++i) genForces[i] = new Variable(nodeName + "_Q" + i.ToString());

            fext = new Function[,] { { new Variable(nodeName + "_Fx") }, { new Variable(nodeName + "_Fy") }, { new Variable(nodeName + "_Fz") }, { 0 } };
            tauext = new Function[,] { { new Variable(nodeName + "_Tx") }, { new Variable(nodeName + "_Ty") }, { new Variable(nodeName + "_Tz") }, { 0 } };

            VM.checkSize<Function>(centerOfMass, 4, 1);
            if ((centerOfMass[3, 0] as Constant).leafValue != 1) { throw new Exception("not valid homogenous point"); }
            r = VM.copy<Function>(centerOfMass);//4x1 matrix

            VM.checkSize<Function>(inertiaMatrix, 4, 4);
            I = VM.copy<Function>(inertiaMatrix);
            m = mass;
            name = nodeName;
            t = time; //parameterizing variable

            //these variables will only be set once but not here
            nodeIndex = -1; //set to illegal values so it will throw an exception if an uninitialized value is used
            Wi = null;
        }

        public double[] centerOfMass
        {
            get
            {
                double[] result = new double[4];
                for (int i = 0; i < result.Length; i++) { result[i] = (r[i, 0] as Constant).val; }
                return result;
            }
        }

        public double mass()
        {
            return (m as Constant).val;
        }

        public Function[,] force
        {
            set
            {
                VM.checkSize<Function>(value, 4, 1);
                if (value[3, 0] != Constant.newConstant(0)) { throw new Exception("force must be a vector, with 0 as last element"); } //this works because there is only one Constant object with any given value
                fext = VM.copy<Function>(value);
            }
            get { return VM.copy<Function>(fext); }
        }

        public Function[,] torque
        {
            set
            {
                VM.checkSize<Function>(value, 4, 1);
                if (value[3, 0] != Constant.newConstant(0)) { throw new Exception("torque must be a vector, with 0 as last element"); } //this works because there is only one Constant object with any given value
                tauext = VM.copy<Function>(value);
            }
            get { return VM.copy<Function>(tauext); }
        }

    }

    public class MultiBodySystem
    {
        public Variable time;
        Graph<Link> systemGraph; //graph of joint nodes
        Link[] systemNodes; //array of joint nodes  
        public UnspecifiedFunction[] genCoords1D; //generalized coordinates stored as a 1d array
        public Variable[] force1D; //external forces stored as a 1d array
        internal Variable[] torque1D;//external torques stored as a 1d array
        internal Variable[] genForce1D; //generalized forces stored as a 1d array    
        internal Function[,] explicitConstraints; //explicit constraints. see summary of the constructor.
        public Function[,] M; //mass matrix (symbolic)
        public Function[] b;  //Q-c (symbolic)
        public Function invDyn; // symbolic inverse dynamics equation

        // input to dynamics
        public double[] genForce1Dv; //generalized forces
        public double[] force1Dv; //external forces 
        public double[] torque1Dv; //external torques 
        public double[] q; //generalized coordinates
        public double[] dq;//velocity of generalized coordinates
        // equations of motion
        public double[,] Mv; //mass matrix 
        public double[] bv;  //Q-c 

        internal bool[] forDynInputVarIndices; // corresponds to [genForce1Dv,force1Dv,torque1Dv]. each element indicates wheter corresponding variable is included in inpurVariables or not.
        internal bool[] forDynInputFunIndices; // same thing for UnspecifiedFunctions
        internal double[] forDynInputs;        // input to dynamics equation from [genForce1Dv, force1Dv, torques1Dv, q, dq] (in that order, with non-domain variables being removed)

        internal bool[] invDynInputVarIndices; 
        internal bool[] invDynInputFunIndices;
        internal double[] invDynInputs;
        
        // output
        public double[] ddq; //acceleration of generalized coordinates
        public double[][,] W; //transformation of each link

        // runtime functions for forward/inverse dynamics and forward kinematics
        RuntimeFunction dynamicsEval; // runtime function for dynamics equation evaluation
        RuntimeFunction inverseDynamicsEval; // runtime function for inverse dynamics evaluation
        RuntimeFunction kinematicEval;// runtime function for kinematics evaluation (Wi given q)
               
        /// <summary>
        /// construct dynamics and kinematic equations
        /// </summary>
        /// <param name="_systemGraph">multibody system</param>
        /// <param name="t">time variable</param>
        /// <param name="_explicitConstraints">explicit constraints. 
        /// it must be a form of {{q_1,y_1},{q_2,y_2}...}
        /// where q_i is the generalized coordinate, defined as a UnspecifiedFunction, that 
        /// will be substituted with a constraint equation y_i. 
        /// _explicitConstraints must be null if there exist no explicit constraints.</param>
        public MultiBodySystem(Graph<Link> _systemGraph, Variable t, Function[,] _explicitConstraints)
        {
            time = t;
            systemGraph = _systemGraph;
            explicitConstraints = _explicitConstraints;
            Link[] nodes = systemGraph.allNodes;
            for (int i = 0; i < nodes.GetLength(0); i++) { nodes[i].nodeIndex = systemGraph.nodeIndex(nodes[i]); }

            UnspecifiedFunction[][] generalizedCoordinates = new UnspecifiedFunction[nodes.GetLength(0)][];
            Variable[][] genForce = new Variable[nodes.GetLength(0)][];
            Variable[][] force = new Variable[nodes.GetLength(0)][];
            Variable[][] torque = new Variable[nodes.GetLength(0)][];

            for (int i = 0; i < nodes.GetLength(0); i++)
            {
                generalizedCoordinates[i] = new UnspecifiedFunction[nodes[i].q.GetLength(0)];
                genForce[i] = new Variable[nodes[i].q.GetLength(0)];
                force[i] = new Variable[3];
                torque[i] = new Variable[3];
                for (int j = 0; j < nodes[i].q.GetLength(0); j++)
                {
                    generalizedCoordinates[i][j] = nodes[i].q[j];
                    genForce[i][j] = (Variable)nodes[i].genForces[j];
                }
                force[i][0] = (Variable)nodes[i].force[0, 0];
                force[i][1] = (Variable)nodes[i].force[1, 0];
                force[i][2] = (Variable)nodes[i].force[2, 0];

                torque[i][0] = (Variable)nodes[i].torque[0, 0];
                torque[i][1] = (Variable)nodes[i].torque[1, 0];
                torque[i][2] = (Variable)nodes[i].torque[2, 0];
            }

            genCoords1D = VM.jaggedToLinear<UnspecifiedFunction>(generalizedCoordinates);
            genForce1D = VM.jaggedToLinear<Variable>(genForce);
            force1D = VM.jaggedToLinear<Variable>(force);
            torque1D = VM.jaggedToLinear<Variable>(torque);

            //generate inverse/forward dynamics functions
            generateDynamics();

            //generate forward kinematics function
            generateKinematics();

            force1Dv = new double[3 * nodes.GetLength(0)];
            torque1Dv = new double[3 * nodes.GetLength(0)];
            genForce1Dv = new double[genCoords1D.Length];
            q = new double[genCoords1D.Length];
            dq = new double[genCoords1D.Length];

            //W
            kinematicsUpdate(q, out W); 
        }

        /// <summary>
        /// return number of generalized coordinates
        /// </summary>
        public int numberOfCoords { get { return genCoords1D.GetLength(0); } }

        /// <summary>
        /// return joint node array
        /// </summary>
        public Link[] allNodes
        {
            get
            {
                seeIfThere();
                return systemNodes;
            }
        }

        /// <summary>
        /// return the number of nodes (rigid bodies)
        /// </summary>
        public int numberOfNodes
        {
            get
            {
                seeIfThere();
                return systemNodes.Length;
            }
        }
        
        /// <summary>
        /// set initial state
        /// </summary>
        /// <param name="qinit">initial value of the generalized coordinates (e.g., inital joint angles) </param>
        /// <param name="dqinit">initial value of the time derivative of generalized coordinates (e.g., inital joint velocity) </param>
        public void setInitialState(double[] qinit, double[] dqinit)
        {
            setGenCoords(qinit);
            setGenVelocity(dqinit);
        }

        /// <summary>
        /// set generalized coordinates to new values
        /// </summary>
        /// <param name="qnew"></param>
        public void setGenCoords(double[] qnew)
        {
            if (qnew.Length != numberOfCoords) { throw new Exception(); }
            for (int i = 0; i < qnew.Length; i++)
            {
                q[i] = qnew[i];
            }
        }

        /// <summary>
        /// set time derivative of generalized coordinates to new values
        /// </summary>
        /// <param name="dqnew"></param>
        public void setGenVelocity(double[] dqnew)
        {
            if (dqnew.Length != numberOfCoords) { throw new Exception(); }
            for (int i = 0; i < dqnew.Length; i++)
            {
                dq[i] = dqnew[i];
            }
        }

        /// <summary>
        /// set zero to all external forces/torques and generalized forces
        /// </summary>
        public void setZeroAllForces()
        {
            int i;
            for (i = 0; i < force1Dv.Length; ++i) force1Dv[i] = 0;
            for (i = 0; i < torque1Dv.Length; ++i) torque1Dv[i] = 0;
            for (i = 0; i < genForce1Dv.Length; ++i) genForce1Dv[i] = 0;
        }

        /// <summary>
        /// set external forces
        /// </summary>
        /// <param name="idx">link index</param>
        /// <param name="f">external force</param>
        public void setExternalForce(int idx, double[] f)
        {
            if (idx < 0 || idx >= numberOfNodes) { throw new Exception("idx must be in range of [0,numberOfNodes-1]"); }
            force1Dv[3 * idx] = f[0];
            force1Dv[3 * idx + 1] = f[1];
            force1Dv[3 * idx + 2] = f[2];
        }

        /// <summary>
        /// add to external forces
        /// </summary>
        /// <param name="idx">link index</param>
        /// <param name="f">external force</param>
        public void addExternalForce(int idx, double[] f)
        {
            if (idx < 0 || idx >= numberOfNodes) { throw new Exception("idx must be in range of [0,numberOfNodes-1]"); }
            force1Dv[3 * idx] += f[0];
            force1Dv[3 * idx + 1] += f[1];
            force1Dv[3 * idx + 2] += f[2];
        }

        /// <summary>
        /// set external torques
        /// </summary>
        /// <param name="idx">link index</param>
        /// <param name="tau">torque</param>
        public void setExternalTorque(int idx, double[] tau)
        {
            if (idx < 0 || idx >= numberOfNodes) { throw new Exception("idx must be in range of [0,numberOfNodes-1]"); }
            torque1Dv[3 * idx] = tau[0];
            torque1Dv[3 * idx + 1] = tau[1];
            torque1Dv[3 * idx + 2] = tau[2];
        }

        /// <summary>
        /// add to external torques
        /// </summary>
        /// <param name="idx">link index</param>
        /// <param name="tau">torque</param>
        public void addExternalTorque(int idx, double[] tau)
        {
            if (idx < 0 || idx >= numberOfNodes) { throw new Exception("idx must be in range of [0,numberOfNodes-1]"); }
            torque1Dv[3 * idx] += tau[0];
            torque1Dv[3 * idx + 1] += tau[1];
            torque1Dv[3 * idx + 2] += tau[2];
        }

        /// <summary>
        /// set generalized forces
        /// </summary>
        /// <param name="idx">index in generalized force vector</param>
        /// <param name="Q">amount of force to apply</param>
        public void setGenForce(int i, double Q)
        {
            if (i < 0 || i >= numberOfCoords) { throw new Exception("i must be in range of [0,numberOfCoords-1]"); }
            genForce1Dv[i] = Q;
        }

        public void setGenForce(double[] Q)
        {
            if (numberOfCoords != Q.Length) { throw new Exception(); }
            for (int i = 0; i < numberOfCoords; ++i) genForce1Dv[i] = Q[i];
        }

        /// <summary>
        /// add to generalized forces
        /// </summary>
        /// <param name="idx">index in generalized force vector</param>
        /// <param name="Q">amount of force to apply</param>
        public void addGenForce(int i, double Q)
        {
            if (i < 0 || i >= numberOfCoords) { throw new Exception("i must be in range of [0,numberOfCoords-1]"); }
            genForce1Dv[i] += Q;
        }

        public void addGenForce(double[] Q)
        {
            if (numberOfCoords != Q.Length) { throw new Exception(); }
            for (int i = 0; i < numberOfCoords; ++i) genForce1Dv[i] += Q[i];
        }
             
        /// <summary>
        /// update inputVariables from genForce1Dv, force1Dv, torque1Dv, q, dq
        /// </summary>
        internal void updateForDynInputs()
        {
            int i, idx = 0, j = 0, n = numberOfCoords, l = numberOfNodes;
            for (i = 0; i < n; ++i)
            {
                if (forDynInputVarIndices[i]) forDynInputs[j++] = genForce1Dv[i];
            }
            idx = n;
            for (i = idx; i < idx + 3 * l; ++i)
            {
                if (forDynInputVarIndices[i]) forDynInputs[j++] = force1Dv[i - idx];
            }
            idx += 3 * l;
            for (i = idx; i < idx + 3 * l; ++i)
            {
                if (forDynInputVarIndices[i]) forDynInputs[j++] = torque1Dv[i - idx];
            }
            idx = 0;
            for (i = 0; i < n; ++i)
            {
                if (forDynInputFunIndices[i]) forDynInputs[j++] = q[i];
            }
            idx += n;
            for (i = idx; i < idx + n; ++i)
            {
                if (forDynInputFunIndices[i]) forDynInputs[j++] = dq[i-idx];
            }
        }



        /// <summary>
        /// update inputVariables from force1Dv, torque1Dv, q, dq, qddt
        /// </summary>
        internal void updateInvDynInputs(double[] qddt)
        {
            int i, idx = 0, j = 0, n = numberOfCoords, l = numberOfNodes;
            for (i = 0; i < 3 * l; ++i)
            {
                if (invDynInputVarIndices[i]) invDynInputs[j++] = force1Dv[i];
            }
            idx += 3 * l;
            for (i = idx; i < idx + 3 * l; ++i)
            {
                if (invDynInputVarIndices[i]) invDynInputs[j++] = torque1Dv[i - idx];
            }
            idx = 0;
            for (i = 0; i < n; ++i)
            {
                if (invDynInputFunIndices[i]) invDynInputs[j++] = q[i];
            }
            idx += n;
            for (i = idx; i < idx+n; ++i)
            {
                if (invDynInputFunIndices[i]) invDynInputs[j++] = dq[i-idx];
            }
            idx += n;
            for (i = idx; i < idx+n; ++i)
            {
                if (invDynInputFunIndices[i]) invDynInputs[j++] = qddt[i-idx];
            }
        }

        /// <summary>
        /// perform inverse dynamics: given current state and input force/torque,
        /// compute genForce to create desired acceleration.
        /// </summary>
        /// <param name="qddt"></param>
        /// <param name="genForce"></param>
        public void inverseDynamics(double[] qddt, out double[] genForce)
        {
            updateInvDynInputs(qddt);
            genForce = new double[genForce1Dv.Length];
            inverseDynamicsEval.eval(genForce, invDynInputs);
        }
        
        /// <summary>
        /// peform forward dynamics: compute acceleration (qddt) and transformation matrices (W) 
        /// from inputs ( genForce, force, torque, q, dq ).
        /// </summary>
        /// <param name="qddt"></param>
        /// <param name="W"></param>
        public void forwardDynamics(out double[] qddt, out double[][,] W)
        {
            evalDynamicsMatrices(out W);
            //TODO: use better method for solving linear algebra (e.g., Cholesky decomposition).
            //FortranArray fA = new FortranArray(Mv); //change to Fortran array
            //qddt = Lapack.solve(fA, bv); // SVD is used.
           
            double[,] fA = VM.copy(Mv);
            qddt = VM.JacobiSolve(fA, bv,10000);
        }

        /// <summary>
        /// evaluate Mv, bv
        /// </summary>
        /// <param name="W"></param>
        public void evalDynamicsMatrices(out double[][,] W)
        {
            updateForDynInputs();
            double[] evalResult = new double[dynamicsEval.rangeDimension];
            dynamicsEval.eval(evalResult, forDynInputs);
            int n = numberOfCoords;
            Mv = VM.vectorToMatrix<double>(n, n, 0, evalResult);   //first n*n elements in evalResult
            bv = VM.subArray<double>(evalResult, n * n, n * n + n); //next n elements
            W = VM.vectorToMatrix<double>(evalResult, 4, 4, numberOfNodes, n * n + n); //next numberOfNodes*(4*4) elements
        }

        /// <summary>
        /// compute W as a function of q (generalized coordinates)
        /// </summary>
        /// <param name="q"></param>
        /// <param name="W"></param>
        public void kinematicsUpdate(double[] q, out double[][,] W)
        {
            double[] evalResult = new double[kinematicEval.rangeDimension];
            kinematicEval.eval(evalResult, q);
            W = new double[numberOfNodes][,];
            W = VM.vectorToMatrix<double>(evalResult, 4, 4, numberOfNodes, 0);
        }

        /// <summary>
        /// evolve to the next time step using symplectic Euler method
        /// </summary>
        /// <param name="h">time step</param>
        public void nextStepSymplecticEulerMethod(double h)
        {
            forwardDynamics(out ddq, out W);
            for (int i = 0; i < dq.GetLength(0); i++)
            {
                dq[i] += h * ddq[i];
                q[i] += h * dq[i];
            }
        }

        /// <summary>
        /// evolve to the next time step using forward Euler method
        /// </summary>
        /// <param name="h">time step</param>
        public void nextStepForwardEulerMethod(double h)
        {
            forwardDynamics(out ddq, out W);
            for (int i = 0; i < dq.GetLength(0); i++)
            {
                q[i] += h * dq[i];
                dq[i] += h * ddq[i];
            }
        }

        /// <summary>
        /// evolve to the next time step using 4th order Runge Kutta Method.
        /// refer to http://en.wikipedia.org/wiki/Runge-Kutta
        /// </summary>
        /// <param name="h">time step</param>
        public void nextStepRK4(double h)
        {
            double h2 = h * 0.5;
            double h3 = h / 3.0;
            double h6 = h / 6.0;

            double[] q0 = new double[numberOfCoords];
            double[] dq0 = new double[numberOfCoords];
            q.CopyTo(q0, 0);
            dq.CopyTo(dq0, 0);

            double[] qu = new double[numberOfCoords];
            double[] dqu = new double[numberOfCoords];
            q.CopyTo(qu, 0);
            dq.CopyTo(dqu, 0);

            //k1
            forwardDynamics(out ddq, out W); 
            qu = VM.plus(qu, VM.mult(dq, h6));
            dqu = VM.plus(dqu, VM.mult(ddq, h6));

            //k2
            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            forwardDynamics(out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            //k3
            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            forwardDynamics(out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            //k4
            q = VM.plus(q0, VM.mult(dq, h));
            dq = VM.plus(dq0, VM.mult(ddq, h));
            forwardDynamics(out ddq, out W);
            q = VM.plus(qu, VM.mult(dq, h6));
            dq = VM.plus(dqu, VM.mult(ddq, h6));
        }

        /// <summary>
        /// return array of Wi's. all the Wi must be computed before calling this function
        /// </summary>
        public Function[][,] transformFunction()
        {
            Link[] nodes = systemGraph.allNodes;
            Function[][,] result = new Function[nodes.Length][,];

            for (int i = 0; i < nodes.Length; i++) { result[i] = nodes[i].Wi; }
            return result;
        }

        

        internal Function[][] generalizedForces()
        {
            Func<Graph<Link>, Function[,], Link, Function[,]> treeProduct = null;

            treeProduct = (Graph<Link> g, Function[,] Wi, Link node) =>
            {
                Function[,] localTerms = VM.plus(linearComponent(node, Wi), rotationalComponent(node, Wi));

                Function[,] childrenSum = VM.constant<Function>(4, 4, Constant.newConstant(0.0));

                foreach (Link child in g.childrenNodes(node)) { childrenSum = VM.plus(childrenSum, VM.mult(child.A, treeProduct(g, VM.mult(Wi, child.A), child))); }
                return VM.plus(localTerms, childrenSum);
            };

            treeProduct = treeProduct.Memoize();

            Function[][] result = new Function[systemGraph.numberOfNodes][];
            Link[] root = systemGraph.roots();
            if (root.GetLength(0) != 1) { throw new Exception("can only have one root in a tree structured system"); }

            generalizedForces(systemGraph, root[0], VM.identity<Function>(4, 0.0, 1.0), result, treeProduct);
            return result;

        }

        /// <summary>
        /// computes generalized forces corresponding to each of the generalized coordinates
        /// </summary>
        internal void generalizedForces(Graph<Link> g, Link node, Function[,] Wiminus1, Function[][] Qi, Func<Graph<Link>, Function[,], Link, Function[,]> treeProduct)
        {
            Function[,] Wi = VM.mult(Wiminus1, node.A);
            Function[,] xi = VM.mult(Wi, node.r);
            int forceIndex = g.nodeIndex(node);
            Qi[forceIndex] = new Function[node.q.GetLength(0)];
            node.Wi = (Function[,])Wi.Clone();


            for (int i = 0; i < node.q.Length; i++)
            {
                Function trace = VM.trace(VM.mult(Wiminus1, VM.mult(Function.D(node.A, node.q[i]), treeProduct(g, Wi, node)))); //Wi-1*dA/dq*((ri*m*xidd) + Ai+1(ri+1*mi+1*xi+1dd + Ai+2(.....)
                //Qi[forceIndex][i] = trace - node.genForces[i]; //subtract user specified generalized force so end up with an equation like this: sum(acc - force)*dxdqi - Qi = 0
                Qi[forceIndex][i] = trace; // will subtract node.genForces[i] later (in order to create inverse dynamics runtime funtion)
            }
            foreach (Link childNode in g.childrenNodes(node)) { generalizedForces(g, childNode, Wi, Qi, treeProduct); }
        }

        internal void seeIfThere() { if (systemNodes == null) { systemNodes = systemGraph.allNodes; } }

        //internal string packedMatricesCCode()
        //{
        //    return createPackedMatricesFunction().CProgramString();
        //}

        internal void createDynamicsFunctions()
        {
            Function[][] forces = generalizedForces();
            Function[] Q = VM.jaggedToLinear<Function>(forces); // Q = M*ddq + c

            //if exists, substitute generalized coordinates with the constraint expressions
            if (explicitConstraints != null)
            {
                //record indices of removed generalized coordinates
                List<int> removedGenCoordIndices = new List<int>();
                List<UnspecifiedFunction> list = genCoords1D.ToList();
                for (int i = 0; i < explicitConstraints.GetLength(0); ++i)
                {
                    int idx = list.IndexOf((UnspecifiedFunction)explicitConstraints[i, 0]);
                    removedGenCoordIndices.Add(idx);
                }

                //reduce Q, genCoords1D, genForce1D                
                List<UnspecifiedFunction> listGenCoord = new List<UnspecifiedFunction>();
                List<Variable> listGenforce = new List<Variable>();
                List<Function> listQ = new List<Function>();
                for (int i = 0; i < genCoords1D.Length; ++i)
                {
                    if (!removedGenCoordIndices.Contains(i))
                    {
                        listGenCoord.Add(genCoords1D[i]);
                        listGenforce.Add(genForce1D[i]);
                        listQ.Add(Q[i]);
                    }
                }
                genCoords1D = listGenCoord.ToArray();
                genForce1D = listGenforce.ToArray();
                Q = listQ.ToArray();


                //substitute with explicitConstraints
                for (int l = 0; l < explicitConstraints.GetLength(0); l++)
                {
                    Function[,] con = new Function[,] { { explicitConstraints[l, 0], explicitConstraints[l, 1] } };

                    for (int i = 0; i < Q.Length; ++i)
                    {
                        Q[i] = Function.substitute(Q[i], con);
                    }
                    Link[] nodes = systemGraph.allNodes;
                    for (int i = 0; i < nodes.Length; i++)
                    {
                        for (int j = 0; j < 3; j++)
                        {
                            for (int k = 0; k < 4; k++)
                            {
                                nodes[i].A[j, k] = Function.substitute(nodes[i].A[j, k], con);
                                nodes[i].Wi[j, k] = Function.substitute(nodes[i].Wi[j, k], con);
                            }
                        }
                    }
                }
            }

            Function derivQ = Function.derivative(Q);

            Function qdt = Function.derivative(Function.D(genCoords1D, time));
            Function qddt = Function.derivative(Function.D(genCoords1D, time, time));
            
            ///
            ///inverse dynamics
            ///
            UnspecifiedFunction[] tempqdot = new UnspecifiedFunction[qdt.rangeDimension];
            for (int i = 0; i < qdt.rangeDimension; i++) { tempqdot[i] = (UnspecifiedFunction)qdt[i]; }
            UnspecifiedFunction[] tempqddot = new UnspecifiedFunction[qddt.rangeDimension];
            for (int i = 0; i < qddt.rangeDimension; i++) { tempqddot[i] = (UnspecifiedFunction)qddt[i]; }
            UnspecifiedFunction[] inputStates = VM.concatenate<UnspecifiedFunction>(genCoords1D, tempqdot, tempqddot);

            invDyn = derivQ; //symbolic inverse dynamics equation
            //create invDynInputVarIndices
            Variable[] variables = VM.concatenate<Variable>(force1D, torque1D);
            variables = initializeInputVariables(invDyn, variables, out invDynInputVarIndices);
            //create invDynInputFunIndices
            inputStates = initializedInputUnspecifiedFunctions(invDyn, inputStates, out invDynInputFunIndices);

            invDynInputs = new double[variables.Length + inputStates.Length]; 
            invDyn.orderVariablesInDomain(variables, inputStates); //input order: force1D, torque1D, q, dq, ddq
            inverseDynamicsEval = invDyn.compile();

            ///
            ///forward dynamics: create Matrix equations 
            ///
            Function[,] substitutionMap = new Function[genCoords1D.GetLength(0), 2];
            for (int i = 0; i < substitutionMap.GetLength(0); i++)
            {
                substitutionMap[i, 0] = qddt[i];
                substitutionMap[i, 1] = 0;
            }

            M = new Function[Q.GetLength(0), Q.GetLength(0)];
            b = new Function[M.GetLength(0)];

            for (int i = 0; i < M.GetLength(0); i++)
            {
                b[i] = -Function.substitute(derivQ[i], substitutionMap); //need to get Ax=b but have the equation in this form: Ax + b = 0 so have to flip sign.
                for (int j = i; j < M.GetLength(1); j++)
                {
                    M[j, i] = M[i, j] = derivQ[i].coefficient(qddt[j]); //use symmetry of M
                }
            }
            
            for (int i = 0; i < b.Length; ++i)
            {
                b[i] = b[i] + genForce1D[i]; //add genForce to b so that b = genForce - c 
            }
            
            dynamicsEval = pack(M, b).compile();
            
        }

        /// <summary>
        /// pack matrix equation into a single function: A[0,0],A[0,1],...,A[1,0], A[1,1],..., b[0], b[1], ..., W0, W1, ...
        /// </summary>
        /// <param name="A"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        internal Function pack(Function[,] A, Function[] b)
        {
            Function[] packed = VM.concatenate<Function>(VM.matrixToVector<Function>(A), b, VM.matricesToVector<Function>(transformFunction()));
            Function temp = new Function(packed);
//            temp = Function.derivative(temp);
            
            //create forDynInputFunIndices
            Function qdot = Function.derivative(Function.D(genCoords1D, time));
            UnspecifiedFunction[] tempqdot = new UnspecifiedFunction[qdot.rangeDimension];
            for (int i = 0; i < qdot.rangeDimension; i++) { tempqdot[i] = (UnspecifiedFunction)qdot[i]; }
            UnspecifiedFunction[] inputStates = VM.concatenate<UnspecifiedFunction>(genCoords1D, tempqdot);
            inputStates = initializedInputUnspecifiedFunctions(temp, inputStates, out forDynInputFunIndices);
            
            //create forDynInputVarIndices
            Variable[] variables = VM.concatenate<Variable>(genForce1D, force1D, torque1D);
            variables = initializeInputVariables(temp, variables, out forDynInputVarIndices);
            
            //order variables
            forDynInputs = new double[variables.Length + inputStates.Length];             
            temp.orderVariablesInDomain(variables, inputStates);
            return temp;
        }

        /// <summary>
        /// this is used at compile time
        /// </summary>
        /// <param name="resultOfEval"></param>
        /// <param name="A"></param>
        /// <param name="b"></param>
        //void unpack(Function packed, out Function[,] A, out Function[] b) {
        //  int n = numberOfCoords, n2 = n * n;
        //  A = new Function[n, n];
        //  b = new Function[n];

        //  for (int i = 0; i < n; i++) {
        //    b[i] = packed[n2 + i];
        //    for (int j = 0; j < n; j++) { A[i, j] = packed[i * n + j]; }
        //  }
        //}

        /// <summary>
        /// this is used at runtime
        /// </summary>
        /// <param name="resultOfEval"> Vector containing A and b in packed form: 
        /// A is n by n. n2 = n*n.
        /// A[i, j] = resultOfEval[i * n + j] 
        /// b[i] = resultOfEval[n2 + i]. b follows A terms in packed result</param>
        /// <param name="A"> Matrix of coefficients of qddt terms</param>
        /// <param name="b"> vector of everything product term that doesn't have a qddt element in it</param>
        void unpack(double[] resultOfEval, out double[,] A, out double[] b)
        {
            int n = numberOfCoords, n2 = n * n;
            A = new double[n, n];
            b = new double[n];

            for (int i = 0; i < n; i++)
            {
                b[i] = resultOfEval[n2 + i];
                for (int j = 0; j < n; j++) { A[i, j] = resultOfEval[i * n + j]; }
            }
        }

        /// <summary>
        /// generate kinematics function
        /// </summary>
        internal void generateKinematics()
        {
            Function kinematics = new Function(VM.matricesToVector<Function>(transformFunction()));
            UnspecifiedFunction[] inputVariables = VM.concatenate<UnspecifiedFunction>(genCoords1D);
            kinematics.orderVariablesInDomain(inputVariables);
            kinematicEval = kinematics.compile();
        }

        /// <summary>
        /// generate dynamics function
        /// </summary>
        internal void generateDynamics()
        {
            createDynamicsFunctions();            
        }

        #region static functions

        /// <summary>
        /// from testInput, collect only elements that are the domain of func and store their indices to ivi
        /// return domain variables.
        /// </summary>
        public static Variable[] initializeInputVariables(Function func, Variable[] testInput, out bool[] ivi)
        {
            int i, j = 0;
            Variable[] tDom = func.trueDomain();
            ivi = new bool[testInput.GetLength(0)];
            Variable[] inputVar = new Variable[tDom.GetLength(0)];
            for (i = 0; i < testInput.GetLength(0); ++i)
            {
                if (tDom.Contains<Variable>(testInput[i]))
                {
                    ivi[i] = true;
                    inputVar[j++] = testInput[i];
                }
                else
                    ivi[i] = false;
            }
            return inputVar;
        }

        /// <summary>
        /// from testInput, collect only UnspecifiedFunctions that are used by func and store their indices to iui
        /// return actually used functions.
        /// </summary>
        public static UnspecifiedFunction[] initializedInputUnspecifiedFunctions(Function func, UnspecifiedFunction[] testInput, out bool[] iui)
        {
            int i, j = 0;
            UnspecifiedFunction[] dom = func.unspecifiedFunctionVariables;
            iui = new bool[testInput.GetLength(0)];
            UnspecifiedFunction[] inputFunc = new UnspecifiedFunction[dom.GetLength(0)];
            for (i = 0; i < testInput.GetLength(0); ++i)
            {
                if (dom.Contains<UnspecifiedFunction>(testInput[i]))
                {
                    iui[i] = true;
                    inputFunc[j++] = testInput[i];
                }
                else
                    iui[i] = false;
            }
            return inputFunc;
        }

        //public static void projectionAxis(JointNode curr, Function[,] Wi, out Function[,] omegaQ, out Function[,] projectionAxis)
        //{
        //    Function[,] x = VM.mult(Wi, curr.r);

        //    Function[,] dist = VM.minus(x, VM.mult(Wi, new Function[,] { { 0 }, { 0 }, { 0 }, { 1 } })); //distance = x - world origin
        //    Function[,] Q = rotationPart(Wi);
        //    omegaQ = VM.mult(VM.transpose(Q), Function.D(Q, curr.q)); //because this is derived from a product of rotation matrices its magnitude will always be one
        //    projectionAxis = VM.mult(Q,VM.mult(omegaQ, dist)); //omega crossproduct dist. omegaQ gives angular velocity in wrong coord frame so have to transform back
        //}

        internal static Function[,] rotationalComponent(Link curr, Function[,] Wi)
        {
            //need to compute tilde(QT (Iwd - tau))
            Function[,] Q = rotationPart(Wi), QT = VM.transpose(Q);
            Function[,] omega = tilde(VM.mult(VM.transpose(Q), Function.D(Q, curr.t)));
            Function[,] Hdt = Function.D(VM.mult(Q, VM.mult(curr.I, omega)), curr.t); //d(angular momentum)/dt in world coordinates
            return VM.mult(VM.mult(QT,tilde(VM.minus(Hdt, curr.tauext))), -.5);
        }

        internal static Function[,] linearComponent(Link curr, Function[,] Wi)
        {
            //need to compute r(mxdd - f)T
            Function[,] x = VM.mult(Wi, curr.r);
            Function[,] xdd = Function.D(x, curr.t, curr.t);
            return VM.mult(curr.r, VM.transpose<Function>(VM.minus(VM.mult(xdd, curr.m), curr.force)));
        }

        /// <summary>
        /// return only the rotation part of a transformation matrix
        /// </summary>
        /// <param name="mat"></param>
        /// <returns></returns>
        public static Function[,] rotationPart(Function[,] mat)
        {
            Function[,] result = VM.constant<Function>(4, 4, 0);
            for (int i = 0; i < 3; i++)
            {
                for (int j = 0; j < 3; j++)
                {
                    result[i, j] = mat[i, j];
                }
            }
            return result;
        }

        /// <summary>
        /// return only the translation part of a tranformation matrix
        /// </summary>
        /// <param name="mat"></param>
        /// <returns></returns>
        public static Function[] translationPart(Function[,] mat)
        {
            return new Function[] { mat[0, 3], mat[1, 3], mat[2, 3] };
        }

        /// <summary>
        /// switch between a skew symmetric matrix and a vector
        /// </summary>
        /// <param name="vectorOrMatrix"></param>
        /// <returns></returns>
        public static Function[,] tilde(Function[,] vectorOrMatrix)
        {
            Function[,] m = vectorOrMatrix;
            if (vectorOrMatrix.GetLength(1) == 1)
            {//input is vector so return 3x3 antisymmetric matrix embedded in 4x4 matrix
                VM.checkSize<Function>(m, 4, 1);
                return new Function[,]{
                    {0,-m[2,0], m[1,0], 0},
                    {m[2,0],0,-m[0,0], 0},
                    {-m[1,0],m[0,0],0,0},
                    {0,0,0,0}};
            }
            else
            {
                VM.checkSize<Function>(m, 4, 4);
                return new Function[,] { 
                  {m[2, 1]}, 
                  {m[0,2]}, 
                  {m[1, 0]},
                  {0}};
            } //{x,y,z} embedded in 4 vector
        }

        /// <summary>
        /// returns the inertia matrix of a box
        /// </summary>
        public static Function[,] inertiaMatrixBox(double mass, double lenX, double lenY, double lenZ)
        {
            // moment of inertia:
            return new Function[,]{{1/12.0*mass*((lenY*lenY) + (lenZ*lenZ)), 0 ,0 ,0},
                         {0, 1/12.0*mass*((lenX*lenX)+(lenZ*lenZ)), 0 ,0},
                         {0, 0, 1/12.0*mass*((lenX*lenX) + (lenY*lenY)), 0},
                         {0,0,0,0}};
        }

        /// <summary>
        /// returns the inertia matrix of a cylinder (Y is axis)
        /// </summary>
        public static Function[,] inertiaMatrixCylinder(double mass, double radius, double height)
        {
            // moment of inertia:
            return new Function[,]{{1/12.0*mass*(3*(radius*radius) + (height*height)), 0 ,0 ,0},
                         {0, .5*mass*(radius*radius), 0 ,0},
                         {0, 0, 1/12.0*mass*(3*(radius*radius) + (height*height)), 0},
                         {0,0,0,0}};
        }

        /// <summary>
        /// returns the inertia matrix of a torus
        /// </summary>
        public static Function[,] inertiaMatrixTorus(double mass, double largeRadius, double smallRadius)
        {
            return new Function[,]{{1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0 ,0 ,0},
                         {0, .25*mass*(4*largeRadius*largeRadius + 3*smallRadius*smallRadius), 0 ,0},
                         {0, 0, 1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0},
                         {0,0,0,0}};
        }

        /// <summary>
        /// returns the inertia matrix of a sphere
        /// </summary>
        public static Function[,] inertiaMatrixSphere(double mass, double radius)
        {
            return new Function[,]{ {0.4*mass*radius*radius, 0 ,0 ,0},
                                    {0, 0.4*mass*radius*radius, 0 ,0},
                                    {0, 0, 0.4*mass*radius*radius, 0},
                                    {0,0,0,0}};
        }

        #endregion
    }
}
