using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

using matrix;
using utilities;
using DifferentiableFunction;
using Dynamics;

namespace InverseDynamics
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        #region Camera Model
        // The next set of variables are used to control the camera used in the sample. 
        // It is an arc ball camera, so it can rotate in a sphere around the target, and
        // zoom in and out.
        float cameraArc = 30.0f;
        float cameraRotation = -30.0f;
        float cameraDistance = 10.0f;
        float cameraMaxDistance = 20.0f;
        float cameraMinDistance = 0.2f;

        Matrix viewMatrix;
        Matrix projectionMatrix;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;
        #endregion

        MultiBodySystemXNA rbs;

        Model ground;
        Matrix groundMatrix;
        Matrix[] groundABT;

        RuntimeFunction evalJointTrajectory;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // define projection matrix
            float aspectRatio = graphics.GraphicsDevice.Viewport.Width /
                (float)graphics.GraphicsDevice.Viewport.Height;
            projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(45.0f), aspectRatio, .01f, 1000);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            spriteFont = Content.Load<SpriteFont>("SpriteFont1");

            // TODO: use this.Content to load your game content here
            rbs = buildRobotArm(Content);

            // trajectory
            Variable t = new Variable("t");
            Function[] jointTrajectory = new Function[] { Function.sin(t), 0, Function.sin(t) };
            Function[] vel = Function.D(jointTrajectory, t);
            Function[] acc = Function.D(jointTrajectory, t, t);
            Function f = new Function(VM.concatenate<Function>(jointTrajectory,vel,acc));
            f = Function.derivative(f);
            evalJointTrajectory = f.compile();

            //ground
            ground = Content.Load<Model>("unitBox");
            groundMatrix = Matrix.CreateScale(new Vector3(10.0f, 1.0f, 10.0f)) * Matrix.CreateTranslation(new Vector3(0, -0.5f, 0));
            groundABT = new Matrix[ground.Bones.Count];
            ground.CopyAbsoluteBoneTransformsTo(groundABT);

        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here

            NextStep(gameTime);

            UpdateCamera();

            base.Update(gameTime);
        }

        double[] genforce;
        protected void NextStep(GameTime gameTime)
        {
            //state update
            double[] result = new double[3 * 3];
            evalJointTrajectory.eval(result, gameTime.TotalGameTime.TotalMilliseconds * 0.001);
            double[] q = VM.subArray(result, 0, 3);
            double[] dq = VM.subArray(result, 3, 6);
            double[] qddt = VM.subArray(result, 6, 9);

            rbs.setGenCoords(q);
            rbs.setGenVelocity(dq);
            rbs.kinematicsUpdate(q, out rbs.W);

            //define applied forces
            rbs.setZeroAllForces();
            double gravity = -9.8;
            for (int i = 0; i < rbs.numberOfNodes; ++i)
            {
                rbs.addExternalForce(i, new double[] { 0, gravity * rbs.allNodes[i].mass(), 0 });
            }
   
            //inverse dynamics
            rbs.inverseDynamics(qddt, out genforce);

        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.RenderState.DepthBufferEnable = true;

            //draw ground
            DrawModel(ground, groundMatrix, groundABT);

            //draw robot
            rbs.PrepareToDraw();
            rbs.Draw(viewMatrix, projectionMatrix);

            //draw text
            spriteBatch.Begin();
            DrawText();
            spriteBatch.End();

            base.Draw(gameTime);
        }

        private void DrawText()
        {
            spriteBatch.DrawString(spriteFont, "Inverse dynamics result: " + genforce[0].ToString("0.00") + ", " + genforce[1].ToString("0.00") + ", " + genforce[2].ToString("0.00"), new Vector2(20, 45), Color.White);
        }

        /// <summary>
        /// Handles input for moving the camera.
        /// </summary>
        void UpdateCamera()
        {
            // clamp the camera distance so it doesn't get too close or too far away.
            cameraDistance = MathHelper.Clamp(cameraDistance, cameraMinDistance, cameraMaxDistance);

            Matrix unrotatedView = Matrix.CreateLookAt(
                new Vector3(0, 0, cameraDistance), new Vector3(0, 2, 0), Vector3.Up);

            viewMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(cameraRotation)) *
                          Matrix.CreateRotationX(MathHelper.ToRadians(cameraArc)) *
                          unrotatedView;
        }

        /// <summary>
        /// DrawModel is a helper function that takes a model, world matrix, and
        /// bone transforms. It does just what its name implies, and draws the model.
        /// </summary>
        /// <param name="model">the model to draw</param>
        /// <param name="worldTransform">where to draw the model</param>
        /// <param name="absoluteBoneTransforms">the model's bone transforms. this can
        /// be calculated using the function Model.CopyAbsoluteBoneTransformsTo</param>
        private void DrawModel(Model model, Matrix worldTransform, Matrix[] absoluteBoneTransforms)
        {
            // nothing tricky in here; this is the same model drawing code that we see
            // everywhere. we'll loop over all of the meshes in the model, set up their
            // effects, and then draw them.
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.PreferPerPixelLighting = true;

                    effect.View = viewMatrix;
                    effect.Projection = projectionMatrix;
                    effect.World = absoluteBoneTransforms[mesh.ParentBone.Index] * worldTransform;
                }
                mesh.Draw();
            }
        }

        public MultiBodySystemXNA buildRobotArm(ContentManager Content)
        {
            Function.newContext();
            List<LinkXNA> alljoints = new List<LinkXNA>();
            Variable t = new Variable("t");
            Graph<Link> g = null;

            //dimension of a link
            float mass = 1.0f;
            float lenX = 2.0f;
            float lenY = 0.5f;
            float lenZ = 0.5f;
            Function[,] centerOfMass = { { 0.5 * lenX }, { 0 }, { 0 }, { 1 } };
            Function[,] inertia = MultiBodySystem.inertiaMatrixBox(mass, lenX, lenY, lenZ);

            //link0 - rotation X
            UnspecifiedFunction q0 = UnspecifiedFunction.functionOf("q0", t);
            Function[,] A0 = VM.rotZ(Math.PI * 0.5, 0, 0, 0);
            A0 = VM.mult(A0, VM.rotX(q0, 0, 0, 0));
            LinkXNA rigidBody0 = new LinkXNA(A0, new Function[] { q0 }, centerOfMass, inertia, mass, "rootBody", t);
            alljoints.Add(rigidBody0);
            rigidBody0.SetShape(Content.Load<Model>("robotLink0"));

            //link1 - rotation Z
            UnspecifiedFunction q1 = UnspecifiedFunction.functionOf("q1", t);
            Function[,] A1 = VM.translate(lenX, 0, 0);
            A1 = VM.mult(A1, VM.rotZ(-Math.PI / 4, 0, 0, 0));
            A1 = VM.mult(A1, VM.rotZ(q1, 0, 0, 0));
            LinkXNA rigidBody1 = new LinkXNA(A1, new Function[] { q1 }, centerOfMass, inertia, mass, "link1", t);
            alljoints.Add(rigidBody1);
            rigidBody0.c(rigidBody1);  // declare rigidBody1 as a child of rigidBody0
            rigidBody1.SetShape(Content.Load<Model>("robotLink1"));

            //link2 - rotation Z
            UnspecifiedFunction q2 = UnspecifiedFunction.functionOf("q2", t);
            Function[,] A2 = VM.translate(lenX, 0, 0);
            A2 = VM.mult(A2, VM.rotZ(-Math.PI / 2, 0, 0, 0));
            A2 = VM.mult(A2, VM.rotZ(q2, 0, 0, 0));
            LinkXNA rigidBody2 = new LinkXNA(A2, new Function[] { q2 }, centerOfMass, inertia, mass, "link2", t);
            alljoints.Add(rigidBody2);
            rigidBody1.c(rigidBody2);  // declare rigidBody1 as a child of rigidBody0
            rigidBody2.SetShape(Content.Load<Model>("robotLink2"));

            g = new Graph<Link>(rigidBody0, alljoints.ToArray());

            MultiBodySystemXNA RBSTemp = new MultiBodySystemXNA(g, t, null);

            return RBSTemp;
        }


    }
}
