using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

using matrix;
using utilities;
using DifferentiableFunction;
using Dynamics;
using Interval;

namespace Torus
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        #region Camera Model
        // The next set of variables are used to control the camera used in the sample. 
        // It is an arc ball camera, so it can rotate in a sphere around the target, and
        // zoom in and out.
        float cameraArc = 20.0f;
        float cameraRotation = -20.0f;
        float cameraDistance = 50.0f;//50.0f;
        float cameraMaxDistance = 100.0f;
        float cameraMinDistance = 0.2f;

        Matrix viewMatrix;
        Matrix projectionMatrix;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;
        #endregion

        MultiBodySystemXNA rbs;

        Model ground;
        Matrix groundMatrix;
        Matrix[] groundABT;


        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // define projection matrix
            float aspectRatio = graphics.GraphicsDevice.Viewport.Width /
                (float)graphics.GraphicsDevice.Viewport.Height;
            projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(45.0f), aspectRatio, .01f, 1000);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            spriteFont = Content.Load<SpriteFont>("SpriteFont1");

            // TODO: use this.Content to load your game content here
            rbs = buildTorus(Content);

            //ground
            ground = Content.Load<Model>("unitBox");
            groundMatrix = Matrix.CreateScale(new Vector3(10.0f, 1.0f, 10.0f)) * Matrix.CreateTranslation(new Vector3(0, -4.5f, 0));
            groundABT = new Matrix[ground.Bones.Count];
            ground.CopyAbsoluteBoneTransformsTo(groundABT);

        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            KeyboardInput();
            // TODO: Add your update logic here

            NextStep(gameTime);

            UpdateCamera();

            base.Update(gameTime);
        }

        protected void KeyboardInput()
        {
//            Random r = new Random();
//            double D2R = Math.PI / 180.0;
            KeyboardState keyState = Keyboard.GetState();
            if (keyState.IsKeyDown(Keys.Space))
            {
                //specify what you want to do here when Space key pressed
            }
        }

        protected void NextStep(GameTime gameTime)
        {
            //proceed to the next time step
            double timeStep = 0.002;//0.001
            int stepMult = Math.Max(1, (int)Math.Round(gameTime.ElapsedGameTime.TotalSeconds / timeStep));
            for (int l = 0; l < stepMult; l++)
            {
                //define applied forces
                rbs.setZeroAllForces();
                double gravity = -9.8;
                
                for (int i = 0; i < rbs.numberOfNodes; ++i)
                {
                    rbs.addExternalForce(i, new double[] { 0, gravity * rbs.allNodes[i].mass(), 0 });
                }

                const double Kd = 1;
                for (int i = 0; i < rbs.numberOfCoords; ++i)
                {
                    rbs.setGenForce(i, -Kd * rbs.dq[i]);
                }

                rbs.nextStepSymplecticEulerMethod(timeStep);
            }
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.RenderState.DepthBufferEnable = true;

            //draw ground
            DrawModel(ground, groundMatrix, groundABT);

            //draw robot
            rbs.PrepareToDraw();
            rbs.Draw(viewMatrix, projectionMatrix);

            //draw text
            spriteBatch.Begin();
            DrawText();
            spriteBatch.End();

            base.Draw(gameTime);
        }

        private void DrawText()
        {
//            spriteBatch.DrawString(spriteFont, "Press Spacebar to set random target pose", new Vector2(20, 45), Color.White);
        }

        /// <summary>
        /// Handles input for moving the camera.
        /// </summary>
        void UpdateCamera()
        {
            // clamp the camera distance so it doesn't get too close or too far away.
            cameraDistance = MathHelper.Clamp(cameraDistance, cameraMinDistance, cameraMaxDistance);

            Matrix unrotatedView = Matrix.CreateLookAt(
                new Vector3(0, 0, cameraDistance), new Vector3(0, 2, 0), Vector3.Up);

            viewMatrix = Matrix.CreateRotationY(MathHelper.ToRadians(cameraRotation)) *
                          Matrix.CreateRotationX(MathHelper.ToRadians(cameraArc)) *
                          unrotatedView;
        }

        /// <summary>
        /// DrawModel is a helper function that takes a model, world matrix, and
        /// bone transforms. It does just what its name implies, and draws the model.
        /// </summary>
        /// <param name="model">the model to draw</param>
        /// <param name="worldTransform">where to draw the model</param>
        /// <param name="absoluteBoneTransforms">the model's bone transforms. this can
        /// be calculated using the function Model.CopyAbsoluteBoneTransformsTo</param>
        private void DrawModel(Model model, Matrix worldTransform, Matrix[] absoluteBoneTransforms)
        {
            // nothing tricky in here; this is the same model drawing code that we see
            // everywhere. we'll loop over all of the meshes in the model, set up their
            // effects, and then draw them.
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.PreferPerPixelLighting = true;

                    effect.View = viewMatrix;
                    effect.Projection = projectionMatrix;
                    effect.World = absoluteBoneTransforms[mesh.ParentBone.Index] * worldTransform;
                }
                mesh.Draw();
            }
        }

        public MultiBodySystemXNA buildTorus(ContentManager Content)
        {
            Function.newContext();
            List<LinkXNA> alljoints = new List<LinkXNA>();
            Variable t = new Variable("t");
            Graph<Link> g = null;

            double m = 1.0;
            Function[,] r0 = { {0},{0},{0},{1}};
            double largeRad = 1, smallRad = .2;

            Function[,] I0 = MultiBodySystem.inertiaMatrixTorus(m, largeRad, smallRad);
            Torus tor=new Torus(largeRad,smallRad);
            Function[,] flip={{0,1,0,0},{1,0,0,0},{0,0,-1,0},{0,0,0,1}};

            LinkXNA ji = null, jiMinus1 = null, jhead = null;
            int numJoints=10;
            
            for (int i = 0; i <numJoints; i++)
            {
                Function[,] A0;
                UnspecifiedFunction qi_0 = UnspecifiedFunction.functionOf("q" + i + "_0", t), qi_1 = UnspecifiedFunction.functionOf("q" + i + "_1", t);
                Function[,] RBi0 = tor.SurfaceTangentFrame(qi_0+Math.PI, qi_1+Math.PI);
                UnspecifiedFunction qi_2 = UnspecifiedFunction.functionOf("q" + i + "_2", t), qi_3 = UnspecifiedFunction.functionOf("q" + i + "_3", t);
                Function[,] RBi1 = tor.SurfaceTangentFrame(qi_2, qi_3+Math.PI);

                if (i == 0)
                {
                    A0 = VM.mult(VM.translate(0, 2 * numJoints-4, 0), VM.rigidBodyInverse(RBi0));
                    ji = new LinkXNA(A0, new Function[] { qi_0, qi_1 }, r0, I0, 1, "node"+i.ToString(), t);
                    ji.SetShape(Content.Load<Model>("torus"));
                    jhead = ji;
                }
                else
                {
                    A0 = VM.mult(RBi0, VM.mult(flip, VM.rigidBodyInverse(RBi1)));
                    ji = new LinkXNA(A0, new Function[] { qi_0, qi_1, qi_2, qi_3 }, r0, I0, 1, "node"+i.ToString(), t);
                    ji.SetShape(Content.Load<Model>("torus"));
                }
                

                alljoints.Add(ji);
                if (jiMinus1 != null)
                    jiMinus1.c(ji);
                jiMinus1 = ji;
            }
            

           
            

            g = new Graph<Link>(jhead, alljoints.ToArray());       
            MultiBodySystemXNA RBSTemp = new MultiBodySystemXNA(g, t, null);   
            
            return RBSTemp;
        }


    }

    public abstract class Geometry
    {
        protected Variable[] variables;
        protected IR[] domainBounds;  //u and v bounds
        protected bool[] periodicity;

        protected RuntimeFunction rtSurface;
        protected RuntimeFunction rtSurfaceAndNormal;

        public Geometry(Variable[] _variables, IR[] _domainBounds, bool[] _periodicity)
        {
            variables = new Variable[_variables.Length];
            _variables.CopyTo(variables,0);

            domainBounds = new IR[_domainBounds.Length];
            _domainBounds.CopyTo(domainBounds,0);

            periodicity = new bool[_periodicity.Length];
            _periodicity.CopyTo(periodicity,0);
        }

        public IR Domain(int i)
        {
            return domainBounds[i].copy();
        }

        //protected void CheckDomainBounds(double[] inputs)
        //{
        //    for (int i = 0; i < inputs.Length; i++)
        //    {
        //        if (!domainBounds[i].contains(inputs[i]))
        //            throw new ArgumentOutOfRangeException("inInvDyn out of domainBounds");
        //    }
        //}

        public void Surface(double[] outputs, params double[] inputs)
        {
            rtSurface.eval(outputs, inputs);
        }

        public void SurfaceAndNormal(double[] outputs, params double[] inputs)
        {
            rtSurfaceAndNormal.eval(outputs, inputs);
        }

        public abstract Function[,] SurfaceTangentFrame(params Function[] coord);

        ///// <summary>
        ///// p0->p1,p1-> p0, n->-n
        ///// </summary>
        ///// <param name="frame"></param>
        ///// <returns></returns>
        //public Function[,] swappedTangentFrame(Function[,] frame){
        //    return new Function[4, 4]{
        //    {frame[0,1],frame[0,0], -frame[0,2],frame[0,3]},
        //    {frame[1,1],frame[1,0], -frame[1,2],frame[1,3]},
        //    {frame[2,1],frame[2,0], -frame[2,2],frame[2,3]},
        //    {0,0,0,1}};
        //}

        /// <summary>
        /// Compute cross produce of Function
        /// </summary>
        /// <param name="v1"></param>
        /// <param name="v2"></param>
        /// <returns></returns>
        protected static Function[] Cross(Function[] v1, Function[] v2)
        {
            if (v1.Length != 3 || v2.Length != 3) throw new ArgumentOutOfRangeException("parameters must be 3 dimensional arrays");
            Function[] result = new Function[]
            {
                v1[1]*v2[2]-v1[2]*v2[1],
                v1[2]*v2[0]-v1[0]*v2[2],
                v1[0]*v2[1]-v1[1]*v2[0]
            };
            return result;
        }

        /// <summary>
        /// Normalized a Function vector v
        /// </summary>
        /// <param name="v"></param>
        protected static void Normalize(Function[] v)
        {
            if (v.Length != 3) throw new ArgumentOutOfRangeException("v must be a 3 dimensional array");
            Function scale = 1.0/Function.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
            v[0] = v[0] * scale;
            v[1] = v[1] * scale;
            v[2] = v[2] * scale;
        }

        /// <summary>
        /// Compute normalized normal vector of surface
        /// </summary>
        /// <param name="surface"></param>
        /// <returns></returns>
        protected static Function[] DefineNormalFunction(Function[] surface)
        {
            if (surface.Length != 3) throw new ArgumentOutOfRangeException("surface must be 3 dimensional array");
            
            Function surf = new Function(surface);            
            Variable u = surf.oneDomainVariable(0);
            Variable v = surf.oneDomainVariable(1);
            Function normal = Utilities.surfaceNormal(surf, u, v);
            return new Function[] { normal.range[0], normal.range[1], normal.range[2] };
        }

        
        public static Function[,] InverseOfTransformMatrix(Function[,] mat)
        {
            Function[,] iR = new Function[,]{ 
                { mat[0,0], mat[1,0], mat[2,0] },
                { mat[0,1], mat[1,1], mat[2,1] },
                { mat[0,2], mat[1,2], mat[2,2] },
            }; //inverse (i.e. transpose) of rotation matrix
            Function[] trn = VM.mult(iR, new Function[] { mat[0, 3], mat[1, 3], mat[2, 3] });
            return new Function[,]{
                {iR[0,0],iR[0,1],iR[0,2],-trn[0]},
                {iR[1,0],iR[1,1],iR[1,2],-trn[1]},
                {iR[2,0],iR[2,1],iR[2,2],-trn[2]},
                {0,0,0,1}};
        }

        public static Function[,] Translate(Function x, Function y, Function z)
        {
            return new Function[,]{
                {1,0,0,x},
                {0,1,0,y},
                {0,0,1,z},
                {0,0,0,1}};
        }

        ////refer to http://en.wikipedia.org/wiki/Arctangent
        ////return approximate arctangent
        //public static Function atan_approx(Function z)
        //{
        //    return z - z * z * z / 3.0 + Function.pow(z,5)/5.0 ;
        //}
    }

    class Torus:Geometry
    {
        public double R;
        public double r;

        public Torus(double radius0, double radius1)
            :base(
            new Variable[]{new Variable("u"),new Variable("v")},
            new IR[]{new IR(0,Math.PI*2),new IR(0,Math.PI*2)},
            new bool[]{true,true})
        {
            R=radius0;
            r=radius1;

            Variable u=variables[0];
            Variable v=variables[1];
            Function cu=Function.cos(u);
            Function su=Function.sin(u);
            Function cv=Function.cos(v);
            Function sv=Function.sin(v);
            Function S=(R+r*cv);

            Function[] surface=new Function[]{S*su,r*sv,S*cu};
            Function[] normal=new Function[]{su*cv,sv,cu*cv};

            Function surfaceRoot=new Function(surface);
            surfaceRoot.orderVariablesInDomain(u,v);
            rtSurface=surfaceRoot.compile();

            Function[] surfaceAndNormal=new Function[]{surface[0], surface[1],surface[2],normal[0],normal[1],normal[2]};
            Function surfaceAndNormalRoot=new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u,v);
            rtSurfaceAndNormal=surfaceAndNormalRoot.compile();
        }

        public override Function[,]  SurfaceTangentFrame(params Function[] uv)
        {
            Function u=uv[0];
            Function v=uv[1];
            Function cu=Function.cos(u);
            Function su=Function.sin(u);
            Function cv=Function.cos(v);
            Function sv=Function.sin(v);
            Function S=(R+r*cv);

            Function[] surface=new Function[]{S*su,r*sv,S*cu};
            Function[] basis1=new Function[]{cu,0,-su};
            Function[] normal=new Function[]{su*cv,sv,cu*cv};
            Function[] basis2=new Function[]{-su*sv,cv,-cu*sv};

            Function[,] frame=new Function[,]{
                {basis1[0],basis2[0],normal[0],surface[0]},
                {basis1[1],basis2[1],normal[1],surface[1]},
                {basis1[2],basis2[2],normal[2],surface[2]},
                {0,0,0,1}
            };

            return frame; 	
        }
    }
  
  

  


}
