using System;
using System.Collections.Generic;
using System.Text;
using matrix;
using utilities;
using DifferentiableFunction;
using Interval;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;



namespace Geometries
{
    public abstract class Geometry
    {
        protected Variable[] variables;
        protected IR[] domainBounds;  //u and v bounds
        protected bool[] periodicity;

        protected RuntimeFunction rtSurface;
        protected RuntimeFunction rtSurfaceAndNormal;

        public Geometry(Variable[] _variables, IR[] _domainBounds, bool[] _periodicity)
        {
            variables = new Variable[_variables.Length];
            _variables.CopyTo(variables,0);

            domainBounds = new IR[_domainBounds.Length];
            _domainBounds.CopyTo(domainBounds,0);

            periodicity = new bool[_periodicity.Length];
            _periodicity.CopyTo(periodicity,0);
        }

        public IR Domain(int i)
        {
            return domainBounds[i].copy();
        }

        //protected void CheckDomainBounds(double[] inputs)
        //{
        //    for (int i = 0; i < inputs.Length; i++)
        //    {
        //        if (!domainBounds[i].contains(inputs[i]))
        //            throw new ArgumentOutOfRangeException("inInvDyn out of domainBounds");
        //    }
        //}

        public void Surface(double[] outputs, params double[] inputs)
        {
            rtSurface.eval(outputs, inputs);
        }

        public void SurfaceAndNormal(double[] outputs, params double[] inputs)
        {
            rtSurfaceAndNormal.eval(outputs, inputs);
        }

        public abstract Function[,] SurfaceTangentFrame(params Function[] coord);

        ///// <summary>
        ///// p0->p1,p1-> p0, n->-n
        ///// </summary>
        ///// <param name="frame"></param>
        ///// <returns></returns>
        //public Function[,] swappedTangentFrame(Function[,] frame){
        //    return new Function[4, 4]{
        //    {frame[0,1],frame[0,0], -frame[0,2],frame[0,3]},
        //    {frame[1,1],frame[1,0], -frame[1,2],frame[1,3]},
        //    {frame[2,1],frame[2,0], -frame[2,2],frame[2,3]},
        //    {0,0,0,1}};
        //}

        /// <summary>
        /// Compute cross produce of Function
        /// </summary>
        /// <param name="v1"></param>
        /// <param name="v2"></param>
        /// <returns></returns>
        protected static Function[] Cross(Function[] v1, Function[] v2)
        {
            if (v1.Length != 3 || v2.Length != 3) throw new ArgumentOutOfRangeException("parameters must be 3 dimensional arrays");
            Function[] result = new Function[]
            {
                v1[1]*v2[2]-v1[2]*v2[1],
                v1[2]*v2[0]-v1[0]*v2[2],
                v1[0]*v2[1]-v1[1]*v2[0]
            };
            return result;
        }

        /// <summary>
        /// Normalized a Function vector v
        /// </summary>
        /// <param name="v"></param>
        protected static void Normalize(Function[] v)
        {
            if (v.Length != 3) throw new ArgumentOutOfRangeException("v must be a 3 dimensional array");
            Function scale = 1.0/Function.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
            v[0] = v[0] * scale;
            v[1] = v[1] * scale;
            v[2] = v[2] * scale;
        }

        /// <summary>
        /// Compute normalized normal vector of surface
        /// </summary>
        /// <param name="surface"></param>
        /// <returns></returns>
        protected static Function[] DefineNormalFunction(Function[] surface)
        {
            if (surface.Length != 3) throw new ArgumentOutOfRangeException("surface must be 3 dimensional array");
            
            Function surf = new Function(surface);            
            Variable u = surf.oneDomainVariable(0);
            Variable v = surf.oneDomainVariable(1);
            Function normal = Utilities.surfaceNormal(surf, u, v);
            return new Function[] { normal.range[0], normal.range[1], normal.range[2] };
        }

        /// <summary>
        /// compute surfaceTangentFrame, of which orientation is determined by ds/du, normal, and cross(ds/du,normal)
        /// and position is a where a is surface function.
        /// </summary>
        /// <param name="surface"></param>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        //protected static Function[,] DefineSurfaceTangentFrameFunction(Function[] surface)
        //{
        //    if (surface.Length != 3) throw new ArgumentOutOfRangeException("surface must be 3 dimensional array");

        //    Function surf = new Function(surface);
        //    Variable u = surf.oneDomainVariable(0);
        //    Variable v = surf.oneDomainVariable(1);
        //    Function[] basis1 = new Function[] { Function.D(surf, 0, u), Function.D(surf, 1, u), Function.D(surf, 2, u) };
        //    Normalize(basis1);
        //    Function[] basis2 = new Function[] { Function.D(surf, 0, v), Function.D(surf, 1, v), Function.D(surf, 2, v) };
        //    Normalize(basis2);
        //    Function[] basis3 = Cross(basis1, basis2);
        //    Function[] basis4 = Cross(basis3, basis1);

        //    Function[,] frame = new Function[,]{
        //        {basis1[0],basis1[1],basis1[2],surface[0]},
        //        {basis4[0],basis4[1],basis4[2],surface[1]},
        //        {basis3[0],basis3[1],basis3[2],surface[2]},
        //        {0,0,0,1}
        //    };
        //    return VM.transpose(frame);
        //}

        /// <summary>
        /// compute surfaceTangentFrame, of which orientation is determined by ds/du, normal, and cross(ds/du,normal)
        /// and position is a where a is surface function.
        /// </summary>
        /// <param name="surface"></param>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        //protected static Function[,] DefineSurfaceTangentFrameFunction(Function[] surface, Function[] normal)
        //{
        //    if (surface.Length != 3 || normal.Length != 3) 
        //        throw new ArgumentOutOfRangeException("surface must be 3 dimensional array");

        //    Function surf = new Function(surface);
        //    Variable u = surf.oneDomainVariable(0);
        //    Variable v = surf.oneDomainVariable(1);
        //    Function[] basis1 = new Function[] { Function.D(surf, 0, u), Function.D(surf, 1, u), Function.D(surf, 2, u) };
        //    Normalize(basis1);
        //    Function[] basis2 = new Function[normal.Length];
        //    normal.CopyTo(basis2, 0);
        //    Normalize(basis2);
        //    Function[] basis3 = Cross(basis2, basis1);

        //    Function[,] frame = new Function[,]{
        //        {basis1[0],basis1[1],basis1[2],0},
        //        {basis3[0],basis3[1],basis3[2],0},
        //        {basis2[0],basis2[1],basis2[2],0},
        //        {surface[0],surface[1],surface[2],1}
        //    };
        //    return VM.transpose(frame);
        //}

        public static Function[,] InverseOfTransformMatrix(Function[,] mat)
        {
            Function[,] iR = new Function[,]{ 
                { mat[0,0], mat[1,0], mat[2,0] },
                { mat[0,1], mat[1,1], mat[2,1] },
                { mat[0,2], mat[1,2], mat[2,2] },
            }; //inverse (i.e. transpose) of rotation matrix
            Function[] trn = VM.mult(iR, new Function[] { mat[0, 3], mat[1, 3], mat[2, 3] });
            return new Function[,]{
                {iR[0,0],iR[0,1],iR[0,2],-trn[0]},
                {iR[1,0],iR[1,1],iR[1,2],-trn[1]},
                {iR[2,0],iR[2,1],iR[2,2],-trn[2]},
                {0,0,0,1}};
        }

        public static Function[,] Translate(Function x, Function y, Function z)
        {
            return new Function[,]{
                {1,0,0,x},
                {0,1,0,y},
                {0,0,1,z},
                {0,0,0,1}};
        }

        ////refer to http://en.wikipedia.org/wiki/Arctangent
        ////return approximate arctangent
        //public static Function atan_approx(Function z)
        //{
        //    return z - z * z * z / 3.0 + Function.pow(z,5)/5.0 ;
        //}
    }

    //class WigglyPlane : Geometry
    //{
    //    public WigglyPlane(double x0, double x1, double z0, double z1, double wiggle)
    //        : base(
    //        new Variable[] { new Variable("u"), new Variable("v"), new Variable("w") },
    //        new IR[] { new IR(x0, x1), new IR(z0, z1) },
    //        new bool[] { false, false })
    //    {
    //        //shapeParams[0] = wiggle;

    //        Variable u = variables[0];
    //        Variable v = variables[1];
    //        Variable w = variables[2];

    //        Function[] surface = new Function[] { u, v, Function.sin(u * w) * Function.sin(v * w) };
    //        Function[] normal = DefineNormalFunction(surface);

    //        //surface
    //        Function surfaceRoot = new Function(surface);
    //        surfaceRoot.order(u, v, w);
    //        rtSurface = surfaceRoot.compile();

    //        //surface and normal            
    //        Function[] surfaceAndNormal = new Function[] { surface[0], surface[1], surface[2], normal[0], normal[1], normal[2] };
    //        Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
    //        surfaceAndNormalRoot.order(u, v, w);
    //        rtSurfaceAndNormal = surfaceAndNormalRoot.compile();  
    //    }

    //    public override void Surface(double[] outputs, params double[] inputs)
    //    {            
    //        rtSurface.eval(outputs, inputs);
    //    }

    //    public override void SurfaceAndNormal(double[] outputs, params double[] inputs)
    //    {         
    //        rtSurfaceAndNormal.eval(outputs, inputs);
    //    }

    //    public override Function[,] SurfaceTangentFrame(params Function[] uv)
    //    {
    //        Variable u = variables[0];
    //        Variable v = variables[1];
    //        Variable w = variables[2];
    //        Function[] surface = new Function[] { u, v, Function.sin(u * w) * Function.sin(v * w) };
    //        Function[,] frame = DefineSurfaceTangentFrameFunction(surface);
    //        return frame;
    //    }
    //}

    //Refer to http://en.wikipedia.org/wiki/Cl%C3%A9lies
    class Clelies : Geometry
    {
        public double r0, r1, m;
        //r0: radius of greates circle        
        //r2: radius of the tube        



        public Clelies(double r0, double r1, double m)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, 2 * Math.PI), new IR(0, 2 * Math.PI) },
            new bool[] { true, true })
        {
            this.r0 = r0;
            this.r1 = r1;
            this.m = m;

            Variable u = variables[0];
            Variable v = variables[1];
            Function sv = Function.sin(v);
            Function cv = Function.cos(v);
            Function smv = Function.sin(m * v);
            Function cmv = Function.cos(m * v);

            //surface function
            Function[] p = new Function[] { r0 * smv * cv, r0 * smv * sv, r0 * cmv };
            Function[] Rx = new Function[] { m * cmv * cv - smv * sv, m * smv * sv + smv * cv, -m * smv };
            Normalize(Rx);
            Function[] Rz = new Function[] { smv * cv, smv * sv, cmv };            
            Function[] Ry = Geometry.Cross(Rz, Rx);
            Function[,] T = new Function[,]{ 
                {Rx[0], Ry[0], Rz[0], p[0]},
                {Rx[1], Ry[1], Rz[1], p[1]},
                {Rx[2], Ry[2], Rz[2], p[2]},
                {0,0,0,1}};
            T = VM.mult(T, VM.rotX(u, 0, 0, 0));
            T = VM.mult(T, Geometry.Translate(0, 0, r1));

            Function[] surface = new Function[] { T[0, 3], T[1, 3], T[2, 3] };
            Function[] normal = new Function[] { T[0, 2], T[1, 2], T[2, 2] };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();

            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { T[0, 3], T[1, 3], T[2, 3], T[0, 2], T[1, 2], T[2, 2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];
            Function sv = Function.sin(v);
            Function cv = Function.cos(v);
            Function smv = Function.sin(m * v);
            Function cmv = Function.cos(m * v);

            //surface function
            Function[] p = new Function[] { r0 * smv * cv, r0 * smv * sv, r0 * cmv };
            Function[] Rx = new Function[] { m * cmv * cv - smv * sv, m * smv * sv + smv * cv, -m * smv };
            Normalize(Rx);
            Function[] Rz = new Function[] { smv * cv, smv * sv, cmv };            
            Function[] Ry = Geometry.Cross(Rz, Rx);
            Function[,] T = new Function[,]{ 
                {Rx[0], Ry[0], Rz[0], p[0]},
                {Rx[1], Ry[1], Rz[1], p[1]},
                {Rx[2], Ry[2], Rz[2], p[2]},
                {0,0,0,1}};
            T = VM.mult(T, VM.rotX(u, 0, 0, 0));
            T = VM.mult(T, Geometry.Translate(0, 0, r1));



            return T;
        }

    }

    class WavyCircle : Geometry
    {
        public double r0, r1, H;
        public int numWav;
        //r0: radius of greates circle        
        //r2: radius of the tube
        //H: magnitude of wave
        //numWav: number of wave



        public WavyCircle(double r0, double r1, double H, int numWav)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, 2 * Math.PI), new IR(0, 2 * Math.PI) },
            new bool[] { true, true })
        {
            this.r0 = r0;
            this.r1 = r1;
            this.H = H;
            this.numWav = numWav;

            Variable u = variables[0];
            Variable v = variables[1];
            Function sv = Function.sin(v);
            Function cv = Function.cos(v);
            Function snv = Function.sin(numWav*v);
            Function cnv = Function.cos(numWav*v);
            
            //surface function
            Function[] p = new Function[] { r0 * sv, H * snv, r0 * cv };
            Function[] Rx = new Function[] { r0*cv, numWav*H*cnv, -r0*sv };
            Normalize(Rx);
            Function[] Rz = new Function[] { sv, 0, cv };
            Function[] Ry = Geometry.Cross(Rz, Rx);
            Function[,] T = new Function[,]{ 
                {Rx[0], Ry[0], Rz[0], p[0]},
                {Rx[1], Ry[1], Rz[1], p[1]},
                {Rx[2], Ry[2], Rz[2], p[2]},
                {0,0,0,1}};
            T = VM.mult(T, VM.rotX(u, 0, 0, 0));
            T = VM.mult(T, Geometry.Translate(0, 0, r1));

            Function[] surface = new Function[] { T[0, 3], T[1, 3], T[2, 3] };
            Function[] normal = new Function[] { T[0,2], T[1,2], T[2,2] };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();

            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { T[0, 3], T[1, 3], T[2, 3], T[0, 2], T[1, 2], T[2, 2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];
            Function sv = Function.sin(v);
            Function cv = Function.cos(v);
            Function snv = Function.sin(numWav * v);
            Function cnv = Function.cos(numWav * v);
    
            Function[] p = new Function[] { r0 * sv, H * snv, r0 * cv };
            Function[] Rx = new Function[] { r0 * cv, numWav * H * cnv, -r0 * sv };
            Normalize(Rx);
            Function[] Rz = new Function[] { sv, 0, cv };
            Function[] Ry = Geometry.Cross(Rz, Rx);
            Function[,] T = new Function[,]{ 
                {Rx[0], Ry[0], Rz[0], p[0]},
                {Rx[1], Ry[1], Rz[1], p[1]},
                {Rx[2], Ry[2], Rz[2], p[2]},
                {0,0,0,1}};
            T = VM.mult(T, VM.rotX(u, 0, 0, 0));
            T = VM.mult(T, Geometry.Translate(0, 0, r1));


            return T;
        }

    }

    class CircularCoil : Geometry
    {
        public double r0, r1, r2;
        //r0: radius of greates circle
        //r1: radius of second greatest circle
        //r2: radius of the tube
        public int numRev;
        //number of revolutions

        public CircularCoil(double r0, double r1, double r2, int numRev)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, 2 * Math.PI), new IR(0, 2 * Math.PI) },
            new bool[] { true, true })
        {
            this.r0 = r0;
            this.r1 = r1;
            this.r2 = r2;
            this.numRev = numRev;

            Variable u = variables[0];
            Variable v = variables[1];
            
            //surface function
            Function[,] T = VM.mult(VM.rotY(u,0,0,0), Geometry.Translate(0, 0, r2));
            T = VM.mult(VM.mult(VM.rotZ(numRev * v, 0, 0, 0), Geometry.Translate(r1, 0, 0)), T);
            T = VM.mult(VM.mult(VM.rotX(v, 0, 0, 0), Geometry.Translate(0, r0, 0)), T);

            Function[] surface = new Function[] { T[0, 3], T[1, 3], T[2, 3] };
            Function[] normal = new Function[] { T[0,2], T[1,2], T[2,2] };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();

            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { T[0, 3], T[1, 3], T[2, 3], T[0, 2], T[1, 2], T[2, 2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];

            //surface function
            Function[,] T = VM.mult(VM.rotY(u, 0, 0, 0), Geometry.Translate(0, 0, r2));
            T = VM.mult(VM.mult(VM.rotZ(numRev * v, 0, 0, 0), Geometry.Translate(r1, 0, 0)), T);
            T = VM.mult(VM.mult(VM.rotX(v, 0, 0, 0), Geometry.Translate(0, r0, 0)), T);
            return T;
        }

    }

    class XZDisk : Geometry
    {
        public double height, inRad, outRad; //height, inner radius, outer radius

        public XZDisk(double r0, double r1, double height)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, 2*Math.PI), new IR(r0, r1) },
            new bool[] { true, false })
        {
            this.height = height;
            inRad = r0; 
            outRad = r1;

            Variable u = variables[0];
            Variable v = variables[1];
            Function su = Function.sin(u);
            Function cu = Function.cos(u);

            //surface function
            Function[] surface = new Function[] { v*su, height, v*cu };
            Function[] normal = new Function[] { 0, 1, 0 };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();

            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { v * su, height, v * cu, 0, 1, 0 };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];
            Function su = Function.sin(u);
            Function cu = Function.cos(u);

            //surface function
            Function[] surface = new Function[] { v * su, height, v * cu };

            //basis1..dsurface/du
            Function[] basis1 = new Function[] { cu, 0, -su };
            Function[] normal = new Function[] { 0, 1, 0 };
            Function[] basis2 = new Function[] { su, 0, cu };


            Function[,] frame = new Function[,]{
                {basis1[0],basis2[0],normal[0],surface[0]},
                {basis1[1],basis2[1],normal[1],surface[1]},
                {basis1[2],basis2[2],normal[2],surface[2]},
                {0,0,0,1},
            };
            return frame;
        }
    }

    class XZPlane : Geometry
    {
        public double height;

        public XZPlane(double x0, double x1, double z0, double z1, double height)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(x0, x1), new IR(z0,z1) },
            new bool[] { false, false })
        {
            this.height = height;

            Variable u = variables[0];
            Variable v = variables[1];

            //surface function
            Function[] surface = new Function[] { u , height, v };
            Function[] normal = new Function[] { 0, 1, 0 };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();
            
            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { u, height, v, 0, 1, 0 };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];

            //surface function
            Function[] surface = new Function[] { u, height, v };
                        
            Function[,] frame = new Function[,]{
                {1,0,0,surface[0]},
                {0,1,0,surface[1]},
                {0,0,1,surface[2]},
                {0,0,0,1},
            };

            return frame;
        }
    }

    public class Torus : Geometry
    {
        public double R; //R: distance from the center to the tube center
        public double r; //r: radius of the tube

        public Torus(double radius0, double radius1)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, Math.PI * 2), new IR(0, Math.PI * 2) },
            new bool[] { true, true })
        {
            R = radius0;
            r = radius1;

            Variable u = variables[0];
            Variable v = variables[1];
            Function cu = Function.cos(u);
            Function su = Function.sin(u);
            Function cv = Function.cos(v);
            Function sv = Function.sin(v);
            Function S = (R+r*cv);

            //surface function
            Function[] surface = new Function[] { S * su, r * sv, S * cu };
            //normal function
            Function[] normal = new Function[] { su * cv, sv, cu * cv };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();


            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { surface[0], surface[1], surface[2], normal[0], normal[1], normal[2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];
            Function cu = Function.cos(u);
            Function su = Function.sin(u);
            Function cv = Function.cos(v);
            Function sv = Function.sin(v);
            Function S = (R + r * cv);

            //surface
            Function[] surface = new Function[] { S * su, r * sv, S * cu };
            //basis1..dsurface/du
            Function[] basis1 = new Function[] { cu, 0, -su }; 
            //normal function
            Function[] normal = new Function[] { su * cv, sv, cu * cv };
            Function[] basis2 = new Function[] { -su * sv, cv, -cu * sv };


            Function[,] frame = new Function[,]{
                {basis1[0],basis2[0],normal[0],surface[0]},
                {basis1[1],basis2[1],normal[1],surface[1]},
                {basis1[2],basis2[2],normal[2],surface[2]},
                {0,0,0,1},
            };

            return frame;
        }
    }

    public class Sphere : Geometry
    {
        public double r; //radius
        public Variable vR; //radius
        
        public Sphere(Variable r)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, Math.PI * 2), new IR(-0.5 * Math.PI, 0.5 * Math.PI) },
            new bool[] { true, true })
        {
            this.vR = r;

            Variable u = variables[0];
            Variable v = variables[1];
            Function cu = Function.cos(u);
            Function su = Function.sin(u);
            Function cv = Function.cos(v);
            Function sv = Function.sin(v);

            //surface function
            Function[] surface = new Function[] { r * su * cv, r * sv, r * cu * cv };
            //NOTE that normal vanishes at v = +- 0.5*PI
            Function[] normal = new Function[] { r * r * su * (cv * cv), r * r * sv * cv, r * r * cu * (cv * cv) };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();


            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { surface[0], surface[1], surface[2], normal[0], normal[1], normal[2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }
        
        public Sphere(double r)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, Math.PI * 2), new IR(-0.5 * Math.PI, 0.5 * Math.PI) },
            new bool[] { true, true })
        {
            this.r = r;

            Variable u = variables[0];
            Variable v = variables[1];
            Function cu = Function.cos(u);
            Function su = Function.sin(u);
            Function cv = Function.cos(v);
            Function sv = Function.sin(v);

            //surface function
            Function[] surface = new Function[] { r * su * cv, r * sv, r * cu * cv };
            //NOTE that normal vanishes at v = +- 0.5*PI
            Function[] normal = new Function[] { r * r * su * (cv * cv), r * r * sv * cv, r * r * cu * (cv * cv) };

            //surface
            Function surfaceRoot = new Function(surface);
            
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();


            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { surface[0], surface[1], surface[2], normal[0], normal[1], normal[2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }
        
        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];
            Function cu = Function.cos(u);
            Function su = Function.sin(u);
            Function cv = Function.cos(v);
            Function sv = Function.sin(v);

            //surface function
            Function[] surface = new Function[] { r * (cv * su), r * sv, r * (cv * cu) };

            Function[] basis1 = new Function[] { r * cv * cu, 0, -r * (cv * su) }; //ds/du             
            Geometry.Normalize(basis1);
            Function[] normal = new Function[] { r * r * (su * cv) * cv, r * r * sv * cv, r * r * (cu * cv) * cv };
            Geometry.Normalize(normal);
            Function[] basis2 = Geometry.Cross(normal, basis1);

            Function[,] frame = new Function[,]{
                {basis1[0],basis2[0],normal[0],surface[0]},
                {basis1[1],basis2[1],normal[1],surface[1]},
                {basis1[2],basis2[2],normal[2],surface[2]},
                {0,0,0,1},
            };

            return frame;
        }
        
    }

    public class Ellipsoid : Geometry
    {
        public double a,b,c; //semi_axes. ellipsoid equation is (x/a)^2 + (y/b)^2 + (z/c)^2  = 1
        public bool outwardNormal; //if true, normal is outward

        public Ellipsoid(double axisX, double axisY, double axisZ, double u0, double u1, double v0, double v1, bool outwardNormal) 
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(u0, u1), new IR(v0, v1) }, //for complete ellipsoid, set u0=0,u1=2Pi,v0=-0.5Pi, v1=0.5Pi
            new bool[] {true,true} )
        {
            a = axisX;
            b = axisY;
            c = axisZ;
            this.outwardNormal = outwardNormal;

            Variable u = variables[0];
            Variable v = variables[1];
            Function cu = Function.cos(u);
            Function su = Function.sin(u);
            Function cv = Function.cos(v);
            Function sv = Function.sin(v);

            //surface function
            Function[] surface = new Function[] { a * cv * su, b * sv, c * cv * cu };
            //NOTE that normal vanishes at v = +- 0.5*PI
            Function[] normal;
            if(outwardNormal)
                normal = new Function[] { b * c * su * cv * cv, a * c * sv * cv, a * b * cu * cv * cv };
            else
                normal = new Function[] { -b * c * su * cv * cv, -a * c * sv * cv, -a * b * cu * cv * cv };

            //surface
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();
                       
            
            //surface and normal            
            Function[] surfaceAndNormal = new Function[] { surface[0], surface[1], surface[2], normal[0], normal[1], normal[2] };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();
        }

         public override Function[,] SurfaceTangentFrame(params Function[] uv)
         {
             Function u = uv[0];
             Function v = uv[1];
             Function cu = Function.cos(u);
             Function su = Function.sin(u);
             Function cv = Function.cos(v);
             Function sv = Function.sin(v);

             //surface function
             Function[] surface = new Function[] { a * ( cv * su ), b * sv, c * (cv * cu )};

             Function[] basis1 = new Function[] { a * cv * cu, 0, -c * (cv * su) }; //ds/du             
             Geometry.Normalize(basis1);
             Function[] normal;
             if (outwardNormal)
                 normal = new Function[] { b * c * (su * cv) * cv, a * c * sv * cv, a * b * (cu * cv) * cv };
             else
                 normal = new Function[] { -b * c * (su * cv) * cv, -a * c * sv * cv, -a * b * (cu * cv) * cv };
             
             Geometry.Normalize(normal);
             Function[] basis2 = Geometry.Cross(normal, basis1);             

             Function[,] frame = new Function[,]{
                {basis1[0],basis2[0],normal[0],surface[0]},
                {basis1[1],basis2[1],normal[1],surface[1]},
                {basis1[2],basis2[2],normal[2],surface[2]},
                {0,0,0,1},
            };

            return frame;
         }
    }

    public class Cone : Geometry
    {
        public double a; //slope factor
        public Cone(double x0, double x1, double slope)
            : base(
            new Variable[] { new Variable("u"), new Variable("v") },
            new IR[] { new IR(0, Math.PI * 2), new IR(x0, x1) },
            new bool[] { true, false })
        {
            this.a = slope;

            Variable u = variables[0];
            Variable v = variables[1];
            Function su = Function.sin(u);
            Function cu = Function.cos(u);

            //surface function
            Function[] surface = new Function[] { v, -a*v*su, a*v*cu };
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();

            //surface and normal function
            double scale = 1.0 / Math.Sqrt(1 + a*a);
            Function[] surfaceAndNormal = new Function[] { v, -a*v*su, a*v*cu, -scale*a, -scale*su, scale*cu };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();  
        }

        
        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];            
            Function su = Function.sin(u);
            Function cu = Function.cos(u);
            double scale = 1.0 / Math.Sqrt(1 + a*a);

            Function[,] A = new Function[,]
            {
                {0,-cu,-su,0},
                {scale, -scale*a*su, scale*a*cu,0},
                {-scale*a,-scale*su,scale*cu,0},
                {v,-a*v*su,a*v*cu,1}
            };
            return VM.transpose(A);
        }



    }

    public class Cylinder:Geometry
    {
        public double radius;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="args">args: thetaLow thetaHigh heightLow heightHigh radius
        /// for full cylinder thetaLow = 0, thetaHigh = 2pi
        /// </param>
        public Cylinder(double thetaLow, double thetaHigh, double heightLow, double heightHigh, double _radius)
            : base(
            new Variable[]{new Variable("u"), new Variable("v")},
            new IR[] { new IR(thetaLow, thetaHigh), new IR(heightLow, heightHigh) },
            new bool[]{true,false}
            )
        {
            this.radius = _radius;
            
            Variable u = variables[0];
            Variable v = variables[1];
            Function su = Function.sin(u);
            Function cu = Function.cos(u);

            //surface function
            Function[] surface = new Function[] { v, -radius * su , -radius * cu };
            Function surfaceRoot = new Function(surface);
            surfaceRoot.orderVariablesInDomain(u, v);
            rtSurface = surfaceRoot.compile();

            //surface and normal function
            Function[] surfaceAndNormal = new Function[] { v, -radius * su, -radius * cu, 0, -su, -cu };
            Function surfaceAndNormalRoot = new Function(surfaceAndNormal);
            surfaceAndNormalRoot.orderVariablesInDomain(u, v);
            rtSurfaceAndNormal = surfaceAndNormalRoot.compile();      
        }

        public override Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Function u = uv[0];
            Function v = uv[1];            
            Function su = Function.sin(u);
            Function cu = Function.cos(u);

            Function[,] A = new Function[,]
            {
                {1,0,0,v},
                {0,-1*cu,-1*su,-radius*su},
                {0,su,-1*cu,-radius*cu},
                {0,0,0,1}
            };
            return A;
        }

    }

    


}
