﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Diagnostics;
using DifferentiableFunction;
using matrix;
using utilities;
using lapack;
using System.IO;


namespace Dynamics
{

    public class Link : GraphNode
    {

        internal Function[,] A; //transformation from its parent
        internal Function[,] R; //Rotation from its parent       
        internal Function[,] p; //Translation from its parent, stored as a 3x1 matrix
        internal Function[,] r; //position of center of mass, stored as 3x1 matrix
        internal Function[,] r4;//position of center of mass, stored as 4x1 matrix
        internal Function[,] I; //inertia matrix, stored as a 3x3 matrix
        internal Function m; //mass of object
        internal string name; //user readable name
        internal Variable t; //time variable
        internal int nodeIndex;

        internal UnspecifiedFunction[] q; //generalized coordinates
        internal Function[,] fext; // externally applied force, assumed to pass through center of mass, stored as 3x1 matrix
        internal Function[,] tauext; //externally applied torque, in the local coordinate frame, stored as 3x1 matrix
        public Function[,] spatialVel; //spatial velocity
        internal Function[,] J; //generalized inertia
        internal Function[,] Ad; //Adjoint mapping
        public Function[,] S; //Joint Jacobian
        public Function[,] Wi;

        internal int parent;
        internal int[] childlist;

        public Link(Function[,] transform, Function[] generalizedCoordinates,
            Function[,] centerOfMass, Function[,] inertiaMatrix, Function mass, string nodeName, Variable time)
        {
            VM.checkSize<Function>(transform, 4, 4);
            A = VM.copy<Function>(transform);

            R = VM.subArray<Function>(A, 0, 0, 3, 3);
            p = VM.subArray<Function>(A, 0, 3, 3, 4);

            if (generalizedCoordinates.GetLength(0) > 6) { throw new Exception("too many generalized coordinates for one joint"); }
            q = new UnspecifiedFunction[generalizedCoordinates.Length];
            for (int i = 0; i < q.Length; i++)
            {
                q[i] = (UnspecifiedFunction)generalizedCoordinates[i];
            }
            fext = new Function[,] { { new Variable(nodeName + "_Fx") }, { new Variable(nodeName + "_Fy") }, { new Variable(nodeName + "_Fz") } };
            tauext = new Function[generalizedCoordinates.Length, 1];
            for (int i = 0; i < generalizedCoordinates.Length; i++)
                tauext[i, 0] = new Variable(nodeName + "_tau" + i.ToString());

            VM.checkSize<Function>(centerOfMass, 4, 1);
            r = VM.subArray<Function>(centerOfMass, 0, 0, 3, 1);
            r4 = VM.constant<Function>(4, 1, 1.0);
            VM.copySubMatrix<Function>(r4, r, 0, 0);            

            VM.checkSize<Function>(inertiaMatrix, 4, 4);
            I = VM.subArray<Function>(inertiaMatrix, 0, 0, 3, 3);

            m = mass;
            name = nodeName;
            t = time;

            nodeIndex = -1;
            Wi = null;
        }

        public double mass()
        {
            return (m as Constant).val;
        }

        public void computeQuantities()
        {
            computeJ();
            computeAd();
            computeS();
        }

        //function to compute the generalized inertia matrix
        internal void computeJ()
        {
            J = VM.constant<Function>(6, 6, Constant.newConstant(0.0));
            J[0, 0] = I[0, 0] + m * (Function.square(r[2, 0]) + Function.square(r[1, 0]));
            J[0, 1] = I[0, 1] - m * r[1, 0] * r[0, 0];
            J[1, 0] = J[0, 1];
            J[0, 2] = I[0, 2] - m * r[2, 0] * r[0, 0];
            J[2, 0] = J[0, 2];
            J[1, 1] = I[1, 1] + m * (Function.square(r[2, 0]) + Function.square(r[0, 0]));
            J[1, 2] = I[1, 2] - m * r[2, 0] * r[1, 0];
            J[2, 1] = J[1, 2];
            J[2, 2] = I[2, 2] + m * (Function.square(r[1, 0]) + Function.square(r[0, 0]));

            Function[,] mr = VM.mult(tilde3(r), m);
            VM.copySubMatrix<Function>(J, mr, 0, 3);
            VM.copySubMatrix<Function>(J, VM.transpose<Function>(mr), 3, 0);
            VM.copySubMatrix<Function>(J, VM.mult(VM.identity<Function>(3, 0.0, 1.0),m), 3, 3);

        }

        //function to compute the adjoint mapping matrix
        internal void computeAd()
        {
            Function[,] invR = VM.transpose<Function>(R);
            Function[,] invp = VM.mult(VM.mult(invR, p), -1);
            Ad = VM.constant<Function>(6, 6, Constant.newConstant(0.0));
            VM.copySubMatrix<Function>(Ad, invR, 0, 0);
            VM.copySubMatrix<Function>(Ad, VM.mult(tilde3(invp), invR), 3, 0);
            VM.copySubMatrix<Function>(Ad, invR, 3, 3);
        }

        //function to compute the joint Jacobian matrix
        internal void computeS()
        {
            Function[,] G = VM.identity<Function>(4, 0.0, 1.0);
            VM.copySubMatrix<Function>(G, R, 0, 0);
            VM.copySubMatrix<Function>(G, p, 0, 3);

            Function[,] RT = VM.transpose<Function>(R);
            Function[,] negp = VM.mult(p, -1);
            Function[,] invG = VM.identity<Function>(4, 0.0, 1.0);
            VM.copySubMatrix<Function>(invG, RT, 0, 0);
            VM.copySubMatrix<Function>(invG, VM.mult(RT, negp), 0, 3);
            Function[,] SHat_j = VM.constant<Function>(4, 4, Constant.newConstant(0.0));
            S = VM.constant<Function>(6, q.Length, Constant.newConstant(0.0));
            for (int i = 0; i < q.Length; i++)
            {
                SHat_j = VM.mult(invG, Function.D(G, q[i]));
                VM.copySubMatrix<Function>(S, deTwist(SHat_j), 0, i);
            }
        }

        public static Function[,] tilde3(Function[,] vector)
        {
            Function[,] m = vector;
            VM.checkSize<Function>(m, 3, 1);
            return new Function[,]{
                    {0,-m[2,0],m[1,0]},
                    {m[2,0],0,-m[0,0]},
                    {-m[1,0],m[0,0],0}};
        }

        //function to compute the adjoint mapping matrix of some vector
        public static Function[,] ad(Function[,] vector)
        {
            Function[,] m = vector;
            VM.checkSize<Function>(m, 6, 1);
            Function[,] w = VM.subArray<Function>(m, 0, 0, 3, 1);
            Function[,] v = VM.subArray<Function>(m, 3, 0, 6, 1);

            Function[,] phi = VM.constant<Function>(6, 6, Constant.newConstant(0.0));

            Function[,] wHat = tilde3(w); Function[,] vHat = tilde3(v);
            VM.copySubMatrix<Function>(phi, wHat, 0, 0);
            VM.copySubMatrix<Function>(phi, vHat, 3, 0);
            VM.copySubMatrix<Function>(phi, wHat, 3, 3);
            return phi;
        }

        public static Function[,] deTwist(Function[,] twist)
        {
            Function[,] m = twist;
            VM.checkSize<Function>(m, 4, 4);

            Function[,] aVector = VM.constant<Function>(6, 1, Constant.newConstant(0.0));
            aVector[0, 0] = m[2, 1]; aVector[1, 0] = m[0, 2]; aVector[2, 0] = m[1, 0];
            aVector[3, 0] = m[0, 3]; aVector[4, 0] = m[1, 3]; aVector[5, 0] = m[2, 3];
            return aVector;
        }

        public Function[,] force
        {
            set
            {
                fext = VM.copy<Function>(fext);
            }
            get { return VM.copy<Function>(fext); }
        }
        public Function[,] torque
        {
            set
            {
                tauext = VM.copy<Function>(value);
            }
            get { return VM.copy<Function>(tauext); }
        }



    }

    public class MultiBodySystem
    {
        public static int WORLDINDEX = -1;
        public static int TIPINDEX = -10;

        public Variable time;
        Graph<Link> systemGraph;
        Link[] systemNodes;
        public UnspecifiedFunction[] genCoords1D;//generalized coordinates stored as a 1d array
        public Variable[] force1D;//external force applied to each joint
        internal Variable[] torque1D;//external torque applied to each degree of freedom
        public Function[] c;//an array to store some important intervening variables

        public Function invDyn;

        public double[] force1Dv;
        public double[] torque1Dv;

        public double[] q;
        public double[] dq;
        public double[] cv;

        internal bool[] forDynInputVarIndices;
        internal bool[] forDynInputFunIndices;
        internal double[] forDynInputs;

        internal bool[] invDynInputVarIndices;
        internal bool[] invDynInputFunIndices;
        internal double[] invDynInputs;

        public double[] ddq;
        public double[][,] W;

        RuntimeFunction dynamicsEval;
        RuntimeFunction inverseDynamicsEval;
        RuntimeFunction kinematicEval;


        public MultiBodySystem(Graph<Link> _systemGraph, Variable t, Function[,] _explicitConstraints)
        {
            time = t;
            systemGraph = _systemGraph;
            Link[] nodes = systemGraph.allNodes;
            seeIfThere();
            for (int i = 0; i < nodes.GetLength(0); i++)
                nodes[i].computeQuantities();


            UnspecifiedFunction[][] generalizedCoordinates = new UnspecifiedFunction[nodes.GetLength(0)][];
            Variable[][] force = new Variable[nodes.GetLength(0)][];
            Variable[][] torque = new Variable[nodes.GetLength(0)][];


            for (int i = 0; i < nodes.GetLength(0); i++)
            {
                nodes[i].nodeIndex = systemGraph.nodeIndex(nodes[i]);
                nodes[i].childlist = systemGraph.childrenIndices(nodes[i].nodeIndex);
                int[] tempindex = systemGraph.parentIndices(nodes[i].nodeIndex);

                if (tempindex == null)
                    nodes[i].parent = WORLDINDEX;
                else
                    nodes[i].parent = tempindex[0];


            }


            for (int i = 0; i < nodes.GetLength(0); i++)
            {
                generalizedCoordinates[i] = new UnspecifiedFunction[nodes[i].q.GetLength(0)];
                force[i] = new Variable[3];
                torque[i] = new Variable[nodes[i].q.GetLength(0)];

                force[i][0] = (Variable)nodes[i].force[0, 0];
                force[i][1] = (Variable)nodes[i].force[1, 0];
                force[i][2] = (Variable)nodes[i].force[2, 0];

                for (int j = 0; j < nodes[i].q.GetLength(0); j++)
                {
                    generalizedCoordinates[i][j] = nodes[i].q[j];
                    torque[i][j] = (Variable)nodes[i].torque[j, 0];
                }
            }

            genCoords1D = VM.jaggedToLinear<UnspecifiedFunction>(generalizedCoordinates);
            force1D = VM.jaggedToLinear<Variable>(force);
            torque1D = VM.jaggedToLinear<Variable>(torque);

            generateTransformation();
            generateDynamics();
            generateKinematics();

            force1Dv = new double[force1D.Length];
            torque1Dv = new double[torque1D.Length];
            q = new double[genCoords1D.Length];
            dq = new double[genCoords1D.Length];

            kinematicsUpdate(q, out W);
        }

        internal Function[][] generalizedForces()
        {
            Func<Graph<Link>, Function[,], Link, Function[,]> treeProduct = null;

            treeProduct = (Graph<Link> g, Function[,] Wi, Link node) =>
            {
                Function[,] localTerms = VM.plus(linearComponent(node, Wi), rotationalComponent(node, Wi));

                Function[,] childrenSum = VM.constant<Function>(4, 4, Constant.newConstant(0.0));

                foreach (Link child in g.childrenNodes(node)) { childrenSum = VM.plus(childrenSum, VM.mult(child.A, treeProduct(g, VM.mult(Wi, child.A), child))); }
                return VM.plus(localTerms, childrenSum);
            };

            treeProduct = treeProduct.Memoize();

            Function[][] result = new Function[systemGraph.numberOfNodes][];
            Link[] root = systemGraph.roots();
            if (root.GetLength(0) != 1) { throw new Exception("can only have one root in a tree structured system"); }

            generalizedForces(systemGraph, root[0], VM.identity<Function>(4, 0.0, 1.0), result, treeProduct);
            return result;

        }

        internal void generalizedForces(Graph<Link> g, Link node, Function[,] Wiminus1, Function[][] Qi, Func<Graph<Link>, Function[,], Link, Function[,]> treeProduct)
        {
            Function[,] Wi = VM.mult(Wiminus1, node.A);
            Function[,] xi = VM.mult(Wi, node.r4);
            int forceIndex = g.nodeIndex(node);
            Qi[forceIndex] = new Function[node.q.GetLength(0)];
            node.Wi = (Function[,])Wi.Clone();


            for (int i = 0; i < node.q.Length; i++)
            {
                Function trace = VM.trace(VM.mult(Wiminus1, VM.mult(Function.D(node.A, node.q[i]), treeProduct(g, Wi, node)))); //Wi-1*dA/dq*((ri*m*xidd) + Ai+1(ri+1*mi+1*xi+1dd + Ai+2(.....)
                //Qi[forceIndex][i] = trace - node.genForces[i]; //subtract user specified generalized force so end up with an equation like this: sum(acc - force)*dxdqi - Qi = 0
                Qi[forceIndex][i] = trace; // will subtract node.genForces[i] later (in order to create inverse dynamics runtime funtion)
            }
            foreach (Link childNode in g.childrenNodes(node)) { generalizedForces(g, childNode, Wi, Qi, treeProduct); }
        }



        internal void seeIfThere()
        {
            if (systemNodes == null)
                systemNodes = systemGraph.allNodes;
        }

        internal void updateInvDynInputs(double[] qddt)
        {
            int i, idx = 0, j = 0, n = numberOfCoords, l = numberOfNodes;
            for (i = 0; i < 3 * l; i++)
            {
                if (invDynInputVarIndices[i])
                    invDynInputs[j++] = force1Dv[i];
            }
            idx = 0;
            for (i = 0; i < n; i++)
            {
                if (invDynInputFunIndices[i])
                    invDynInputs[j++] = q[i];
            }
            idx += n;
            for (i = idx; i < idx + n; i++)
            {
                if (invDynInputFunIndices[i])
                    invDynInputs[j++] = dq[i - idx];
            }
            idx += n;
            for (i = idx; i < idx + n; i++)
            {
                if (invDynInputFunIndices[i])
                    invDynInputs[j++] = qddt[i - idx];
            }

        }


        internal void updateForDynInputs()
        {
            int i, idx = 0, j = 0, n = numberOfCoords, l = numberOfNodes;
            for (i = idx; i < idx + 3 * l; i++)
            {
                if (forDynInputVarIndices[i])
                    forDynInputs[j++] = force1Dv[i - idx];
            }
            idx += 3 * l;
            for (i = idx; i < idx + torque1D.Length; i++)
            {
                if (forDynInputVarIndices[i])
                    forDynInputs[j++] = torque1Dv[i - idx];
            }
            idx = 0;
            for (i = 0; i < n; i++)
            {
                if (forDynInputFunIndices[i])
                    forDynInputs[j++] = q[i];
            }
            idx += n;
            for (i = idx; i < idx + n; i++)
            {
                if (forDynInputFunIndices[i])
                    forDynInputs[j++] = dq[i - idx];
            }            

        }
        internal void generateDynamics()
        {
            createDynamicsFunctions();
        }

        internal void generateTransformation()
        {

            for (int i = 0; i < systemNodes.Length; i++)
            {
                int parent = systemNodes[i].parent;
                if (parent == WORLDINDEX)
                {
                    systemNodes[i].Wi = systemNodes[i].A;
                }
                else
                {
                    systemNodes[i].Wi = VM.mult(systemNodes[parent].Wi, systemNodes[i].A);
                }

            }
        }
        internal void generateKinematics()
        {

            Function kinematics = new Function(VM.matricesToVector<Function>(transformFunction()));
            UnspecifiedFunction[] inputVariables = VM.concatenate<UnspecifiedFunction>(genCoords1D);
            kinematics.orderVariablesInDomain(inputVariables);
            kinematicEval = kinematics.compile();
        }

        internal void createDynamicsFunctions()
        {
            Function qdt = Function.derivative(Function.D(genCoords1D, time));

            Function[][,] qdot = new Function[systemNodes.Length][,];

            int tempindex=0;
            for (int i = 0; i < systemNodes.Length; i++)
            {
                qdot[i]=VM.constant<Function>(systemNodes[i].q.Length,1,Constant.newConstant(0.0));

                for(int j=0;j<systemNodes[i].q.Length;j++)                
                    qdot[i][j,0]=qdt[tempindex+j];

                tempindex +=systemNodes[i].q.Length;                   
            }
             

            int num_node = systemNodes.Length;
            int num_dof = genCoords1D.Length;

            Function[][,] a = new Function[systemNodes.Length][,];
            Function[][,] b = new Function[systemNodes.Length][,];
            Function[,] temp1 = VM.constant<Function>(6, 1, Constant.newConstant(0.0));
            

            int parent_i = 0;

            for (int i = 0; i < systemNodes.Length; i++)
            {
                systemNodes[i].spatialVel = VM.constant<Function>(6, 1, Constant.newConstant(0.0));
                parent_i = systemNodes[i].parent;
                if (parent_i != WORLDINDEX)
                {
                    systemNodes[i].spatialVel = VM.mult(systemNodes[i].Ad, systemNodes[parent_i].spatialVel);
                }
                temp1 = VM.mult(systemNodes[i].S, qdot[i]);
                a[i] = VM.mult(VM.mult(Link.ad(temp1), systemNodes[i].spatialVel), -1);

                Function[,] Sdot = VM.constant<Function>(systemNodes[i].S.GetLength(0),systemNodes[i].S.GetLength(1), Constant.newConstant(0.0));
                for (int j = 0; j < systemNodes[i].q.Length; j++)
                {
                    Sdot = VM.plus(Sdot, VM.mult(Function.D(systemNodes[i].S, systemNodes[i].q[j]), qdot[i][j,0]));
                }
                a[i]=VM.plus(a[i],VM.mult(Sdot,qdot[i]));

                systemNodes[i].spatialVel = VM.plus(systemNodes[i].spatialVel, temp1);
                b[i] = VM.mult(VM.mult(VM.transpose<Function>(Link.ad(systemNodes[i].spatialVel)), VM.mult(systemNodes[i].J, systemNodes[i].spatialVel)), -1);

                Function[,] fi = VM.constant<Function>(3, 1, Constant.newConstant(0.0));
                fi[0, 0] = force1D[i * 3]; fi[1, 0] = force1D[i * 3 + 1]; fi[2, 0] = force1D[i * 3 + 2];

                fi = VM.mult(VM.transpose<Function>(VM.subArray<Function>(systemNodes[i].Wi, 0, 0, 3, 3)), fi);

                Function[,] mi = VM.constant<Function>(3, 1, Constant.newConstant(0.0));
                mi=VM.mult(Link.tilde3(systemNodes[i].r),fi);
                Function[,] Fe = VM.constant<Function>(6, 1, Constant.newConstant(0.0));
                VM.copySubMatrix<Function>(Fe, mi, 0, 0);
                VM.copySubMatrix<Function>(Fe, fi, 3, 0);
                b[i] = VM.minus(b[i], Fe);
            }

            Function[][,] alpha = new Function[systemNodes.Length][,];
            Function[][,] alpha_score = new Function[systemNodes.Length][,];

            for (int i = 0; i < systemNodes.Length; i++)
            {
                parent_i = systemNodes[i].parent;
                alpha[i] = a[i];
                alpha_score[i] = b[i];
                if (parent_i != WORLDINDEX)
                {
                    alpha[i] = VM.plus(VM.mult(systemNodes[i].Ad, alpha[parent_i]), alpha[i]);
                }

                alpha_score[i] = VM.plus(VM.mult(systemNodes[i].J, alpha[i]), alpha_score[i]);
            }

            Function[][,] P = new Function[systemNodes.Length][,];
            Function[][,] tau_hat = new Function[systemNodes.Length][,];

            for (int i = systemNodes.Length - 1; i >= 0; i--)
            {
                P[i] = alpha_score[i];
                for (int j = 0; j < systemNodes[i].childlist.Length; j++)
                {
                    int childindex=systemNodes[i].childlist[j];
                    P[i] = VM.plus(P[i], VM.mult(VM.transpose<Function>(systemNodes[childindex].Ad), P[childindex]));
                }
                tau_hat[i] = VM.minus(systemNodes[i].tauext, VM.mult(VM.transpose<Function>(systemNodes[i].S), P[i]));
            }

            Function[] c = new Function[genCoords1D.Length];
            tempindex = 0;
            for (int i = 0; i < systemNodes.Length; i++)
            {
                for (int j = 0; j < systemNodes[i].q.Length; j++)
                    c[tempindex + j] = tau_hat[i][j, 0];

                tempindex += systemNodes[i].q.Length;
            }

//          Function.printCompilerSource = true;

            dynamicsEval = pack(c).compile();

            //Inverse Dynamics
            Function qddt = Function.derivative(Function.D(genCoords1D, time, time));
            Function[][] forces = generalizedForces();
            Function[] Q = VM.jaggedToLinear<Function>(forces);


            Function derivQ = Function.derivative(Q);

            UnspecifiedFunction[] tempqdot = new UnspecifiedFunction[qdt.rangeDimension];
            for (int i = 0; i < qdt.rangeDimension; i++) { tempqdot[i] = (UnspecifiedFunction)qdt[i]; }
            UnspecifiedFunction[] tempqddot = new UnspecifiedFunction[qddt.rangeDimension];
            for (int i = 0; i < qddt.rangeDimension; i++) { tempqddot[i] = (UnspecifiedFunction)qddt[i]; }
            UnspecifiedFunction[] inputStates = VM.concatenate<UnspecifiedFunction>(genCoords1D, tempqdot, tempqddot);

            invDyn = derivQ;

            Variable[] variables = force1D;//VM.concatenate<Variable>(force1D, torque1D);
            variables = initializeInputVariables(invDyn, variables, out invDynInputVarIndices);
            inputStates = initializedInputUnspecifiedFunctions(invDyn, inputStates, out invDynInputFunIndices);

            invDynInputs = new double[variables.Length + inputStates.Length];
            invDyn.orderVariablesInDomain(variables, inputStates);
            inverseDynamicsEval = invDyn.compile();

        }

        internal Function pack(Function[] b)
        {
            Function[][,] AdMatrix = new Function[systemNodes.Length][,];
            Function[][,] JMatrix = new Function[systemNodes.Length][,];

            for (int i = 0; i < systemNodes.Length; i++)
            {
                AdMatrix[i] = systemNodes[i].Ad;
                JMatrix[i] = systemNodes[i].J;
            }
            Function[] SVector = VM.matrixToVector<Function>(systemNodes[0].S);
            for (int i = 1; i < systemNodes.Length; i++)
            {
                SVector=VM.concatenate<Function>(SVector,VM.matrixToVector<Function>(systemNodes[i].S));
            }



            Function[] packed = VM.concatenate<Function>(b, VM.matricesToVector<Function>(AdMatrix),VM.matricesToVector<Function>(JMatrix),SVector,VM.matricesToVector<Function>(transformFunction()));
            Function temp = Function.derivative(packed);

            //you have to do initialize variables after the derivative
            Function qdot = Function.derivative(Function.D(genCoords1D, time));
            UnspecifiedFunction[] tempqdot = new UnspecifiedFunction[qdot.rangeDimension];
            for (int i = 0; i < qdot.rangeDimension; i++)
                tempqdot[i] = (UnspecifiedFunction)qdot[i];
            UnspecifiedFunction[] inputStates = VM.concatenate<UnspecifiedFunction>(genCoords1D, tempqdot);
            inputStates = initializedInputUnspecifiedFunctions(temp, inputStates, out forDynInputFunIndices);

            Variable[] variables = VM.concatenate<Variable>(force1D, torque1D);
            variables = initializeInputVariables(temp, variables, out forDynInputVarIndices);

            forDynInputs = new double[variables.Length + inputStates.Length];
            temp.orderVariablesInDomain(variables, inputStates);

            return temp;
        }


        public Link[] allNodes
        {
            get
            {
                seeIfThere();
                return systemNodes;
            }
        }

        public int numberOfNodes
        {
            get
            {
                seeIfThere();
                return systemNodes.Length;
            }
        }

        public int numberOfCoords
        {
            get
            {
                return genCoords1D.GetLength(0);
            }
        }

        /// <summary>
        /// set initial state
        /// </summary>
        /// <param name="qinit">initial value of the generalized coordinates (e.g., inital joint angles) </param>
        /// <param name="dqinit">initial value of the time derivative of generalized coordinates (e.g., inital joint velocity) </param>
        public void setInitialState(double[] qinit, double[] dqinit)
        {
            setGenCoords(qinit);
            setGenVelocity(dqinit);
        }

        /// <summary>
        /// set generalized coordinates to new values
        /// </summary>
        /// <param name="qnew"></param>
        public void setGenCoords(double[] qnew)
        {
            if (qnew.Length != numberOfCoords) { throw new Exception(); }
            for (int i = 0; i < qnew.Length; i++)
            {
                q[i] = qnew[i];
            }
        }

        /// <summary>
        /// set time derivative of generalized coordinates to new values
        /// </summary>
        /// <param name="dqnew"></param>
        public void setGenVelocity(double[] dqnew)
        {
            if (dqnew.Length != numberOfCoords) { throw new Exception(); }
            for (int i = 0; i < dqnew.Length; i++)
            {
                dq[i] = dqnew[i];
            }
        }

        public void setZeroAllForces()
        {
            int i;
            for (i = 0; i < force1Dv.Length; i++)
                force1Dv[i] = 0;

            for (i = 0; i < torque1Dv.Length; i++)
                torque1Dv[i] = 0;
        }

        public void setExternalForce(int idx, double[] f)
        {
            if (idx < 0 || idx >= numberOfNodes) { throw new Exception("idx must be in range of [0,numberOfNodes-1]"); }
            force1Dv[3 * idx] = f[0];
            force1Dv[3 * idx + 1] = f[1];
            force1Dv[3 * idx + 2] = f[2];
        }


        public void addExternalForce(int idx, double[] f)
        {
            if (idx < 0 || idx >= numberOfNodes)
            { throw new Exception("idx must be in range of [0,numberOfNodes-1]"); }
            force1Dv[3 * idx] += f[0];
            force1Dv[3 * idx + 1] += f[1];
            force1Dv[3 * idx + 2] += f[2];
        }


        /// <summary>
        /// set generalized forces
        /// </summary>
        /// <param name="idx">index in generalized force vector</param>
        /// <param name="Q">amount of force to apply</param>
        public void setGenForce(int i, double Q)
        {
            if (i < 0 || i >= numberOfCoords) { throw new Exception("i must be in range of [0,numberOfCoords-1]"); }
            torque1Dv[i] = Q;
        }

        public void setGenForce(double[] Q)
        {
            if (numberOfCoords != Q.Length) { throw new Exception(); }
            for (int i = 0; i < numberOfCoords; ++i) torque1Dv[i] = Q[i];
        }

        /// <summary>
        /// add to generalized forces
        /// </summary>
        /// <param name="idx">index in generalized force vector</param>
        /// <param name="Q">amount of force to apply</param>
        public void addGenForce(int i, double Q)
        {
            if (i < 0 || i >= numberOfCoords) { throw new Exception("i must be in range of [0,numberOfCoords-1]"); }
            torque1Dv[i] += Q;
        }

        public void addGenForce(double[] Q)
        {
            if (numberOfCoords != Q.Length) { throw new Exception(); }
            for (int i = 0; i < numberOfCoords; ++i) torque1Dv[i] += Q[i];
        }




        public void kinematicsUpdate(double[] q, out double[][,] W)
        {
            double[] evalResult = new double[kinematicEval.rangeDimension];
            kinematicEval.eval(evalResult, q);
            W = new double[numberOfNodes][,];
            W = VM.vectorToMatrix<double>(evalResult, 4, 4, numberOfNodes, 0);
        }

        public void inverseDynamics(double[] qddt, out double[] genForce)
        {
            updateInvDynInputs(qddt);
            genForce = new double[torque1Dv.Length];
            inverseDynamicsEval.eval(genForce, invDynInputs);
        }


        public void forwardDynamics(out double[] qddt, out double[][,] W)
        {
            double[][,]AdMatrix;
            double[][,]JMatrix;
            double[][,]SMatrix;
            evalDynamicsMatrices(out AdMatrix, out JMatrix, out SMatrix, out W);

            double[][,] tau_hat=new double[systemNodes.Length][,];
            int tempindex=0;
            for(int i=0;i<systemNodes.Length;i++)
            {
                tau_hat[i]=new double[systemNodes[i].q.Length,1];
                for(int j=0;j<systemNodes[i].q.Length;j++)
                    tau_hat[i][j,0]=cv[tempindex+j];

                tempindex+=systemNodes[i].q.Length;
            }

            double[][,] J_hat = new double[systemNodes.Length][,];
            double[][,] Omega = new double[systemNodes.Length][,];
            double[][,] OmegaInverse = new double[systemNodes.Length][,];
            double[,] temp3 = VM.constant<double>(6, 6, 0.0);

            for (int i = systemNodes.Length - 1; i >= 0; i--)
            {
                J_hat[i] = JMatrix[i];
                for (int j = 0; j < systemNodes[i].childlist.Length; j++)
                {
                    int childindex = systemNodes[i].childlist[j];
                    temp3 = VM.mult(VM.transpose(AdMatrix[childindex]), VM.mult(J_hat[childindex], AdMatrix[childindex]));
                    double[,] temp4 = VM.mult(VM.transpose(AdMatrix[childindex]), VM.mult(J_hat[childindex], SMatrix[childindex]));
                    temp4 = VM.mult(temp4, VM.mult(OmegaInverse[childindex], VM.transpose(temp4)));

                    temp3 = VM.minus(temp3, temp4);
                    J_hat[i] = VM.plus(J_hat[i], temp3);
                }

                Omega[i] = VM.mult(VM.transpose(SMatrix[i]), VM.mult(J_hat[i], SMatrix[i]));

                double[,] U ;
                double[,] V ;
                double[] Sigma;

                VM.JacobiSVD(Omega[i], out U, out Sigma,out V);
                for (int j = 0; j < Sigma.Length; j++)
                {
                    if (Sigma[0]/Sigma[j] > 10000)
                    {
                        Sigma[j] = 0;
                    }
                    else
                        Sigma[j] = 1.0 / Sigma[j];
                        
                }
                
                OmegaInverse[i] = VM.mult(V, VM.mult(VM.diagonal(Sigma,V.GetLength(0),U.GetLength(0)), VM.transpose(U)));


 
            }

            double[][,] Y = new double[systemNodes.Length][,];
            double[][,] PI = new double[systemNodes.Length][,];

            for (int i = 0; i < systemNodes.Length; i++)
            {
                PI[i] = VM.mult(VM.mult(VM.transpose(AdMatrix[i]), J_hat[i]), VM.mult(SMatrix[i], OmegaInverse[i]));
                Y[i] = VM.mult(PI[i], VM.transpose(SMatrix[i]));
                Y[i] = VM.minus(VM.transpose(AdMatrix[i]), Y[i]);
            }

            double[][,] z_hat = new double[systemNodes.Length][,];
            double[][,] c_hat = new double[systemNodes.Length][,];

            double[,] temp1 = VM.constant<double>(6, 1, 0.0);
            for (int i = systemNodes.Length - 1; i >= 0; i--)
            {
                z_hat[i] = VM.constant<double>(6, 1, 0);
                for (int j = 0; j < systemNodes[i].childlist.Length; j++)
                {
                    int childindex = systemNodes[i].childlist[j];
                    temp1 = VM.plus(VM.mult(Y[childindex], z_hat[childindex]), VM.mult(PI[childindex], tau_hat[childindex]));
                    z_hat[i] = VM.plus(z_hat[i], temp1);                    
                }

                c_hat[i] = VM.minus(tau_hat[i], VM.mult(VM.transpose(SMatrix[i]), z_hat[i]));
                c_hat[i] = VM.mult(OmegaInverse[i], c_hat[i]);
            }

            double[][,] qddot = new double[systemNodes.Length][,];
            double[][,] eita = new double[systemNodes.Length][,];
            int parent_i = 0;
            for (int i = 0; i < systemNodes.Length; i++)
            {
                parent_i = systemNodes[i].parent;
                eita[i] = VM.mult(SMatrix[i], c_hat[i]);
                qddot[i] = c_hat[i];
                if (parent_i != WORLDINDEX)
                {
                    eita[i] = VM.plus(VM.mult(VM.transpose(Y[i]), eita[parent_i]), eita[i]);
                    qddot[i] = VM.minus(qddot[i], VM.mult(VM.transpose(PI[i]), eita[parent_i]));

                }
            }
            tempindex = 0;
            qddt = new double[genCoords1D.Length];
            for (int i = 0; i < systemNodes.Length; i++)
            {
                for (int j = 0; j < systemNodes[i].q.Length; j++)
                    qddt[tempindex + j] = qddot[i][j, 0];

                tempindex += systemNodes[i].q.Length;
            }
        }


        public void evalDynamicsMatrices(out double[][,] AdMatrix, out double[][,] JMatrix, out double[][,] SMatrix, out double[][,] W)
        {
            updateForDynInputs();
            double[] evalResult = new double[dynamicsEval.rangeDimension];
            dynamicsEval.eval(evalResult, forDynInputs);
            int n = numberOfCoords;    
            int tempindex=0;
            cv = VM.subArray<double>(evalResult, 0, n);
            tempindex +=n;
            AdMatrix = VM.vectorToMatrix<double>(evalResult, 6, 6, numberOfNodes, tempindex);
            tempindex += 6 * 6 * numberOfNodes;
            JMatrix = VM.vectorToMatrix<double>(evalResult, 6, 6, numberOfNodes, tempindex);
            tempindex += 6 * 6 * numberOfNodes;
            SMatrix=new double[numberOfNodes][,];
            for (int i = 0; i < numberOfNodes; i++)
            {
                SMatrix[i] = VM.vectorToMatrix<double>(systemNodes[i].q.Length, 6, tempindex,evalResult);
                tempindex += 6 * systemNodes[i].q.Length;
            }
            W = VM.vectorToMatrix<double>(evalResult, 4, 4, numberOfNodes, tempindex);
        }



        public void nextStepSymplecticEulerMethod(double h)
        {
//            Stopwatch s = new Stopwatch();
//            s.Start();
            forwardDynamics(out ddq, out W);
//            s.Stop();
//            Console.WriteLine(s.ElapsedMilliseconds);
            for (int i = 0; i < dq.GetLength(0); i++)
            {
                dq[i] += h * ddq[i];
                q[i] += h * dq[i];
            }
        }

        public void nextStepForwardEulerMethod(double h)
        {
            forwardDynamics(out ddq, out W);
            for (int i = 0; i < dq.GetLength(0); i++)
            {
                q[i] += h * dq[i];
                dq[i] += h * ddq[i];
            }
        }
        public void nextStepRK4(double h)
        {
            double h2 = h * 0.5;
            double h3 = h / 3.0;
            double h6 = h / 6.0;

            double[] q0 = new double[numberOfCoords];
            double[] dq0 = new double[numberOfCoords];
            q.CopyTo(q0, 0);
            dq.CopyTo(dq0, 0);

            double[] qu = new double[numberOfCoords];
            double[] dqu = new double[numberOfCoords];
            q.CopyTo(qu, 0);
            dq.CopyTo(dqu, 0);

            forwardDynamics(out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h6));
            dqu = VM.plus(dqu, VM.mult(ddq, h6));

            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            forwardDynamics(out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            forwardDynamics(out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            q = VM.plus(q0, VM.mult(dq, h));
            dq = VM.plus(dq0, VM.mult(ddq, h));
            forwardDynamics(out ddq, out W);
            q = VM.plus(qu, VM.mult(dq, h6));
            dq = VM.plus(dqu, VM.mult(ddq, h6));
        }

        public Function[][,] transformFunction()
        {
            Link[] nodes = systemGraph.allNodes;
            Function[][,] result = new Function[nodes.Length][,];

            for (int i = 0; i < nodes.Length; i++) { result[i] = nodes[i].Wi; }
            return result;
        }

        public static Variable[] initializeInputVariables(Function func, Variable[] testInput, out bool[] ivi)
        {
            int i, j = 0;
            Variable[] tDom = func.trueDomain();
            ivi = new bool[testInput.GetLength(0)];
            Variable[] inputVar = new Variable[tDom.GetLength(0)];
            for (i = 0; i < testInput.GetLength(0); i++)
            {
                if (tDom.Contains<Variable>(testInput[i]))
                {
                    ivi[i] = true;
                    inputVar[j++] = testInput[i];
                }
                else
                    ivi[i] = false;
            }
            return inputVar;
        }

        public static UnspecifiedFunction[] initializedInputUnspecifiedFunctions(Function func, UnspecifiedFunction[] testInput, out bool[] iui)
        {
            int i, j = 0;
            UnspecifiedFunction[] dom = func.unspecifiedFunctionVariables;
            iui = new bool[testInput.GetLength(0)];
            UnspecifiedFunction[] inputFunc = new UnspecifiedFunction[dom.GetLength(0)];
            for (i = 0; i < testInput.GetLength(0); i++)
            {
                if (dom.Contains<UnspecifiedFunction>(testInput[i]))
                {
                    iui[i] = true;
                    inputFunc[j++] = testInput[i];
                }
                else
                    iui[i] = false;
            }
            return inputFunc;
        }


        public static Function[,] rotationPart(Function[,] mat)
        {
            Function[,] result = VM.constant<Function>(4, 4, 0);
            for (int i = 0; i < 3; i++)
            {
                for (int j = 0; j < 3; j++)
                {
                    result[i, j] = mat[i, j];
                }
            }
            return result;
        }

        /// <summary>
        /// return only the translation part of a tranformation matrix
        /// </summary>
        /// <param name="mat"></param>
        /// <returns></returns>
        public static Function[] translationPart(Function[,] mat)
        {
            return new Function[] { mat[0, 3], mat[1, 3], mat[2, 3] };
        }

        public static Function[,] tilde(Function[,] vectorOrMatrix)
        {
            Function[,] m = vectorOrMatrix;
            if (vectorOrMatrix.GetLength(1) == 1)
            {//input is vector so return 3x3 antisymmetric matrix embedded in 4x4 matrix
                VM.checkSize<Function>(m, 4, 1);
                return new Function[,]{
                    {0,-m[2,0], m[1,0], 0},
                    {m[2,0],0,-m[0,0], 0},
                    {-m[1,0],m[0,0],0,0},
                    {0,0,0,0}};
            }
            else
            {
                VM.checkSize<Function>(m, 4, 4);
                return new Function[,] { 
                  {m[2, 1]}, 
                  {m[0,2]}, 
                  {m[1, 0]},
                  {0}};
            } //{x,y,z} embedded in 4 vector
        }

        internal static Function[,] rotationalComponent(Link curr, Function[,] Wi)
        {
            //need to compute tilde(QT (Iwd - tau))
            Function[,] Q = rotationPart(Wi), QT = VM.transpose(Q);
            Function[,] omega = tilde(VM.mult(VM.transpose(Q), Function.D(Q, curr.t)));
            Function[,] I4x4 = VM.constant<Function>(4, 4, 0);
            VM.copySubMatrix<Function>(I4x4, curr.I, 0, 0);
            Function[,] Hdt = Function.D(VM.mult(Q, VM.mult(I4x4, omega)), curr.t); //d(angular momentum)/dt in world coordinates
            Function[,] tauext4 = VM.constant<Function>(4, 1, 0);
            //            VM.copySubMatrix<Function>(tauext4, curr.tauext, 0, 0);
            return VM.mult(VM.mult(QT, tilde(VM.minus(Hdt, tauext4))), -.5);
        }

        internal static Function[,] linearComponent(Link curr, Function[,] Wi)
        {
            //need to compute r(mxdd - f)T
            Function[,] x = VM.mult(Wi, curr.r4);
            Function[,] xdd = Function.D(x, curr.t, curr.t);
            Function[,] force4 = VM.constant<Function>(4, 1, 0);
            VM.copySubMatrix<Function>(force4, curr.force, 0, 0);
            return VM.mult(curr.r4, VM.transpose<Function>(VM.minus(VM.mult(xdd, curr.m), force4)));
        }


        public static Function[,] inertiaMatrixBox(double mass, double lenX, double lenY, double lenZ)
        {
            return new Function[,] {{1/12.0*mass*((lenY*lenY)+(lenZ*lenZ)),0,0,0},
                {0,1/12.0*mass*((lenX*lenX)+(lenZ*lenZ)),0,0},
                {0,0,1/12.0*mass*((lenX*lenX)+(lenY*lenY)),0},
                {0,0,0,1}};
        }

        /// <summary>
        /// returns the inertia matrix of a cylinder (Y is axis)
        /// </summary>
        public static Function[,] inertiaMatrixCylinder(double mass, double radius, double height)
        {
            // moment of inertia:
            return new Function[,]{{1/12.0*mass*(3*(radius*radius) + (height*height)), 0 ,0,0},
                         {0, .5*mass*(radius*radius), 0,0},
                         {0, 0, 1/12.0*mass*(3*(radius*radius) + (height*height)),0},
                         {0,0,0,1}};
        }

        /// <summary>
        /// returns the inertia matrix of a torus
        /// </summary>
        public static Function[,] inertiaMatrixTorus(double mass, double largeRadius, double smallRadius)
        {
            return new Function[,]{{1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0 ,0,0},
                         {0, .25*mass*(4*largeRadius*largeRadius + 3*smallRadius*smallRadius), 0,0},
                         {0, 0, 1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius),0},
                         {0,0,0,1}
                            };
        }

        /// <summary>
        /// returns the inertia matrix of a sphere
        /// </summary>
        public static Function[,] inertiaMatrixSphere(double mass, double radius)
        {
            return new Function[,]{ {0.4*mass*radius*radius, 0 ,0,0},
                                    {0, 0.4*mass*radius*radius, 0,0},
                                    {0, 0, 0.4*mass*radius*radius,0},
                                    {0,0,0,1}};
        }

    }

}



