//Written by Leonid Velikovich and Benjamin Hernandez Arreguin

//+-----------------------------------------------------------------------------
//
//  Class BSplineCurve2DRealization.  Implements glue and UI code for editing
//  and displaying a BSplineCurve2D on the screen.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Diagnostics;
using System.Drawing;
//using System.Drawing.Drawing2D;  Avoids a namespace collision with XNA
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;

namespace ParametricModelEditor
{
    public class BSplineCurve2DRealization : BSplineCurve2D
    {
        public static bool curveInvalidated = false;

        public BSplineCurve2DRealization()
        {
        }

        public ISplineControlPointHost Parent
        {
            set { _parent = value; }
        }

        public override ParametricCurve2D GetSerialization()
        {
            BSplineCurve2D serialization = new BSplineCurve2D();
            serialization.IsClosedCurve = IsClosedCurve;
            serialization.ControlPointCount = ControlPointCount;
            serialization.ControlPoints = ControlPoints;
            return serialization;
        }

        protected override ControlPoint[] GetControlPointSerialization()
        {
            ControlPoint[] controlPoints = new ControlPoint[_controlObjectList.Count];

            for (int n = 0; n < _controlObjectList.Count; n++)
            {
                controlPoints[n].x = _controlObjectList[n].Center.X - _origin.X;
                controlPoints[n].y = _controlObjectList[n].Center.Y - _origin.Y;
                controlPoints[n].weight = _controlObjectList[n].Weight;
            }

            return controlPoints;
        }

        public void Draw(Graphics g)
        {
            Pen pen = new Pen(_curveColor);
            //pen.Width = 1.5f;

            for (int n = 0; n < _samplingPointList.Count - 1; n++)
            {
                float x1 = _samplingPointList[n].X;
                float y1 = _samplingPointList[n].Y;
                float x2 = _samplingPointList[n+1].X;
                float y2 = _samplingPointList[n+1].Y;

                g.DrawLine(pen, x1, y1, x2, y2);
            }

            pen.Dispose();

            /*pen = new Pen(System.Drawing.Color.Linen);

            Vector2 a, b;
            Vector2 normal;  // unused
            Evaluate(0, out b, out normal);
            for (int n = 1; n <= 100; n++)
            {
                float t = .01f * ((float)n);

                a = b;
                Evaluate(t, out b, out normal);

                g.DrawLine(pen, a.X, a.Y, b.X, b.Y);
            }

            pen.Dispose();*/
        }

        public void OnControlPointAdded(SplineControlPointControl control)
        {
            Debug.Assert(!_controlObjectList.Contains(control));

            _controlObjectList.Add(control);
            _controlPointCount += control.Weight;

            OnCurveInvalidated();
        }

        public void OnControlPointInserted(SplineControlPointControl previousControl, SplineControlPointControl newControl)
        {
            Debug.Assert(_controlObjectList.Contains(previousControl));
            Debug.Assert(!_controlObjectList.Contains(newControl));

            int index = _controlObjectList.IndexOf(previousControl);
            _controlObjectList.Insert(index + 1, newControl);

            _controlPointCount += newControl.Weight;

            OnCurveInvalidated();
        }

        public void OnControlPointRemoved(SplineControlPointControl control)
        {
            Debug.Assert(_controlObjectList.Contains(control));

            _controlObjectList.Remove(control);
            _controlPointCount -= control.Weight;

            OnCurveInvalidated();
        }

        public void OnControlPointWeightIncremented(SplineControlPointControl control)
        {
            Debug.Assert(_controlObjectList.Contains(control));

            _controlPointCount++;

            OnCurveInvalidated();
        }

        public void OnControlPointWeightDecremented(SplineControlPointControl control)
        {
            Debug.Assert(_controlObjectList.Contains(control));

            _controlPointCount--;

            OnCurveInvalidated();
        }

        public void OnControlPointMoved(SplineControlPointControl control)
        {
            Debug.Assert(_controlObjectList.Contains(control));

            OnCurveInvalidated();
        }

        public void OnControlPointsCleared()
        {
            _controlObjectList.Clear();
            _controlPointCount = 0;

            _samplingPointList.Clear();
        }


        protected override void OnCurveInvalidated()
        {
            if (_controlPointCount >= 4)
            {
                ComputeControlPointList();
                ComputeSamplingPointList();
            }

            curveInvalidated = true;

            _parent.InvalidateIfNotBusy();

            base.OnCurveInvalidated();
        }

        // Turn the list of Point controls into a list of point Vectors.
        // This means we take Point controls with weight > 1 and expand
        // into a corresponding number of same-valued point Vectors
        // (repeating control points).
        private void ComputeControlPointList()
        {
            Debug.Assert(_controlPointCount >= 4);

            _controlPointList.Clear();

            int currentObjectIndex = 0;
            int currentWeightIndex = _controlObjectList[0].Weight + 1;

            // For each control point, figure out if we are repeating the same
            // source point (when its weight > 1), or moving to the next one.
            for (int n = 0; n < _controlPointCount; n++)
            {
                if (currentWeightIndex > 1)
                {
                    currentWeightIndex--;
                }
                else
                {
                    currentObjectIndex++;
                    currentWeightIndex = _controlObjectList[currentObjectIndex].Weight;
                }

                _controlPointList.Add(_controlObjectList[currentObjectIndex].Center - _origin);
            }

            Debug.Assert(_controlPointCount == _controlPointList.Count);
        }


        private void ComputeSamplingPointList()
        {
            _samplingPointList.Clear();

            Debug.Assert(_controlPointCount >= 4);

            Vector2 c0, c1, c2, c3;

            c1 = _controlPointList[0];
            c2 = _controlPointList[1];
            c3 = _controlPointList[2];

            // For a closed curve, we act like we append the first 3 points at the end of the list
            int appliedControlPointCount = _controlPointCount + (_isClosedCurve ? 3 : 0);
            for (int q = 3; q < appliedControlPointCount; q++)  // For each set of 4 control points...
            {
                c0 = c1;
                c1 = c2;
                c2 = c3;
                c3 = _controlPointList[q % _controlPointCount];

                Matrix controlMatrix = new Matrix(c0.X, c0.Y, 0, 0,
                                                  c1.X, c1.Y, 0, 0,
                                                  c2.X, c2.Y, 0, 0,
                                                  c3.X, c3.Y, 0, 0);
                Matrix combinedMatrix = c_baseMatrix * controlMatrix;

                // Use 10 sampling values of T over each [0,1) interval
                for (int n = 0; n < 10; n++)
                {
                    float t = 0.1f * ((float)n);

                    Vector4 tVector = new Vector4(t * t * t, t * t, t, 1);

                    Vector4 output = Vector4.Transform(tVector, combinedMatrix);

                    _samplingPointList.Add(_origin + new Vector2(output.X, output.Y));
                }

                // At the end of the last interval, add one last sampling point;
                // Similar as above code, but substituting 1 for t
                if (q + 1 == appliedControlPointCount)
                {
                    float outputX = combinedMatrix.M11 + combinedMatrix.M21 + combinedMatrix.M31 + combinedMatrix.M41;
                    float outputY = combinedMatrix.M12 + combinedMatrix.M22 + combinedMatrix.M32 + combinedMatrix.M42;
                    _samplingPointList.Add(_origin + new Vector2(outputX, outputY));
                }
            }
        }

        public void OnOriginInvalidated(Vector2 newOrigin)
        {
            _origin = newOrigin;

            OnCurveInvalidated();
        }


        private ISplineControlPointHost _parent;
        protected Vector2 _origin = Vector2.Zero;
        private Color _curveColor = System.Drawing.Color.LightGreen;
        
        protected List<SplineControlPointControl> _controlObjectList = new List<SplineControlPointControl>();
        protected List<Vector2> _samplingPointList = new List<Vector2>();

    }
}
