//+-----------------------------------------------------------------------------
//
//  Class BitmapDataViewer.  Visualizes array information as an image control.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Diagnostics;
using System.Drawing;
//using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;

using Float = ParametricModelEditor.FloatingPointMathHelper;



namespace ParametricModelEditor
{
    [Description("Bitmap Data Viewer")]
    public class BitmapDataViewer : System.Windows.Forms.PictureBox
    {
        public BitmapDataViewer()
        {
            this.OnSizeChanged(null);
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            _bitmap = new Bitmap(this.Width, this.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
            _analyzer = new ParametricSurfaceAnalyzer(_bitmap.Width, _bitmap.Height);
        }


        unsafe public void FillWithNormalData(ParametricSurfaceEffect effect)
        {
            if (effect == null) return;

            _analyzer.FillWithNormalData(effect);

            BitmapData bitmapData = _bitmap.LockBits(
                                        new System.Drawing.Rectangle(0, 0, _bitmap.Width, _bitmap.Height),
                                        ImageLockMode.ReadWrite,
                                        PixelFormat.Format32bppArgb);
            
            Int32* ptr = (Int32*)bitmapData.Scan0.ToPointer();

            Vector4[] data = _analyzer.DataMatrix;
            int index = 0;

            for (int j = 0; j < bitmapData.Height; j++)
            {
                for (int i = 0; i < bitmapData.Width; i++)
                {
                    Vector4 normal = data[index++];

                    *(ptr++) = Vector4ToInt32(new Vector4(.5f)
                        + .5f * new Vector4(1.0f, normal.X, normal.Y, normal.Z));
                }
            }

            _bitmap.UnlockBits(bitmapData);
            _bitmap.RotateFlip(RotateFlipType.Rotate270FlipNone);

            this.Image = _bitmap;
        }

        unsafe public void FillWithCurvatureData(ParametricSurfaceEffect effect)
        {
            if (effect == null)  return;

            _analyzer.FillWithCurvatureData(effect);

            BitmapData bitmapData = _bitmap.LockBits(
                                        new System.Drawing.Rectangle(0, 0, _bitmap.Width, _bitmap.Height),
                                        ImageLockMode.ReadWrite,
                                        PixelFormat.Format32bppArgb);
            

            Int32* ptr = (Int32*)bitmapData.Scan0.ToPointer();
            
            Vector4[] data = _analyzer.DataMatrix;
            Vector4 maxValue = _analyzer.DataMatrixMaxValue;
            int index = 0;

            for (int j = 0; j < bitmapData.Height; j++)
            {
                for (int i = 0; i < bitmapData.Width; i++)
                {
                    // Normalize the data before drawing to bitmap
                    Vector4 current = data[index++];

                    float scaledCurvature_du = Float.Pow(current.X / maxValue.X, .3f);
                    float scaledCurvature_dv = Float.Pow(current.Y / maxValue.Y, .3f);

                    *(ptr++) = Vector4ToInt32(new Vector4(1.0f, 0.0f, scaledCurvature_du, scaledCurvature_dv));
                }
            }

            _bitmap.UnlockBits(bitmapData);
            _bitmap.RotateFlip(RotateFlipType.Rotate270FlipNone);

            this.Image = _bitmap;
        }


        unsafe public void FillWithNormalAngleData(ParametricSurfaceEffect effect)
        {
            if (effect == null) return;

            _analyzer.GenerateDensityWarpingFunctions(effect);

            BitmapData bitmapData = _bitmap.LockBits(
                                        new System.Drawing.Rectangle(0, 0, _bitmap.Width, _bitmap.Height),
                                        ImageLockMode.ReadWrite,
                                        PixelFormat.Format32bppArgb);
            Int32* ptr = (Int32*)bitmapData.Scan0.ToPointer();

            Vector4[] data = _analyzer.DataMatrix;
            Vector4 maxValue = _analyzer.DataMatrixMaxValue;
            int index = 0;

            for (int j = 0; j < bitmapData.Height; j++)
            {
                for (int i = 0; i < bitmapData.Width; i++)
                {
                    Vector4 currentValue = data[index++];
                    /*
                    currentValue.X = _analyzer.DataVectorX[i];
                    currentValue.Y = _analyzer.DataVectorY[j];
                    */
                    float uValue = Float.Pow(currentValue.X / maxValue.X, .5f);
                    float vValue = Float.Pow(currentValue.Y / maxValue.Y, .5f);

                    *(ptr++) = Vector4ToInt32(new Vector4(0.0f)
                        + 1.0f * new Vector4(1.0f, 0.0f, uValue, vValue));
                }
            }

            _bitmap.UnlockBits(bitmapData);
            _bitmap.RotateFlip(RotateFlipType.Rotate270FlipNone);

            this.Image = _bitmap;
        }

        // Convert vector with values in [0..1] range into 32-bit ARGB value with each channel in [0..255]
        private static Int32 Vector4ToInt32(Vector4 vector)
        {
            byte A = (byte)(255.0f * MathHelper.Clamp(vector.X, 0.0f, 1.0f));
            byte R = (byte)(255.0f * MathHelper.Clamp(vector.Y, 0.0f, 1.0f));
            byte G = (byte)(255.0f * MathHelper.Clamp(vector.Z, 0.0f, 1.0f));
            byte B = (byte)(255.0f * MathHelper.Clamp(vector.W, 0.0f, 1.0f));

            return ((A << 24) | (R << 16) | (G << 8) | B);
        }

        private Bitmap _bitmap;
        private ParametricSurfaceAnalyzer _analyzer;
    }
}
