//+-----------------------------------------------------------------------------
//
//  Class EditorForm.  The main window of the application.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;


namespace ParametricModelEditor
{
    public partial class EditorForm : Form
    {
        public const int c_maxIndexBufferSize = 65543;

        public EditorForm()
        {
            InitializeComponent();
        }

        private void OnLoad(object sender, EventArgs e)
        {
            PresentationParameters pp = new PresentationParameters();
            pp.BackBufferCount = 1;
            pp.IsFullScreen = false;
            pp.SwapEffect = SwapEffect.Discard;
            pp.BackBufferWidth = panelViewport.Width;
            pp.BackBufferHeight = panelViewport.Height;

            pp.AutoDepthStencilFormat = DepthFormat.Depth24Stencil8;
            pp.EnableAutoDepthStencil = true;
            pp.PresentationInterval = PresentInterval.Immediate;
            pp.BackBufferFormat = SurfaceFormat.Unknown;
            pp.MultiSampleType = MultiSampleType.EightSamples;
            pp.MultiSampleQuality = 2;

            _device = new GraphicsDevice(GraphicsAdapter.DefaultAdapter,
                DeviceType.Hardware, this.panelViewport.Handle, pp);
            
            _parametricModel3DVisualizer = new ParametricModel3DVisualizer(_device, 100, 100);

            _parametricModelDStar3DVisualizer = new ParametricModelDStar3DVisualizer(_device, 50, 50);

            ResetGraphicsDevice();
        }

        private void OnViewportResize(object sender, EventArgs e)
        {
            ResetGraphicsDevice();
        }

        private void ResetGraphicsDevice()
        {
            if (_redrawTimer != null)
            {
                _redrawTimer.Stop();
            }

            if (panelViewport.Width == 0 || panelViewport.Height == 0)
            {
                return;
            }

            _parametricModel3DVisualizer.OnDisposeGraphicsDevice();
            _parametricModelDStar3DVisualizer.OnDisposeGraphicsDevice();

            _device.PresentationParameters.BackBufferWidth = panelViewport.Width;
            _device.PresentationParameters.BackBufferHeight = panelViewport.Height;
            _device.Reset();
            _aspectRatio = (float)panelViewport.Width / (float)panelViewport.Height;

            _parametricModel3DVisualizer.OnResetGraphicsDevice(_device);
            _parametricModelDStar3DVisualizer.OnResetGraphicsDevice(_device);

            if (_redrawTimer == null)
            {
                _redrawTimer = new Timer();
                _redrawTimer.Interval = 12;
                _redrawTimer.Tick += new EventHandler(OnRedrawTimerTick);
            }

            _redrawTimer.Start();

            OnViewportPaint(null, null);
        }

        private void OnRedrawTimerTick(object sender, EventArgs e)
        {
            OnViewportPaint(null, null);
        }

        private void OnViewportPaint(object sender, PaintEventArgs e)
        {
            if (BSplineCurve2DRealization.curveInvalidated  && splineEditorGridControl1.BSplineCurve.ControlPointCount != 0 && splineEditorGridControl2.BSplineCurve.ControlPointCount != 0)
            {
                /*
                _parametricModel3DVisualizer.ApplyProfileProductEffect(splineEditorGridControl1.BSplineCurve,
                                                             splineEditorGridControl2.BSplineCurve);
                */
                /*
                 _parametricModelDStar3DVisualizer.ApplyProfileProductEffect(splineEditorGridControl1.BSplineCurve,
                                                                splineEditorGridControl2.BSplineCurve);
                 * */
                _parametricModelDStar3DVisualizer.upDateCtrlPoints(splineEditorGridControl1.BSplineCurve,
                                                                splineEditorGridControl2.BSplineCurve);
                BSplineCurve2DRealization.curveInvalidated = false;
            }
            _device.Clear(Microsoft.Xna.Framework.Graphics.Color.LightBlue);

            _device.RenderState.FillMode = _wireframeMode ? FillMode.WireFrame : FillMode.Solid;
            _device.RenderState.CullMode = CullMode.None;//.CullCounterClockwiseFace;            
            //_device.RenderState.CullMode = CullMode.CullClockwiseFace;// CullCounterClockwiseFace;            

            /*_device.RenderState.AlphaBlendEnable = true;
            _device.RenderState.SourceBlend = Blend.SourceAlpha;
            _device.RenderState.DestinationBlend = Blend.InverseSourceAlpha;
            _device.RenderState.BlendFunction = BlendFunction.Add;*/

            //_parametricModel3DVisualizer.OnRender(_device, _aspectRatio);
            _parametricModelDStar3DVisualizer.OnRender(_device, _aspectRatio);
            
            _device.Present();

            UpdateFPS();
        }

        private void UpdateFPS()
        {
            _framesRenderedWithinSecond++;

            double newTime = HighPerformanceTimer.GetTimeInSeconds();
            double elapsedTime = newTime - _timeLastFrameMeasured;

            if (elapsedTime >= 1.0)
            {
                double newFps = _framesRenderedWithinSecond / elapsedTime;

                label1.Text = newFps.ToString("F") + " FPS";

                _framesRenderedWithinSecond = 0;
                _timeLastFrameMeasured = newTime;
            }
        }

        private void OnScrollRotation(object sender, ScrollEventArgs e)
        {
            // e.NewValue can be from 0 to 360
            _parametricModel3DVisualizer.Rotation = MathHelper.ToRadians(e.NewValue);
            _parametricModelDStar3DVisualizer.Rotation = MathHelper.ToRadians(e.NewValue);
            OnViewportPaint(null, null);
        }

        private void vScrollBar2_Scroll(object sender, ScrollEventArgs e)
        {
            // e.NewValue can be from 0 to 360
            _parametricModel3DVisualizer.Distance = (.2f + .005f * e.NewValue);
            _parametricModelDStar3DVisualizer.Distance = (.2f + .005f * e.NewValue);

            OnViewportPaint(null, null);
        }

        private void vScrollBar3_Scroll(object sender, ScrollEventArgs e)
        {
            // e.NewValue can be from 0 to 360
            _parametricModel3DVisualizer.LeanAngle = MathHelper.ToRadians(.5f * (float)(e.NewValue - 180));
            _parametricModelDStar3DVisualizer.LeanAngle = MathHelper.ToRadians(.5f * (float)(e.NewValue - 180));

            OnViewportPaint(null, null);
        }

        private void trackBarAmbient_Scroll(object sender, EventArgs e)
        {
            // e.NewValue can be from 0 to 100
            _parametricModel3DVisualizer.AmbientFactor = .01f * (float)trackBarAmbient.Value;
            _parametricModelDStar3DVisualizer.AmbientFactor = .01f * (float)trackBarAmbient.Value;

            OnViewportPaint(null, null);
        }

        private void trackBarDiffuse_Scroll(object sender, EventArgs e)
        {
            // e.NewValue can be from 0 to 100
            _parametricModel3DVisualizer.DiffuseFactor = .01f * (float)trackBarDiffuse.Value;
            _parametricModelDStar3DVisualizer.DiffuseFactor = .01f * (float)trackBarDiffuse.Value;

            OnViewportPaint(null, null);
        }

        private void trackBarSpecular_Scroll(object sender, EventArgs e)
        {
            // e.NewValue can be from 0 to 100+
            _parametricModel3DVisualizer.SpecularFactor = .01f * (float)trackBarSpecular.Value;
            _parametricModelDStar3DVisualizer.SpecularFactor = .01f * (float)trackBarSpecular.Value;

            OnViewportPaint(null, null);
        }

        /*private void trackBarLOD_Scroll(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.DomainLodFraction = ((float)trackBarLOD.Value) / (float)trackBarLOD.Maximum;
        }*/



        private void comboBoxLOD_SelectedIndexChanged(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.SetLodTechnique(comboBoxLOD.SelectedIndex);
            OnViewportPaint(null, null);
        }

        private void comboBoxMethod_SelectedIndexChanged(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.SetCurveMode(comboBoxMethod.SelectedIndex);
        }

/*
        private void checkBoxHardwareCurves_CheckedChanged(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.SetHardwareCurveMode(checkBoxHardwareCurves.Checked);
            OnViewportPaint(null, null);
        }
*/


        static public bool SnapToGrid
        {
            get { return _snapToGrid; }

            // NOTE Remove setter when possible - can result in inconsistent state.
            // Currently only used to suppress SnapToGrid during model loading.
            set { _snapToGrid = value; }
        }



        private void snapToGridCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            _snapToGrid = snapToGridCheckBox.Checked;
        }




        private void buttonSurfaceOfRevolution_Click(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.ApplySurfaceOfRevolutionEffect(splineEditorGridControl1.BSplineCurve);
            _parametricModelDStar3DVisualizer.ApplySurfaceOfRevolutionEffect(splineEditorGridControl1.BSplineCurve);
        }

        private void buttonProfileProduct_Click(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.ApplyProfileProductEffect(splineEditorGridControl1.BSplineCurve,
                                                         splineEditorGridControl2.BSplineCurve);
            _parametricModelDStar3DVisualizer.ApplyProfileProductEffect(splineEditorGridControl1.BSplineCurve,
                                                            splineEditorGridControl2.BSplineCurve);
        }

        private void buttonWireProduct_Click(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.ApplyWireProductEffect(splineEditorGridControl1.BSplineCurve,
                                                      splineEditorGridControl2.BSplineCurve);
            _parametricModelDStar3DVisualizer.ApplyWireProductEffect(splineEditorGridControl1.BSplineCurve,
                                                        splineEditorGridControl2.BSplineCurve);
        }




        private void buttonCurve1Clear_Click(object sender, EventArgs e)
        {
            splineEditorGridControl1.ClearControlPoints();
        }

        private void buttonCurve2Clear_Click(object sender, EventArgs e)
        {
            splineEditorGridControl2.ClearControlPoints();
        }

        private void buttonCurve3Clear_Click(object sender, EventArgs e)
        {
            splineEditorGridControl3.ClearControlPoints();
        }

        private void buttonCurve4Clear_Click(object sender, EventArgs e)
        {
            splineEditorGridControl4.ClearControlPoints();
        }



        private void checkboxCurve1isClosed_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl1.IsClosedCurve = checkboxCurve1isClosed.Checked;
            _parametricModelDStar3DVisualizer._isClosedProfileCurve = checkboxCurve1isClosed.Checked;
        }

        private void checkboxCurve2isClosed_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl2.IsClosedCurve = checkboxCurve2isClosed.Checked;
            _parametricModelDStar3DVisualizer._isClosedCrossSectionCurve = checkboxCurve2isClosed.Checked;
        }

        private void checkboxCurve3isClosed_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl3.IsClosedCurve = checkboxCurve3isClosed.Checked;
        }

        private void checkboxCurve4isClosed_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl4.IsClosedCurve = checkboxCurve4isClosed.Checked;
        }




        private void checkboxCurve1isCentered_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl1.IsCentered = checkboxCurve1isCentered.Checked;
        }

        private void checkboxCurve2isCentered_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl2.IsCentered = checkboxCurve2isCentered.Checked;
        }

        private void checkboxCurve3isCentered_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl3.IsCentered = checkboxCurve3isCentered.Checked;
        }

        private void checkboxCurve4isCentered_CheckedChanged(object sender, EventArgs e)
        {
            splineEditorGridControl4.IsCentered = checkboxCurve4isCentered.Checked;
        }




        private void buttonSaveModelNew_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "xml model files (*.3dmodel.xml)|*.3dmodel.xml|xml files (*.xml)|*.xml|All files (*.*)|*.*";
            saveFileDialog.SupportMultiDottedExtensions = true;
            saveFileDialog.FilterIndex = 1;
            saveFileDialog.RestoreDirectory = true;
            saveFileDialog.FileName = _currentFileName;

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                XmlSerializer serializer = new XmlSerializer(typeof(ParametricModel3DData));
                TextWriter textWriter = new StreamWriter(saveFileDialog.OpenFile());

                ParametricModel3DData data = _parametricModel3DVisualizer.GenerateParametricModel3DData();
                //ParametricModel3DData data = _parametricModelDStar3DVisualizer.GenerateParametricModel3DData();
                data.Name = saveFileDialog.FileName.Substring(_currentFileName.LastIndexOf('\\') + 1);

                // TODO save _parametricModelDstar3DVisualizer

                serializer.Serialize(textWriter, data);
                textWriter.Close();
            }
        }

        private void buttonLoadModelNew_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "xml model files (*.3dmodel.xml)|*.3dmodel.xml|xml files (*.xml)|*.xml|All files (*.*)|*.*";
            openFileDialog.FilterIndex = 1;
            openFileDialog.SupportMultiDottedExtensions = true;
            openFileDialog.RestoreDirectory = true;
            

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                XmlSerializer serializer = new XmlSerializer(typeof(ParametricModel3DData));
                ParametricModel3DData parametricModel3DData = new ParametricModel3DData();
                TextReader textReader = new StreamReader(openFileDialog.OpenFile());
                parametricModel3DData = (ParametricModel3DData)serializer.Deserialize(textReader);

                // Update window title
                _currentFileName = openFileDialog.FileName;
                this.Text = "SplineEditor - " + _currentFileName;

                if (parametricModel3DData.MaterialProperties.InternalColor != -Vector4.One)
                {
                    _parametricModel3DVisualizer.MaterialColor = parametricModel3DData.MaterialProperties.InternalColor;

                    _parametricModelDStar3DVisualizer.MaterialColor = parametricModel3DData.MaterialProperties.InternalColor;
                }
                if (parametricModel3DData.MaterialProperties.AmbientFactor != -1.0f)
                {
                    _parametricModel3DVisualizer.AmbientFactor = parametricModel3DData.MaterialProperties.AmbientFactor;
                    trackBarAmbient.Value = (int)(100.0f * parametricModel3DData.MaterialProperties.AmbientFactor);

                    _parametricModelDStar3DVisualizer.AmbientFactor = parametricModel3DData.MaterialProperties.AmbientFactor;
                }
                if (parametricModel3DData.MaterialProperties.DiffuseFactor != -1.0f)
                {
                    _parametricModel3DVisualizer.DiffuseFactor = parametricModel3DData.MaterialProperties.DiffuseFactor;
                    trackBarDiffuse.Value = (int)(100.0f * parametricModel3DData.MaterialProperties.DiffuseFactor);

                    _parametricModelDStar3DVisualizer.DiffuseFactor = parametricModel3DData.MaterialProperties.DiffuseFactor;
                }
                if (parametricModel3DData.MaterialProperties.SpecularFactor != -1.0f)
                {
                    _parametricModel3DVisualizer.SpecularFactor = parametricModel3DData.MaterialProperties.SpecularFactor;
                    trackBarSpecular.Value = (int)(100.0f * parametricModel3DData.MaterialProperties.SpecularFactor);

                    _parametricModelDStar3DVisualizer.SpecularFactor = parametricModel3DData.MaterialProperties.SpecularFactor;
                }

                int count = parametricModel3DData.ParametricCurves.Count;
                if (count >= 1)
                {
                    splineEditorGridControl1.LoadCurve((BSplineCurve2D)parametricModel3DData.ParametricCurves[0]);
                    _parametricModelDStar3DVisualizer._isClosedProfileCurve = splineEditorGridControl1.IsClosedCurve;
                }
                else
                {
                    splineEditorGridControl1.ClearControlPoints();
                }

                if (count >= 2)
                {
                    splineEditorGridControl2.LoadCurve((BSplineCurve2D)parametricModel3DData.ParametricCurves[1]);
                    _parametricModelDStar3DVisualizer._isClosedCrossSectionCurve = splineEditorGridControl2.IsClosedCurve;
                }
                else
                {
                    splineEditorGridControl2.ClearControlPoints();
                }

                if (count >= 3)
                {
                    splineEditorGridControl3.LoadCurve((BSplineCurve2D)parametricModel3DData.ParametricCurves[2]);
                }
                else
                {
                    splineEditorGridControl3.ClearControlPoints();
                }

                if (count >= 4)
                {
                    splineEditorGridControl4.LoadCurve((BSplineCurve2D)parametricModel3DData.ParametricCurves[3]);
                }
                else
                {
                    splineEditorGridControl4.ClearControlPoints();
                }


                switch (parametricModel3DData.ParametricEffectType)
                {
                    case ParametricEffectType.SurfaceOfRevolution:
                        {
                            splineEditorGridControl1.IsCentered = false;
                            _parametricModel3DVisualizer.ApplySurfaceOfRevolutionEffect(
                                                        splineEditorGridControl1.BSplineCurve);

                            _parametricModelDStar3DVisualizer.ApplySurfaceOfRevolutionEffect(
                                                        splineEditorGridControl1.BSplineCurve);
                            break;
                        }
                    case ParametricEffectType.ProfileProduct:
                        {
                            splineEditorGridControl1.IsCentered = false;
                            splineEditorGridControl2.IsCentered = true;
                            _parametricModel3DVisualizer.ApplyProfileProductEffect(
                                                         splineEditorGridControl1.BSplineCurve,
                                                         splineEditorGridControl2.BSplineCurve);

                            _parametricModelDStar3DVisualizer.ApplyProfileProductEffect(
                                                        splineEditorGridControl1.BSplineCurve,
                                                        splineEditorGridControl2.BSplineCurve);
                            break;
                        }
                    case ParametricEffectType.WireProduct:
                        {
                            splineEditorGridControl1.IsCentered = true;
                            splineEditorGridControl2.IsCentered = true;
                            _parametricModel3DVisualizer.ApplyWireProductEffect(
                                                            splineEditorGridControl1.BSplineCurve,
                                                            splineEditorGridControl2.BSplineCurve);

                            _parametricModelDStar3DVisualizer.ApplyWireProductEffect(
                                                            splineEditorGridControl1.BSplineCurve,
                                                            splineEditorGridControl2.BSplineCurve);
                            break;
                        }
                    default:
                        {
                            _parametricModel3DVisualizer.ApplyNullEffect();

                            _parametricModelDStar3DVisualizer.ApplyNullEffect();
                            break;
                        }
                }

                

                bitmapDataViewer1.Visible = false;
            }
        }




        private void checkboxWireframeMode_CheckedChanged(object sender, EventArgs e)
        {
            _wireframeMode = this.checkboxWireframeMode.Checked;
            OnViewportPaint(null, null);
        }

        private void checkBoxHardwareCurves_CheckedChanged(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.SetHardwareCurveMode(checkBoxHardwareCurves.Checked);
            OnViewportPaint(null, null);
        }


        private void checkBoxEnableOptimization1_CheckedChanged(object sender, EventArgs e)
        {
            _parametricModel3DVisualizer.EnableWarpOptimization = checkBoxEnableOptimization1.Checked;
            OnViewportPaint(null, null);
        }

        private void buttonPickColor_Click(object sender, EventArgs e)
        {
            ColorDialog colorDialog = new ColorDialog();
            colorDialog.AnyColor = true;
            colorDialog.Color = System.Drawing.Color.FromArgb(
                (int)(_parametricModel3DVisualizer.MaterialColor.W * (float)byte.MaxValue),
                (int)(_parametricModel3DVisualizer.MaterialColor.X * (float)byte.MaxValue),
                (int)(_parametricModel3DVisualizer.MaterialColor.Y * (float)byte.MaxValue),
                (int)(_parametricModel3DVisualizer.MaterialColor.Z * (float)byte.MaxValue));

            if (colorDialog.ShowDialog() == DialogResult.OK)
            {
                System.Drawing.Color color = colorDialog.Color;
                _parametricModel3DVisualizer.MaterialColor = new Vector4(
                    ((float)color.R) / (float)byte.MaxValue,
                    ((float)color.G) / (float)byte.MaxValue,
                    ((float)color.B) / (float)byte.MaxValue,
                    ((float)color.A) / (float)byte.MaxValue);

                _parametricModelDStar3DVisualizer.MaterialColor = new Vector4(
                    ((float)color.R) / (float)byte.MaxValue,
                    ((float)color.G) / (float)byte.MaxValue,
                    ((float)color.B) / (float)byte.MaxValue,
                    ((float)color.A) / (float)byte.MaxValue);


                OnViewportPaint(null, null);
            }
        }

        private void buttonAnalyze_Click(object sender, EventArgs e)
        {
            if (!bitmapDataViewer1.Visible)
            {
                bitmapDataViewer1.FillWithNormalData(_parametricModel3DVisualizer.ParametricSurfaceEffect);
                bitmapDataViewer1.Visible = true;
            }
            else
            {
                bitmapDataViewer1.Visible = false;
            }
        }

        private void buttonAnalyze2_Click(object sender, EventArgs e)
        {
            if (!bitmapDataViewer2.Visible)
            {
                bitmapDataViewer2.FillWithCurvatureData(_parametricModel3DVisualizer.ParametricSurfaceEffect);
                bitmapDataViewer2.Visible = true;
            }
            else
            {
                bitmapDataViewer2.Visible = false;
            }
        }

        private void buttonAnalyze3_Click(object sender, EventArgs e)
        {
            /*if (!bitmapDataViewer3.Visible)
            {
                bitmapDataViewer3.FillWithNormalAngleData(_parametricModel3DRealization.ParametricSurfaceEffect);
                bitmapDataViewer3.Visible = true;
            }
            else
            {
                bitmapDataViewer3.Visible = false;
            }*/

            PerfBenchmark();
        }

        private void PerfBenchmark()
        {
            int framesToRender = 1000;
            double beginTime = HighPerformanceTimer.GetTimeInSeconds();

            for (int n = 0; n < framesToRender; n++)
            {
                _parametricModel3DVisualizer.Rotation = ((float)n) / ((float)framesToRender) * FloatingPointMathHelper.TwoPi;
                OnViewportPaint(null, null);
            }


            double endTime = HighPerformanceTimer.GetTimeInSeconds();
            double elapsedTime = endTime - beginTime;

            if (elapsedTime > 0)
            {
                double fps = framesToRender / elapsedTime;

                MessageBox.Show(framesToRender + " frames drawn in " + elapsedTime + " seconds,\nframerate: " + fps + " FPS");
            }
        }




        private GraphicsDevice _device;
        private float _aspectRatio;
        private bool _wireframeMode;

        private Timer _redrawTimer;

        private ParametricModel3DVisualizer _parametricModel3DVisualizer;
        private string _currentFileName = String.Empty;

        private ParametricModelDStar3DVisualizer _parametricModelDStar3DVisualizer;

        private double _timeLastFrameMeasured = Double.MinValue;
        private int _framesRenderedWithinSecond = 0;

        // Static field, so ensure we don't reuse Form1
        static private bool _snapToGrid = true;

     
    }
}