//Written by Leonid Velikovich and Benjamin Hernandez Arreguin

//+-----------------------------------------------------------------------------
//
//  Class ParametricModel3DVisualizer.  Provides realtime display of the
//  parametric model, using software or hardware parametric evaluation.
//
//------------------------------------------------------------------------------


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
//using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using Float = ParametricModelEditor.FloatingPointMathHelper;

using WarpLookupType = Microsoft.Xna.Framework.Graphics.PackedVector.HalfSingle;

namespace ParametricModelEditor
{
    class ParametricModel3DVisualizer
    {
        // Lookup table parameters
        public const int c_bSplineLookupTableSize = 128;
        public const int c_domainWarpLookupTableSize = 256;

        private const SurfaceFormat c_bSplineLookupTableFormat = SurfaceFormat.Vector2;
        private const SurfaceFormat c_domainWarpLookupTableFormat = SurfaceFormat.HalfSingle;

        private const int c_triangleCount = 30000;



        public ParametricModel3DVisualizer(GraphicsDevice device, int xSamples, int ySamples)
        {
            Debug.Assert(xSamples > 0 && ySamples > 0);

            _device = device;
            _strEffect = File.ReadAllText("ParametricSurfaceEffect.fx");

            _analyzer = new ParametricSurfaceAnalyzer(500, 500);

            Resample(2 * xSamples * ySamples);

            InitializeLookupTextures();
        }

        private void InitializeLookupTextures()
        {
            
            _bSplineControlPointTexture = new Texture2D(_device, c_bSplineLookupTableSize, 1, 1,
                                              TextureUsage.None, c_bSplineLookupTableFormat);
            
            _bSplineControlPointRT2D = new RenderTarget2D(_device, c_bSplineLookupTableSize, 1, 1,
                                              c_bSplineLookupTableFormat);
            
            _domainWarpFunctionTexture = new Texture2D(_device, c_domainWarpLookupTableSize, 1, 1,
                                              TextureUsage.None, c_domainWarpLookupTableFormat);
        }


        public ParametricModel3DData GenerateParametricModel3DData()
        {
            ParametricModel3DData parametricModel3DData = new ParametricModel3DData();
            parametricModel3DData.MaterialProperties.InternalColor = _materialColor;
            parametricModel3DData.MaterialProperties.AmbientFactor = _materialAmbientFactor;
            parametricModel3DData.MaterialProperties.DiffuseFactor = _materialDiffuseFactor;
            parametricModel3DData.MaterialProperties.SpecularFactor = _materialSpecularFactor;

            if (_parametricSurfaceEffect != null)
            {
                parametricModel3DData.ParametricEffectType = _parametricSurfaceEffect.ParametricEffectType;
                parametricModel3DData.ParametricCurves = _parametricSurfaceEffect.GetSerialization();
            }
            return parametricModel3DData;
        }

        public void Resample(int totalTriangles)
        {
            ResetGeometry();  // Deallocate what is used

            _analyzer.GenerateDensityWarpingFunctions(_parametricSurfaceEffect);

            _domainWarpFunctionU = new float[_analyzer.DataVectorX.Length];
            Array.Copy(_analyzer.DataVectorX, _domainWarpFunctionU, _domainWarpFunctionU.Length);

            _domainWarpFunctionV = new float[_analyzer.DataVectorY.Length];
            Array.Copy(_analyzer.DataVectorY, _domainWarpFunctionV, _domainWarpFunctionV.Length);

            float variationU = _analyzer.DataVectorXTotalVariation;
            float variationV = _analyzer.DataVectorYTotalVariation;

            float root = Float.Sqrt(.5f * (float)totalTriangles);  // # of grid squares = 1/2 * #triangles

            float ratio = (variationV > 0.0f)
                        ? Float.Sqrt(variationU / variationV)
                        : 1.0f;

            if (!(ratio > 0.0f))  // Handle degenerate cases (e.g. null effect)
            {
                ratio = 1.0f;
            }

            _numUSamples = (int)(root * ratio + .5f);
            _numVSamples = (int)(root / ratio + .5f);

            InitializeVertexBuffer();
            InitializeIndexBuffer();
        }

        private void ResetGeometry()
        {
            if (_vertexBufferPositionNormal != null)
            {
                _vertexBufferPositionNormal.Dispose();
                _vertexBufferPositionNormal = null;
            }
            if (_indexBuffer != null)
            {
                _indexBuffer.Dispose();
                _indexBuffer = null;
            }
        }

        // // This function was modified to run under XNA 3.0
        private void InitializeVertexBuffer()
        {
            _numVertices = (_numUSamples + 1) * (_numVSamples + 1);
            _verticesPosition2D = new VertexPosition2D[_numVertices];
            _verticesPositionNormal = new VertexPositionNormal[_numVertices];

            int index = 0;
            for (int y = 0; y <= _numVSamples; y++)
            {
                for (int x = 0; x <= _numUSamples; x++)
                {
                    // fx and fy follow the sequence: 0, .5, 1, 1.5, 2, ...
                    // This simplifies the GPU algorithm for geomorphing alternating points.
                    float fx = .5f * (float)x;
                    float fy = .5f * (float)(_numVSamples - y);

                    //Debug.Assert((0.0f <= fx) && (fx <= 1.0f));
                    //Debug.Assert((0.0f <= fy) && (fy <= 1.0f));

                    _verticesPosition2D[index++].Position = new Vector2(fx, fy);  // Left-right, top-down vertex order
                }
            }

            _vertexBufferPosition2D = new VertexBuffer(_device, typeof(VertexPosition2D), _numVertices,
                                             BufferUsage.WriteOnly);
            /*
            _vertexBufferPositionNormal = new VertexBuffer(_device, typeof(VertexPositionNormal), _numVertices,
                                             BufferUsage.Dynamic | BufferUsage.WriteOnly,
                                             ResourceManagementMode.Manual);
             * */
            _vertexBufferPositionNormal = new VertexBuffer(_device, typeof(VertexPositionNormal), _numVertices,
                                 BufferUsage.WriteOnly);

        }

        private float LinearSampleFromArray(float index, float[] source, int sourceLength)
        {
            index *= (float)sourceLength;
            int index1 = (int)index;
            float returnValue;

            if (index1 + 1 < sourceLength)
            {
                float weight = index - (float)index1;
                returnValue = MathHelper.Lerp(source[index1], source[index1 + 1], weight);
            }
            else
            {
                returnValue = source[source.Length - 1];  // Clamp to last value
            }

            return returnValue;
        }

        
        // Using triangle strips
        // // This function was modified to run under XNA 3.0
        private void InitializeIndexBuffer()
        {
            _numIndices = _numVSamples * 2 * (_numUSamples + 2);   // Conservative estimate
            Debug.Assert(_numIndices <= EditorForm.c_maxIndexBufferSize);  // Max legal index buffer size
            _indices = new short[_numIndices];

            int index = 0;

            for (int y = 0; y < _numVSamples; y++)
            {
                if (y % 2 == 0)
                {
                    for (int x = 0; x <= _numUSamples; x++)
                    {
                        _indices[index++] = (short)((y + 1) * (_numUSamples + 1) + x);
                        _indices[index++] = (short)((y    ) * (_numUSamples + 1) + x);
                    }

                    // Degenerate point on the tristrip
                    _indices[index++] = (short)((y + 2) * (_numUSamples + 1) - 1);
                    _indices[index++] = (short)((y + 2) * (_numUSamples + 1) - 1);
                }
                else
                {
                    for (int x = _numUSamples; x >= 0; x--)
                    {
                        _indices[index++] = (short)((y    ) * (_numUSamples + 1) + x);
                        _indices[index++] = (short)((y + 1) * (_numUSamples + 1) + x);
                    }
                    
                    // Degenerate points on the tristrip
                    _indices[index++] = (short)((y + 1) * (_numUSamples + 1));
                    _indices[index++] = (short)((y + 2) * (_numUSamples + 1));
                }
            }


            _numIndices = index;
            _numTriangles = index - 2;

            /*
            _indexBuffer = new IndexBuffer(_device, sizeof(short) * indices.Length, ResourceUsage.WriteOnly,
                                           IndexElementSize.SixteenBits);
             * */

            _indexBuffer = new IndexBuffer(_device, sizeof(short) * _indices.Length, BufferUsage.WriteOnly,
                                           IndexElementSize.SixteenBits);

            _indexBuffer.SetData<short>(_indices);
        }


        public void OnDisposeGraphicsDevice()
        {
            if (_hardwareEffect != null)
            {
                _hardwareEffect.Dispose();
                _hardwareEffect = null;
            }
            if (_vertexBufferPositionNormal != null)
            {
                _vertexBufferPositionNormal.Dispose();
                _vertexBufferPositionNormal = null;
            }
            _device = null;
        }


        // This function was modified to run under XNA 3.0
        public void OnResetGraphicsDevice(GraphicsDevice device)
        {
            _device = device;

            CompiledEffect compEffect = Effect.CompileEffectFromSource(_strEffect,
                null, null, CompilerOptions.Debug, TargetPlatform.Windows);

            Debug.Assert(compEffect.Success == true, compEffect.ErrorsAndWarnings);

            byte[] effectCode = compEffect.GetEffectCode();

            _hardwareEffect = new Effect(device, effectCode, CompilerOptions.Debug, null);

            _hardwareEffect.CurrentTechnique = _hardwareEffect.Techniques[_enableHardwareParametrization ? "HardwareTechnique_ProfileProduct" : "SoftwareTechnique"];

            _hardwareEffect.Parameters["BSpline_ControlPointTexture"].SetValue(_bSplineControlPointTexture);
            _hardwareEffect.Parameters["Surface_WarpFunctionTexture"].SetValue(_domainWarpFunctionTexture);

            /*
            _vertexBufferPositionNormal = new VertexBuffer(device, typeof(VertexPositionNormal), _numVertices,
                                             ResourceUsage.Dynamic | ResourceUsage.WriteOnly,
                                             ResourceManagementMode.Manual);
             * */
            _vertexBufferPositionNormal = new VertexBuffer(device, typeof(VertexPositionNormal), _numVertices,
                                             BufferUsage.WriteOnly);

        }



        // This function was modified to run under XNA 3.0
        public void OnRender(GraphicsDevice device, float aspectRatio)
        {
            _device = device;

            Vector3 eyeLocation = new Vector3(-3.0f, 0.0f, 1f) * _distance;
            Vector3 upVector = new Vector3(0, 0, 1);
            Matrix leanTransform = Matrix.CreateRotationY(_leanAngle);
            eyeLocation = Vector3.Transform(eyeLocation, leanTransform);
            upVector = Vector3.Transform(upVector, leanTransform);

            _hardwareEffect.Parameters["World"].SetValue(
                Matrix.CreateRotationZ(_rotation));
            _hardwareEffect.Parameters["View"].SetValue(
                Matrix.CreateLookAt(eyeLocation,
                                    Vector3.Zero,
                                    upVector));
            _hardwareEffect.Parameters["Projection"].SetValue(
                Matrix.CreatePerspectiveFieldOfView(1.25f * MathHelper.Pi / 4.0f,
                                                    aspectRatio,
                                                    .25f, 100.0f));
            _hardwareEffect.Parameters["EyeLocation"].SetValue(eyeLocation);
            
            // TEST: distance-based domain scaling
            //TEST float lodFactorInverse = 1.0f;
            float lodFactorInverse = Math.Min(Math.Max(3.0f * _distance, 1.0f), 10.0f);
            float lodFactorPow2 = Float.NearestPowerOf2(lodFactorInverse);
            float lodFraction = 2.0f - lodFactorInverse / lodFactorPow2;  // E.g. 2-15/8 yields 1/8 remaining till next power of 2

            _domainScalingFactor = new Vector2(2.0f / _numUSamples, 2.0f / _numVSamples) * lodFactorPow2;
            _hardwareEffect.Parameters["Surface_DomainScalingFactor"].SetValue(_domainScalingFactor);
            _domainLodFraction = lodFraction;
            _hardwareEffect.Parameters["Surface_DomainLodFraction"].SetValue(_domainLodFraction);

            _hardwareEffect.Parameters["AmbientFactor"].SetValue(_materialAmbientFactor);
            _hardwareEffect.Parameters["DiffuseFactor"].SetValue(_materialDiffuseFactor);
            _hardwareEffect.Parameters["SpecularFactor"].SetValue(_materialSpecularFactor);
            _hardwareEffect.Parameters["MaterialColor"].SetValue(_materialColor);

            _hardwareEffect.CommitChanges();



            if (_parametricSurfaceEffect != null)
            {
                //UpdateHardwareControlPointTables();

                _parametricSurfaceEffect.Parameters["World"].SetValue(
                    Matrix.CreateRotationZ(_rotation));

                if (!_enableHardwareParametrization)
                {
                
                    for (int n = 0; n < _numVertices; n++)
                    {
                        VertexPosition2D position2D = _verticesPosition2D[n];

                        Vector2 inputPoint = position2D.Position * _domainScalingFactor;
                        inputPoint = Vector2.Clamp(inputPoint, Vector2.Zero, Vector2.One);

                        Vector2 domainPoint;
                        
                        if (_enableWarpOptimization)
                        {
                            domainPoint.X = LinearSampleFromArray(inputPoint.X, _domainWarpFunctionU, _domainWarpFunctionU.Length);
                            domainPoint.Y = LinearSampleFromArray(inputPoint.Y, _domainWarpFunctionV, _domainWarpFunctionV.Length);
                        }
                        else
                        {
                            domainPoint = inputPoint;
                        }
                        Vector3 position3D, dF_du, dF_dv, ddf_duu, ddf_dvv;

                        _parametricSurfaceEffect.VS_SurfaceFunction(domainPoint,
                                                            out position3D,
                                                            out dF_du,
                                                            out dF_dv,
                                                            out ddf_duu,
                                                            out ddf_dvv);
                        _verticesPositionNormal[n].Position = position3D;
                        _verticesPositionNormal[n].Normal = Vector3.Normalize(Vector3.Cross(dF_du, dF_dv));
                        //Console.WriteLine(domainPoint.X.ToString() + ", " + domainPoint.Y.ToString());
                    }
                }
            }

            EffectTechnique technique;
            if (_enableHardwareParametrization)
            {
                technique = _hardwareEffect.Techniques[_currentRenderingTechnique];
                _device.VertexDeclaration = new VertexDeclaration(device, VertexPosition2D.Elements);
                _device.Vertices[0].SetSource(_vertexBufferPosition2D, 0, VertexPosition2D.SizeInBytes);
                _hardwareEffect.Begin();
                technique.Passes[0].Begin();
                _device.DrawUserIndexedPrimitives(PrimitiveType.TriangleStrip, _verticesPosition2D, 0, _numVertices, _indices, 0, _numTriangles);
                technique.Passes[0].End();
                _hardwareEffect.End();

            }
            else
            {
                technique = _hardwareEffect.Techniques["SoftwareTechnique"];
                _device.VertexDeclaration = new VertexDeclaration(device, VertexPositionNormal.Elements);
                _device.Vertices[0].SetSource(_vertexBufferPositionNormal, 0, VertexPositionNormal.SizeInBytes);
                _hardwareEffect.Begin();
                technique.Passes[0].Begin();
                _device.DrawUserIndexedPrimitives(PrimitiveType.TriangleStrip, _verticesPositionNormal, 0, _numVertices, _indices, 0, _numTriangles);
                technique.Passes[0].End();
                _hardwareEffect.End();

            }

        }



        public float Rotation
        {
            set { _rotation = value; }
        }

        public float Distance
        {
            set { _distance = value; }
        }

        public float LeanAngle
        {
            set { _leanAngle = value; }
        }

        public float AmbientFactor
        {
            set { _materialAmbientFactor = value; }
        }

        public float DiffuseFactor
        {
            set { _materialDiffuseFactor = value; }
        }

        public float SpecularFactor
        {
            set { _materialSpecularFactor = value; }
        }

        public Vector2 DomainScalingFactor
        {
            set { _domainScalingFactor = value; }
        }

        public float DomainLodFraction
        {
            set { _domainLodFraction = value; }
        }

        public Vector4 MaterialColor
        {
            get { return _materialColor; }
            set { _materialColor = value; }
        }

        public ParametricSurfaceEffect ParametricSurfaceEffect
        {
            get { return _parametricSurfaceEffect; }
        }
        
        public void ApplySurfaceOfRevolutionEffect(ParametricCurve2D profileCurve)
        {
            _parametricSurfaceEffect = new SurfaceOfRevolutionEffect(profileCurve);

            Resample(c_triangleCount);

            UpdateHardwareControlPointTables();
            UpdateHardwareTechnique();
        }

        public void ApplyProfileProductEffect(ParametricCurve2D profileCurve, ParametricCurve2D crossSectionCurve)
        {
            _parametricSurfaceEffect = new ProfileProductEffect(profileCurve, crossSectionCurve);

            Resample(c_triangleCount);

            UpdateHardwareControlPointTables();
            UpdateHardwareTechnique();
        }

        public void ApplyWireProductEffect(ParametricCurve2D wireCurve, ParametricCurve2D crossSectionCurve)
        {
            _parametricSurfaceEffect = new WireProductEffect(wireCurve, crossSectionCurve);

            Resample(c_triangleCount);

            UpdateHardwareControlPointTables();
            UpdateHardwareTechnique();
        }

        private void UpdateHardwareControlPointTables()
        {
            UpdateHardwareBSplineLookupTables();
            UpdateHardwareWarpLookupTables();

            _hardwareEffect.CommitChanges();
        }

        private void UpdateHardwareBSplineLookupTables()
        {
            int countU, countV;
            float offsetV;
            Vector2[] data = new Vector2[c_bSplineLookupTableSize];

            PackArraysIntoLookupArray<Vector2>(  // Duplicate last control points
                _parametricSurfaceEffect.GetControlPointsU(),
                _parametricSurfaceEffect.GetControlPointsV(),
                true,
                out countU,
                out countV,
                out offsetV,
                ref data);
            
           
            /*_bSplineControlPointTexture = _bSplineControlPointRT2D.GetTexture();
           */
            // Update number of control points in shader
            _bSplineControlPointTexture.GraphicsDevice.Reset();
            _bSplineControlPointTexture.SetData<Vector2>(data);
            _hardwareEffect.Parameters["BSpline1_ControlPointCount"].SetValue(countU);
            _hardwareEffect.Parameters["BSpline2_ControlPointCount"].SetValue(countV);
            _hardwareEffect.Parameters["BSpline2_ControlPointOffset"].SetValue(offsetV);
            _hardwareEffect.Parameters["BSpline_ControlPointTexture"].SetValue(_bSplineControlPointTexture);

            // NOTE Caller is responsible for calling _hardwareEffect.CommitChanges();
        }

        private void UpdateHardwareWarpLookupTables()
        {
            WarpLookupType[] data = new WarpLookupType[c_domainWarpLookupTableSize];

            PackWarpLookupArrays(  // Duplicate last control points
                _domainWarpFunctionU,
                _domainWarpFunctionV,
                ref data);

            // Update number of control points in shader
            _domainWarpFunctionTexture.SetData<WarpLookupType>(data);

            _hardwareEffect.Parameters["Surface_WarpFunctionTexture"].SetValue(_domainWarpFunctionTexture);

            // NOTE Caller is responsible for calling _hardwareEffect.CommitChanges();
        }

        private void PackWarpLookupArrays(float[] source1,
                                          float[] source2,
                                          ref WarpLookupType[] dest)
        {
            int index = 0;
            for (int n = 0; n < source1.Length; n++)
            {
                dest[index++] = new WarpLookupType(source1[n]);
            }
            for (int n = 0; n < source2.Length; n++)
            {
                dest[index++] = new WarpLookupType(source2[n]);
            }
        }

        // Returns number of control points
        private void PackArraysIntoLookupArray<T>(T[] source1,
                                                 T[] source2,
                                                 bool padLastValues,
                                                 out int source1Length,
                                                 out int source2Length,
                                                 out float source2Offset,
                                                 ref T[] dest)
        {
            int padAmount = padLastValues ? 1 : 0;

            source1Length = (source1 != null) ? source1.Length : 0;
            source2Length = (source2 != null) ? source2.Length : 0;
            source2Offset = ((float)(source1Length + padAmount)) / (float)dest.Length;

            if (source1Length > 0)
            {
                Debug.Assert(source1Length + padAmount <= dest.Length);

                Array.Copy(source1, dest, source1Length);

                // Duplicate last point to avoid extra logic in hardware shader that avoids endpoint
                if (padLastValues)
                {
                    dest[source1Length] = dest[source1Length - 1];
                }
            }

            if (source2Length > 0)
            {
                int integerOffset = (source1Length + padAmount);

                Debug.Assert(integerOffset + source2Length + padAmount <= dest.Length);

                Array.Copy(source2, 0, dest, integerOffset, source2Length);

                // Duplicate last point to avoid extra logic in hardware shader that avoids endpoint
                if (padLastValues)
                {
                    dest[integerOffset + source2Length] = dest[integerOffset + source2Length - 1];
                }
            }
        }

        public void ApplyNullEffect()
        {
            _parametricSurfaceEffect = null;
        }

        public void SetHardwareCurveMode(bool enable)
        {
            _enableHardwareParametrization = enable;

            UpdateHardwareTechnique();
        }

        public void SetLodTechnique(int technique)
        {
            _lodTechnique = technique;

            UpdateHardwareTechnique();
        }

        public void SetCurveMode(int mode)
        {
            //_curveMode = mode;
            switch (mode)
            {
                case 0:
                    _curveMode = "SoftwareMode";
                    _enableHardwareParametrization = false;
                    break;
                case 1:
                    _enableHardwareParametrization = true;
                    break;
                case 2:
                    _curveMode = "DstarMode";
                    _enableHardwareParametrization = false;
                    break;
            }
            UpdateHardwareTechnique();
        }

        private void UpdateHardwareTechnique()
        {
            if (_hardwareEffect != null)
            {
                // Set the proper GPU shader technique
                if (_parametricSurfaceEffect == null || !_enableHardwareParametrization)
                {
                    _currentRenderingTechnique = "SoftwareTechnique";
                }
                else if (_parametricSurfaceEffect.ParametricEffectType == ParametricEffectType.ProfileProduct)
                {
                    switch (_lodTechnique)
                    {
                        case 0:
                            _currentRenderingTechnique = "HardwareTechnique_ProfileProduct";
                            break;
                        case 1:
                            _currentRenderingTechnique = "HardwareTechnique_ProfileProduct_Geomorph1";
                            break;
                        case 2:
                            _currentRenderingTechnique = "HardwareTechnique_ProfileProduct_Geomorph2";
                            break;
                        case 3:
                            _currentRenderingTechnique = "HardwareTechnique_ProfileProduct_Geomorph3";
                            break;
                    }
                }
                else if (_parametricSurfaceEffect.ParametricEffectType == ParametricEffectType.SurfaceOfRevolution)
                {
                    _currentRenderingTechnique = "HardwareTechnique_SurfaceOfRevolution";
                }
                else if (_parametricSurfaceEffect.ParametricEffectType == ParametricEffectType.WireProduct)
                {
                    _currentRenderingTechnique = "HardwareTechnique_WireProduct";
                }


                _hardwareEffect.CurrentTechnique = _hardwareEffect.Techniques[_currentRenderingTechnique];
            }
        }

        public bool EnableWarpOptimization
        {
            get { return _enableWarpOptimization; }
            set
            {
                _enableWarpOptimization = value;
                Resample(c_triangleCount);
            }
        }

        private GraphicsDevice _device;
        private string _currentRenderingTechnique = "SoftwareTechnique";
        private string _curveMode = "SoftwareMode";

        private ParametricSurfaceAnalyzer _analyzer;

        private int _numUSamples;  // We have a rectangular numU x numV grid domain
        private int _numVSamples;

        private int _numVertices;
        private int _numIndices;
        private int _numTriangles;

        private bool _enableWarpOptimization = true;
        private bool _enableHardwareParametrization = true;
        private int  _lodTechnique = 0;
        

        private VertexPosition2D[] _verticesPosition2D;
        private VertexBuffer _vertexBufferPosition2D;

        private VertexPositionNormal[] _verticesPositionNormal;
        private VertexBuffer _vertexBufferPositionNormal;

        private IndexBuffer _indexBuffer;
        private short[] _indices;

        private Effect _hardwareEffect;

        private RenderTarget2D _bSplineControlPointRT2D;
        private Texture2D _bSplineControlPointTexture;
        
        private Texture2D _domainWarpFunctionTexture;

        private float[] _domainWarpFunctionU;
        private float[] _domainWarpFunctionV;

        private ParametricSurfaceEffect _parametricSurfaceEffect = null;
        private string _strEffect;

        private float _rotation = 0.0f;
        private float _distance = 1.0f;
        private float _leanAngle = 0.0f;
        private Vector2 _domainScalingFactor = new Vector2(1.0f, 1.0f);
        private float _domainLodFraction = 1.0f;
        
        private float _materialAmbientFactor = .05f;
        private float _materialDiffuseFactor = .90f;
        private float _materialSpecularFactor = 4.0f;

        private Vector4 _materialColor = new Vector4(1.0f, .27f, .20f, 1.0f);

    }
}
