//Written by Leonid Velikovich and Benjamin Hernandez Arreguin

//+-----------------------------------------------------------------------------
//
//  Class ParametricSurfaceAnalyzer.  Used for deriving curvature-based
//  domain warping functions and misc data analyses of a parametric surface.
//
//  NOTE: Curvature-based domain warping functions work on surfaces whose
//  parametrization is highly separable for the U and V domain parameters.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Diagnostics;
using System.Drawing;
//using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;

using Float = ParametricModelEditor.FloatingPointMathHelper;


namespace ParametricModelEditor
{
    public class ParametricSurfaceAnalyzer
    {
        public ParametricSurfaceAnalyzer(int width, int height)
        {
            _width = width;
            _height = height;

            _data = new Vector4[_width * _height];

            _localMaxWeightU = new float[_width];
            _localMaxWeightV = new float[_height];

            _localMaxAngleU = new float[_width];
            _localMaxAngleV = new float[_height];
        }

        private void ResetData()
        {
            _maxValue = Vector4.Zero;

            Array.Clear(_data, 0, _data.Length);

            Array.Clear(_localMaxWeightU, 0, _localMaxWeightU.Length);
            Array.Clear(_localMaxWeightV, 0, _localMaxWeightV.Length);

            Array.Clear(_localMaxAngleU, 0, _localMaxAngleU.Length);
            Array.Clear(_localMaxAngleV, 0, _localMaxAngleV.Length);
        }

        unsafe public void FillWithNormalData(ParametricSurfaceEffect effect)
        {
            ResetData();

            if (effect == null) return;

            float invHeight = 1.0f / _height;
            float invWidth = 1.0f / _width;

            int index = 0;

            for (int j = 0; j < _height; j++)
            {
                float v = invHeight * (.5f + (float)j);

                for (int i = 0; i < _width; i++)
                {
                    float u = invWidth * (.5f + (float)i);

                    Vector3 position, dF_du, dF_dv, ddF_duu, ddF_dvv;

                    effect.VS_SurfaceFunction(new Vector2(u, v),
                                              out position,
                                              out dF_du,
                                              out dF_dv,
                                              out ddF_duu,
                                              out ddF_dvv);

                    Vector4 normal = new Vector4(Vector3.Normalize(Vector3.Cross(dF_du, dF_dv)), 0);

                    _data[index++] = normal;
                }
            }
        }

        unsafe public void FillWithCurvatureData(ParametricSurfaceEffect effect)
        {
            ResetData();

            if (effect == null) return;

            float invHeight = 1.0f / _height;
            float invWidth = 1.0f / _width;

            int index = 0;

            for (int j = 0; j < _height; j++)
            {
                float v = invHeight * (.5f + (float)j);

                for (int i = 0; i < _width; i++)
                {
                    float u = invWidth * (.5f + (float)i);

                    Vector3 position, dF_du, dF_dv, ddF_duu, ddF_dvv;

                    effect.VS_SurfaceFunction(new Vector2(u, v),
                                              out position,
                                              out dF_du,
                                              out dF_dv,
                                              out ddF_duu,
                                              out ddF_dvv);

                    Vector3 normal = Vector3.Normalize(Vector3.Cross(dF_du, dF_dv));
                    float curvature_du, curvature_dv;

                    //
                    // Compute normal curvature from du tangent vector
                    //
                    {
                        Vector3 dF_du_norm = Vector3.Normalize(dF_du);
                        Vector3 sideNormalDu = Vector3.Normalize(Vector3.Cross(dF_du, normal));

                        Matrix planeProjectionMatrixDu = new Matrix(dF_du_norm.X, dF_du_norm.Y, dF_du_norm.Z, 0,
                                                                    sideNormalDu.X, sideNormalDu.Y, sideNormalDu.Z, 0,
                                                                    normal.X, normal.Y, normal.Z, 0,
                                                                    0, 0, 0, 1);
                        planeProjectionMatrixDu = Matrix.Invert(planeProjectionMatrixDu);

                        Vector3 dF_du_new = Vector3.Transform(dF_du, planeProjectionMatrixDu);
                        Vector3 ddF_duu_new = Vector3.Transform(ddF_duu, planeProjectionMatrixDu);

                        // Projected onto XZ plane
                        curvature_du = Float.Abs((dF_du_new.X * ddF_duu_new.Z - dF_du_new.Z * ddF_duu_new.X)
                                     / Float.Pow(Float.Sq(dF_du_new.X) + Float.Sq(dF_du_new.Z), 1.5f));
                        if (curvature_du > _maxValue.X)
                        {
                            _maxValue.X = curvature_du;
                        }
                    }

                    //
                    // Compute normal curvature from dv tangent vector
                    //
                    {
                        Vector3 dF_dv_norm = Vector3.Normalize(dF_dv);
                        Vector3 sideNormalDv = Vector3.Normalize(Vector3.Cross(dF_dv, normal));

                        Matrix planeProjectionMatrixDv = new Matrix(dF_dv_norm.X, dF_dv_norm.Y, dF_dv_norm.Z, 0,
                                                                    sideNormalDv.X, sideNormalDv.Y, sideNormalDv.Z, 0,
                                                                    normal.X, normal.Y, normal.Z, 0,
                                                                    0, 0, 0, 1);
                        planeProjectionMatrixDv = Matrix.Invert(planeProjectionMatrixDv);

                        Vector3 dF_dv_new = Vector3.Transform(dF_dv, planeProjectionMatrixDv);
                        Vector3 ddF_dvv_new = Vector3.Transform(ddF_dvv, planeProjectionMatrixDv);

                        // Projected onto XZ plane
                        curvature_dv = Float.Abs((dF_dv_new.X * ddF_dvv_new.Z - dF_dv_new.Z * ddF_dvv_new.X)
                                     / Float.Pow(Float.Sq(dF_dv_new.X) + Float.Sq(dF_dv_new.Z), 1.5f));
                        if (curvature_dv > _maxValue.Y)
                        {
                            _maxValue.Y = curvature_dv;
                        }
                    }

                    _data[index++] = new Vector4(curvature_du, curvature_dv, 0, 0);
                }
            }
        }


        // This will need some heavy commenting later on
        public void GenerateDensityWarpingFunctions(ParametricSurfaceEffect effect)
        {
            ResetData();

            if (effect == null) return;

            FillWithCurvatureWeightData(effect);

            GenerateInverseDistributionFunctions(effect);
        }

        unsafe private void FillWithCurvatureWeightData(ParametricSurfaceEffect effect)
        {
            int tempArraySize = (_width + 1) * (_height + 1);
            Vector3[] tempPositionData = new Vector3[ tempArraySize ];
            Vector3[] tempNormalData = new Vector3[ tempArraySize ];

            // Fill out positions, normals in domain space
            {
                float invHeight = 1.0f / (float)_height;
                float invWidth = 1.0f / (float)_width;

                int index = 0;
                for (int j = 0; j <= _height; j++)
                {
                    float v = invHeight * (float)j;

                    for (int i = 0; i <= _width; i++)
                    {
                        float u = invWidth * (float)i;

                        Vector3 position, dF_du, dF_dv, ddF_duu, ddF_dvv;

                        effect.VS_SurfaceFunction(new Vector2(u, v),
                                                  out position,
                                                  out dF_du,
                                                  out dF_dv,
                                                  out ddF_duu,
                                                  out ddF_dvv);

                        tempPositionData[index] = position;
                        tempNormalData[index] = Vector3.Normalize(Vector3.Cross(dF_du, dF_dv));

                        index++;
                    }
                }
            }

            // Generate curvature weights from normal derivatives (angles) and position derivatives (distances)
            {
                int index = 0;

                for (int j = 0; j < _height; j++)
                {
                    for (int i = 0; i < _width; i++)
                    {
                        int indexCurrent = (_width + 1) * j + i;
                        Vector3 currentPosition = tempPositionData[indexCurrent];
                        Vector3 currentNormal = tempNormalData[indexCurrent];

                        float angleUDirection, angleVDirection;
                        float weightUDirection, weightVDirection;

                        {
                            int indexNextUDirection = (_width + 1) * j + (i + 1);

                            Vector3 nextNormalUDirection = tempNormalData[indexNextUDirection];
                            angleUDirection = Float.Acos(Vector3.Dot(currentNormal, nextNormalUDirection));

                            Vector3 nextPositionUDirection = tempPositionData[indexNextUDirection];
                            float distanceUDirection = (nextPositionUDirection - currentPosition).Length();

                            // The weight formula.  Can be A, A*sqrt(D), A*D, etc.
                            weightUDirection = angleUDirection * (1f/_width + 10.0f * distanceUDirection);

                            if (weightUDirection > _localMaxWeightU[i])
                            {
                                _localMaxWeightU[i] = weightUDirection;
                                if (weightUDirection > _maxValue.X)
                                {
                                    _maxValue.X = weightUDirection;
                                }
                            }

                            if (angleUDirection > _localMaxAngleU[i])
                            {
                                _localMaxAngleU[i] = angleUDirection;
                            }
                        }

                        {
                            int indexNextVDirection = (_width + 1) * (j + 1) + i;

                            Vector3 nextNormalVDirection = tempNormalData[indexNextVDirection];
                            angleVDirection = Float.Acos(Vector3.Dot(currentNormal, nextNormalVDirection));

                            Vector3 nextPositionVDirection = tempPositionData[indexNextVDirection];
                            float distanceVDirection = (nextPositionVDirection - currentPosition).Length();

                            // The weight formula.  Can be A, A*sqrt(D), A*D, etc.
                            weightVDirection = angleVDirection * (1f/_height + 10.0f * distanceVDirection);

                            if (weightVDirection > _localMaxWeightV[j])
                            {
                                _localMaxWeightV[j] = weightVDirection;
                                if (weightVDirection > _maxValue.Y)
                                {
                                    _maxValue.Y = weightVDirection;
                                }
                            }

                            if (angleVDirection > _localMaxAngleV[j])
                            {
                                _localMaxAngleV[j] = angleVDirection;
                            }
                        }

                        _data[index++] = new Vector4(weightUDirection, weightVDirection,
                                                     angleUDirection,  angleVDirection);
                    }
                }
            }
        }

        private void GenerateInverseDistributionFunctions(ParametricSurfaceEffect effect)
        {
            float totalVariationU = GetArraySum(_localMaxWeightU);
            float totalVariationV = GetArraySum(_localMaxWeightV);

            float totalAngleU = GetArraySum(_localMaxAngleU);
            float totalAngleV = GetArraySum(_localMaxAngleV);

            totalVariationU = Float.Sqrt(totalVariationU * totalAngleU);
            totalVariationV = Float.Sqrt(totalVariationV * totalAngleV);

            float[] singularitiesU = effect.GetSingularPointsU();
            float[] singularitiesV = effect.GetSingularPointsV();

            // We only care about singularities which divide the domain space,
            // they don't include endpoints 0 and 1
            singularitiesU = TrimArrayEndpoints(singularitiesU);
            singularitiesV = TrimArrayEndpoints(singularitiesV);

            // NOTE naive heuristic for how much weight singularities add
            // .04f = 1 / 25 where 25 is an average expected min samples per dimension
            //totalVariationU *= (1.0f + .04f * (float)singularitiesU.Length);
            //totalVariationV *= (1.0f + .04f * (float)singularitiesV.Length);

            int uSamplesMin, vSamplesMin;
            ComputeMinimumSamplesUV(totalVariationU, totalVariationV,
                                    out uSamplesMin, out vSamplesMin);

            // NOTE hardcoded            
            int uSamplesPerSingularity = 2 * (int)(((float)(_lookupU.Length - 1)) / (float)uSamplesMin + .5f) + 1;
            int vSamplesPerSingularity = 2 * (int)(((float)(_lookupV.Length - 1)) / (float)vSamplesMin + .5f) + 1;
            
            // Split each value into (value - halfTolerance) and (value + halfTolerance)
            // NOTE assumption - we are limiting ourselves to 500 B-spline samples here
            float halfSingularityTolerance = .5f * (BSplineCurve2D.c_singularityTolerance / 500.0f);
            float[] paddedSingularitiesU = PadSingularities(singularitiesU, uSamplesPerSingularity, halfSingularityTolerance);
            float[] paddedSingularitiesV = PadSingularities(singularitiesV, vSamplesPerSingularity, halfSingularityTolerance);

            int uRegularSamples = _lookupU.Length - singularitiesU.Length * uSamplesPerSingularity;
            int vRegularSamples = _lookupV.Length - singularitiesV.Length * vSamplesPerSingularity;

            GenerateDensityDistributionFunction(_localMaxWeightU,
                                                _lookupU,
                                                uRegularSamples);
            GenerateDensityDistributionFunction(_localMaxWeightV,
                                                _lookupV,
                                                vRegularSamples);
            
            if (paddedSingularitiesU.Length > 0)
            {
                Debug.Assert(uRegularSamples < (_lookupU.Length - 1));
                Debug.Assert(_lookupU[uRegularSamples - 1] == 1.0f);  // Compensate for accumulated error

                _lookupU = MergeSortedArrays(_lookupU,             uRegularSamples,
                                             paddedSingularitiesU, paddedSingularitiesU.Length);
            }

            if (paddedSingularitiesV.Length > 0)
            {
                Debug.Assert(vRegularSamples < (_lookupV.Length - 1));
                Debug.Assert(_lookupV[vRegularSamples - 1] == 1.0f);  // Compensate for accumulated error

                _lookupV = MergeSortedArrays(_lookupV, vRegularSamples,
                                             paddedSingularitiesV, paddedSingularitiesV.Length);
            }

            _maxWeightU = totalVariationU;
            _maxWeightV = totalVariationV;

            /*
            // FOR DEBUGGING ONLY
            for (int n = 0; n < _lookupU.Length; n++)
            {
                _lookupU[n] = Float.Sq(((float)n) / (float)(_lookupU.Length - 1));
            }
            for (int n = 0; n < _lookupV.Length; n++)
            {
                _lookupV[n] = Float.Sq(((float)n) / (float)(_lookupV.Length - 1));
            }*/
        }

        private static float[] TrimArrayEndpoints(float[] source)
        {
            if (source.Length > 0)  // NOTE we assume all singularities are unique
            {
                int destinationLength = source.Length;
                int sourceOffset = 0;
                if (source[0] == 0.0f)
                {
                    destinationLength--;
                    sourceOffset++;
                }
                if (source[source.Length - 1] == 1.0f)
                {
                    destinationLength--;
                }

                float[] dest = new float[destinationLength];
                Array.Copy(source, sourceOffset, dest, 0, destinationLength);
                return dest;
            }
            else
            {
                return source;
            }
        }

        // Split each value into (value - halfTolerance) and (value + halfTolerance)
        private static float[] PadSingularities(float[] source, int samplesPerSingularity, float halfDelta)
        {
            float[] dest = new float[samplesPerSingularity * source.Length];
            int halfSamples = samplesPerSingularity / 2;

            for (int n = 0; n < source.Length; n++)
            {
                int m;
                for (m = 0; m < halfSamples; m++)
                {
                    dest[samplesPerSingularity * n + m] = source[n] - halfDelta;
                }
                for (; m < samplesPerSingularity; m++)
                {
                    dest[samplesPerSingularity * n + m] = source[n] + halfDelta;
                }
            }

            // NOTE should check that we are non-decreasing

            return dest;
        }

        private float[] MergeSortedArrays(float[] source1, int source1length,
                                          float[] source2, int source2length)
        {
            float[] dest = new float[source1length + source2length];

            int source1Index = 0;
            int source2Index = 0;
            int destIndex = 0;

            // Perform array merge
            while (destIndex < dest.Length)
            {
                float nextValue1 = (source1Index < source1length)
                                   ? source1[source1Index]
                                   : float.MaxValue;
                float nextValue2 = (source2Index < source2length)
                                   ? source2[source2Index]
                                   : float.MaxValue;
                Debug.Assert((nextValue1 != float.MaxValue) || (nextValue2 != float.MaxValue));

                if (nextValue1 < nextValue2)
                {
                    dest[destIndex++] = nextValue1;
                    source1Index++;
                }
                else
                {
                    dest[destIndex++] = nextValue2;
                    source2Index++;
                }
            }

            Debug.Assert(destIndex == source1length + source2length);

            return dest;
        }

        public static void ComputeMinimumSamplesUV(float totalVariationU, float totalVariationV,
                                                   out int uSamplesMin, out int vSamplesMin)
        {
            const int guaranteedMinTriangles = 2000;

            ComputeRecommendedSamplesUV(guaranteedMinTriangles, totalVariationU, totalVariationV,
                                        out uSamplesMin, out vSamplesMin);
        }

        public static void ComputeRecommendedSamplesUV(
                                                int numTriangles,
                                                float totalVariationU, float totalVariationV,
                                                out int uSamples, out int vSamples)
        {
            float root = Float.Sqrt(.5f * (float)numTriangles);
            float ratio = Float.Sqrt(totalVariationU / totalVariationV);

            if (!(ratio > 0.0f))  // Handle degenerate cases (e.g. null effect)
            {
                ratio = 1.0f;
            }

            uSamples = (int)(root * ratio + .5f);
            vSamples = (int)(root / ratio + .5f);
        }

        private float GetArraySum(float[] array)
        {
            float sum = 0;
            for (int n = 0; n < array.Length; n++)
            {
                sum += array[n];
            }
            return sum;
        }


        /// <summary>
        /// Given a distribution in histogram form, generate a function that will
        /// map an even domain to an output that matches the histogram density function.
        /// </summary>
        /// <param name="distribution"></param>
        /// <param name="output"></param>
        /// <returns>Total amount of angle change recorded.</returns>
        private float GenerateDensityDistributionFunction(float[] distribution,
                                                          float[] output,
                                                          int outputLength)
        {
            int distributionLength = distribution.Length;

            float totalSum = GetArraySum(distribution);

            // WARNING destructive operation - pad the angle information a little
            /*float avgElement = totalSum / (float)distributionLength;
            for (int n = 0; n < distributionLength; n++)
            {
                distribution[n] += .25f * avgElement;
            }

            // Now we must exactly recompute total sum
            totalSum = 0;
            for (int n = 0; n < distributionLength; n++)
            {
                totalSum += distribution[n];
            }*/

            int index = 0;
            float currentSum = 0;

            for (int n = 0; n < outputLength; n++)
            {
                float fn = ((float)n) / (float)(outputLength - 1);  // n normalized to [0..1] range
                float expectedSum = fn * totalSum;  // Function must progress this far

                float prevSum = currentSum;
                // NOTE: the index range check is only necessary because currentSum may not
                // exactly reach the value of expectedSum due to accumulated error from
                // additions performed in the loop.
                while (index < distributionLength && currentSum < expectedSum)
                {
                    prevSum = currentSum;
                    currentSum += distribution[index++];
                }

                Debug.Assert(expectedSum <= currentSum);
                Debug.Assert((index == 0) || (currentSum - distribution[index - 1]) < expectedSum);

                float fractionalIndex;
                if (currentSum > prevSum)  // We advanced in the distribution table
                {
                    float interpolateFactor = (expectedSum - prevSum) / (currentSum - prevSum);

                    fractionalIndex = MathHelper.Lerp((float)(index - 1), (float)index,
                                                      interpolateFactor);

                    currentSum -= distribution[--index];     // Roll back 1 step to ensure that
                    Debug.Assert(currentSum < expectedSum);  // interpolation will work next time.
                }
                else
                {
                    Debug.Assert(currentSum == prevSum && currentSum == expectedSum);
                    fractionalIndex = (float)index;
                }

                output[n] = fractionalIndex / (float)distributionLength;
            }

            // NOTE: Ensure we cover the full range
            output[outputLength - 1] = 1.0f;

            return totalSum;
        }

        public int Width
        {
            get { return _width; }
        }

        public int Height
        {
            get { return _height; }
        }

        public Vector4[] DataMatrix
        {
            get { return _data; }
        }

        public Vector4 DataMatrixMaxValue
        {
            get { return _maxValue; }
        }

        public float[] DataVectorX
        {
            get { return _lookupU; }
        }

        public float[] DataVectorY
        {
            get { return _lookupV; }
        }

        public float DataVectorXTotalVariation
        {
            get { return _maxWeightU; }
        }

        public float DataVectorYTotalVariation
        {
            get { return _maxWeightV; }
        }


        private int _width, _height;
        private Vector4[] _data;
        private Vector4 _maxValue;

        private float _maxWeightU;
        private float _maxWeightV;

        private float[] _localMaxWeightU;
        private float[] _localMaxWeightV;

        private float[] _localMaxAngleU;
        private float[] _localMaxAngleV;

        private float[] _lookupU = new float[ParametricModel3DVisualizer.c_domainWarpLookupTableSize / 2];
        private float[] _lookupV = new float[ParametricModel3DVisualizer.c_domainWarpLookupTableSize / 2];
    }
}
