//Written by Leonid Velikovich and Benjamin Hernandez Arreguin

//+-----------------------------------------------------------------------------
//
//  Enum ParametricEffectType.  Enumerates 3D operations we can perform on 2D curves.
//
//  Class ParametricSurfaceEffect.  Base class for a software-based
//  parametric surface implementation.  It also emulates a subset of the
//  Microsoft.Xna.Framework.Graphics.Effect API for easier interchangeability
//  with a hardware effect.
//
//  Likewise, SoftwareEffectParameter class emulates EffectParameter.  This ad-hoc
//  emulation will probably go away.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;


namespace ParametricModelEditor
{
    public enum ParametricEffectType
    {
        None = 0,
        SurfaceOfRevolution = 1,
        ProfileProduct = 2,
        WireProduct = 3,
    }

    public abstract class ParametricSurfaceEffect
    {
        public class SoftwareEffectParameter
        {
            private SoftwareEffectParameter() { }

            public SoftwareEffectParameter(ParametricSurfaceEffect host, string name)
            {
                _host = host;
                _name = name;
            }

            public void SetValue(Matrix value)
            {
                switch (_name)
                {
                    case "World":
                        _host.World = value;
                        break;
                    case "View":
                        _host.View = value;
                        break;
                    case "Projection":
                        _host.Projection = value;
                        break;
                }
            }

            private ParametricSurfaceEffect _host;
            private string _name;
        }


        public ParametricSurfaceEffect()
        {
            Parameters.Add("World", new SoftwareEffectParameter(this, "World"));
            Parameters.Add("View", new SoftwareEffectParameter(this, "View"));
            Parameters.Add("Projection", new SoftwareEffectParameter(this, "Projection"));
        }

        public ParametricEffectType ParametricEffectType
        {
            get { return _parametricEffectType; }
        }

        // Actual surface function
        // Returns {x,y,z}, {x',y',z'}, {x",y",z"} at {u,v}
        public abstract void VS_SurfaceFunction(Vector2 input, out Vector3 position,
                                                               out Vector3 dF_du,
                                                               out Vector3 dF_dv,
                                                               out Vector3 ddF_duu,
                                                               out Vector3 ddF_dvv);

        public virtual float[] GetSingularPointsU() { return null; }
        public virtual float[] GetSingularPointsV() { return null; }

        public virtual Vector2[] GetControlPointsU() { return null; }
        public virtual Vector2[] GetControlPointsV() { return null; }

        // Used for serialization
        public abstract List<ParametricCurve2D> GetSerialization();
        public abstract void SetSerialization(List<ParametricCurve2D> list);


        public Matrix World;
        public Matrix View;
        public Matrix Projection;

        public Dictionary<String, SoftwareEffectParameter> Parameters = new Dictionary<string, SoftwareEffectParameter>();

        protected ParametricEffectType _parametricEffectType = ParametricEffectType.None;

        protected const float PI = 3.141592654f;
        protected const float TWOPI = 6.283185307f;
    }
}
