//+-----------------------------------------------------------------------------
//
// ParametricSurfaceEffect shader.  Allows either software-based or GPU/HLSL-
// based parametric surface evaluation and domain surface warping.
//
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
//
// Global definitions and variables
//
//------------------------------------------------------------------------------

const float    PI = 3.141592654;
const float TWOPI = 6.283185307;


//------------------------------------------------------------------------------
//
// BSpline 2D curve definitions
//
//------------------------------------------------------------------------------

struct PARAMETRICCURVE2D_OUTPUT
{
    float2 Position     : POSITION;         // 2D position on the curve
    float2 Tangent      : TANGENT;          // 1st derivative at that position
    float2 Acceleration;                    // 2nd derivative at that position
};

const float BSpline_TextureSizeInverse = .0078125;  // = 1/128
const float BSpline_SingularityTolerance = .01;

float BSpline1_ControlPointCount;
float BSpline2_ControlPointCount;
float BSpline2_ControlPointOffset;  // Where the coordinates for the 2nd BSpline begin in the lookup texture.

texture2D BSpline_ControlPointTexture;

sampler1D BSpline_ControlPointSampler = sampler_state
{
    Texture = (BSpline_ControlPointTexture);
    
    MinFilter = None;
    MagFilter = None;
    AddressU  = Clamp;
    AddressV  = Clamp;
};

const matrix BSpline_BaseMatrix = matrix(-1,  3, -3,  1,
                                          3, -6,  3,  0,
                                         -3,  0,  3,  0,
                                          1,  4,  1,  0) / 6;


//------------------------------------------------------------------------------
//
// BSpline 2D curve methods
//
//------------------------------------------------------------------------------

// [t] ---> [x,y, x',y', x",y"]
PARAMETRICCURVE2D_OUTPUT BSpline_Evaluate( const float input,
                                           const float samplerOffset,
                                           const float controlPointCount)
{
    float scaledT = float(controlPointCount - 3) * input;
    float currentSegment = floor(scaledT);
    
    // NOTE skipping if(t==1) {currentSegment--;}
    // Prevent bad values by assuming an extra control point at the end
    
    float index0 = samplerOffset + currentSegment * BSpline_TextureSizeInverse;
    float index1 = index0 +     BSpline_TextureSizeInverse;
    float index2 = index0 + 2 * BSpline_TextureSizeInverse;
    float index3 = index0 + 3 * BSpline_TextureSizeInverse;
    
    float2 c0 = tex1Dlod(BSpline_ControlPointSampler, index0).xy;
    float2 c1 = tex1Dlod(BSpline_ControlPointSampler, index1).xy;
    float2 c2 = tex1Dlod(BSpline_ControlPointSampler, index2).xy;
    float2 c3 = tex1Dlod(BSpline_ControlPointSampler, index3).xy;
    
    float relativeT = frac(scaledT);  // relativeT is in [0..1] range
    
    // GPU-optimized singularity treatment: to avoid dynamic branching and testing
    // control points for the occurence of a singularity, simply clamp relative T
    // to the [0+delta, 1-delta] range.
    relativeT = clamp(relativeT, BSpline_SingularityTolerance, 1-BSpline_SingularityTolerance);
    
    float relativeT2 = relativeT * relativeT;
    float relativeT3 = relativeT2 * relativeT;
    
    matrix controlMatrix = matrix(c0.xy, 0, 0,
                                  c1.xy, 0, 0,
                                  c2.xy, 0, 0,
                                  c3.xy, 0, 0);
                           
    matrix combinedMatrix = mul(BSpline_BaseMatrix, controlMatrix);
    
    PARAMETRICCURVE2D_OUTPUT output;
    
    float4 tVector0 = float4(relativeT3, relativeT2, relativeT, 1);
    output.Position = mul(tVector0, combinedMatrix);
    
    float4 tVector1 = float4(3 * relativeT2, 2 * relativeT, 1, 0);
    output.Tangent  = mul(tVector1, combinedMatrix);
    
    float4 tVector2 = float4(6 * relativeT, 2, 0, 0);
    output.Acceleration = mul(tVector2, combinedMatrix);
    
    return output;
}


//------------------------------------------------------------------------------
//
// 3D parametric surface definitions
//
//------------------------------------------------------------------------------

struct PARAMETRICSURFACE3D_OUTPUT
{
    float3 Position     : POSITION;         // 3D position on the plane
    float3 dF_du        : TANGENT;          // 1st-order partials
    float3 dF_dv        : TANGENT;
    float3 ddF_duu;                         // 2nd-order partials
    float3 ddF_dvv;
};


//------------------------------------------------------------------------------
//
// 3D parametric surface methods
//
//------------------------------------------------------------------------------

PARAMETRICSURFACE3D_OUTPUT ProfileProduct_Evaluate( const float2 input )
{    
    PARAMETRICCURVE2D_OUTPUT profileCurve = BSpline_Evaluate( input.x,
                                                              0,
                                                              BSpline1_ControlPointCount);                                                               
    profileCurve.Position *= .004;
  
    PARAMETRICCURVE2D_OUTPUT crossSectionCurve = BSpline_Evaluate( input.y,
                                                                   BSpline2_ControlPointOffset,
                                                                   BSpline2_ControlPointCount);
    crossSectionCurve.Position *= .01;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = crossSectionCurve.Position.x * profileCurve.Position.x;
    output.Position.y = crossSectionCurve.Position.y * profileCurve.Position.x;
    output.Position.z = .7 - profileCurve.Position.y;
    
    //
    // Compute 1st-order partials
    //
    output.dF_du.x = crossSectionCurve.Position.x * profileCurve.Tangent.x;
    output.dF_dv.x = crossSectionCurve.Tangent.x  * profileCurve.Position.x;
    
    output.dF_du.y = crossSectionCurve.Position.y * profileCurve.Tangent.x;
    output.dF_dv.y = crossSectionCurve.Tangent.y  * profileCurve.Position.x;
    
    output.dF_du.z = -profileCurve.Tangent.y;
    output.dF_dv.z = 0;
    
    //
    // Compute 2nd-order partials
    //
    output.ddF_duu.x = crossSectionCurve.Position.x * profileCurve.Acceleration.x;
    output.ddF_dvv.x = crossSectionCurve.Acceleration.x * profileCurve.Position.x;

    output.ddF_duu.y = crossSectionCurve.Position.y * profileCurve.Acceleration.x;
    output.ddF_dvv.y = crossSectionCurve.Acceleration.y * profileCurve.Position.x;

    output.ddF_duu.z = -profileCurve.Acceleration.y;
    output.ddF_dvv.z = 0;    
    
    return output;
}


PARAMETRICSURFACE3D_OUTPUT SurfaceOfRevolution_Evaluate( const float2 input )
{    
    PARAMETRICCURVE2D_OUTPUT profileCurve = BSpline_Evaluate( input.x,
                                                              0,
                                                              BSpline1_ControlPointCount);                                                               
    profileCurve.Position *= .004;
  
    float theta = TWOPI * input.y;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = cos(theta) * profileCurve.Position.x;
    output.Position.y =-sin(theta) * profileCurve.Position.x;
    output.Position.z = .7 - profileCurve.Position.y;
    
    //
    // Compute 1st-order partials
    //
    output.dF_du.x = cos(theta) * profileCurve.Tangent.x;
    output.dF_dv.x = -TWOPI * sin(theta) * profileCurve.Position.x;
    
    output.dF_du.y = -sin(theta) * profileCurve.Tangent.x;
    output.dF_dv.y = -TWOPI * cos(theta) * profileCurve.Position.x;
    
    output.dF_du.z = -profileCurve.Tangent.y;
    output.dF_dv.z = 0;
    
    //
    // Compute 2nd-order partials
    //
    output.ddF_duu.x = cos(theta) * profileCurve.Acceleration.x;
    output.ddF_dvv.x = -TWOPI * TWOPI * cos(theta) * profileCurve.Position.x;

    output.ddF_duu.y = -sin(theta) * profileCurve.Acceleration.x;
    output.ddF_dvv.y = TWOPI * TWOPI * sin(theta) * profileCurve.Position.x;

    output.ddF_duu.z = -profileCurve.Acceleration.y;
    output.ddF_dvv.z = 0;    
    
    return output;
}


PARAMETRICSURFACE3D_OUTPUT WireProduct_Evaluate( const float2 input )
{    
    PARAMETRICCURVE2D_OUTPUT wireCurve = BSpline_Evaluate( input.x,
                                                           0,
                                                           BSpline1_ControlPointCount);                                                               
    wireCurve.Position *= .005;
  
    PARAMETRICCURVE2D_OUTPUT crossSectionCurve = BSpline_Evaluate( input.y,
                                                                   BSpline2_ControlPointOffset,
                                                                   BSpline2_ControlPointCount);
    crossSectionCurve.Position *= .0005;
    
    float2 normVector = normalize(float2(-wireCurve.Tangent.y, -wireCurve.Tangent.x));
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = wireCurve.Position.x + normVector.x * crossSectionCurve.Position.x;
    output.Position.y =-wireCurve.Position.y + normVector.y * crossSectionCurve.Position.x;
    output.Position.z =-crossSectionCurve.Position.y;
    
    //
    // Compute 1st-order partials
    //
    output.dF_du.x = wireCurve.Tangent.x;  // + d/dU(normVector.X) * crossSectionCurvePoint2D.X;
    output.dF_dv.x = normVector.x * crossSectionCurve.Tangent.x;
    
    output.dF_du.y =-wireCurve.Tangent.y;  // + d/dU(normVector.Y) * crossSectionCurvePoint2D.X;
    output.dF_dv.y = normVector.y * crossSectionCurve.Tangent.x;
    
    output.dF_du.z = 0;
    output.dF_dv.z = -crossSectionCurve.Tangent.y;
    
    //
    // Compute 2nd-order partials
    //
    output.ddF_duu.x = wireCurve.Acceleration.x;  // + d/dU( d/dU(normVector.X) * crossSectionCurvePoint2D.X );
    output.ddF_dvv.x = normVector.x * crossSectionCurve.Acceleration.x;

    output.ddF_duu.y =-wireCurve.Acceleration.y;  // + d/DU( d/dU(normVector.Y) * crossSectionCurvePoint2D.X );
    output.ddF_dvv.y = normVector.y * crossSectionCurve.Acceleration.x;

    output.ddF_duu.z = 0;
    output.ddF_dvv.z = -crossSectionCurve.Acceleration.y;    
    
    return output;
}


//------------------------------------------------------------------------------
//
// Vertex Shader - definitions and variables
//
//------------------------------------------------------------------------------

// TODO remove default values so everything is always set via software
float3 LightVector = -normalize(float3(.8, .7, -1));
float4 MaterialColor = float4(1.00, .27, .20, 1);

float AmbientFactor = 0.05f;
float DiffuseFactor = 0.9f;
float SpecularFactor = 4.0f;
float SpecularPower = 16;


matrix World;
matrix View;
matrix Projection;

float3 EyeLocation;

struct VSHARDWARE_INPUT
{
    float2 Position : POSITION;         // Position
};

struct VSPOINT_INPUT
{
    float3 Position : POSITION;         // Position
    float3 Normal   : NORMAL;           // Normal
};

struct PS_INPUT_HARDWARE
{
    float4 Position  : POSITION0;        // Screen space
    float2 PositionP : POSITION1;        // Parametric space
};

struct PS_INPUT_SOFTWARE
{
    float4 Position  : POSITION0;        // Screen space
    float4 PositionW : POSITION1;        // World space
    float3 Normal    : NORMAL;
    float4 Color     : COLOR;
};

// Forward declaration
PS_INPUT_HARDWARE VS_Hardware(const PARAMETRICSURFACE3D_OUTPUT surfacePoint, float2 domainPoint);


texture2D Surface_WarpFunctionTexture;

sampler1D Surface_WarpFunctionSampler = sampler_state
{
    Texture = (Surface_WarpFunctionTexture);
    
    MinFilter = None;
    MagFilter = None;
    MipFilter = None;
    AddressU  = Clamp;   // NOTE per NVIDIA GPU Programming Guide P.48, in DX9.0c
    AddressV  = Clamp;   // wrapping/mipmapping capability is not exposed on floats.
};

const float Surface_WrapTextureSize = 256;
const float Surface_WrapTextureSizeInverse = .00390625;  // = 1/256

// Scaling variable to achieve dynamic LOD.
float2 Surface_DomainScalingFactor = float2(1, 1);

// Variable in [0..1] range.  1 means use maximum geometric detail at current LOD resolution.
// 0 means pretend the current resolution is a lower power of 2, to smoothly blend with lower LOD.
float Surface_DomainLodFraction = 1.0;

//------------------------------------------------------------------------------
//
// Vertex Shader - hardware parametrization
//
//------------------------------------------------------------------------------
float2 Surface_WarpCoordinates(float2 input)
{
    float2 output;

    // Clamp input.x to [0, .5) and input.y to [.5, 1) for correct texture lookup
    input = clamp(.5 * input * Surface_DomainScalingFactor, 0, .5 - Surface_WrapTextureSizeInverse) + float2(0, .5);

    // Perform linearly interpolated sampling of an array element

    float sample1 = tex1Dlod(Surface_WarpFunctionSampler, input.x).x;
    //float sample2 = sample1;
    float sample2 = tex1Dlod(Surface_WarpFunctionSampler, input.x + Surface_WrapTextureSizeInverse).x;

    float sample3 = tex1Dlod(Surface_WarpFunctionSampler, input.y).x;
    //float sample4 = sample3;
    float sample4 = tex1Dlod(Surface_WarpFunctionSampler, input.y + Surface_WrapTextureSizeInverse).x;

    output.x = lerp(sample1, sample2, frac(Surface_WrapTextureSize * input.x));
    output.y = lerp(sample3, sample4, frac(Surface_WrapTextureSize * input.y));

    return output;
}


PS_INPUT_HARDWARE VS_Hardware_ProfileProduct(const VSHARDWARE_INPUT input)
{
    float2 warpedPosition = Surface_WarpCoordinates(input.Position);

    PARAMETRICSURFACE3D_OUTPUT surfacePoint = ProfileProduct_Evaluate(warpedPosition);
    
    return VS_Hardware(surfacePoint, warpedPosition);
}


PS_INPUT_HARDWARE VS_Hardware_ProfileProduct_Geomorph1(VSHARDWARE_INPUT input)
{
    float2 point   = input.Position;

    // 0.0 ->   0
    // 0.5 -> +.5
    // 1.0 ->   0
    // 1.5 -> -.5
    // 2.0 ->   0
    float2 fracPart = frac(input.Position);
    float2 fracMove = 2 - 4 * frac(.5 * input.Position);
    float2 fracTotal = fracPart * fracMove;

    float2 morphedInput = input.Position + fracTotal * (1 - Surface_DomainLodFraction);
    
    float2 warpedPosition = Surface_WarpCoordinates(morphedInput);

    PARAMETRICSURFACE3D_OUTPUT surfacePoint = ProfileProduct_Evaluate(warpedPosition);
    
    return VS_Hardware(surfacePoint, warpedPosition);
}


PS_INPUT_HARDWARE VS_Hardware_ProfileProduct_Geomorph2(VSHARDWARE_INPUT input)
{
    float2 point   = input.Position;

    // For points with odd indices, find the two even-indexed points on whose line
    // (possibly diagonal) our odd point initially lies on.
    float2 point00 = float2(floor(point.x), ceil(point.y));
    float2 point11 = float2(ceil(point.x), floor(point.y));

    float2 warpedPoint = Surface_WarpCoordinates(point);
    PARAMETRICSURFACE3D_OUTPUT surface = ProfileProduct_Evaluate(warpedPoint);

    float2 warpedPoint00 = Surface_WarpCoordinates(point00);
    PARAMETRICSURFACE3D_OUTPUT surface00 = ProfileProduct_Evaluate(warpedPoint00);
    
    float2 warpedPoint11 = Surface_WarpCoordinates(point11);
    PARAMETRICSURFACE3D_OUTPUT surface11 = ProfileProduct_Evaluate(warpedPoint11);
    
    float weightCenter = Surface_DomainLodFraction;
    float weightSide = .5 * (1 - Surface_DomainLodFraction);

    float2 averagePoint = weightCenter * warpedPoint + weightSide * (warpedPoint00 + warpedPoint11);

    PARAMETRICSURFACE3D_OUTPUT averageSurface;
    averageSurface.Position = weightCenter * surface.Position
                            + weightSide   *(surface00.Position + surface11.Position);
    averageSurface.dF_du = weightCenter * surface.dF_du
                         + weightSide   *(surface00.dF_du + surface11.dF_du);
    averageSurface.dF_dv = weightCenter * surface.dF_dv
                         + weightSide   *(surface00.dF_dv + surface11.dF_dv);
    averageSurface.ddF_duu = surface.ddF_duu;
    averageSurface.ddF_dvv = surface.ddF_dvv;  // NOTE shortcut - not weighing the acceleration.

    return VS_Hardware(averageSurface, averagePoint);
}


PS_INPUT_HARDWARE VS_Hardware_ProfileProduct_Geomorph3(VSHARDWARE_INPUT input)
{
    float2 point   = input.Position;

    // For points with odd indices, find the two even-indexed points on whose line
    // (possibly diagonal) our odd point initially lies on.
    float2 point00 = float2(floor(point.x), ceil(point.y));
    float2 point11 = float2(ceil(point.x), floor(point.y));

    float2 warpedPoint = Surface_WarpCoordinates(point);
    PARAMETRICSURFACE3D_OUTPUT surface = ProfileProduct_Evaluate(warpedPoint);

    float2 warpedPoint00 = Surface_WarpCoordinates(point00);
    PARAMETRICSURFACE3D_OUTPUT surface00 = ProfileProduct_Evaluate(warpedPoint00);
    
    float2 warpedPoint11 = Surface_WarpCoordinates(point11);
    PARAMETRICSURFACE3D_OUTPUT surface11 = ProfileProduct_Evaluate(warpedPoint11);
    
    float weightCenter = Surface_DomainLodFraction;
    float weightSides = 1 - Surface_DomainLodFraction;
    
    // Drop the altitude in the triangle <surface00, surface11, surface> and see how
    // close the altitude falls between surface00 and surface11.  This is the weight
    // we will use in our interpolation scheme.
    float2 vector00_11 = surface11.Position - surface00.Position;
    float2 vector00_CC = surface.Position - surface00.Position;
    float weight11 = weightSides * saturate(dot(vector00_11, vector00_CC) / dot(vector00_11, vector00_11));
    float weight00 = weightSides - weight11;

    float2 averagePoint = weightCenter * warpedPoint
                        + weight00 * warpedPoint00
                        + weight11 * warpedPoint11;

    PARAMETRICSURFACE3D_OUTPUT averageSurface;
    averageSurface.Position = weightCenter * surface.Position
                            + weight00 * surface00.Position
                            + weight11 * surface11.Position;
    averageSurface.dF_du = weightCenter * surface.dF_du
                         + weight00 * surface00.dF_du
                         + weight11 * surface11.dF_du;
    averageSurface.dF_dv = weightCenter * surface.dF_dv
                         + weight00 * surface00.dF_dv
                         + weight11 * surface11.dF_dv;
    averageSurface.ddF_duu = surface.ddF_duu;
    averageSurface.ddF_dvv = surface.ddF_dvv;  // NOTE shortcut - not weighing the acceleration.

    return VS_Hardware(averageSurface, averagePoint);
}


PS_INPUT_HARDWARE VS_Hardware_SurfaceOfRevolution(const VSHARDWARE_INPUT input)
{
    float2 warpedPosition = Surface_WarpCoordinates(input.Position);

    PARAMETRICSURFACE3D_OUTPUT surfacePoint = SurfaceOfRevolution_Evaluate(warpedPosition);
    
    return VS_Hardware(surfacePoint, warpedPosition);
}


PS_INPUT_HARDWARE VS_Hardware_WireProduct(const VSHARDWARE_INPUT input)
{
    float2 warpedPosition = Surface_WarpCoordinates(input.Position);

    PARAMETRICSURFACE3D_OUTPUT surfacePoint = WireProduct_Evaluate(warpedPosition);
    
    return VS_Hardware(surfacePoint, warpedPosition);
}


PS_INPUT_HARDWARE VS_Hardware(const PARAMETRICSURFACE3D_OUTPUT surfacePoint, float2 domainPoint)
{
    float4 positionWorld = mul( float4(surfacePoint.Position, 1), World );
    float4 positionView  = mul( positionWorld, View );
    float4 positionScreen= mul( positionView, Projection );
    
    PS_INPUT_HARDWARE output;
    output.Position  = positionScreen;
    output.PositionP = domainPoint;
    
    return output;
}


//------------------------------------------------------------------------------
//
// Vertex Shader - software parametrization
//
//------------------------------------------------------------------------------
PS_INPUT_SOFTWARE VS_Software(const VSPOINT_INPUT input)
{
    float4 positionWorld = mul( float4(input.Position, 1), World );
    float4 positionView  = mul( positionWorld, View );
    float4 positionScreen= mul( positionView, Projection );
    
    PS_INPUT_SOFTWARE output;
   output.Position  = positionScreen;
    output.PositionW = positionWorld;
    output.Normal = normalize(mul( input.Normal, World ));
    
    output.Color = MaterialColor;
    /*  Specifically for rose model
    float blendingFactor = (input.Position.z - 0.025)/.25;
    float4 redTone = lerp(float4(1.30, .87, .30, 1), float4(1.00, .27, .20, 1), blendingFactor);
    output.Color = redTone * step(.025, input.Position.z)
                 + .5f * float4(.20, 1.00, .27, 1) * (1 - step(.025, input.Position.z));
                 */
    
    return output;
}


//------------------------------------------------------------------------------
//
// Pixel Shader - hardware parametrization
//
//------------------------------------------------------------------------------
float4 PS_Hardware(const PS_INPUT_HARDWARE input, const PARAMETRICSURFACE3D_OUTPUT surfacePoint) : COLOR0
{
    float3 norm = normalize(cross(surfacePoint.dF_du, surfacePoint.dF_dv));
    norm = mul( norm, World );
    
    float fDiffuse = saturate(.5 + .5 * dot(norm, LightVector ) );
    
    float3 eyeVector = normalize(EyeLocation - surfacePoint.Position);
    float3 halfwayVector = normalize(eyeVector + LightVector);
    
    float fSpecular = pow(saturate(dot(halfwayVector, norm)), SpecularPower);
    
    float fTotalIntensity = AmbientFactor
                          + DiffuseFactor  * fDiffuse
                          + SpecularFactor * fSpecular;
    float4 outputColor = MaterialColor * fTotalIntensity;
    
    return outputColor;
}

float4 PS_Hardware_ProfileProduct(const PS_INPUT_HARDWARE input) : COLOR0
{
    PARAMETRICSURFACE3D_OUTPUT surfacePoint = ProfileProduct_Evaluate( input.PositionP );

    return PS_Hardware(input, surfacePoint);
}

float4 PS_Hardware_SurfaceOfRevolution(const PS_INPUT_HARDWARE input) : COLOR0
{
    PARAMETRICSURFACE3D_OUTPUT surfacePoint = SurfaceOfRevolution_Evaluate( input.PositionP );

    return PS_Hardware(input, surfacePoint);
}

float4 PS_Hardware_WireProduct(const PS_INPUT_HARDWARE input) : COLOR0
{
    PARAMETRICSURFACE3D_OUTPUT surfacePoint = WireProduct_Evaluate( input.PositionP );

    return PS_Hardware(input, surfacePoint);
}


//------------------------------------------------------------------------------
//
// Pixel Shader - software parametrization
//
//------------------------------------------------------------------------------
float4 PS_Software(const PS_INPUT_SOFTWARE input) : COLOR0
{
    float3 norm = normalize(input.Normal);
    
    float fDiffuse = saturate(.5 + .5 * dot(norm, LightVector ) );
    
    float3 eyeVector = normalize(EyeLocation - input.PositionW);
    float3 halfwayVector = normalize(eyeVector + LightVector);
    
    float fSpecular = pow(saturate(dot(halfwayVector, norm)), SpecularPower);
    
    float fTotalIntensity = AmbientFactor
                          + DiffuseFactor  * fDiffuse
                          + SpecularFactor * fSpecular;
   float4 outputColor = input.Color * fTotalIntensity;
    
    return outputColor;
}


//------------------------------------------------------------------------------
//
// Shader Techniques
//
//------------------------------------------------------------------------------
technique SoftwareTechnique
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Software();
        PixelShader  = compile ps_3_0 PS_Software();
        //SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique HardwareTechnique_ProfileProduct
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Hardware_ProfileProduct();
        PixelShader  = compile ps_3_0 PS_Hardware_ProfileProduct();
    }
}

technique HardwareTechnique_ProfileProduct_Geomorph1
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Hardware_ProfileProduct_Geomorph1();
        PixelShader  = compile ps_3_0 PS_Hardware_ProfileProduct();
    }
}

technique HardwareTechnique_ProfileProduct_Geomorph2
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Hardware_ProfileProduct_Geomorph2();
        PixelShader  = compile ps_3_0 PS_Hardware_ProfileProduct();
    }
}

technique HardwareTechnique_ProfileProduct_Geomorph3
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Hardware_ProfileProduct_Geomorph3();
        PixelShader  = compile ps_3_0 PS_Hardware_ProfileProduct();
    }
}

technique HardwareTechnique_SurfaceOfRevolution
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Hardware_SurfaceOfRevolution();
        PixelShader  = compile ps_3_0 PS_Hardware_SurfaceOfRevolution();
    }
}

technique HardwareTechnique_WireProduct
{
    Pass P0
    {
        VertexShader = compile vs_3_0 VS_Hardware_WireProduct();
        PixelShader  = compile ps_3_0 PS_Hardware_WireProduct();
    }
}


