//+-----------------------------------------------------------------------------
//
// ProfileProductEffect class.  Software implementation for profile product effect.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;


namespace ParametricModelEditor
{
    public class ProfileProductEffect : ParametricSurfaceEffect
    {
        public ProfileProductEffect(ParametricCurve2D profileCurve, ParametricCurve2D crossSectionCurve)
        {
            Debug.Assert(profileCurve != null && crossSectionCurve != null);            
            _profileCurve = profileCurve;
            _crossSectionCurve = crossSectionCurve;

            _parametricEffectType = ParametricEffectType.ProfileProduct;
        }

        // input.X is the B-Spline's parametric t parameter
        // input.Y is the theta angle parameter
        public override void VS_SurfaceFunction(Vector2 input, out Vector3 position,
                                                               out Vector3 dF_du,
                                                               out Vector3 dF_dv,
                                                               out Vector3 ddF_duu,
                                                               out Vector3 ddF_dvv)
        {
            //
            // Compute surface of revolution parameters: curve t and angle theta
            //
            Vector2 profileCurvePosition2D, profileCurveTangent2D, profileCurveAcceleration2D;
            _profileCurve.Evaluate(input.X, out profileCurvePosition2D,
                                            out profileCurveTangent2D,
                                            out profileCurveAcceleration2D);
            profileCurvePosition2D *= .004f;  // Scaled to small size

            Vector2 crossSectionCurvePosition2D, crossSectionCurveTangent2D, crossSectionCurveAcceleration2D;
            _crossSectionCurve.Evaluate(input.Y, out crossSectionCurvePosition2D,
                                                 out crossSectionCurveTangent2D,
                                                 out crossSectionCurveAcceleration2D);
            crossSectionCurvePosition2D *= .01f;

            //
            // Compute the position
            //
            position.X = crossSectionCurvePosition2D.X * profileCurvePosition2D.X;
            position.Y = crossSectionCurvePosition2D.Y * profileCurvePosition2D.X;
            position.Z = .7f - profileCurvePosition2D.Y;

            //
            // Compute partials
            //
            dF_du.X = crossSectionCurvePosition2D.X * profileCurveTangent2D.X;
            dF_dv.X = crossSectionCurveTangent2D.X * profileCurvePosition2D.X;

            dF_du.Y = crossSectionCurvePosition2D.Y * profileCurveTangent2D.X;
            dF_dv.Y = crossSectionCurveTangent2D.Y * profileCurvePosition2D.X;

            dF_du.Z = -profileCurveTangent2D.Y;
            dF_dv.Z = 0;

            //
            // Compute 2nd-order partials
            //
            ddF_duu.X = crossSectionCurvePosition2D.X * profileCurveAcceleration2D.X;
            ddF_dvv.X = crossSectionCurveAcceleration2D.X * profileCurvePosition2D.X;

            ddF_duu.Y = crossSectionCurvePosition2D.Y * profileCurveAcceleration2D.X;
            ddF_dvv.Y = crossSectionCurveAcceleration2D.Y * profileCurvePosition2D.X;

            ddF_duu.Z = -profileCurveAcceleration2D.Y;
            ddF_dvv.Z = 0;
        }

        public override float[] GetSingularPointsU()
        {
            return _profileCurve.GetSingularPoints();
        }

        public override float[] GetSingularPointsV()
        {
            return _crossSectionCurve.GetSingularPoints();
        }

        public override Vector2[] GetControlPointsU()
        {
            return _profileCurve.GetControlPoints();
        }

        public override Vector2[] GetControlPointsV()
        {
            return _crossSectionCurve.GetControlPoints();
        }


        public override List<ParametricCurve2D> GetSerialization()
        {
            List<ParametricCurve2D> list = new List<ParametricCurve2D>();
            list.Add(_profileCurve.GetSerialization());
            list.Add(_crossSectionCurve.GetSerialization());
            return list;
        }


        public override void SetSerialization(List<ParametricCurve2D> list)
        {
            Debug.Assert(list.Count >= 1);
            _profileCurve = list[0];
            _crossSectionCurve = list[1];
        }


        private ParametricCurve2D _profileCurve, _crossSectionCurve;
    }
}
