//+-----------------------------------------------------------------------------
//
// SplineControlPointControl class.  The circular dot visualization of B-Spline
// control points visible inside the SplineEditorGridControl.  Draggable.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;

namespace ParametricModelEditor
{
    [Description("Spline Control Point Control")]
    public class SplineControlPointControl : System.Windows.Forms.Control
    {
        public SplineControlPointControl(SplineEditorGridControl parent, int weight)
        {
            SetStyle(ControlStyles.SupportsTransparentBackColor, true);
            this.BackColor = Color.Transparent;
            this.Width = 7;
            this.Height = 7;

            _weight = weight;

            _parent = parent;
        }

        public Vector2 Center
        {
            get
            {
                return _center;
            }
            set
            {
                _center = value;
            }
        }

        public int Weight
        {
            get { return _weight; }
        }


        [Description("Point color")]
        [Category("Appearance")]
        [DefaultValue(typeof(Color), "System.Drawing.Color.Yellow")]
        public Color Color
        {
            get { return _color[_weight - 1]; }
            set { _color[_weight - 1] = value; }
        }


        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if (e.Button == MouseButtons.Left)
            {
                _mousePressed = true;
                _mouseMoved = false;
                _mousePosition.X = this.Location.X + e.X;
                _mousePosition.Y = this.Location.Y + e.Y;
            }
            else if (e.Button == MouseButtons.Right)
            {
                DecrementWeight();
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if (_mousePressed && (e.Button & MouseButtons.Left) != 0)
            {
                _mouseMoved = true;

                int xParentCoords = this.Location.X + e.X;
                int yParentCoords = this.Location.Y + e.Y;

                _parent.SetControlPointLocation(this, xParentCoords, yParentCoords);
                
                _mousePosition.X = xParentCoords;
                _mousePosition.Y = yParentCoords;

                Invalidate();
                Update();

                _parent.OnControlPointMoved(this);
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            if (_mousePressed)
            {
                _mousePressed = false;

                if (!_mouseMoved)
                {
                    if (e.Button == MouseButtons.Left)
                    {
                        if (Control.ModifierKeys == Keys.Control)
                        {
                            _parent.InsertControlPoint(this, this.Center.X, this.Center.Y, 1);
                        }
                        else
                        {
                            IncrementWeight();
                        }
                    }
                }
                else
                {
                    Invalidate();
                    Update();

                    _parent.OnControlPointMoved(this);
                }
            }
        }

        private void DecrementWeight()
        {
            if (0 == --_weight)  // If we are down to 0 points, remove the control
            {
                _weight = 1;  // This is the weight we had before removing ourselves
                _parent.OnControlPointRemoved(this);

                Control parent = Parent;
                parent.SuspendLayout();
                parent.Controls.Remove(this);
                parent.ResumeLayout(false);
                parent.PerformLayout();
                this.Dispose();
            }
            else
            {
                _parent.OnControlPointWeightDecremented(this);

                Invalidate();
                Update();
            }
        }

        private void IncrementWeight()
        {
            if (_weight < 3)
            {
                _weight++;

                _parent.OnControlPointWeightIncremented(this);

                Invalidate();
                Update();
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            if (_weight > 0)  // Prevent reentrancy when we destroy the point but try to redraw
            {
                Graphics g = e.Graphics;
                g.Clear(BackColor);
                Draw(g);
            }
        }

        protected virtual void Draw(Graphics g)
        {
            SolidBrush brush = new SolidBrush(_color[_weight - 1]);

            RectangleF rect = new RectangleF(.5f, .5f, Size.Width-1, Size.Height-1);

            g.FillEllipse(brush, rect);

            brush.Dispose();
        }

        private SplineEditorGridControl _parent;


        private Color[] _color = {System.Drawing.Color.Yellow,
                                  System.Drawing.Color.Orange,
                                  System.Drawing.Color.Red};
        private int _weight;
        protected Vector2 _center;


        // Mouse tracking
        protected System.Drawing.Point _mousePosition;
        protected bool _mousePressed, _mouseMoved;

    }
}
