//+-----------------------------------------------------------------------------
//
// SplineEditorGridControl class.  A control that displays a grid and allows
// the user to add and manipulate B-Spline control points, with the resulting
// B-Spline curve drawn on top of the grid.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;

namespace ParametricModelEditor
{
    public interface ISplineControlPointHost
    {
        void InvalidateIfNotBusy();
        void CreateControlPoint(float x, float y, int weight);
        void InsertControlPoint(SplineControlPointControl previousControl, float x, float y, int weight);
    }

    [Description("Spline Editor Grid Control")]
    public class SplineEditorGridControl : System.Windows.Forms.Panel,
                                           ISplineControlPointHost
    {
        public SplineEditorGridControl()
        {
            _bSplineCurve2DRealization = new BSplineCurve2DRealization();
            _bSplineCurve2DRealization.Parent = this;
        }

        public void LoadCurve(BSplineCurve2D newCurve)
        {
            _isBusy = true;
            bool previousValueOfSnapToGrid = EditorForm.SnapToGrid;
            EditorForm.SnapToGrid = false;  // Temporarily disable snap to grid

            ClearControlPoints();
            _bSplineCurve2DRealization.IsClosedCurve = newCurve.IsClosedCurve;

            bool usesCenterCoords = false;
            for (int n = 0; n < newCurve.ControlPoints.Length; n++)
            {
                if ((newCurve.ControlPoints[n].x < -5) || (newCurve.ControlPoints[n].y < -5))
                {
                    usesCenterCoords = true;
                    break;
                }
            }

            IsCentered = usesCenterCoords;

            for (int n = 0; n < newCurve.ControlPoints.Length; n++)
            {
                CreateControlPoint(newCurve.ControlPoints[n].x + _origin.X,
                                   newCurve.ControlPoints[n].y + _origin.Y,
                                   newCurve.ControlPoints[n].weight);
            }

            _isBusy = false;
            EditorForm.SnapToGrid = previousValueOfSnapToGrid;
            InvalidateIfNotBusy();
        }

        public void InvalidateIfNotBusy()
        {
            if (!_isBusy)
            {
                this.Invalidate();
                this.Update();
            }
        }


        [Description("How many pixels between major grid lines")]
        [Category("Appearance")]
        [DefaultValue(50.0)]
        public float GridSide
        {
            get { return _gridSide; }
            set
            {
                _gridSide = value;
                InvalidateIfNotBusy();
            }
        }

        [Description("How many minor grid divisions per every major grid division")]
        [Category("Appearance")]
        [DefaultValue(5)]
        public int GridSubdivisions
        {
            get { return _gridSubdivisions; }
            set
            {
                _gridSubdivisions = value;
                InvalidateIfNotBusy();
            }
        }


        [Description("Major axis color")]
        [Category("Appearance")]
        [DefaultValue(typeof(Color), "System.Drawing.Color.Yellow")]
        public Color MajorAxisColor
        {
            get { return _majorAxisColor; }
            set
            {
                _majorAxisColor = value;
                InvalidateIfNotBusy();
            }
        }


        [Description("Major grid color")]
        [Category("Appearance")]
        [DefaultValue(typeof(Color), "System.Drawing.Color.LightYellow")]
        public Color MajorGridColor
        {
            get { return _majorGridColor; }
            set
            {
                _majorGridColor = value;
                InvalidateIfNotBusy();
            }
        }


        [Description("Minor grid color")]
        [Category("Appearance")]
        [DefaultValue(typeof(Color), "System.Drawing.Color.DarkOrange")]
        public Color MinorGridColor
        {
            get { return _minorGridColor; }
            set
            {
                _minorGridColor = value;
                InvalidateIfNotBusy();
            }
        }

        [Description("Represents a closed curve")]
        [Category("Behavior")]
        [DefaultValue(false)]
        public bool IsClosedCurve
        {
            get { return _isClosedCurve; }
            set
            {
                _isClosedCurve = value;
                _bSplineCurve2DRealization.IsClosedCurve = _isClosedCurve;
            }
        }

        [Description("Center coordinates")]
        [Category("Behavior")]
        [DefaultValue(false)]
        public bool IsCentered
        {
            get { return _isCentered; }
            set
            {
                _isCentered = value;
                OnOriginInvalidated();
                InvalidateIfNotBusy();
            }
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            OnOriginInvalidated();
            base.OnSizeChanged(e);
        }

        private void OnOriginInvalidated()
        {
            _origin = _isCentered ? new Vector2((float)(int)(.5f * (float)Size.Width),
                                                (float)(int)(.5f * (float)Size.Height))
                                  : Vector2.Zero;
            _bSplineCurve2DRealization.OnOriginInvalidated(_origin);
        }



        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if (e.Button == MouseButtons.Left)
            {
                _mousePressed = true;
                _mousePosition.X = e.X;
                _mousePosition.Y = e.Y;

                CreateControlPoint(e.X, e.Y, 1);

            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if (_mousePressed && (e.Button & MouseButtons.Left) != 0)
            {
                _mousePosition.X = e.X;
                _mousePosition.Y = e.Y;
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            if (_mousePressed)
            {
                _mousePressed = false;

                InvalidateIfNotBusy();
            }
        }

        public void OnControlPointAdded(SplineControlPointControl control)
        {
            _bSplineCurve2DRealization.OnControlPointAdded(control);
        }
        public void OnControlPointRemoved(SplineControlPointControl control)
        {
            _bSplineCurve2DRealization.OnControlPointRemoved(control);
        }
        public void OnControlPointWeightIncremented(SplineControlPointControl control)
        {
            _bSplineCurve2DRealization.OnControlPointWeightIncremented(control);
        }
        public void OnControlPointWeightDecremented(SplineControlPointControl control)
        {
            _bSplineCurve2DRealization.OnControlPointWeightDecremented(control);
        }
        public void OnControlPointMoved(SplineControlPointControl control)
        {
            _bSplineCurve2DRealization.OnControlPointMoved(control);
        }
        public void OnControlPointsCleared()
        {
            _bSplineCurve2DRealization.OnControlPointsCleared();
            InvalidateIfNotBusy();
        }


        public void CreateControlPoint(float x, float y, int weight)
        {
            SplineControlPointControl pointControl = new SplineControlPointControl(this, weight);

            SetControlPointLocation(pointControl, x, y);

            this.SuspendLayout();
            this.Controls.Add(pointControl);
            this.ResumeLayout(false);
            this.PerformLayout();

            pointControl.BringToFront();

            _bSplineCurve2DRealization.OnControlPointAdded(pointControl);
        }

        public void InsertControlPoint(SplineControlPointControl previousControl, float x, float y, int weight)
        {
            SplineControlPointControl newControl = new SplineControlPointControl(this, weight);

            SetControlPointLocation(newControl, x, y);

            this.SuspendLayout();
            this.Controls.Add(newControl);
            this.ResumeLayout(false);
            this.PerformLayout();

            newControl.BringToFront();

            _bSplineCurve2DRealization.OnControlPointInserted(previousControl, newControl);
        }

        // Applies snapToGrid as needed to a control point
        public void SetControlPointLocation(SplineControlPointControl pointControl, float x, float y)
        {
            float cx, cy;  // Center coords
            if (EditorForm.SnapToGrid)
            {
                float minorDivision = _gridSide / _gridSubdivisions;
                cx = minorDivision * (float)(int)(x / minorDivision + .5f);
                cy = minorDivision * (float)(int)(y / minorDivision + .5f);
            }
            else
            {
                cx = x;
                cy = y;
            }
            x = cx - .5f * (float)pointControl.Size.Width;
            y = cy - .5f * (float)pointControl.Size.Height;

            pointControl.Center = new Vector2(cx, cy);
            pointControl.Location = new System.Drawing.Point((int)(x + .5f), (int)(y + .5f));
        }

        // Clears of control points
        public void ClearControlPoints()
        {
            for (int n = 0; n < this.Controls.Count; )
            {
                SplineControlPointControl pointControl = this.Controls[n] as SplineControlPointControl;

                if (pointControl != null)
                {
                    this.Controls.Remove(pointControl);
                    pointControl.Dispose();
                }
                else
                {
                    n++;
                }
            }

            _bSplineCurve2DRealization.OnControlPointsCleared();

            InvalidateIfNotBusy();
        }


        protected override void OnPaint(PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            //g.Clear(Parent.BackColor);
            Draw(g);
        }

        protected virtual void Draw(Graphics g)
        {
            float height = (float)Size.Height;
            float width = (float)Size.Width;

            //
            // Draw minor grid
            //
            Pen pen = new Pen(_minorGridColor);
            float minorGridSide = _gridSide / _gridSubdivisions;

            int minorDivisionsAboveOrigin  = (int)(_origin.Y / minorGridSide);
            float minorInitialY = _origin.Y - minorDivisionsAboveOrigin * minorGridSide;

            for (float y = minorInitialY; y < height; y += minorGridSide)
            {
                if (Math.Abs((y - _origin.Y) % _gridSide) > .01f)
                {
                    g.DrawLine(pen, 0, y, width, y);
                }
            }

            int minorDivisionsLeftOfOrigin = (int)(_origin.X / minorGridSide);
            float minorInitialX = _origin.X - minorDivisionsLeftOfOrigin * minorGridSide;

            for (float x = minorInitialX; x < width; x += minorGridSide)
            {
                if (Math.Abs((x - _origin.X) % _gridSide) > .01f)
                {
                    g.DrawLine(pen, x, 0, x, height);
                }
            }

            pen.Dispose();

            //
            // Draw major grid
            //
            pen = new Pen(_majorGridColor);

            int divisionsAboveOrigin = (int)(_origin.Y / _gridSide);
            float initialY = _origin.Y - divisionsAboveOrigin * _gridSide;

            for (float y = initialY; y < height; y += _gridSide)
            {
                if (Math.Abs(y - _origin.Y) > .01f)
                {
                    g.DrawLine(pen, 0, y, width, y);
                }
            }

            int divisionsLeftOfOrigin = (int)(_origin.X / _gridSide);
            float initialX = _origin.X - divisionsLeftOfOrigin * _gridSide;

            for (float x = initialX; x < width; x += _gridSide)
            {
                if (Math.Abs(x - _origin.X) > .01f)
                {
                    g.DrawLine(pen, x, 0, x, height);
                }
            }

            pen.Dispose();

            //
            // Draw X, Y axes
            //
            pen = new Pen(_majorAxisColor);

            g.DrawLine(pen, 0, _origin.Y, width, _origin.Y);
            g.DrawLine(pen, _origin.X, 0, _origin.X, height);

            pen.Dispose();

            //
            // Draw B-Spline curve
            //
            if (_bSplineCurve2DRealization != null)
            {
                _bSplineCurve2DRealization.Draw(g);
            }
        }

        public BSplineCurve2D BSplineCurve
        {
            get { return _bSplineCurve2DRealization; }
        }

        // Presentation parameters
        private float _gridSide = 50.0f;
        private int _gridSubdivisions = 5;
        private Color _majorAxisColor = System.Drawing.Color.Yellow;
        private Color _majorGridColor = System.Drawing.Color.LightYellow;
        private Color _minorGridColor = System.Drawing.Color.DarkOrange;

        private bool _isClosedCurve = false;
        private bool _isCentered = false;
        private Vector2 _origin = Vector2.Zero;

        // Mouse tracking
        protected System.Drawing.Point _mousePosition;
        protected bool  _mousePressed;

        protected bool _isBusy = false;


        private BSplineCurve2DRealization _bSplineCurve2DRealization = null;
    }
}
