//Written by Leonid Velikovich and Benjamin Hernandez Arreguin

//+-----------------------------------------------------------------------------
//
// SurfaceOfRevolutionEffect class.  Software implementation for surface
// of revolution effect.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using Float = ParametricModelEditor.FloatingPointMathHelper;


namespace ParametricModelEditor
{
    public class SurfaceOfRevolutionEffect : ParametricSurfaceEffect
    {
        public SurfaceOfRevolutionEffect(ParametricCurve2D profileCurve)
        {
            Debug.Assert(profileCurve != null);
            _profileCurve = profileCurve;

            _parametricEffectType = ParametricEffectType.SurfaceOfRevolution;
        }

        // input.X is the B-Spline's parametric t parameter
        // input.Y is the theta angle parameter
        public override void VS_SurfaceFunction(Vector2 input, out Vector3 position,
                                                               out Vector3 dF_du,
                                                               out Vector3 dF_dv,
                                                               out Vector3 ddF_duu,
                                                               out Vector3 ddF_dvv)
        {
            //
            // Compute surface of revolution parameters: curve t and angle theta
            //
            float curvet = input.X;
            float theta = TWOPI * input.Y;

            Vector2 curvePoint2D, curveTangent2D, curveAcceleration2D;
            _profileCurve.Evaluate(curvet, out curvePoint2D,
                                           out curveTangent2D,
                                           out curveAcceleration2D);
            curvePoint2D = .004f * curvePoint2D;

            //
            // Compute the position
            //
            position.X = Float.Cos(theta) * curvePoint2D.X;
            position.Y =-Float.Sin(theta) * curvePoint2D.X;
            position.Z = .7f - curvePoint2D.Y;

            //
            // Compute partials
            //
            dF_du.X = Float.Cos(theta) * curveTangent2D.X;
            dF_dv.X = -TWOPI * Float.Sin(theta) * curvePoint2D.X;

            dF_du.Y =-Float.Sin(theta) * curveTangent2D.X;
            dF_dv.Y = -TWOPI * Float.Cos(theta) * curvePoint2D.X;

            dF_du.Z = -curveTangent2D.Y;
            dF_dv.Z = 0;

            //
            // Compute 2nd-order partials
            //
            ddF_duu.X = Float.Cos(theta) * curveAcceleration2D.X;
            ddF_dvv.X = -TWOPI * TWOPI * Float.Cos(theta) * curvePoint2D.X;

            ddF_duu.Y = -Float.Sin(theta) * curveAcceleration2D.X;
            ddF_dvv.Y = TWOPI * TWOPI * Float.Sin(theta) * curvePoint2D.X;

            ddF_duu.Z = -curveAcceleration2D.Y;
            ddF_dvv.Z = 0;
        }

        public override float[] GetSingularPointsU()
        {
            return _profileCurve.GetSingularPoints();
        }

        public override float[] GetSingularPointsV()
        {
            // No singularities on circular V curve
            return new float[0];
        }

        public override Vector2[] GetControlPointsU()
        {
            return _profileCurve.GetControlPoints();
        }

        public override List<ParametricCurve2D> GetSerialization()
        {
            List<ParametricCurve2D> list = new List<ParametricCurve2D>();
            list.Add(_profileCurve.GetSerialization());
            return list;
        }

        public override void SetSerialization(List<ParametricCurve2D> list)
        {
            Debug.Assert(list.Count >= 1);
            _profileCurve = list[0];
        }

        private ParametricCurve2D _profileCurve;
    }
}
