//Written by Leonid Velikovich and Benjamin Hernandez Arreguin

//+-----------------------------------------------------------------------------
//
// WireProductEffect class.  Software implementation for wire product effect.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;


namespace ParametricModelEditor
{
    public class WireProductEffect : ParametricSurfaceEffect
    {
        public WireProductEffect(ParametricCurve2D wireCurve, ParametricCurve2D crossSectionCurve)
        {
            Debug.Assert(wireCurve != null && crossSectionCurve != null);
            _wireCurve = wireCurve;
            _crossSectionCurve = crossSectionCurve;

            _parametricEffectType = ParametricEffectType.WireProduct;
        }

        // input.X is the B-Spline's parametric t parameter
        // input.Y is the theta angle parameter
        public override void VS_SurfaceFunction(Vector2 input, out Vector3 position,
                                                               out Vector3 dF_du,
                                                               out Vector3 dF_dv,
                                                               out Vector3 ddF_duu,
                                                               out Vector3 ddF_dvv)
        {
            //
            // Compute surface of revolution parameters: curve t and angle theta
            //
            Vector2 wireCurvePoint2D, wireCurveTangent2D, wireCurveAcceleration2D;
            _wireCurve.Evaluate(input.X, out wireCurvePoint2D,
                                         out wireCurveTangent2D,
                                         out wireCurveAcceleration2D);
            wireCurvePoint2D *= .005f;

            Vector2 crossSectionCurvePoint2D, crossSectionCurveTangent2D, crossSectionCurveAcceleration2D;
            _crossSectionCurve.Evaluate(input.Y, out crossSectionCurvePoint2D,
                                                 out crossSectionCurveTangent2D,
                                                 out crossSectionCurveAcceleration2D);
            crossSectionCurvePoint2D *= .0005f;

            //
            // Compute the position
            //
            Vector2 normVector = Vector2.Normalize(new Vector2(-wireCurveTangent2D.Y, -wireCurveTangent2D.X));

            position.X = wireCurvePoint2D.X + normVector.X * crossSectionCurvePoint2D.X;
            position.Y =-wireCurvePoint2D.Y + normVector.Y * crossSectionCurvePoint2D.X;
            position.Z =-crossSectionCurvePoint2D.Y;


            //
            // Compute partials.  NOTE: this is not 100% accurate because of commented out section:
            //
            dF_du.X = wireCurveTangent2D.X;// + d/dU(normVector.X) * crossSectionCurvePoint2D.X;
            dF_dv.X = normVector.X * crossSectionCurveTangent2D.X;

            dF_du.Y =-wireCurveTangent2D.Y;// + d/dU(normVector.Y) * crossSectionCurvePoint2D.X;
            dF_dv.Y = normVector.Y * crossSectionCurveTangent2D.X;

            dF_du.Z = 0;
            dF_dv.Z = -crossSectionCurveTangent2D.Y;

            //
            // Compute 2nd-order partials
            //
            ddF_duu.X = wireCurveAcceleration2D.X;// + d/dU( d/dU(normVector.X) * crossSectionCurvePoint2D.X );
            ddF_dvv.X = normVector.X * crossSectionCurveAcceleration2D.X;

            ddF_duu.Y =-wireCurveAcceleration2D.Y;// + d/DU( d/dU(normVector.Y) * crossSectionCurvePoint2D.X );
            ddF_dvv.Y = normVector.Y * crossSectionCurveAcceleration2D.X;

            ddF_duu.Z = 0;
            ddF_dvv.Z = -crossSectionCurveAcceleration2D.Y;
        }

        public override float[] GetSingularPointsU()
        {
            return _wireCurve.GetSingularPoints();
        }

        public override float[] GetSingularPointsV()
        {
            return _crossSectionCurve.GetSingularPoints();
        }

        public override Vector2[] GetControlPointsU()
        {
            return _wireCurve.GetControlPoints();
        }

        public override Vector2[] GetControlPointsV()
        {
            return _crossSectionCurve.GetControlPoints();
        }


        public override List<ParametricCurve2D> GetSerialization()
        {
            List<ParametricCurve2D> list = new List<ParametricCurve2D>();
            list.Add(_wireCurve.GetSerialization());
            list.Add(_crossSectionCurve.GetSerialization());
            return list;
        }



        public override void SetSerialization(List<ParametricCurve2D> list)
        {
            Debug.Assert(list.Count >= 1);
            _wireCurve = list[0];
            _crossSectionCurve = list[1];
        }

        private ParametricCurve2D _wireCurve, _crossSectionCurve;
    }
}
