//+-----------------------------------------------------------------------------
//
//  Class BSplineCurve2D.  Essential BSpline 2D curve implementation.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Diagnostics;
using System.Drawing;
//using System.Drawing.Drawing2D;  Avoids a namespace collision with XNA
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using Microsoft.Xna.Framework;
using matrix;
using utilities;
using DifferentiableFunction;
using Interval;


namespace WinFormsGraphicsDevice
{
    public struct ControlPoint
    {
        [XmlAttribute]
        public float x, y;

        [XmlAttribute]
        public int weight;
    }

    
    public class BSplineCurve2D : ParametricCurve2D
    {
        public BSplineCurve2D()
        {
        }
        
        // Given t in the range [0, 1], produces {x, y}, {x', y'}, and {x", y"} of t.
        public override void Evaluate(float t, out Vector2 position,
                                               out Vector2 tangent,
                                               out Vector2 acceleration)
        {


            // DStar
            //Variable t = new Variable("t");





            if (_controlPointCount < 4)
            {
                position = tangent = acceleration = Vector2.Zero;
                return;
            }
        
            // For a closed curve, we act like we append the first 3 points at the end of the list
            int appliedControlPointCount = _controlPointCount + (_isClosedCurve ? 3 : 0);
            float scaledT = (appliedControlPointCount - 3) * t;
            int currentSegment = (int)scaledT;

            if (t == 1)  // The last point on the curve
            {
                currentSegment--;  // This is the only case when relativeT will be set to 1
            }

            float relativeT  = scaledT - (float)currentSegment;  // In [0..1] range

            Vector2 c0, c1, c2, c3;
            c0 = _controlPointList[(currentSegment    ) % _controlPointCount];
            c1 = _controlPointList[(currentSegment + 1) % _controlPointCount];
            c2 = _controlPointList[(currentSegment + 2) % _controlPointCount];
            c3 = _controlPointList[(currentSegment + 3) % _controlPointCount];

            // Singularity test, clamp T to a safe area in the [0..1] range
            // NOTE: This safety mechanism does not work against quadruple points (e.g. c0==c3)
            if (c1 == c2)
            {
                // Entering triple point
                if (c1 == c3)
                {
                    if (relativeT > (1.0f - c_singularityTolerance))
                    {
                        relativeT = (1.0f - c_singularityTolerance);
                    }
                }
                // Departing triple point
                else if (c0 == c2)
                {
                    if (relativeT < c_singularityTolerance)
                    {
                        relativeT = c_singularityTolerance;
                    }
                }
            }

            float relativeT2 = relativeT * relativeT;
            float relativeT3 = relativeT2 * relativeT;

            Matrix controlMatrix = new Matrix(c0.X, c0.Y, 0, 0,
                                              c1.X, c1.Y, 0, 0,
                                              c2.X, c2.Y, 0, 0,
                                              c3.X, c3.Y, 0, 0);

            Matrix combinedMatrix = c_baseMatrix * controlMatrix;

            Vector4 tVector0 = new Vector4(relativeT3, relativeT2, relativeT, 1);
            Vector4 position4 = Vector4.Transform(tVector0, combinedMatrix);

            position.X = position4.X;
            position.Y = position4.Y;

            Vector4 tVector1 = new Vector4(3 * relativeT2, 2 * relativeT, 1, 0);
            Vector4 tangent3 = Vector4.Transform(tVector1, combinedMatrix);

            tangent.X = tangent3.X;
            tangent.Y = tangent3.Y;

            Vector4 tVector2 = new Vector4(6 * relativeT, 2, 0, 0);
            Vector4 acceleration4 = Vector4.Transform(tVector2, combinedMatrix);

            acceleration.X = acceleration4.X;
            acceleration.Y = acceleration4.Y;
        }

        protected virtual void OnCurveInvalidated()
        {
        }

        [XmlAttribute()]
        public bool IsClosedCurve
        {
            get { return _isClosedCurve; }
            set
            {
                _isClosedCurve = value;
                OnCurveInvalidated();
            }
        }
        [XmlAttribute()]
        public int ControlPointCount
        {
            get { return _controlPointCount; }
            set { _controlPointCount = value; }
        }

        [XmlElement("ControlPoint")]
        public ControlPoint[] ControlPoints
        {
            get
            {
                return GetControlPointSerialization();
            }
            set
            {
                ControlPoint[] points = (ControlPoint[])value;
                _controlPointList.Clear();
                _controlPointCount = 0;
                _controlPointListSerialized = value;

                foreach (ControlPoint point in points)
                {
                    _controlPointCount += point.weight;
                    for (int n = 0; n < point.weight; n++)
                    {
                        _controlPointList.Add(new Vector2(point.x , point.y));
                    }
                }

                OnCurveInvalidated();  // Probably does nothing since this is only for serialization
            }
        }

        protected virtual ControlPoint[] GetControlPointSerialization()
        {
            return _controlPointListSerialized;
        }

        public override float[] GetSingularPoints()
        {
            List<float> singularPointList = new List<float>();

            int appliedCount = _controlPointCount - (_isClosedCurve ? 0 : 2);
            for (int n = 0; n < appliedCount; n++)
            {
                // Detect when we have a triple point
                if (_controlPointList[n] == _controlPointList[(n + 2) % _controlPointCount])
                {
                    float newIndex = InverseOfIndex(n);
                    if (!singularPointList.Contains(newIndex))
                    {
                        singularPointList.Add(newIndex);
                    }
                }
            }

            return singularPointList.ToArray();
        }

        public override Vector2[] GetControlPoints()
        {
            // Duplicates 3 last points for closed curves
            List<Vector2> output = new List<Vector2>(_controlPointList);
            if (_isClosedCurve)
            {
                output.Add(_controlPointList[0]);
                output.Add(_controlPointList[1]);
                output.Add(_controlPointList[2]);
            }
            return output.ToArray();
        }

        

        /// <summary>
        /// Returns the value of t which maps to control point at index n as its first of 4 points
        /// </summary>
        /// <param name="n"></param>
        /// <returns></returns>
        private float InverseOfIndex(int n)
        {
            // For a closed curve, we act like we append the first 3 points at the end of the list
            int appliedControlPointCount = _controlPointCount + (_isClosedCurve ? 3 : 0);

            return ((float)n) / (float)(appliedControlPointCount - 3);
        }




        protected bool _isClosedCurve = false;

        protected int _controlPointCount = 0;

        protected List<Vector2> _controlPointList = new List<Vector2>();

         protected ControlPoint[] _controlPointListSerialized = null;



        protected static Microsoft.Xna.Framework.Matrix c_baseMatrix =
            new Microsoft.Xna.Framework.Matrix(-1,  3, -3,  1,
                                                3, -6,  3,  0,
                                               -3,  0,  3,  0,
                                                1,  4,  1,  0) * (1.0f/6.0f);

        public const float c_singularityTolerance = .01f;

    }
}
