﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using matrix;
using utilities;
using DifferentiableFunction;
using Interval;

using Float = WinFormsGraphicsDevice.FloatingPointMathHelper;

namespace WinFormsGraphicsDevice
{
    public class ParametricModelDStar3DVisualizer
    {
        private const int c_triangleCount = 30000;
        
        public ParametricModelDStar3DVisualizer (GraphicsDevice device, int xSamples, int ySamples)
        {
            _device = device;
            Resample(2 * xSamples * ySamples);
        }

        public void OnDisposeGraphicsDevice()
        {
            _device = null;
        }

        public void OnResetGraphicsDevice(GraphicsDevice device)
        {
            _device = device;
            _vertexBufferPositionNormal = new VertexBuffer(device, typeof(VertexPositionNormal), _numVertices,
                                             BufferUsage.WriteOnly);
        }

        private void ResetGeometry()
        {
            if (_vertexBufferPositionNormal != null)
            {
                _vertexBufferPositionNormal.Dispose();
                _vertexBufferPositionNormal = null;
            }
        }

        private void Resample(int totalTriangles)
        {
            float root = Float.Sqrt(.5f * (float)totalTriangles);  // # of grid squares = 1/2 * #triangles


            _numUSamples = (int)(root + .5f);
            _numVSamples = (int)(root + .5f);

            InitializeVertexBuffer();
            InitializeIndices();

        }
        
        public static Function normalize(Function a)
        {
            int len = a.rangeDimension;
            List<Function> res = new List<Function>();
            Function sum = 0.0d;

            foreach (Function subfunc in a) { sum += subfunc * subfunc; }
            sum = 1.0d / Function.sqrt(sum);
            foreach (Function subfunc in a) { res.Add(a * sum); }
            if (res.Count == 1)
            {
                return res[0];
            }
            else
                return new Function(res);
        }

        private Function cross(Function a, Function b)
        {
            return new Function(a[1] * b[2] - a[2] * b[1],
                                a[2] * b[0] - a[0] * b[2],
                                a[0] * b[1] - a[1] * b[0]);
        }
        
        public void upDateCtrlPoints(ParametricCurve2D profileCurve, ParametricCurve2D crossSectionCurve)
        {
            int i;
            Vector2[] ctrlPointsProfileCurve = profileCurve.GetControlPoints();
            Vector2[] ctrlPointsCrossSectionCurve = crossSectionCurve.GetControlPoints();
            _ctrlPointsProfileCurveX = new double
                                                [ctrlPointsProfileCurve.Length];
            _ctrlPointsProfileCurveY = new double
                                                [ctrlPointsProfileCurve.Length];
            _ctrlCrossSectionCurveX = new double
                                                [ctrlPointsCrossSectionCurve.Length];
            _ctrlCrossSectionCurveY = new double
                                                [ctrlPointsCrossSectionCurve.Length];

            for (i = 0; i < ctrlPointsProfileCurve.Length; i++)
            {
                _ctrlPointsProfileCurveX[i] = ctrlPointsProfileCurve[i].X;
                _ctrlPointsProfileCurveY[i] = ctrlPointsProfileCurve[i].Y;
            }
            for (i = 0; i < ctrlPointsCrossSectionCurve.Length; i++)
            {
                _ctrlCrossSectionCurveX[i] = ctrlPointsCrossSectionCurve[i].X;
                _ctrlCrossSectionCurveY[i] = ctrlPointsCrossSectionCurve[i].Y;
            }
        }

        public Function[,] SurfaceTangentFrame(params Function[] uv)
        {
            Variable[] inputVars;
            Function[,] frame;
            Function[] fposition=null, fdu=null, fdv=null;


            if (_parametricSurfaceEffect == "ProfileProductEffect")
            {
                //PROFILE CURVE             PROFILE CURVE
                BSpline profileCurveX = new BSpline(_vCtrlPointsProfileCurveX);
                Function fProfileCurvePositionX = profileCurveX.curveValue(uv[0]) * 0.008;
                Function fProfileCurveTangentX = profileCurveX.tangent(uv[0]);
                
                BSpline profileCurveY = new BSpline(_vCtrlPointsProfileCurveY);
                Function fProfileCurvePositionY = profileCurveY.curveValue(uv[0]) * 0.008;
                Function fProfileCurveTangentY = profileCurveY.tangent(uv[0]);
                

                //CROSS SECTION CURVE           CROSS SECTION CURVE 
                BSpline crossSectionCurveX = new BSpline(_vCtrlCrossSectionCurveX);
                Function fCrossSectionCurvePositionX = crossSectionCurveX.curveValue(uv[1]) * 0.015;
                Function fCrossSectionCurveTangentX = crossSectionCurveX.tangent(uv[1]);
                
                BSpline crossSectionCurveY = new BSpline(_vCtrlCrossSectionCurveY);
                Function fCrossSectionCurvePositionY = crossSectionCurveY.curveValue(uv[1]) * 0.015;
                Function fCrossSectionCurveTangentY = crossSectionCurveY.tangent(uv[1]);
                //Function fCrossSectionCurveAccelerationY = crossSectionCurveY.acceleration(t[1]);

                fposition = new Function [] { fCrossSectionCurvePositionX * fProfileCurvePositionX,
                                       fProfileCurvePositionX*fCrossSectionCurvePositionY,
                                        2.0 -fProfileCurvePositionY };
                fdu = new Function[] {fCrossSectionCurvePositionX * fProfileCurveTangentX, 
                                 fCrossSectionCurvePositionY * fProfileCurveTangentX, 
                                        -fProfileCurveTangentY };
                fdv = new Function [] {  fCrossSectionCurveTangentX * fProfileCurvePositionX,
                                 fCrossSectionCurveTangentY * fProfileCurvePositionX,
                                        0 };   
            }
            else if (_parametricSurfaceEffect == "WireProductEffect")
            {
                //PROFILE CURVE             PROFILE CURVE
                BSpline wireCurveX = new BSpline(_vCtrlPointsWireCurveX);
                Function fWireCurvePositionX = wireCurveX.curveValue(uv[0]) * 0.005;
                Function fWireCurveTangentX = wireCurveX.tangent(uv[0]);
                //Function fProfileCurveAccelerationX = profileCurveX.acceleration(t[0]);

                BSpline wireCurveY = new BSpline(_vCtrlPointsWireCurveY);
                Function fWireCurvePositionY = wireCurveY.curveValue(uv[0]) * 0.005;
                Function fWireCurveTangentY = wireCurveY.tangent(uv[0]);
                //Function fProfileCurveAccelerationY = profileCurveY.acceleration(t[0]); 

                //CROSS SECTION CURVE           CROSS SECTION CURVE 
                BSpline crossSectionCurveX = new BSpline(_vCtrlCrossSectionCurveX);
                Function fCrossSectionCurvePositionX = crossSectionCurveX.curveValue(uv[1]) * 0.0005;
                Function fCrossSectionCurveTangentX = crossSectionCurveX.tangent(uv[1]);
                //Function fCrossSectionCurveAccelerationX = crossSectionCurveX.acceleration(t[1]);

                BSpline crossSectionCurveY = new BSpline(_vCtrlCrossSectionCurveY);
                Function fCrossSectionCurvePositionY = crossSectionCurveY.curveValue(uv[1]) * 0.0005;
                Function fCrossSectionCurveTangentY = crossSectionCurveY.tangent(uv[1]);
                //Function fCrossSectionCurveAccelerationY = crossSectionCurveY.acceleration(t[1]);

                //those should be normalized

                Function normalX = normalize(-fWireCurveTangentY);
                Function normalY = normalize(-fWireCurveTangentX);

                fposition = new Function[] { fWireCurvePositionX + normalX * fCrossSectionCurvePositionX,
                                       -fWireCurvePositionY + normalY * fCrossSectionCurvePositionX,
                                   -fCrossSectionCurvePositionY};

                fdu = new Function [] { fWireCurveTangentX, -fWireCurveTangentY, 0 };

                fdv = new Function [] { normalX * fCrossSectionCurveTangentX, 
                             normalY * fCrossSectionCurveTangentX,
                             -fCrossSectionCurveTangentY};
            }
            else
            {
            }

            Function fPosition = new Function(fposition);
            Function fDu = new Function(fdu);
            Function fDv = new Function(fdv);
            Function fNormal = cross(fDu, fDv);
            frame = new Function[,]{
                    {fdu[0],fdv[0],fNormal[0],fPosition[0]},
                    {fdu[1],fdv[1],fNormal[1],fPosition[1]},
                    {fdu[2],fdv[2],fNormal[2],fPosition[2]},
                    {0,0,0,1},
                };

            return frame;
        }

        public void ApplySurfaceOfRevolutionEffect(ParametricCurve2D profileCurve, double width, double height)
        {
            double profileSize = height * 0.004;
            _profileCurve = profileCurve;
            int i;
            Variable[] t = { new Variable("t"), new Variable("tetha") };
            Variable[] inputVars;

            Vector2[] ctrlPointsProfileCurve = profileCurve.GetControlPoints();
            
            _ctrlPointsProfileCurveX = new double
                                                [ctrlPointsProfileCurve.Length];
            _ctrlPointsProfileCurveY = new double
                                                [ctrlPointsProfileCurve.Length];
            
            _vCtrlPointsProfileCurveX = new Variable[4];
            _vCtrlPointsProfileCurveY = new Variable[4];
            
            for (i = 0; i < 4; i++)
            {
                _vCtrlPointsProfileCurveX[i] = new Variable("profileX" + i);
                _vCtrlPointsProfileCurveY[i] = new Variable("profileY" + i);
            }
            for (i = 0; i < ctrlPointsProfileCurve.Length; i++)
            {
                _ctrlPointsProfileCurveX[i] = ctrlPointsProfileCurve[i].X;
                _ctrlPointsProfileCurveY[i] = ctrlPointsProfileCurve[i].Y;
            }
            
            //PROFILE CURVE             PROFILE CURVE
            BSpline profileCurveX = new BSpline(_vCtrlPointsProfileCurveX);
            Function fProfileCurvePositionX = profileCurveX.curveValue(t[0]) * profileSize;
            Function fProfileCurveTangentX = profileCurveX.tangent(t[0]);
            //Function fProfileCurveAccelerationX = profileCurveX.acceleration(t[0]);

            BSpline profileCurveY = new BSpline(_vCtrlPointsProfileCurveY);
            Function fProfileCurvePositionY = profileCurveY.curveValue(t[0]) * profileSize;
            Function fProfileCurveTangentY = profileCurveY.tangent(t[0]);
            //Function fProfileCurveAccelerationY = profileCurveY.acceleration(t[0]); 
            
            Function[] fposition = { width * Function.cos(t[1]) * fProfileCurvePositionX,
                                       - fProfileCurvePositionY,
                                       width * Function.sin(t[1]) * fProfileCurvePositionX,
                                       
                                   };
            Function[] fdu = { Function.cos(t[1]) * fProfileCurveTangentX,
                                 -Function.sin(t[1]) * fProfileCurveTangentX,
                                 -fProfileCurveTangentY
                             };
            Function[] fdv = { -TWOPI * Function.sin(t[1]) * fProfileCurvePositionX,
                                 -TWOPI * Function.cos(t[1]) * fProfileCurvePositionX,
                                 0
                                };

            /* 
             * Acceleration Computation, we are not going to use this right now?
            
            Function[] fduu = { fCrossSectionCurvePositionX * fProfileCurveAccelerationX, fCrossSectionCurvePositionY * fProfileCurveAccelerationX,
                                        -fProfileCurveAccelerationY};
            Function[] fdvv = {  fCrossSectionCurveAccelerationX * fProfileCurvePositionX, fCrossSectionCurveAccelerationY * fProfileCurvePositionX,
                                        0 };
             * */
            Function fPosition = new Function(fposition);
            inputVars = VM.concatenate<Variable>(t, _vCtrlPointsProfileCurveX, _vCtrlPointsProfileCurveY);
            fPosition.orderVariablesInDomain(inputVars);
            _rtPosition = fPosition.compile();

            Function fDu = new Function(fdu);
            Function fDv = new Function(fdv);
            Function fNormal = cross(fDu, fDv);
            fNormal.orderVariablesInDomain(inputVars);
            _rtNormal = fNormal.compile();
            
            _parametricSurfaceEffect = "SurfaceOfRevolutionEffect";
            InitializeVertexBuffer();
            InitializeIndices();
            GenerateMesh();
            

            _eqCalculated = true;
            
        }
        
        
        public void ApplyProfileProductEffect(ParametricCurve2D profileCurve, ParametricCurve2D crossSectionCurve, double width, double height)
        {            
            int i;
            _profileCurve = profileCurve;
            _crossSectionCurve = crossSectionCurve;

            Variable[] t = { new Variable("t"), new Variable("tetha")};
            Variable[] inputVars;

            Vector2[] ctrlPointsProfileCurve = profileCurve.GetControlPoints();
            Vector2[] ctrlPointsCrossSectionCurve = crossSectionCurve.GetControlPoints();
            _ctrlPointsProfileCurveX = new double
                                                [ctrlPointsProfileCurve.Length];
            _ctrlPointsProfileCurveY = new double
                                                [ctrlPointsProfileCurve.Length];
            _ctrlCrossSectionCurveX = new double
                                                [ctrlPointsCrossSectionCurve.Length];
            _ctrlCrossSectionCurveY = new double
                                                [ctrlPointsCrossSectionCurve.Length];
            
            _vCtrlPointsProfileCurveX = new Variable[4];
            _vCtrlPointsProfileCurveY = new Variable[4];
            _vCtrlCrossSectionCurveX  = new Variable[4];
            _vCtrlCrossSectionCurveY  = new Variable[4];

            for (i = 0; i < 4; i++)
            {
                _vCtrlPointsProfileCurveX[i] = new Variable("profileX" + i);
                _vCtrlPointsProfileCurveY[i] = new Variable("profileY" + i);
                _vCtrlCrossSectionCurveX[i] = new Variable("crossSectionX" + i);
                _vCtrlCrossSectionCurveY[i] = new Variable("crossSectionY" + i);
            }            
            for (i = 0; i < ctrlPointsProfileCurve.Length; i++)
            {
                _ctrlPointsProfileCurveX[i] = ctrlPointsProfileCurve[i].X;
                _ctrlPointsProfileCurveY[i] = ctrlPointsProfileCurve[i].Y;
            }
            for (i = 0; i < ctrlPointsCrossSectionCurve.Length; i++)
            {
                _ctrlCrossSectionCurveX[i] = ctrlPointsCrossSectionCurve[i].X;
                _ctrlCrossSectionCurveY[i] = ctrlPointsCrossSectionCurve[i].Y;
            }

            //PROFILE CURVE             PROFILE CURVE
            BSpline profileCurveX = new BSpline(_vCtrlPointsProfileCurveX);
            Function fProfileCurvePositionX = profileCurveX.curveValue(t[0]) * 0.004;
            Function fProfileCurveTangentX = profileCurveX.tangent(t[0]);
            //Function fProfileCurveAccelerationX = profileCurveX.acceleration(t[0]);
            
            BSpline profileCurveY = new BSpline(_vCtrlPointsProfileCurveY);
            Function fProfileCurvePositionY = profileCurveY.curveValue(t[0]) * 0.004;
            Function fProfileCurveTangentY = profileCurveY.tangent(t[0]);
            //Function fProfileCurveAccelerationY = profileCurveY.acceleration(t[0]); 

            //CROSS SECTION CURVE           CROSS SECTION CURVE 
            BSpline crossSectionCurveX = new BSpline(_vCtrlCrossSectionCurveX);
            Function fCrossSectionCurvePositionX = crossSectionCurveX.curveValue(t[1]) * 0.01;
            Function fCrossSectionCurveTangentX = crossSectionCurveX.tangent(t[1]);
            //Function fCrossSectionCurveAccelerationX = crossSectionCurveX.acceleration(t[1]);
            
            BSpline crossSectionCurveY = new BSpline(_vCtrlCrossSectionCurveY);
            Function fCrossSectionCurvePositionY = crossSectionCurveY.curveValue(t[1]) * 0.01;
            Function fCrossSectionCurveTangentY = crossSectionCurveY.tangent(t[1]);
            //Function fCrossSectionCurveAccelerationY = crossSectionCurveY.acceleration(t[1]);

            Function[] fposition = { width*fCrossSectionCurvePositionX * fProfileCurvePositionX,
                                       -height*fProfileCurvePositionY,
                                      width*fProfileCurvePositionX*fCrossSectionCurvePositionY
                                         };
            Function[] fdu = {fCrossSectionCurvePositionX * fProfileCurveTangentX, 
                                 fCrossSectionCurvePositionY * fProfileCurveTangentX, 
                                  
                                        -fProfileCurveTangentY };
            Function[] fdv = {  fCrossSectionCurveTangentX * fProfileCurvePositionX,
                                 fCrossSectionCurveTangentY * fProfileCurvePositionX,
                                 
                                        0 };

            /* 
             * Acceleration Computation, we are not going to use this right now?
            
            Function[] fduu = { fCrossSectionCurvePositionX * fProfileCurveAccelerationX, fCrossSectionCurvePositionY * fProfileCurveAccelerationX,
                                        -fProfileCurveAccelerationY};
            Function[] fdvv = {  fCrossSectionCurveAccelerationX * fProfileCurvePositionX, fCrossSectionCurveAccelerationY * fProfileCurvePositionX,
                                        0 };
             * */
            Function fPosition = new Function(fposition);
            inputVars = VM.concatenate<Variable>(t, _vCtrlPointsProfileCurveX, _vCtrlPointsProfileCurveY, _vCtrlCrossSectionCurveX, _vCtrlCrossSectionCurveY);
            fPosition.orderVariablesInDomain(inputVars);
            Function.printCompilerSource = true;
            _rtPosition = fPosition.compile();

            Function fDu = new Function(fdu);
            Function fDv = new Function(fdv);
            Function fNormal = cross(fDu, fDv);
            fNormal.orderVariablesInDomain(inputVars);
            _rtNormal = fNormal.compile();
            
            _parametricSurfaceEffect = "ProfileProductEffect";
            InitializeVertexBuffer();
            InitializeIndices();
            GenerateMesh();
            

            _eqCalculated = true;
        }

        private void GenerateMesh()
        {
            int i, j;
            double dt = 1.0 / (_numUSamples-1);
            double dQ = 1.0 / (_numVSamples-1);
            double t=0, tetha=0;
            double scaledT, scaledTetha;
            double[] position3D = new double[3]; 
            double[] normal3D = new double[3];

            double[] inputVars;
            double[] profileCX = new double[4];
            double[] profileCY = new double[4];
            double[] crossSectionCX = new double[4];
            double[] crossSectionCY = new double[4];
            double[] domain = new double[2];

            int n = 0;
            int appliedControlPoinProfile = _ctrlPointsProfileCurveX.Length;/* +(_isClosedProfileCurve ? 3 : 0);*/
            int appliedControlPointCrossSection = _ctrlCrossSectionCurveX.Length;/* +(_isClosedCrossSectionCurve ? 3 : 0);*/
            int currentSegmentU = 0;
            int currentSegmentV = 0;

            for (i = 0; i < _numUSamples; i++)
            {
                scaledT = (appliedControlPoinProfile - 3) *t;   
                currentSegmentU = (int)scaledT;

                if (scaledT > 1)
                {
                    domain[0] = (scaledT - Math.Floor(scaledT));
                }
                else
                {
                    domain[0] = scaledT;
                }
                profileCX[0] = _ctrlPointsProfileCurveX[currentSegmentU % (_ctrlPointsProfileCurveX.Length)];
                profileCX[1] = _ctrlPointsProfileCurveX[(currentSegmentU + 1) % (_ctrlPointsProfileCurveX.Length)];
                profileCX[2] = _ctrlPointsProfileCurveX[(currentSegmentU + 2) % (_ctrlPointsProfileCurveX.Length)];
                profileCX[3] = _ctrlPointsProfileCurveX[(currentSegmentU + 3) % (_ctrlPointsProfileCurveX.Length)];
                profileCY[0] = _ctrlPointsProfileCurveY[currentSegmentU % (_ctrlPointsProfileCurveY.Length)];
                profileCY[1] = _ctrlPointsProfileCurveY[(currentSegmentU + 1) % (_ctrlPointsProfileCurveY.Length)];
                profileCY[2] = _ctrlPointsProfileCurveY[(currentSegmentU + 2) % (_ctrlPointsProfileCurveY.Length)];
                profileCY[3] = _ctrlPointsProfileCurveY[(currentSegmentU + 3) % (_ctrlPointsProfileCurveY.Length)];

                for (j = 0; j < _numVSamples; j++)
                {
                    if (domain[0] == 1) domain[0] -= 0.00000001;
                    if (_parametricSurfaceEffect != "SurfaceOfRevolutionEffect")
                    {
                        scaledTetha = (appliedControlPointCrossSection - 3) * tetha;
                        currentSegmentV = (int)scaledTetha;
                        if (scaledTetha > 1)
                        {
                            domain[1] = (scaledTetha - Math.Floor(scaledTetha));
                        }
                        else
                        {
                            domain[1] = scaledTetha;
                        }
                        
                        crossSectionCX[0] = _ctrlCrossSectionCurveX[currentSegmentV % (_ctrlCrossSectionCurveX.Length)];
                        crossSectionCX[1] = _ctrlCrossSectionCurveX[(currentSegmentV + 1) % (_ctrlCrossSectionCurveX.Length)];
                        crossSectionCX[2] = _ctrlCrossSectionCurveX[(currentSegmentV + 2) % (_ctrlCrossSectionCurveX.Length)];
                        crossSectionCX[3] = _ctrlCrossSectionCurveX[(currentSegmentV + 3) % (_ctrlCrossSectionCurveX.Length)];
                        crossSectionCY[0] = _ctrlCrossSectionCurveY[(currentSegmentV) % (_ctrlCrossSectionCurveY.Length)];
                        crossSectionCY[1] = _ctrlCrossSectionCurveY[(currentSegmentV + 1) % (_ctrlCrossSectionCurveY.Length)];
                        crossSectionCY[2] = _ctrlCrossSectionCurveY[(currentSegmentV + 2) % (_ctrlCrossSectionCurveY.Length)];
                        crossSectionCY[3] = _ctrlCrossSectionCurveY[(currentSegmentV + 3) % (_ctrlCrossSectionCurveY.Length)];
                        inputVars = VM.concatenate<double>(domain, profileCX, profileCY, crossSectionCX, crossSectionCY);
                    }
                    else
                    {
                        domain[1] = (TWOPI * tetha );
                        inputVars = VM.concatenate<double>(domain, profileCX, profileCY);
                    }
                    _rtPosition.eval(position3D, inputVars);
                    

                    _verticesPositionNormal[n].Position.X = (float)position3D[0];
                    _verticesPositionNormal[n].Position.Y = (float)position3D[1];
                    _verticesPositionNormal[n].Position.Z = (float)position3D[2];

                    _rtNormal.eval(normal3D, inputVars);

                    _verticesPositionNormal[n].Normal.X = (float)normal3D[0];
                    _verticesPositionNormal[n].Normal.Y = (float)normal3D[1];
                    _verticesPositionNormal[n].Normal.Z = (float)normal3D[2];
                    _verticesPositionNormal[n++].Normal.Normalize();
                
                    tetha += dQ;   
                }
                t += dt;
                tetha = 0;
            }
            
        }


        public void ApplyWireProductEffect(ParametricCurve2D wireCurve, ParametricCurve2D crossSectionCurve, double width, double height)
        {
            int i;
            _wireCurve = wireCurve;
            _crossSectionCurve = crossSectionCurve;

            //Function.newContext();
            //Function.printCompilerSource = true;
            Variable[] t = { new Variable("t"), new Variable("tetha") };
            Variable[] inputVars;

            //_isClosedCrossSectionCurve = crossSectionCurve.

            Vector2[] ctrlPointsWireCurve = wireCurve.GetControlPoints();
            Vector2[] ctrlPointsCrossSectionCurve = crossSectionCurve.GetControlPoints();
            _ctrlPointsProfileCurveX = new double
                                                [ctrlPointsWireCurve.Length];
            _ctrlPointsProfileCurveY = new double
                                                [ctrlPointsWireCurve.Length];
            _ctrlCrossSectionCurveX = new double
                                                [ctrlPointsCrossSectionCurve.Length];
            _ctrlCrossSectionCurveY = new double
                                                [ctrlPointsCrossSectionCurve.Length];
            _vCtrlPointsWireCurveX = new Variable[4];
            _vCtrlPointsWireCurveY = new Variable[4];
            _vCtrlCrossSectionCurveX = new Variable[4];
            _vCtrlCrossSectionCurveY = new Variable[4];

            for (i = 0; i < 4; i++)
            {
                _vCtrlPointsWireCurveX[i] = new Variable("profileX" + i);
                _vCtrlPointsWireCurveY[i] = new Variable("profileY" + i);
                _vCtrlCrossSectionCurveX[i] = new Variable("crossSectionX" + i);
                _vCtrlCrossSectionCurveY[i] = new Variable("crossSectionY" + i);
            }
            for (i = 0; i < ctrlPointsWireCurve.Length; i++)
            {
                _ctrlPointsProfileCurveX[i] = ctrlPointsWireCurve[i].X;
                _ctrlPointsProfileCurveY[i] = ctrlPointsWireCurve[i].Y;
            }
            for (i = 0; i < ctrlPointsCrossSectionCurve.Length; i++)
            {
                _ctrlCrossSectionCurveX[i] = ctrlPointsCrossSectionCurve[i].X;
                _ctrlCrossSectionCurveY[i] = ctrlPointsCrossSectionCurve[i].Y;
            }
            //PROFILE CURVE             PROFILE CURVE
            BSpline wireCurveX = new BSpline(_vCtrlPointsWireCurveX);
            Function fWireCurvePositionX = wireCurveX.curveValue(t[0]) * 0.005;
            Function fWireCurveTangentX = wireCurveX.tangent(t[0]);
            //Function fProfileCurveAccelerationX = profileCurveX.acceleration(t[0]);

            BSpline wireCurveY = new BSpline(_vCtrlPointsWireCurveY);
            Function fWireCurvePositionY = wireCurveY.curveValue(t[0]) * 0.005;
            Function fWireCurveTangentY = wireCurveY.tangent(t[0]);
            //Function fProfileCurveAccelerationY = profileCurveY.acceleration(t[0]); 

            //CROSS SECTION CURVE           CROSS SECTION CURVE 
            BSpline crossSectionCurveX = new BSpline(_vCtrlCrossSectionCurveX);
            Function fCrossSectionCurvePositionX = crossSectionCurveX.curveValue(t[1]) * 0.0005;
            Function fCrossSectionCurveTangentX = crossSectionCurveX.tangent(t[1]);
            //Function fCrossSectionCurveAccelerationX = crossSectionCurveX.acceleration(t[1]);

            BSpline crossSectionCurveY = new BSpline(_vCtrlCrossSectionCurveY);
            Function fCrossSectionCurvePositionY = crossSectionCurveY.curveValue(t[1]) * 0.0005;
            Function fCrossSectionCurveTangentY = crossSectionCurveY.tangent(t[1]);
            //Function fCrossSectionCurveAccelerationY = crossSectionCurveY.acceleration(t[1]);

            //those should be normalized
            
            Function normalX = normalize(-fWireCurveTangentY);
            Function normalY = normalize(-fWireCurveTangentX);

            Function[] fposition = { width*(fWireCurvePositionX + normalX * fCrossSectionCurvePositionX),
                                       -height*fCrossSectionCurvePositionY,
                                       width*(-fWireCurvePositionY + normalY * fCrossSectionCurvePositionX)
                                   };
            Function[] fdu = { fWireCurveTangentX, -fWireCurveTangentY, 0 };
            
            Function[] fdv = { normalX * fCrossSectionCurveTangentX, 
                             normalY * fCrossSectionCurveTangentX,
                             -fCrossSectionCurveTangentY};

            Function fPosition = new Function(fposition);
            inputVars = VM.concatenate<Variable>(t, _vCtrlPointsWireCurveX, _vCtrlPointsWireCurveY, _vCtrlCrossSectionCurveX, _vCtrlCrossSectionCurveY);
            fPosition.orderVariablesInDomain(inputVars);
            _rtPosition = fPosition.compile();

            Function fDu = new Function(fdu);
            Function fDv = new Function(fdv);
            Function fNormal = cross(fDu, fDv);
            fNormal.orderVariablesInDomain(inputVars);
            _rtNormal = fNormal.compile();

            _parametricSurfaceEffect = "WireProductEffect";
            InitializeVertexBuffer();
            InitializeIndices();
            GenerateMesh();
            

            _eqCalculated = true;
        }



        public void ApplyNullEffect()
        {
            _parametricSurfaceEffect = null;
        }


        public float Rotation
        {
            set { _rotation = value; }
        }

        public float Distance
        {
            set { _distance = value; }
        }

        public float LeanAngle
        {
            set { _leanAngle = value; }
        }

        public float AmbientFactor
        {
            set { _materialAmbientFactor = value; }
        }

        public float DiffuseFactor
        {
            set { _materialDiffuseFactor = value; }
        }

        public float SpecularFactor
        {
            set { _materialSpecularFactor = value; }
        }

        public Vector2 DomainScalingFactor
        {
            set { _domainScalingFactor = value; }
        }

        public float DomainLodFraction
        {
            set { _domainLodFraction = value; }
        }

        public Vector4 MaterialColor
        {
            get { return _materialColor; }
            set { _materialColor = value; }
        }

        private void InitializeVertexBuffer()
        {
            _numVertices = (_numUSamples + 1) * (_numVSamples + 1);
            _verticesPosition2D = new VertexPosition2D[_numVertices];
            _verticesPositionNormal = new VertexPositionNormal[_numVertices];

            int index = 0;
            for (int y = 0; y <= _numVSamples; y++)
            {
                for (int x = 0; x <= _numUSamples; x++)
                {
                    float fx = .5f * (float)x;
                    float fy = .5f * (float)(_numVSamples - y);
                    _verticesPosition2D[index++].Position = new Vector2(fx, fy);  // Left-right, top-down vertex order
                }
            }
            _vertexBufferPositionNormal = new VertexBuffer(_device, typeof(VertexPositionNormal), _numVertices,
                                 BufferUsage.WriteOnly);
        }

        // Using triangle strips
        private void InitializeIndices()
        {
          
            //setup indices for TriangleList
            _indices = new short[2 * 3 * (_numVSamples ) * (_numUSamples )];
            int cnt = 0;
            for (int j = 0; j < _numVSamples ; j++)
            {
                for (int i = 0; i < _numUSamples ; i++)
                {
                    int idx = j * _numUSamples + i;
                    if (true)
                    {
                        _indices[cnt++] = (short)idx;
                        _indices[cnt++] = (short)(idx + _numUSamples);
                        _indices[cnt++] = (short)(idx + _numUSamples + 1);

                        _indices[cnt++] = (short)(idx);
                        _indices[cnt++] = (short)(idx + _numUSamples + 1);
                        _indices[cnt++] = (short)(idx + 1);
                    }
                    else
                    {
                        _indices[cnt++] = (short)idx;
                        _indices[cnt++] = (short)(idx + _numUSamples + 1);
                        _indices[cnt++] = (short)(idx + _numUSamples);


                        _indices[cnt++] = (short)idx;
                        _indices[cnt++] = (short)(idx + 1);
                        _indices[cnt++] = (short)(idx + _numUSamples + 1);
                    }
                }
            }
            _numTriangles = cnt - 2;
          
            /*
            _numIndices = _numVSamples * 2 * (_numUSamples+2 );  // Conservative estimate
            Debug.Assert(_numIndices <= EditorForm.c_maxIndexBufferSize);  // Max legal index buffer size
            _indices = new short[_numIndices];

            int index = 0;

            for (int y = 0; y < _numVSamples; y++)
            {
                if (y % 2 == 0)
                {
                    for (int x = 0; x < _numUSamples; x++)
                    {
                        _indices[index++] = (short)((y + 1) * (_numUSamples + 1) + x);
                        _indices[index++] = (short)((y) * (_numUSamples + 1) + x);
                    }

                    // Degenerate point on the tristrip
                    _indices[index++] = (short)((y + 2) * (_numUSamples + 1) - 1);
                    _indices[index++] = (short)((y + 2) * (_numUSamples + 1) - 1);
                }
                else
                {
                    for (int x = _numUSamples; x > 0; x--)
                    {
                        _indices[index++] = (short)((y) * (_numUSamples + 1) + x);
                        _indices[index++] = (short)((y + 1) * (_numUSamples + 1) + x);
                    }

                    // Degenerate points on the tristrip
                    _indices[index++] = (short)((y + 1) * (_numUSamples + 1));
                    _indices[index++] = (short)((y + 2) * (_numUSamples + 1));
                }
            }
            _numTriangles = index - 2;
             * */
             
             
        }


        public bool _isClosedProfileCurve = false;
        public bool _isClosedCrossSectionCurve = true;
        public const float c_singularityTolerance = .01f;

        public short[] _indices;
        public int _numTriangles;
        public int _numVertices;
                

        public VertexPosition2D[] _verticesPosition2D;
        public VertexPositionNormal[] _verticesPositionNormal;
        public VertexBuffer _vertexBufferPositionNormal;

        private bool _eqCalculated = false;
        private GraphicsDevice _device;
        //private Effect _hardwareEffect=null;

        private int _numUSamples;  // We have a rectangular numU x numV grid domain
        private int _numVSamples;


        private ParametricCurve2D _profileCurve;
        private ParametricCurve2D _crossSectionCurve;
        private ParametricCurve2D _wireCurve;

        // Control Points read from the GUI/file 
        private double[] _ctrlPointsProfileCurveX;
        private double[] _ctrlPointsProfileCurveY;
        private double[] _ctrlCrossSectionCurveX;
        private double[] _ctrlCrossSectionCurveY;

        private Variable[] _vCtrlPointsProfileCurveX;
        private Variable[] _vCtrlPointsProfileCurveY;
        private Variable[] _vCtrlCrossSectionCurveX;
        private Variable[] _vCtrlCrossSectionCurveY;
        private Variable[] _vCtrlPointsWireCurveX;
        private Variable[] _vCtrlPointsWireCurveY;


        private RuntimeFunction _rtPosition;
        private RuntimeFunction _rtNormal;
        /*
        private RuntimeFunction _rtDu;
        private RuntimeFunction _rtDv;
        private RuntimeFunction _rtDuu;
        private RuntimeFunction _rtDvv;
        */
        public string _parametricSurfaceEffect = null;
        private string _strEffect;

        private float _rotation = 0.0f;
        private float _distance = 1.0f;
        private float _leanAngle = 0.0f;
        private Vector2 _domainScalingFactor = new Vector2(1.0f, 1.0f);
        private float _domainLodFraction = 1.0f;

        public float _materialAmbientFactor = .05f;
        public float _materialDiffuseFactor = .90f;
        public float _materialSpecularFactor = 4.0f;

        public Vector4 _materialColor = new Vector4(1.0f, .27f, .20f, 1.0f);

        protected const float PI = 3.141592654f;
        protected const float TWOPI = 6.283185307f;


    }
}
