#region Using Statements
using System;
using System.Diagnostics;
using System.IO;
//using System.Xml.Serialization;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using matrix;
using utilities;
using DifferentiableFunction;
using Dynamics;
using WinFormsGraphicsDevice.Geometries;
#endregion

namespace WinFormsGraphicsDevice {

    public enum INTEGRATION_METHOD {
        EXPLICIT_EULER, // explicit Euler method
        RK4,            // 4th order Runge-Kutta method
        IMPLICIT_EULER  // not implemented because I cannot apply symbolic differentiation to Composite Inertia Method.
    }

    public enum MATRIX_SOLVER {
        LU,     // LU decomposition
        SVD     // Singular value decomposition
    }

    public enum FORWARD_DYNAMICS_METHOD {
        COMPOSITE_RIGID_BODY_METHOD,
        NEW_METHOD
    }

    public enum RUNTIME_FUNCTION_MODE {
        FULL,   // TheRuntimeFunction will contain Jacobians, dynamics, energy, constraint
        MINIMAL,// will contain only dynamics and constraint (if any). userSpring MUST NOT BE USED for MINIMAL mode
    }

    public class RigidBodySystemUtitlities {
        /// <summary>
        /// Copy a matrix represented by src to Xna.Framework.Matrix
        /// Note that Xna.Framework.Matrix is TRANSPOSE of homogeneous transformation matrix
        /// </summary>
        /// /// <param name="src"> represents 4x4 matrix row-wise. i.e. [ src[0]..src[3]; src[4]..src[7]; src[8]..src[11]; src[12]..src[16] ]</param>
        /// <param name="dst">is target matrix that elements of m will be copied to.</param>
        /// <param name="start">is first index in src to be copied</param>
        public static void CopyRowWiseDouble16ToXnaMatrix(double[] src, int start, out Matrix dst) {
            dst.M11 = (float)src[start];
            dst.M21 = (float)src[start + 1];
            dst.M31 = (float)src[start + 2];
            dst.M41 = (float)src[start + 3];

            dst.M12 = (float)src[start + 4];
            dst.M22 = (float)src[start + 5];
            dst.M32 = (float)src[start + 6];
            dst.M42 = (float)src[start + 7];

            dst.M13 = (float)src[start + 8];
            dst.M23 = (float)src[start + 9];
            dst.M33 = (float)src[start + 10];
            dst.M43 = (float)src[start + 11];

            dst.M14 = (float)src[start + 12];
            dst.M24 = (float)src[start + 13];
            dst.M34 = (float)src[start + 14];
            dst.M44 = (float)src[start + 15];
        }

        /// <summary>
        /// string out a matrix to a vector
        /// </summary>
        /// <param name="fn"></param>
        /// <returns></returns>
        public static Function[] RowWiseStringOut(Function[,] mat, out int length) {
            int h = mat.GetLength(0);
            int w = mat.GetLength(1);
            Function[] re = new Function[h * w];
            for (int i = 0; i < h; i++) {
                for (int j = 0; j < w; j++) {
                    re[j + w * i] = mat[i, j];
                }
            }
            length = h * w;
            return re;
        }

        /// <summary>
        /// Gets matrix dst of from src starting at startIdx
        /// </summary>
        public static void GetMatrixFromArray(double[,] dst, double[] src, int startIdx) {
            int cnt = 0;
            int h = dst.GetLength(0);
            int w = dst.GetLength(1);
            for (int i = 0; i < h; i++) {
                for (int j = 0; j < w; j++) {
                    dst[i, j] = src[startIdx + cnt++];
                }
            }
        }

        /// <summary>
        /// check if f is the element of farray
        /// </summary>        
        public static bool IsIncludedInFunctionArray(Function f, Function[] farray) {
            foreach (Function fn in farray) {
                if (fn == f) return true;
            }
            return false;
        }

        /// <summary>
        /// create a 4x4 transformation matrix of translation
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        /// <returns></returns>
        public static Function[,] makeTrans(double x, double y, double z) {
            return new Function[,] { { 1, 0, 0, x }, { 0, 1, 0, y }, { 0, 0, 1, z }, { 0, 0, 0, 1 } };
        }
    }

    /// 
    /// specification of rolling constraint
    /// Let's2 suppose two objects L1 and L2 are surface-constrained and only rolling can occur,
    /// and the surface function (R^2->R^3) is s2(u1,v1) and s2(u2,v2) respectively.
    /// Also A denotes the transformation from L1 to L2.
    /// Then rolling constraint can be written as v'*((dA/dt)*s2) = 0
    /// where v is a tangent vector of L1, i.e., v = ds1/dq where q is either u1 or v1
    /// 
    /// NEED TO DOUBLE CHECK THIS DERIVATION.
    /// 
    public struct RollingConstraint {
        public Function q;
        public Function[] s1;
        public Function[,] A;
        public Function[] s2;
    }

    public class RigidBody : JointNode {


        public RigidBody(Function[,] transform, Function[] generalizedCoordinates, Function[] generalizedForces,
            Function[,] centerOfMass, Function[,] inertiaMatrix, Function mass, string nodeName, Variable time)
            : base(transform, generalizedCoordinates, generalizedForces,
            centerOfMass, inertiaMatrix, mass, nodeName, time) {
            name = nodeName;
            colorWhenSelected.X = colorWhenSelected.Y = colorWhenSelected.Z = colorWhenSelected.W = 1.0f;
            displayAxis = false;
        }

        // name        
        public string name;

        //visualization
        //Some of below members are XNA dependent.
        public Geometry geometry;   // geometry. we assume that only one geometry will be assigned to each link.
        public VertexBuffer vertexBuffer;       // VertexBuffer for geometry
        public IndexBuffer indexBuffer;         // IndexBuffer for geometry
        private VertexDeclaration vertexDeclaration; //VertexDeclaration
        public Texture2D texture;
        public int numU, numV;
        public Matrix matWorld;
        public BoundingSphere bsphere;
        public Vector4 colorWhenSelected;
        public bool displayAxis;
        VertexPositionColor[] axis;
        public Variable angleX, angleY, angleZ;

        public bool isParametric = false;
        public ParametricModelDStar3DVisualizer parametricModel;

        //internal Variable[] angles; // orientation of the object
        //internal Variable[] pos; // position of the object;





        public float reflectivity;        //reflectivity

        public void CreateGeometry(GraphicsDevice device, Geometry _geometry, Texture2D texture) {
            CreateGeometry(device, _geometry, 40, 40, texture);
        }

        public void CreateGeometry(GraphicsDevice device, Geometry inputGeometry, int numU, int numV, Texture2D texture) {
            const bool ccw = true;
            this.numU = numU;
            this.numV = numV;
            this.texture = texture;
            CreateMeshFromGeometry(device, inputGeometry, ccw);
        }


        /// <summary>
        /// creates a triangle list mesh given geometry.
        /// </summary>
        /// <param name="device"></param>
        /// <param name="geometry"></param>
        /// <param name="ccw">creates counterclockwise triangles if ccw is true. clockwise triangles if false.</param>
        /// <param name="args">additional input to geometry'a runtimeFunction</param>
        void CreateMeshFromGeometry(GraphicsDevice device, Geometry geometry, bool ccw, params double[] args) {
            //setup vertices
            int cnt = 0;
            double[] result = new double[6];
            VertexPositionNormalTexture[] vertices = new VertexPositionNormalTexture[numU * numV];
            double uLow = geometry.Domain(0).low;
            double vLow = geometry.Domain(1).low;
            double uRng = geometry.Domain(0).up - uLow;
            double vRng = geometry.Domain(1).up - vLow;
            double minX = double.MaxValue;
            double minY = double.MaxValue;
            double minZ = double.MaxValue;
            double maxX = double.MinValue;
            double maxY = double.MinValue;
            double maxZ = double.MinValue;

            double[] input = new double[2 + args.Length]; // two for uv + additional coordinates

            for (int i = 0; i < numV; i++) {
                double v = vLow + vRng * (double)i / (numV - 1);
                for (int j = 0; j < numU; j++) {
                    double u = uLow + uRng * (double)j / (numU - 1);

                    input[0] = u;
                    input[1] = v;
                    for (int k = 0; k < args.Length; k++) input[k + 2] = args[k];

                    geometry.SurfaceAndNormal(result, input);

                    minX = Math.Min(minX, result[0]);
                    maxX = Math.Max(maxX, result[0]);
                    minY = Math.Min(minY, result[1]);
                    maxY = Math.Max(maxY, result[1]);
                    minZ = Math.Min(minZ, result[2]);
                    maxZ = Math.Max(maxZ, result[2]);

                    vertices[cnt].Position = new Vector3((float)result[0], (float)result[1], (float)result[2]);
                    vertices[cnt].Normal = new Vector3((float)result[3], (float)result[4], (float)result[5]);
                    vertices[cnt].Normal.Normalize(); //or, normalization can be performed in GPU
                    vertices[cnt].TextureCoordinate.X = (float)j / (float)(numU - 1);
                    vertices[cnt].TextureCoordinate.Y = (float)i / (float)(numV - 1);
                    cnt++;
                }
            }

            //vertexBuffer = new VertexBuffer(device, VertexPositionNormalTexture.SizeInBytes * numU * numV, ResourceUsage.WriteOnly);
            ///FOR PERFORMANCE, IT IS BETTER TO USE "WriteOnly" Usage, but then we cannot read back the vertex buffer, which
            ///is necessary for intersection testing between mouse cursor and rigid bodies.
            vertexBuffer = new VertexBuffer(device, VertexPositionNormalTexture.SizeInBytes * numU * numV, BufferUsage.None);
            vertexBuffer.SetData(vertices);
            //initialize boundingSphere
            Vector3 minv = new Vector3((float)minX, (float)minY, (float)minZ);
            Vector3 maxv = new Vector3((float)maxX, (float)maxY, (float)maxZ);
            Vector3 center = 0.5f * (minv + maxv);
            bsphere = new BoundingSphere(center, 0.5f * Vector3.Distance(maxv, minv));

            axis = new VertexPositionColor[6];
            axis[0] = new VertexPositionColor(center, Color.Red);
            axis[1] = new VertexPositionColor(center + new Vector3(2.1f, 0, 0), new Color(255, 0, 0));
            axis[2] = new VertexPositionColor(center, Color.Green);
            axis[3] = new VertexPositionColor(center + new Vector3(0, 2.1f, 0), new Color(0, 255, 0));
            axis[4] = new VertexPositionColor(center, Color.Blue);
            axis[5] = new VertexPositionColor(center + new Vector3(0, 0, 2.1f), new Color(0, 0, 255));

            //setup indices for TriangleList
            int[] indices = new int[2 * 3 * (numV - 1) * (numU - 1)];
            cnt = 0;
            for (int j = 0; j < numV - 1; j++) {
                for (int i = 0; i < numU - 1; i++) {
                    int idx = j * numU + i;
                    if (ccw) {
                        indices[cnt++] = idx;
                        indices[cnt++] = idx + numU;
                        indices[cnt++] = idx + numU + 1;

                        indices[cnt++] = idx;
                        indices[cnt++] = idx + numU + 1;
                        indices[cnt++] = idx + 1;
                    }
                    else {
                        indices[cnt++] = idx;
                        indices[cnt++] = idx + numU + 1;
                        indices[cnt++] = idx + numU;

                        indices[cnt++] = idx;
                        indices[cnt++] = idx + 1;
                        indices[cnt++] = idx + numU + 1;
                    }
                }
            }
            indexBuffer = new IndexBuffer(device, typeof(int), indices.Length, BufferUsage.WriteOnly);
            indexBuffer.SetData(indices);
            vertexDeclaration = new VertexDeclaration(device, VertexPositionNormalTexture.VertexElements);
        }

        /// <summary>
        /// draw using custom effect
        /// </summary>
        public void Draw(GraphicsDevice device, Effect effect, PipelineMatrices pm, Vector4 color) {
            effect.Parameters["Texture"].SetValue(texture);
            effect.Parameters["World"].SetValue(matWorld);
            effect.Parameters["WorldViewProjection"].SetValue(matWorld * pm.viewMatrix * pm.projectionMatrix);
            effect.Parameters["LightWorldViewProjection"].SetValue(matWorld * pm.lightViewMatrix * pm.projectionMatrix);
            effect.Parameters["Reflectance"].SetValue(reflectivity);
            effect.Parameters["objectSelectedColor"].SetValue(color);
            effect.Begin();

            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
                device.VertexDeclaration = new VertexDeclaration(device, VertexPositionNormalTexture.VertexElements);
                device.Indices = indexBuffer;
                device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, numU * numV, 0, 2 * (numU - 1) * (numV - 1));
                pass.End();
            }
            effect.Parameters["objectSelectedColor"].SetValue(new Vector4(1, 1, 1, 1));
            effect.End();
        }

        public void DrawParametricModel(GraphicsDevice device, Effect effect, PipelineMatrices pm, Vector4 color) {
            if (texture != null)
                effect.Parameters["Texture"].SetValue(texture);
            effect.Parameters["World"].SetValue(matWorld);
            effect.Parameters["WorldViewProjection"].SetValue(matWorld * pm.viewMatrix * pm.projectionMatrix);
            effect.Parameters["LightWorldViewProjection"].SetValue(matWorld * pm.lightViewMatrix * pm.projectionMatrix);
            effect.Parameters["Reflectance"].SetValue(reflectivity);
            if (color.Equals(new Vector4(1, 0, 0, (float)0.5)))
                effect.Parameters["objectSelectedColor"].SetValue(color);
            else
                effect.Parameters["materialColor"].SetValue(parametricModel._materialColor);
            effect.Begin();

            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(parametricModel._vertexBufferPositionNormal, 0, VertexPositionNormal.SizeInBytes);
                device.VertexDeclaration = new VertexDeclaration(device, VertexPositionNormal.Elements);

                device.Indices = indexBuffer;
                device.DrawUserIndexedPrimitives(PrimitiveType.TriangleStrip, parametricModel._verticesPositionNormal
                                        , 0, parametricModel._numVertices, parametricModel._indices, 0, parametricModel._numTriangles);
                pass.End();
            }
            effect.Parameters["objectSelectedColor"].SetValue(new Vector4(1, 1, 1, 1));
            effect.Parameters["materialColor"].SetValue(new Vector4(1, 1, 1, 1));
            effect.End();

        }



        BasicEffect basicEffect = null;

        //draw Axis using BasicEffect
        public void DrawAxis(GraphicsDevice device, PipelineMatrices pm) {
            device.VertexDeclaration = new VertexDeclaration(device, VertexPositionColor.VertexElements);
            if (basicEffect == null) {
                basicEffect = new BasicEffect(device, null);
                basicEffect.LightingEnabled = false;
                basicEffect.VertexColorEnabled = true;
            }

            basicEffect.World = this.matWorld;
            basicEffect.View = pm.viewMatrix;
            basicEffect.Projection = pm.projectionMatrix;

            basicEffect.Begin();
            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes) {
                pass.Begin();



                device.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, axis, 0, 3);

                pass.End();
            }
            basicEffect.End();
        }

        //draw using BasicEffect
        public void DrawBasic(GraphicsDevice device, BasicEffect effect, PipelineMatrices pm) {
            effect.World = pm.worldMatrix * this.matWorld;
            effect.Texture = texture;

            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
                device.VertexDeclaration = vertexDeclaration;
                device.Indices = indexBuffer;
                device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, numU * numV, 0, 2 * (numU - 1) * (numV - 1));
                pass.End();
            }
            effect.End();
        }



        /// <summary>
        /// compute the world space coordinate of a point given its local space coordinate
        /// </summary>
        /// <param name="pointLocal"></param>
        /// <returns></returns>
        public Vector3 pointInWorldFrame(Vector3 pointLocal) {
            return Vector3.Transform(pointLocal, matWorld);
        }

        //public Variable[] position {
        //    set {
        //        pos = (Variable[])value.Clone();
        //    }
        //}

        //public Variable[] orientation {
        //    set {
        //        angles = (Variable[])value.Clone();
        //    }
        //}



    }


    /// <summary>
    /// Rigid body system
    /// </summary>
    public class RigidBodySystem : MultiBodyDynamics {
        //Ground ground;
        public INTEGRATION_METHOD integrationMethod = INTEGRATION_METHOD.EXPLICIT_EULER; //default

        public MATRIX_SOLVER matrixSolver = MATRIX_SOLVER.LU; //default
        public double epsSVD;  //parameter exclusively when matrixSolver is SVD

        public FORWARD_DYNAMICS_METHOD forwardDynMethod = FORWARD_DYNAMICS_METHOD.COMPOSITE_RIGID_BODY_METHOD;
        //public FORWARD_DYNAMICS_METHOD forwardDynMethod = FORWARD_DYNAMICS_METHOD.NEW_METHOD;

        public RUNTIME_FUNCTION_MODE runtimeFunctionMode = RUNTIME_FUNCTION_MODE.FULL;
        //public RUNTIME_FUNCTION_MODE runtimeFunctionMode = RUNTIME_FUNCTION_MODE.MINIMAL; //since Jacobian is not defined, userSpring must not used at all!!
        /// 
        /// constants
        /// 
        public double[] gravity = new double[3];    // defining per system gravity is useful
        public float timeStep = .016f;

        ///
        /// coordinates and torques
        /// 
        public double[] q;      // current value of generalized coordinates (STATE) 
        public double[] dq;     // current value of generalized velocities (STATE)
        public double[] mass;   // current value of mass (STATE) values obtained from the GUI
        public double[] pos;    // current value of position (STATE) values obtained from the GUI
        public double[] angle; // current value of orientation(STATE) values obtained from the GUI
        public double[] force; // current value of force(STATE) values obtained from the GUI
        public double[] torque; // current value of torque(STATE) values obtained from the GUI
        public double[] centerOfMass; // current value of center of mass (STATE) values obtained from the GUI
        public double[] inputGUIVariables;
        public bool motion = false;


        //
        public double[] ddq;    // current value of generalized accelerations (OUTPUT OF THE FORWARYD DYNAMICS)


        public int nCon = 0;    // number of constraints (induced by closed-loops or rolling contact conditions)
        public double[][,] W;   // transforms for each rigid body]


        public RigidBodySystem(Graph<JointNode> _systemGraph, Variable t, GraphicsDevice device)
            : base(_systemGraph, t) {
            //ddq = new double[_systemGraph.numberOfNodes];
        }

        public RigidBodySystem(Graph<JointNode> _systemGraph, Variable t, GraphicsDevice device, double[] mass,
            double[] pos, double[] angles, double[] force, double[] torque, double[] centerOfMass)
            : base(_systemGraph, t, mass) {

            int i = 0, j = 0, tempSize = 0;
            this.mass = mass;
            this.pos = pos;
            this.angle = angles;
            this.force = force;
            this.torque = torque;
            this.centerOfMass = centerOfMass;
            inputGUIVariables = VM.concatenate<double>(mass, pos, angle, force, torque, centerOfMass);

            if (deletedGuiVarsIndices != null) {
                for (i = 0; i < deletedGuiVarsIndices.GetLength(0); ++i) {
                    if (deletedGuiVarsIndices[i] != -1) {
                        inputGUIVariables.SetValue(-1, deletedGuiVarsIndices[i]);
                    }
                    else {
                        tempSize++;
                    }
                }

                double[] temp = new double[tempSize];

                for (i = 0; i < inputGUIVariables.GetLength(0); ++i) {
                    if (inputGUIVariables[i] != -1) {
                        temp[j++] = inputGUIVariables[i];
                    }
                }

                inputGUIVariables = temp;
            }

        }


        //public double[] initialQ
        //{
        //    get { return (double[]) q.Clone(); }
        //    set
        //    {
        //        if (value.Length == q.Length) { Array.Copy(value, q, q.Length); }
        //        else { throw new Exception("incorrect length array for initial q values"); }
        //    }
        //}
        public void initialize(double[] qinit, double[] qdinit) {
            q = new double[numberOfCoords];
            dq = new double[numberOfCoords];
            if (qinit.Length != numberOfCoords || qdinit.Length != numberOfCoords || qinit.Length != qdinit.Length) { throw new Exception(); }
            for (int i = 0; i < qinit.Length; i++) {
                q[i] = qinit[i];
                dq[i] = qdinit[i];
            }
        }

        public void initialize(double[] qinit) {
            q = new double[numberOfCoords];
            dq = new double[numberOfCoords];
            //if (qinit.Length != numberOfCoords ||  qinit.Length != qdinit.Length) { throw new Exception(); }
            for (int i = 0; i < qinit.Length; i++) {
                q[i] = qinit[i];
                //dq[i] = qdinit[i];
            }
        }

        public void initialize() {
            int k = 0;
            q = new double[numberOfCoords];
            dq = new double[numberOfCoords];
        }

        public void updateInputGUIArray() {
            int i = 0, j = 0, tempSize = 0;
            inputGUIVariables = VM.concatenate<double>(mass, pos, angle, force, torque, centerOfMass);

            for (i = 0; i < deletedGuiVarsIndices.GetLength(0); ++i) {
                if (deletedGuiVarsIndices[i] != -1) {
                    inputGUIVariables.SetValue(-1, deletedGuiVarsIndices[i]);
                }
                else {
                    tempSize++;
                }
            }

            double[] temp = new double[tempSize];

            for (i = 0; i < inputGUIVariables.GetLength(0); ++i) {
                if (inputGUIVariables[i] != -1) {
                    temp[j++] = inputGUIVariables[i];
                }
            }
            inputGUIVariables = temp;
            //UnspecifiedFunction[] t1 = genCoords1D;

        }

        public void updateMass(int index, double mass) {
            this.mass[index] = mass;
            updateInputGUIArray();

        }

        public void updatePos(int index, Vector3 pos) {
            this.pos[index * 3] = pos.X;
            this.pos[index * 3 + 1] = pos.Y;
            this.pos[index * 3 + 2] = pos.Z;
            updateInputGUIArray();
        }

        public void updatePos(int index, double x, double y, double z) {
            this.pos[index * 3] = x;
            this.pos[index * 3 + 1] = y;
            this.pos[index * 3 + 2] = z;
            updateInputGUIArray();
        }

        public void updateOrientation(int index, Vector3 orientation) {
            this.angle[index * 3] = orientation.X * Math.PI / 180.0;
            this.angle[index * 3 + 1] = orientation.Y * Math.PI / 180.0;
            this.angle[index * 3 + 2] = orientation.Z * Math.PI / 180.0;
            updateInputGUIArray();
        }

        public void updateForces(int index, Vector3 force) {
            this.force[index * 3] = force.X;
            this.force[index * 3 + 1] = force.Y;
            this.force[index * 3 + 2] = force.Z;
            updateInputGUIArray();
        }

        public void updateTorque(int index, Vector3 torque) {
            this.torque[index * 3] = torque.X;
            this.torque[index * 3 + 1] = torque.Y;
            this.torque[index * 3 + 2] = torque.Z;
            updateInputGUIArray();
        }

        public void updateCenterOfMass(int index, Vector3 com) {
            this.centerOfMass[index * 3] = com.X;
            this.centerOfMass[index * 3 + 1] = com.Y;
            this.centerOfMass[index * 3 + 2] = com.Z;
            updateInputGUIArray();
        }


        public void updateGenCoords(int index, double q, string qName) {
            int i = Convert.ToInt32(qName);
            string genCoordName;
            if (i < this.generalizedCoordinates[index].Length) {
                genCoordName = this.generalizedCoordinates[index][i].ToString();
                genCoordName = genCoordName.Substring(1);
                this.q[Convert.ToInt32(genCoordName)] = q;
            }

        }

        //proceed to the next timestep
        public void NextStep(double displayTimeStep, UserSpring userSpring) {
            const int stepMult = 2;
            for (int i = 0; i < stepMult; i++) {

                integrate(displayTimeStep / (double)stepMult, q, dq, out q, out dq, out W);
            }
            UpdateKinematics(W);
            //for (int i = 0; i < W.Length; i++) { links[i].matWorld = stupidProgramTheXNAPeopleShouldHaveWritten(W[i]);}
        }

        public void NextStepUsingGUIVars(double displayTimeStep, UserSpring userSpring) {
            const int stepMult = 2;
            for (int i = 0; i < stepMult; i++) {

                integrate(displayTimeStep / (double)stepMult, inputGUIVariables, q, dq, out q, out dq, out W);
            }
            UpdateKinematics(W);
            //for (int i = 0; i < W.Length; i++) { links[i].matWorld = stupidProgramTheXNAPeopleShouldHaveWritten(W[i]);}
        }
        ///
        ///this version only uses kinematics
        ///
        public void NextStepUsingGUIVars() {
            kinematicsUpdate(inputGUIVariables, q, out W);
            UpdateKinematics(W);
        }




        public void UpdateKinematics(double[][,] W) {
            RigidBody[] links = (RigidBody[])this.allNodes;
            for (int i = 0; i < W.Length; i++) { links[i].matWorld = stupidProgramTheXNAPeopleShouldHaveWritten(W[i]); }
        }

        protected Matrix stupidProgramTheXNAPeopleShouldHaveWritten(double[,] transform) {
            return new Matrix(
                (float)transform[0, 0], (float)transform[1, 0], (float)transform[2, 0], (float)transform[3, 0],
                (float)transform[0, 1], (float)transform[1, 1], (float)transform[2, 1], (float)transform[3, 1],
                (float)transform[0, 2], (float)transform[1, 2], (float)transform[2, 2], (float)transform[3, 2],
                (float)transform[0, 3], (float)transform[1, 3], (float)transform[2, 3], (float)transform[3, 3]);
        }
        //int nIter = Math.Max(1, (int)(displayTimeStep / timeStep) + 1);
        //if (integrationMethod == INTEGRATION_METHOD.EXPLICIT_EULER)
        //{
        //    for (int i = 0; i < nIter; ++i)
        //    {
        //        NextStepExplicitEuler(userSpring);
        //    }
        //}
        //else if (integrationMethod == INTEGRATION_METHOD.RK4)
        //{
        //    for (int i = 0; i < nIter; ++i)
        //    {
        //        NextStepRK4(userSpring);
        //    }
        //}
        //else
        //    throw new Exception("Unsupported integration method specified");
        /// <summary>
        /// Implementation of 4th order Runge Kutta Method.
        /// Refer to http://en.wikipedia.org/wiki/Runge-Kutta
        /// </summary>
        /// <param name="userSpring"></param>
        /// 
        public void NextStepRK4(UserSpring userSpring) {

            double[] ddq;
            double[][,] W;

            double h2 = timeStep * 0.5;
            double h3 = timeStep / 3.0;
            double h6 = timeStep / 6.0;


            double[] q0 = new double[numberOfCoords];
            double[] dq0 = new double[numberOfCoords];
            q.CopyTo(q0, 0);
            dq.CopyTo(dq0, 0);


            double[] qu = new double[numberOfCoords];
            double[] dqu = new double[numberOfCoords];
            q.CopyTo(qu, 0);
            dq.CopyTo(dqu, 0);

            //k1
            // SetTorque(userSpring);
            //  ComputeAccelerations(matrixSolver);
            //now q,dq are k1. Update qu     
            qddt(q, dq, out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h6));
            dqu = VM.plus(dqu, VM.mult(ddq, h6));

            //k2
            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            UpdateKinematics(W);
            //SetTorque(userSpring);
            //ComputeAccelerations(matrixSolver);
            //now q,dq are k2. Update qu
            qddt(q, dq, out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            //k3
            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            UpdateKinematics(W);
            //SetTorque(userSpring);
            // ComputeAccelerations(matrixSolver);
            //now q,dq are k3. Update qu
            qddt(q, dq, out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            //k4
            q = VM.plus(q0, VM.mult(dq, timeStep));
            dq = VM.plus(dq0, VM.mult(ddq, timeStep));
            UpdateKinematics(W);
            //SetTorque(userSpring);
            //ComputeAccelerations(matrixSolver);
            //now q,dq are k4. Update qu = q
            qddt(q, dq, out ddq, out W);
            q = VM.plus(qu, VM.mult(dq, h6));
            dq = VM.plus(dqu, VM.mult(ddq, h6));
            UpdateKinematics(W);
        }

        public void NextStepRK4usingGUIVars(UserSpring userSpring) {

            double[] ddq;
            double[][,] W;

            double h2 = timeStep * 0.5;
            double h3 = timeStep / 3.0;
            double h6 = timeStep / 6.0;


            double[] q0 = new double[numberOfCoords];
            double[] dq0 = new double[numberOfCoords];
            q.CopyTo(q0, 0);
            dq.CopyTo(dq0, 0);


            double[] qu = new double[numberOfCoords];
            double[] dqu = new double[numberOfCoords];
            q.CopyTo(qu, 0);
            dq.CopyTo(dqu, 0);

            //k1
            // SetTorque(userSpring);
            //  ComputeAccelerations(matrixSolver);
            //now q,dq are k1. Update qu     
            qddt(inputGUIVariables, q, dq, out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h6));
            dqu = VM.plus(dqu, VM.mult(ddq, h6));

            //k2
            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            UpdateKinematics(W);
            //SetTorque(userSpring);
            //ComputeAccelerations(matrixSolver);
            //now q,dq are k2. Update qu
            qddt(inputGUIVariables, q, dq, out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            //k3
            q = VM.plus(q0, VM.mult(dq, h2));
            dq = VM.plus(dq0, VM.mult(ddq, h2));
            UpdateKinematics(W);
            //SetTorque(userSpring);
            // ComputeAccelerations(matrixSolver);
            //now q,dq are k3. Update qu
            qddt(inputGUIVariables, q, dq, out ddq, out W);
            qu = VM.plus(qu, VM.mult(dq, h3));
            dqu = VM.plus(dqu, VM.mult(ddq, h3));

            //k4
            q = VM.plus(q0, VM.mult(dq, timeStep));
            dq = VM.plus(dq0, VM.mult(ddq, timeStep));
            UpdateKinematics(W);
            //SetTorque(userSpring);
            //ComputeAccelerations(matrixSolver);
            //now q,dq are k4. Update qu = q
            qddt(inputGUIVariables, q, dq, out ddq, out W);
            q = VM.plus(qu, VM.mult(dq, h6));
            dq = VM.plus(dqu, VM.mult(ddq, h6));
            UpdateKinematics(W);
        }

        //public void NextStepExplicitEuler(UserSpring userSpring)
        //{
        //    /// set torque                
        //    //SetTorque(userSpring);

        //    //compute accelerations
        //    //ComputeAccelerations(matrixSolver);

        //    //explicit Euler method
        //    for (int i = 0; i < nDof; i++)
        //    {
        //        q[i] += timeStep * dq[i];
        //        dq[i] += timeStep * ddq[i];
        //    }

        //    //update kinematics
        //    //UpdateKinematics();
        //}

        /// <summary>
        /// draw with custom effect
        /// </summary>
        /// <param name="device"></param>
        /// <param name="effect"></param>
        /// <param name="pm"></param>
        public void Draw(GraphicsDevice device, Effect effect, PipelineMatrices pm) {
            foreach (RigidBody rb in this.allNodes) {
                if (rb.isParametric) {
                    rb.DrawParametricModel(device, effect, pm, rb.colorWhenSelected);
                    //rb.parametricModel.OnRender(device, 1);
                }
                else {
                    if (rb.vertexBuffer != null) {
                        rb.Draw(device, effect, pm, rb.colorWhenSelected);
                    }

                }
            }
        }

        public void DrawAxis(GraphicsDevice device, PipelineMatrices pm) {
            foreach (RigidBody rb in this.allNodes) {
                if (rb.vertexBuffer != null) {
                    if (rb.geometry != null) {
                        if (rb.displayAxis == true)
                            rb.DrawAxis(device, pm);
                    }
                    else if (rb.parametricModel != null) {
                        if (rb.displayAxis == true)
                            rb.DrawAxis(device, pm);
                    }
                }
            }
        }

        /// <summary>
        /// draw with BasicEffect
        /// </summary>
        /// <param name="device"></param>
        /// <param name="effect"></param>
        /// <param name="pm"></param>
        public void DrawBasic(GraphicsDevice device, BasicEffect effect, PipelineMatrices pm) {
            foreach (RigidBody rb in this.allNodes) {
                if (rb.vertexBuffer != null) rb.DrawBasic(device, effect, pm);
            }
            //foreach (Link link in links)
            //{
            //    if (link.vertexBuffer != null) link.DrawBasic(device, effect, pm);
            //}
        }

        /*
        private void DrawConnectors(GraphicsDevice device, Effect effect, PipelineMatrices pm)
        {
            EffectTechnique oldTechnique = effect.CurrentTechnique;
            foreach (LineSegments c in connectors)
            {
                //generate vertices every time...inefficient.
                VertexPositionColor[] vertices = new VertexPositionColor[c.viaPoints.Count];
                for (int i = 0; i < c.viaPoints.Count; i++)
                {
                    vertices[i] = new VertexPositionColor(
                        Vector3.Transform(c.viaPoints[i].localPos, c.viaPoints[i].link.matWorld),
                        Color.Brown);
                }

                effect.CurrentTechnique = effect.Techniques["Colored"];                
                Matrix matWorld = Matrix.Identity;
                effect.Parameters["World"].SetValue(matWorld);
                effect.Parameters["WorldViewProjection"].SetValue(matWorld * pm.viewMatrix * pm.projectionMatrix);
                effect.Parameters["LightWorldViewProjection"].SetValue(matWorld * pm.lightViewMatrix * pm.projectionMatrix);
           
                effect.Begin();
                foreach (EffectPass pass in effect.CurrentTechnique.Passes)
                {
                    pass.Begin();

                    device.VertexDeclaration = new VertexDeclaration(device, VertexPositionColor.VertexElements);
                    device.DrawUserPrimitives(PrimitiveType.LineStrip, vertices, 0, c.viaPoints.Count - 1);

                    pass.End();
                }
                effect.End();
            }
            effect.CurrentTechnique = oldTechnique;
        }
         * */
    }


    /// <summary>
    /// This is a rigid body in the rigid body system
    /// </summary>

    public class Link {
        // name        
        public string name;



        //visualization
        //Some of below members are XNA dependent.
        public Geometry geometry;   // geometry. we assume that only one geometry will be assigned to each link.
        public VertexBuffer vertexBuffer;       // VertexBuffer for geometry
        public IndexBuffer indexBuffer;         // IndexBuffer for geometry
        private VertexDeclaration vertexDeclaration; //VertexDeclaration
        public Texture2D texture;
        public int numU, numV;
        public Matrix matWorld;
        public BoundingSphere bsphere;
        public float reflectivity;        //reflectivity

        [ObsoleteAttribute]
        public Link() { }

        public Link(GraphicsDevice device, Geometry _geometry, Texture2D texture) {
            CreateGeometry(device, _geometry, 20, 20, texture);
        }

        public void CreateGeometry(GraphicsDevice device, Geometry inputGeometry, int numU, int numV, Texture2D texture) {
            const bool ccw = true;
            this.numU = numU;
            this.numV = numV;
            this.texture = texture;
            CreateMeshFromGeometry(device, inputGeometry, ccw);
        }

        /// <summary>
        /// geometry variable must be initialized before calling this function
        /// </summary>
        /// <param name="device"></param>
        /// <param name="numU"></param>
        /// <param name="numV"></param>
        /// <param name="texture"></param>
        protected void CreateGeometry(GraphicsDevice device, int numU, int numV, Texture2D texture) {
            const bool ccw = true;
            this.numU = numU;
            this.numV = numV;
            this.texture = texture;
            CreateMeshFromGeometry(device, geometry, ccw);
        }

        /// <summary>
        /// create a triangle mesh 
        /// </summary>
        /// <param name="device"></param>
        /// <param name="type">"Cylinder" for Cylinder, "Sphere" for Sphere, and etc.</param>
        /// <param name="numU">number of vertices in u direction</param>
        /// <param name="numV">numver of vertices in v direction</param>
        /// <param name="texture"></param>
        /// <param name="param">parameters to define Geometry</param>
        public void CreateGeometry(GraphicsDevice device, string type, int numU, int numV, Texture2D texture, bool createMesh, double[] param) {
            this.numU = numU;
            this.numV = numV;
            this.texture = texture;

            const bool ccw = true;
            switch (type) {
                case "CLELIES":
                    geometry = new Clelies(param[0], param[1], param[2]);
                    break;
                case "WAVY_CIRCLE":
                    geometry = new WavyCircle(param[0], param[1], param[2], (int)param[3]);
                    break;
                case "CIRCULAR_COIL":
                    geometry = new CircularCoil(param[0], param[1], param[2], (int)param[3]);
                    break;
                case "SPHERE":
                    geometry = new Sphere(param[0]);
                    break;
                case "CYLINDER":
                    geometry = new Cylinder(param[0], param[1], param[2], param[3], param[4]);
                    break;
                case "CONE":
                    geometry = new Cone(param[0], param[1], param[2]);
                    break;
                case "ELLIPSOID":
                    bool outwardNormal = param[7] == 0 ? false : true;
                    geometry = new Ellipsoid(param[0], param[1], param[2], param[3], param[4], param[5], param[6], outwardNormal);
                    break;
                case "XZ_DISK":
                    geometry = new XZDisk(param[0], param[1], param[2]);
                    break;
                case "XZ_PLANE":
                    geometry = new XZPlane(param[0], param[1],
                        param[2], param[3], param[4]);
                    break;
                case "TORUS":
                    geometry = new Torus(param[0], param[1]);
                    break;
                default:
                    throw new ArgumentException("undefined geometry specified");
            }

            if (createMesh) CreateMeshFromGeometry(device, geometry, ccw);
        }

        /// <summary>
        /// creates a triangle list mesh given geometry.
        /// </summary>
        /// <param name="device"></param>
        /// <param name="geometry"></param>
        /// <param name="ccw">creates counterclockwise triangles if ccw is true. clockwise triangles if false.</param>
        /// <param name="args">additional input to geometry'a runtimeFunction</param>
        void CreateMeshFromGeometry(GraphicsDevice device, Geometry geometry, bool ccw, params double[] args) {
            //setup vertices
            int cnt = 0;
            double[] result = new double[6];
            VertexPositionNormalTexture[] vertices = new VertexPositionNormalTexture[numU * numV];
            double uLow = geometry.Domain(0).low;
            double vLow = geometry.Domain(1).low;
            double uRng = geometry.Domain(0).up - uLow;
            double vRng = geometry.Domain(1).up - vLow;
            double minX = double.MaxValue;
            double minY = double.MaxValue;
            double minZ = double.MaxValue;
            double maxX = double.MinValue;
            double maxY = double.MinValue;
            double maxZ = double.MinValue;

            double[] input = new double[2 + args.Length]; // two for uv + additional coordinates

            for (int i = 0; i < numV; i++) {
                double v = vLow + vRng * (double)i / (numV - 1);
                for (int j = 0; j < numU; j++) {
                    double u = uLow + uRng * (double)j / (numU - 1);

                    input[0] = u;
                    input[1] = v;
                    for (int k = 0; k < args.Length; k++) input[k + 2] = args[k];

                    geometry.SurfaceAndNormal(result, input);

                    minX = Math.Min(minX, result[0]);
                    maxX = Math.Max(maxX, result[0]);
                    minY = Math.Min(minY, result[1]);
                    maxY = Math.Max(maxY, result[1]);
                    minZ = Math.Min(minZ, result[2]);
                    maxZ = Math.Max(maxZ, result[2]);

                    vertices[cnt].Position = new Vector3((float)result[0], (float)result[1], (float)result[2]);
                    vertices[cnt].Normal = new Vector3((float)result[3], (float)result[4], (float)result[5]);
                    vertices[cnt].Normal.Normalize(); //or, normalization can be performed in GPU
                    vertices[cnt].TextureCoordinate.X = (float)j / (float)(numU - 1);
                    vertices[cnt].TextureCoordinate.Y = (float)i / (float)(numV - 1);
                    cnt++;
                }
            }

            //vertexBuffer = new VertexBuffer(device, VertexPositionNormalTexture.SizeInBytes * numU * numV, ResourceUsage.WriteOnly);
            ///FOR PERFORMANCE, IT IS BETTER TO USE "WriteOnly" Usage, but then we cannot read back the vertex buffer, which
            ///is necessary for intersection testing between mouse cursor and rigid bodies.
            vertexBuffer = new VertexBuffer(device, VertexPositionNormalTexture.SizeInBytes * numU * numV, BufferUsage.None);
            vertexBuffer.SetData(vertices);
            //initialize boundingSphere
            Vector3 minv = new Vector3((float)minX, (float)minY, (float)minZ);
            Vector3 maxv = new Vector3((float)maxX, (float)maxY, (float)maxZ);
            bsphere = new BoundingSphere(0.5f * (minv + maxv), 0.5f * Vector3.Distance(maxv, minv));

            //setup indices for TriangleList
            int[] indices = new int[2 * 3 * (numV - 1) * (numU - 1)];
            cnt = 0;
            for (int j = 0; j < numV - 1; j++) {
                for (int i = 0; i < numU - 1; i++) {
                    int idx = j * numU + i;
                    if (ccw) {
                        indices[cnt++] = idx;
                        indices[cnt++] = idx + numU;
                        indices[cnt++] = idx + numU + 1;

                        indices[cnt++] = idx;
                        indices[cnt++] = idx + numU + 1;
                        indices[cnt++] = idx + 1;
                    }
                    else {
                        indices[cnt++] = idx;
                        indices[cnt++] = idx + numU + 1;
                        indices[cnt++] = idx + numU;


                        indices[cnt++] = idx;
                        indices[cnt++] = idx + 1;
                        indices[cnt++] = idx + numU + 1;
                    }
                }
            }
            //indexBuffer = new IndexBuffer(device, typeof(int), indices.Length, BufferUsage.WriteOnly, ResourceManagementMode.Automatic);
            indexBuffer = new IndexBuffer(device, typeof(int), indices.Length, BufferUsage.WriteOnly);
            indexBuffer.SetData(indices);

            vertexDeclaration = new VertexDeclaration(device, VertexPositionNormalTexture.VertexElements);

        }


        /// <summary>
        /// draw using custom effect
        /// </summary>
        public void Draw(GraphicsDevice device, Effect effect, PipelineMatrices pm) {
            effect.Parameters["Texture"].SetValue(texture);
            effect.Parameters["World"].SetValue(matWorld);
            effect.Parameters["WorldViewProjection"].SetValue(matWorld * pm.viewMatrix * pm.projectionMatrix);
            effect.Parameters["LightWorldViewProjection"].SetValue(matWorld * pm.lightViewMatrix * pm.projectionMatrix);
            effect.Parameters["Reflectance"].SetValue(reflectivity);

            effect.Begin();

            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
                device.VertexDeclaration = vertexDeclaration;
                device.Indices = indexBuffer;
                device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, numU * numV, 0, 2 * (numU - 1) * (numV - 1));
                pass.End();
            }

            effect.End();
        }




        //draw using BasicEffect
        public void DrawBasic(GraphicsDevice device, BasicEffect effect, PipelineMatrices pm) {
            effect.World = pm.worldMatrix * this.matWorld;
            effect.Texture = texture;

            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
                device.VertexDeclaration = vertexDeclaration;
                device.Indices = indexBuffer;
                device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, numU * numV, 0, 2 * (numU - 1) * (numV - 1));
                pass.End();
            }
            effect.End();
        }



        /// <summary>
        /// compute the world space coordinate of a point given its local space coordinate
        /// </summary>
        /// <param name="pointLocal"></param>
        /// <returns></returns>
        public Vector3 pointInWorldFrame(Vector3 pointLocal) {
            return Vector3.Transform(pointLocal, matWorld);
        }
    }




    public class Ground {
        public Texture2D texture;          //TexCoord for ground        
        public int numWrapsU = 1;          //number of wraps of ground texture
        public int numWrapsV = 1;          //number of wraps of ground texture
        public Vector3 center;             //position of the center
        public float width, depth;         //lengths in x, z
        public float reflectivity;        //reflectivity
        VertexBuffer vertexBuffer;  //VertexBuffer for ground
        VertexDeclaration vertexDeclaration;


        public void Create(GraphicsDevice device) {
            float xMin = center.X - width / 2;
            float xMax = center.X + width / 2;
            float zMin = center.Z - depth / 2;
            float zMax = center.Z + depth / 2;
            float height = center.Y;
            //create ground
            VertexPositionNormalTexture[] vertices = new VertexPositionNormalTexture[2 * 3]; //two triangles
            vertices[0] = new VertexPositionNormalTexture(new Vector3(xMin, height, zMax), Vector3.UnitY, new Vector2(0, 0));
            vertices[1] = new VertexPositionNormalTexture(new Vector3(xMin, height, zMin), Vector3.UnitY, new Vector2(0, numWrapsV));
            vertices[2] = new VertexPositionNormalTexture(new Vector3(xMax, height, zMin), Vector3.UnitY, new Vector2(numWrapsU, numWrapsV));
            vertices[3] = new VertexPositionNormalTexture(new Vector3(xMin, height, zMax), Vector3.UnitY, new Vector2(0, 0));
            vertices[4] = new VertexPositionNormalTexture(new Vector3(xMax, height, zMin), Vector3.UnitY, new Vector2(numWrapsU, numWrapsV));
            vertices[5] = new VertexPositionNormalTexture(new Vector3(xMax, height, zMax), Vector3.UnitY, new Vector2(numWrapsU, 0));
            vertexBuffer = new VertexBuffer(device, VertexPositionNormalTexture.SizeInBytes * 2 * 3, BufferUsage.WriteOnly);
            vertexBuffer.SetData(vertices);

            vertexDeclaration = new VertexDeclaration(device, VertexPositionNormalTexture.VertexElements);


        }

        public void Draw(GraphicsDevice device, Effect effect, PipelineMatrices pm) {
            effect.Parameters["Texture"].SetValue(texture);
            effect.Parameters["World"].SetValue(Matrix.Identity);
            effect.Parameters["WorldViewProjection"].SetValue(pm.viewMatrix * pm.projectionMatrix);
            effect.Parameters["LightWorldViewProjection"].SetValue(pm.lightViewMatrix * pm.projectionMatrix);
            effect.Parameters["Reflectance"].SetValue(reflectivity);
            effect.Begin();
            //device.RenderState.FillMode = FillMode.WireFrame;
            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
                device.VertexDeclaration = vertexDeclaration;
                device.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);
                pass.End();
            }
            //device.RenderState.FillMode = FillMode.Solid;
            effect.End();
        }

        public void DrawBasic(GraphicsDevice device, BasicEffect effect, PipelineMatrices pm) {
            effect.Texture = texture;
            effect.World = Matrix.Identity;
            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();
                device.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);
                device.VertexDeclaration = vertexDeclaration;
                device.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);
                pass.End();
            }
            effect.End();
        }
    }

    /// <summary>
    /// A spring with which a user can drag objects using a mouse.
    /// </summary>
    public class UserSpring {
        public bool active = false;     // if true, UserSpring is applying a force
        public Vector3 position;        // position of the mouse-side end of the spring
        public float kp = 10;           // spring parameter
        //public float kd = 1;          // damping parameter..ignored. Better userSpring will need this term.
        public RigidBodySystem attachedSystem; //RigidBodySystem that userSpring is attached
        public Link attachedLink;              //Link of RigidBodySystem that userSpring in attached
        public Vector3 attachmentPoint;        // attachment point of a link in local frame     

        public void Draw(GraphicsDevice device, Effect effect) {
            Vector3 pos = Vector3.Transform(attachmentPoint, attachedLink.matWorld);

            VertexPositionColor[] vertices = new VertexPositionColor[2];
            vertices[0] = new VertexPositionColor(position, Color.Yellow);
            vertices[1] = new VertexPositionColor(pos, Color.Yellow);
            EffectTechnique old = effect.CurrentTechnique;
            effect.CurrentTechnique = effect.Techniques["Colored"];
            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes) {
                pass.Begin();

                device.VertexDeclaration = new VertexDeclaration(device, VertexPositionColor.VertexElements);
                device.DrawUserPrimitives(PrimitiveType.LineList, vertices, 0, 1);

                pass.End();
            }
            effect.End();
            effect.CurrentTechnique = old;
        }
    }

    /// <summary>
    /// visual props attached to a Link or World
    /// </summary>
    public struct Prop {
        public Model model;
        public Link parent;             // if null, the model is fixed to world
        public Matrix mat;              // transformation wrt parent
        public float lightIntensity;
        public float reflectivity;
    }
}


