
#define USE_GUI_VARS

#region File Description
//-----------------------------------------------------------------------------
// WindowsFormXNA.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.

//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Timers;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Xml.Serialization;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using DifferentiableFunction;
using Dynamics;
using utilities;
using matrix;


#endregion

namespace WinFormsGraphicsDevice
{
    /// <summary>
    /// Example control inherits from GraphicsDeviceControl, which allows it to
    /// render using a GraphicsDevice. This control shows how to draw animating
    /// 3D graphics inside a WinForms application. It hooks the Application.Idle
    /// event, using this to invalidate the control, which will cause the animation
    /// to constantly redraw.
    /// </summary>
    /// 
    /// <summary>
    /// View, Projection, World matrices of the scene
    /// </summary>
    public class PipelineMatrices
    {
        public Matrix worldMatrix;
        public Matrix viewMatrix;       //view from eye
        public Matrix projectionMatrix;
        public Matrix lightViewMatrix;  //view from the light source

        public void UpdateViewMatrix(Vector3 cameraPos, Vector3 cameraTarget)
        {
            viewMatrix = Matrix.CreateLookAt(cameraPos, cameraTarget, Vector3.Up);
        }

        public void UpdateLightViewMatrix(Vector3 lightPos, Vector3 lightTarget)
        {
            lightViewMatrix = Matrix.CreateLookAt(lightPos, lightTarget, Vector3.Up);
        }
    }

    /// <summary>
    /// Current state of user interaction
    /// </summary>
    public enum UserInputMode
    {
        None,       // user is not applying a force
        Spring      // user is applying a force with a UserSpring
    }


    class WindowsFormXNA : GraphicsDeviceControl
    {
        ContentBuilder contentBuilder;
        ContentManager contentManager;

        VertexDeclaration vertexDeclaration;
        Stopwatch timer;
 
        ///
        /// simulation
        ///
        public enum simState{
            PLAY = 1,
            PAUSE = 0,
            RESTART = -1
        };
        int simulationOn = (int)simState.RESTART;              // simulation begins when simulationOn = 1 after spacebar is pressed         
        double duration;                    // time span of the simuation in milisec   
        
        Stopwatch watch = new Stopwatch();  // stop watch to measure duration
        
        double displayTimeStep = 1.0 / 60.0; //default is 60Hz. When recording, displayTimeStep will be set 1.0/30.0

        long frame = 0, time, timebase = 0;
        double fps = 0.0;


        ///
        /// user input
        ///
        KeyboardState oldKeyState;
        int mouseX, mouseY;                 // mouse position
        UserInputMode userInputMode = UserInputMode.None;

        /// 
        /// objects
        ///  
        Ground GlobalGround = new Ground();                   // ground
        List<RigidBodySystem> rigidBodySystems = null;// = new List<RigidBodySystem>(); // systems
        List<Prop> props = new List<Prop>();            // Microsoft.Xna.Framework.Graphics.Model instances
        UserSpring userSpring = new UserSpring();       // user spring controlled by mouse 
        
        public RigidBodySystem rigidBodySystemUI=null;      // RBS specified using the GUI
        

        /// 
        /// camera / light / world-view-projection
        /// 
        Vector3 cameraTarget, cameraPos, lightPos;
        float cameraTheta, cameraPhi, cameraDist;       // determine cameraPos
        float lightTheta, lightPhi, lightDist;          // determine lightPos        
        PipelineMatrices pm = new PipelineMatrices();   // world/view/projection matrices

        /// 
        /// lighting options
        /// 
        bool highQualityRender = true; // [MODIFIABLE PARAM] set true for many light sources. This will change SurfaceFormat of RenderTargets, which may not be supported by Brian's2 computer.
        float areaLightWidth = 30.0f;   // [MODIFIABLE PARAM] width of area light
        float areaLightHeight = 30.0f;  // [MODIFIABLE PARAM] height of are light
        int numLightsHor = 1;       // [MODIFIABLE PARAM]
        int numLightsVer = 1;       // [MODIFIABLE PARAM] total # lights = numLightsHor*numLightsVer
        float ambient = 0.01f;      // [MODIFIABLE PARAM] ambiency
        float lightPower = 1.3f;    // [MODIFIABLE PARAM] power of light
        Texture2D lightCircleFilter;// texture to make each point light a spot light

        ///
        /// effects 
        /// 
        string effectFile = "Content//ShadowMap";   //[MODIFIABLE PARAM] effect file path
        Effect effect;                      // shadowed environment map effect
        float invMaxShadowDepth = 1.0f / 150f;  //[MODIFIABLE PARAM] inverse of maximum depth. important for shadow quality. Refer to http://www.riemers.net/eng/Tutorials/XNA/Csharp/series3.php
        RenderTarget2D rtShadowMap;         //a render target that a shadow map is drawn.
        RenderTarget2D rtScenePointLight;   //a render target that a point lighted scene is drawn
        RenderTarget2D rtSceneAreaLight;    //a render target that point lighed scenes are accumulated
        Texture2D textureRenderedTo;        //used for rendering to texture
        VertexBuffer quad;                  //a billboard to draw shadowed scene to
        VertexDeclaration vertexDeclarationVPT; //VertexPositionTexture

        ///
        /// basicEffect
        /// 
        bool useBasicEffect = false;        // [MODIFIABLE PARAM] if true, basicEffect will be used instead of effect.
        BasicEffect basicEffect;            // basic effect. only to be use for debugging/performance evaluation. will draw only Link and Ground

        ///
        /// texture / font  
        /// 
        bool enableTexture = true;          // [MODIFIABLE PARAM] if false, draw scene without textures
        TextureCube cubeMap;                //environment map
        //string fileCubeMap = "Content\\seattle"; //environment map texture file        
        //string fileCubeMap = "Content\\envmap1"; //[MODIFIABLE PARAM] environment map texture file. Must set the Content Processor to CubeMapProcessor. To do this, right click on the texture file in Solution Explorer and open the Properties Window. And then set "Content Processor" to CubeMapProcessor, which will generate a cube map from a regular texture map.
        string fileCubeMap = "Content//chipboard"; //[MODIFIABLE PARAM] environment map texture file. Must set the Content Processor to CubeMapProcessor. To do this, right click on the texture file in Solution Explorer and open the Properties Window. And then set "Content Processor" to CubeMapProcessor, which will generate a cube map from a regular texture map.
        SpriteFont arial;
        SpriteBatch spriteBatch;
        string fontFile = "Content//Arial";     //[MODIFIABLE PARAM] font file
        public List<Texture2D> textureMaps;

        ///
        /// etc
        ///         
        BoundingBox bb = new BoundingBox(new Vector3(-30, -2, -30), new Vector3(30, 1, 30)); // bounding box of the scene. will be adjusted by the size of ground.
        CullMode cullMode = CullMode.None;  //modify this to set culling mode to cw or ccw        

        ///
        /// save images
        /// 
        bool saveScreenShots = false;       //[MODIFIABLE PARAM] if true, save screenshots to "imagedump" folder
        int screenShotCounter = 0;          //starting index of stored files.
        ImageFileFormat imageFileFormat = ImageFileFormat.Bmp; // [MODIFIABLE PARAM]select suitable image format
        string imageFileFormatExt = ".bmp"; //[MODIFIABLE PARAM]. specify file type appropriately.


        // Vertex positions and colors used to display a spinning triangle.
        public readonly VertexPositionColor[] Vertices =
        {
            new VertexPositionColor(new Vector3(-1, -1, 0), Color.Red),
            new VertexPositionColor(new Vector3( 1, -1, 0), Color.Black),
            new VertexPositionColor(new Vector3( 0,  1, 0), Color.Black),
        };

        private void setupDevice()
        {
          
            vertexDeclarationVPT = new VertexDeclaration(GraphicsDevice, VertexPositionTexture.VertexElements);
            //set mouse visible
           //IsMouseVisible = true;

            //set mouse visible
            //IsMouseVisible = true;
            GraphicsDevice.RenderState.CullMode = cullMode;

            //rtShadowMap: shadowmap for each lightsource will be drawn here
            rtShadowMap = new RenderTarget2D(
                GraphicsDevice,
                Height,
                Width,
                1,
                highQualityRender ? SurfaceFormat.Single : SurfaceFormat.Single,
                //highQualityRender ? SurfaceFormat.Color : SurfaceFormat.Color, //only need to do this on my old machine because its graphics card wouldn't support the higher resolution format
                GraphicsDevice.DepthStencilBuffer.MultiSampleType,
                GraphicsDevice.DepthStencilBuffer.MultiSampleQuality
                );

            //shadowed scene for each lightsource will be drawn here
            rtScenePointLight = new RenderTarget2D(
                GraphicsDevice,
                Height,
                Width,
                1,
                highQualityRender ? SurfaceFormat.Color : SurfaceFormat.Color, //only need to do this on my old machine because its graphics card wouldn't support the higher resolution format
                //highQualityRender ? SurfaceFormat.Vector4 : SurfaceFormat.Color, //Sung's original code
                GraphicsDevice.DepthStencilBuffer.MultiSampleType,
                GraphicsDevice.DepthStencilBuffer.MultiSampleQuality
                );
            //each shadowed scene will be accumulated here
            rtSceneAreaLight = new RenderTarget2D(
                GraphicsDevice,
                Height,
                Width,
                1,
                highQualityRender ? SurfaceFormat.Color : SurfaceFormat.Color, //only need to do this on my old machine because its graphics card wouldn't support the higher resolution format
                //highQualityRender ? SurfaceFormat.Vector4 : SurfaceFormat.Color, //Sung's original code
                GraphicsDevice.DepthStencilBuffer.MultiSampleType,
                GraphicsDevice.DepthStencilBuffer.MultiSampleQuality
                );
        }

        private void SetUpCameraAndLight()
        {
            Vector3 center = 0.5f * (bb.Max + bb.Min);
            Vector3 v = bb.Max - bb.Min;
            float len = Math.Max(v.X, Math.Max(v.Y, v.Z));

            cameraTarget = Vector3.Multiply(Vector3.Add(bb.Min, bb.Max), 0.5f);
            cameraTarget.Y = 0;                                     //[MODIFIABLE PARAM] Camera will look at height 0. 
            cameraDist = 0.5f * Math.Max(Math.Max(v.X, v.Y), v.Z);  //[MODIFIABLE PARAM] initial camera distance
            cameraTheta = MathHelper.ToRadians(45.0f);              //[MODIFIABLE PARAM] initial longitude of camera
            cameraPhi = MathHelper.ToRadians(25.0f); ;              //[MODIFIABLE PARAM] initial latitude of camera

            pm.worldMatrix = Matrix.Identity;
            pm.projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.PiOver4,
                //(float)this.Window.ClientBounds.Width / (float)this.Window.ClientBounds.Height, //this is the way Sung wrote itwhich I believe is incorrect
                (float)Height / (float) Width,
                0.1f,       //[MODIFIABLE PARAM] near
                100.0f);    //[MODIFIABLE PARAM] far

            lightPos = new Vector3(center.X + len, center.Y + len, center.Z + len);
            lightDist = 1.1f * cameraDist;                  //[MODIFIABLE PARAM] initial light distance
            lightTheta = MathHelper.ToRadians(45.0f);       //[MODIFIABLE PARAM] initial longitude of light
            lightPhi = MathHelper.ToRadians(45.0f);         //[MODIFIABLE PARAM] inital latitude of light
        }

        

        private void loadContent()
        {

          //  contentBuilder = new ContentBuilder();
            contentManager = new ContentManager(Services);
           
            //font
            arial = contentManager.Load<SpriteFont>(fontFile);             // sprite fonts
            spriteBatch = new SpriteBatch(GraphicsDevice);                  // sprites are drawn by SpriteBatch, so create one.
            //environment map
            cubeMap = contentManager.Load<TextureCube>(fileCubeMap);
            //cubeMap = CustomEffectPipeline.CubemapProcessor.load(fileCubeMap);
            //lightCircleFilter texture
            lightCircleFilter = contentManager.Load<Texture2D>("Content//shadow_filter");
            //texture list
  
            textureMaps = loadAllTextures();

            //create a quad which will be used to draw a texture (shadowed scene) to the backbuffer
            VertexPositionTexture[] quadVerts = new VertexPositionTexture[3 * 2];
            quadVerts[0] = new VertexPositionTexture(new Vector3(-1, -1, 0), new Vector2(0, 1));
            quadVerts[1] = new VertexPositionTexture(new Vector3(-1, 1, 0), new Vector2(0, 0));
            quadVerts[2] = new VertexPositionTexture(new Vector3(1, 1, 0), new Vector2(1, 0));
            quadVerts[3] = new VertexPositionTexture(new Vector3(-1, -1, 0), new Vector2(0, 1));
            quadVerts[4] = new VertexPositionTexture(new Vector3(1, 1, 0), new Vector2(1, 0));
            quadVerts[5] = new VertexPositionTexture(new Vector3(1, -1, 0), new Vector2(1, 1));
            quad = new VertexBuffer(GraphicsDevice, VertexPositionTexture.SizeInBytes * 6, BufferUsage.WriteOnly);
            quad.SetData(quadVerts);


            basicEffect = new BasicEffect(GraphicsDevice, null);
            basicEffect.EnableDefaultLighting();
            basicEffect.TextureEnabled = true;
            //effect
            effect = contentManager.Load<Effect>(effectFile);              // effect file
            effect.Parameters["TextureEnabled"].SetValue(enableTexture);
            effect.Parameters["InvMaxDepth"].SetValue(invMaxShadowDepth);
            effect.Parameters["Ambient"].SetValue(ambient / (float)(numLightsHor * numLightsVer));
            effect.Parameters["LightPower"].SetValue(lightPower / (float)(numLightsHor * numLightsVer));
            effect.Parameters["EnvironmentMap"].SetValue(cubeMap);
            effect.Parameters["LightTexture"].SetValue(lightCircleFilter);

            loadRigidBodyDescription(textureMaps);
            InitializeGround(GlobalGround, textureMaps, GraphicsDevice);

            //ignore default effects and set our effect to props
            //this is necessary because we will use ShadowMap.
            //Also, note that each ModelMeshPart should be textured
            //Non-textured ModelMeshPart will be drawn with a undefined color.
            for (int i = 0; i < props.Count; i++)
            {
                foreach (ModelMesh mesh in props[i].model.Meshes)
                {
                    foreach (ModelMeshPart part in mesh.MeshParts)
                    {
                        part.Effect = effect;
                    }
                }
            }

            //camera and light
            bb.Min = GlobalGround.center - new Vector3(GlobalGround.width / 2, 0, GlobalGround.depth / 2);
            bb.Max = GlobalGround.center + new Vector3(GlobalGround.width / 2, 0, GlobalGround.depth / 2);
            SetUpCameraAndLight();
        }
        /// <summary>
        /// Initializes the control.
        /// </summary>
        protected override void Initialize()
        {
            // Create our vertex declaration.
            vertexDeclaration = new VertexDeclaration(GraphicsDevice,
                                                VertexPositionColor.VertexElements);

            // Create our effect.
            basicEffect = new BasicEffect(GraphicsDevice, null);
            basicEffect.VertexColorEnabled = true;

            // Start the animation timer.
            timer = Stopwatch.StartNew();

            // Hook the idle event to constantly redraw our animation.
            Application.Idle += delegate { Invalidate(); };

            setupDevice();

            loadContent();
            oldKeyState = Keyboard.GetState();
            
        }

        /// <summary>
        /// Disposes the control, unloading the ContentManager.
        /// </summary>
        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                contentManager.Unload();
            }

            base.Dispose(disposing);
        }

        private void drawScene(string technique)
        {
            effect.CurrentTechnique = effect.Techniques[technique];

            //draw rigidBodies
            if (rigidBodySystems != null)
            {
                foreach (RigidBodySystem system in rigidBodySystems)
                {
                    system.Draw(GraphicsDevice, effect, pm);
                    
                }
                if (technique != "ShadowMap") // Do not draw Axis' Shadows
                {
                    foreach (RigidBodySystem system in rigidBodySystems)
                    {
                        system.DrawAxis(GraphicsDevice, pm);

                    }
                }
            }
            if (rigidBodySystemUI != null)
            {
                rigidBodySystemUI.Draw(GraphicsDevice, effect, pm);
                if (technique != "ShadowMap")
                {
                    rigidBodySystemUI.DrawAxis(GraphicsDevice, pm);
                }
            }
            
            //draw ground 
            GlobalGround.Draw(GraphicsDevice, effect, pm);

            //draw userSpring. do not draw for ShadowMap
            if (technique != "ShadowMap" && userSpring.active) userSpring.Draw(GraphicsDevice, effect);
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input and playing audio.
        /// </summary>
    
        protected new void Update()
        {
            //handle input
            HandleInput();

            if (simulationOn == (int)simState.PLAY)
            {
                //Window.Title = "Simulation On. Press ESC to close.";
                watch.Reset();
                watch.Start();
                if (rigidBodySystems != null)
                {
                    foreach (RigidBodySystem system in rigidBodySystems)
                    {
#if USE_GUI_VARS
                        system.NextStepRK4usingGUIVars(userSpring);
#else
                        system.NextStep(displayTimeStep, userSpring); //simulate to the next step
#endif
      
                        // system.NextStepRK4(userSpring);
                    }
                }

                if (rigidBodySystemUI != null)
                {
#if USE_GUI_VARS
                    rigidBodySystemUI.NextStepUsingGUIVars(displayTimeStep, userSpring);
#else
                    rigidBodySystemUI.NextStep(displayTimeStep, userSpring);
#endif
                }


                    watch.Stop();
                duration = watch.ElapsedTicks;
            }
            else if (simulationOn == (int)simState.RESTART) //do only one time step...I need to run at least one step to draw systems
            {
                if (rigidBodySystems != null)
                {
                    //Window.Title = "Press SPACEBAR to run or R to begin recording";
                    foreach (RigidBodySystem system in rigidBodySystems)
                    {
#if USE_GUI_VARS
                        system.NextStepRK4usingGUIVars(userSpring);
#else
                        system.NextStep(displayTimeStep, userSpring);
#endif
                        
                    }
                }
                if (rigidBodySystemUI != null)
                {
#if USE_GUI_VARS
                    if (rigidBodySystemUI.motion == true)
                        rigidBodySystemUI.NextStepUsingGUIVars(displayTimeStep, userSpring);
                    else
                        rigidBodySystemUI.NextStepUsingGUIVars();
#else
                     rigidBodySystemUI.NextStep(displayTimeStep, userSpring);
#endif
                    simulationOn = (int)simState.PAUSE; // stop simulation
                }
            }

        }



       
        public void setSimulationState(int state)
        {
            simulationOn = state;
        }
        /// <summary>
        /// Draws the control.
        /// </summary>
        /// 

        protected override void Draw()
        {
          //  watchGameTime.Reset();
           
            frame++;
           
            //handle input
            HandleInput();
            GraphicsDevice.Clear(Color.Black);
            Update();

            ///light position
            Vector3 nominalLightPos;
            nominalLightPos.X = cameraTarget.X + lightDist * (float)Math.Sin(lightTheta) * (float)Math.Cos(lightPhi);
            nominalLightPos.Y = cameraTarget.Y + lightDist * (float)Math.Sin(lightPhi);
            nominalLightPos.Z = cameraTarget.Z + lightDist * (float)Math.Cos(lightTheta) * (float)Math.Cos(lightPhi);


            ///distance between point lights
            float dx = MathHelper.ToRadians(areaLightWidth) / (float)numLightsHor;
            float dy = MathHelper.ToRadians(areaLightHeight) / (float)numLightsVer;

            Vector3 dir = cameraPos - new Vector3(lightPos.X, lightPos.Y, lightPos.Z);
            dir.Normalize();
            Vector3 vx = Vector3.Cross(dir, Vector3.Up);
            Vector3 vy = Vector3.Cross(dir, vx);


            for (int i = 0; i < numLightsHor; i++)
            {
                for (int j = 0; j < numLightsVer; j++)
                {
                    Vector3 del = vx * (dx * i) + vy * (dy * j);
                    lightPos = nominalLightPos + del;
                    pm.UpdateLightViewMatrix(lightPos, cameraTarget + del);
                    effect.Parameters["LightPos"].SetValue(lightPos);
                    effect.Parameters["View"].SetValue(pm.viewMatrix);
                    effect.Parameters["LightDirection"].SetValue(cameraTarget - lightPos);

                    ///
                    ///generate shadowmap to rtShadowMap
                    ///

                    GraphicsDevice.SetRenderTarget(0, rtShadowMap);
                    GraphicsDevice.Clear(Color.Black);
                    drawScene("ShadowMap");
                    //GlobalDevice.ResolveRenderTarget(0);
                    GraphicsDevice.SetRenderTarget(0, null);
                    textureRenderedTo = rtShadowMap.GetTexture();

                    ///
                    ///generate point lighted shadowedscene to rtScenePointLight
                    ///                                        
                    GraphicsDevice.SetRenderTarget(0, rtScenePointLight);
                    GraphicsDevice.Clear(Color.Black);
                    effect.Parameters["ShadowMap"].SetValue(textureRenderedTo);
                    drawScene("ShadowedScene");
                    //GlobalDevice.ResolveRenderTarget(0);
                    GraphicsDevice.SetRenderTarget(0, null);
                    textureRenderedTo = rtScenePointLight.GetTexture();

                    ///
                    ///accumulate shadowedscenes to textureRenderedTo
                    /// 
                    DrawTextureToBackBuffer(textureRenderedTo);
                }

            }

            ///draw to backbuffer
            //GlobalDevice.ResolveRenderTarget(0);
            GraphicsDevice.SetRenderTarget(0, null);

            textureRenderedTo = rtSceneAreaLight.GetTexture();

            /// save the screenshot
            if (saveScreenShots)
            {
                textureRenderedTo.Save("..\\..\\..\\imagedump\\shot_" + screenShotCounter.ToString() + imageFileFormatExt, imageFileFormat);
                screenShotCounter++;
            }

            ///
            ///draw to backbuffer
            ///
            GraphicsDevice.SetRenderTarget(0, null);
            GraphicsDevice.Clear(Color.Black);
            spriteBatch.Begin(SpriteBlendMode.Additive, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            spriteBatch.Draw(textureRenderedTo, Vector2.Zero, Color.White);
            spriteBatch.End();


            //fps Calculation
            if (time - timebase > 1000)
            {
                
                fps = frame * 1000.0 / Math.Max(1, (time - timebase));
                timebase = time;
                frame = 0;
            }


            ///draw text   
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            spriteBatch.DrawString(arial, "Simulation time: " + duration.ToString(), new Vector2(0, 25), Color.Red);
            spriteBatch.DrawString(arial, "fps: " + fps.ToString("F"), Vector2.Zero, Color.White);
            spriteBatch.End();


            time = timer.ElapsedMilliseconds;
          
        }

        //texture: texture to render to back buffer
        void DrawTextureToBackBuffer(Texture2D texture)
        {
            //set Rendertarget to backbuffer
            GraphicsDevice.SetRenderTarget(0, rtSceneAreaLight);
            GraphicsDevice.Clear(Color.Black);
            //enable alpha blending
            GraphicsDevice.RenderState.AlphaBlendEnable = true;
            GraphicsDevice.RenderState.BlendFunction = BlendFunction.Add;
            GraphicsDevice.RenderState.DestinationBlend = Blend.One;
            GraphicsDevice.RenderState.SourceBlend = Blend.One;
            //draw a quad that covers the screen with texture
            effect.CurrentTechnique = effect.Techniques["DrawTexture"];
            effect.Parameters["Texture"].SetValue(texture);
            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Begin();
                GraphicsDevice.VertexDeclaration = vertexDeclarationVPT;
                GraphicsDevice.Vertices[0].SetSource(quad, 0, VertexPositionTexture.SizeInBytes);
                GraphicsDevice.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);
                pass.End();
            }
            effect.End();
            //disable alpha blending
            GraphicsDevice.RenderState.AlphaBlendEnable = false;
        }

        private List<Texture2D> loadAllTextures()
        {
            List<Texture2D> textureMaps = new List<Texture2D>();

            Texture2D texture;
            texture = contentManager.Load<Texture2D>("Content//green"); texture.Name = "green"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//red"); texture.Name = "red"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//blue"); texture.Name = "blue"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//yellow"); texture.Name = "yellow"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//lime"); texture.Name = "lime"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//orange"); texture.Name = "orange"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//purple"); texture.Name = "purple"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//white"); texture.Name = "white"; textureMaps.Add(texture);
            //texture = content.Load<Texture2D>("Content//chipboard"); texture.Name = "chipboard"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//cellceiling"); texture.Name = "crosstile"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//shiny_metal"); texture.Name = "shiny_metal"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//checkerboard_purple"); texture.Name = "checkerboard_purple"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//checkerboard"); texture.Name = "checkerboard"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//lake"); texture.Name = "lake"; textureMaps.Add(texture); //this is a texture map I added as a potential environment candidate
            texture = contentManager.Load<Texture2D>("Content//checkerboardGreen"); texture.Name = "checkerboardGreen"; textureMaps.Add(texture);
            texture = contentManager.Load<Texture2D>("Content//checkerboardYellow"); texture.Name = "checkerboardYellow"; textureMaps.Add(texture);

            return textureMaps;
        }
        
        private Texture2D FindTexture(string name, List<Texture2D> textures)
        {
            //name = name.Substring(1, name.Length - 2);
            foreach (Texture2D t in textures)
            {
                if (t.Name == name) 
                    return t;
            }
            return null;
        }

        void InitializeGround(Ground ground, List<Texture2D> textures, GraphicsDevice device)
        {
            ground.center = new Vector3(0, -1, 0);
            ground.width = 60;
            ground.depth = 60;
            ground.texture = FindTexture("lime", textures);
            ground.numWrapsU = 1;
            ground.numWrapsV = 1;
            ground.reflectivity = 0.0f;
            ground.Create(device);
        }

        //public rbsExample6 _test = new rbsExample6();

        private void loadRigidBodyDescription(List<Texture2D> textureMaps)
        {
          
            rigidBodySystems = new List<RigidBodySystem>(5);
            
            //rigidBodySystems.Add(coilSystem(GraphicsDevice, new Texture2D[] { textureMaps[0], textureMaps[1], textureMaps[2], textureMaps[3] }));
            //rigidBodySystems.Add(ringSystem(GraphicsDevice, new Texture2D[] { textureMaps[12], textureMaps[10], textureMaps[11], textureMaps[14] }, 3));
            //rigidBodySystems.Add(gyratory(GraphicsDevice, new Texture2D[] { textureMaps[0], textureMaps[10], textureMaps[1], textureMaps[14] }));
            //rigidBodySystems.Add(torusLinkSystem(GraphicsDevice, new Texture2D[] { textureMaps[11], textureMaps[12], textureMaps[13], textureMaps[14] }, 4));
            //rigidBodySystems.Add(_test.build(GraphicsDevice, new Texture2D[] { textureMaps[11], textureMaps[11], textureMaps[11], textureMaps[11] }));
            
        }

        public void loadRigidBodyDescription(iRigidBodySystemFunctions rb)
        {

            rigidBodySystems = new List<RigidBodySystem>(5);
            rigidBodySystems.Add(rb.build(GraphicsDevice, new Texture2D[] { this.textureMaps[0], this.textureMaps[1], this.textureMaps[2], this.textureMaps[3] }));

        }

        private RigidBodySystem TraverseTreeView(TreeView tview)
        {
            Function.newContext();
            
            
            Variable t = new Variable("t");
            List<RigidBody> allJoints = new List<RigidBody>();
            Graph<JointNode> g = null;
            long qCont = new long();

            
            //UnspecifiedFunction q0 = UnspecifiedFunction.functionOf("q0", t);

            //Create a TreeNode to hold the Parent Node
            cRigidBodyUI temp = new cRigidBodyUI(); 
            cRigidBodyUI temp2 = new cRigidBodyUI();
            
            List<double> mass = new List<double>();
            List<double> pos = new List<double>();
            List<double> angles = new List<double>();

            List<double> force = new List<double>();
            List<double> torque = new List<double>();

            List<double> lCenterOfMass = new List<double>();


            List<double> qValues = new List<double>();


            //Loop through the Parent Nodes
            for (int k = 0; k < tview.Nodes[0].Nodes.Count; k++)
            {
                
                //Store the Parent Node in temp
                temp = (cRigidBodyUI)tview.Nodes[0].Nodes[k];
                temp.q0 = UnspecifiedFunction.functionOf("q" + qCont.ToString(), t);
                qValues.Add(temp.genCoords.q0);
                temp.vMass = new Variable("m" + qCont.ToString());
                
                Function[,] rb = null;
                //reading shape of the body
                // TODO: Change the way this properties are specified in the PropertyGrid

                Variable posX = new Variable("posX" + qCont);
                Variable posY = new Variable("posY" + qCont);
                Variable posZ = new Variable("posZ" + qCont);

                Variable angleX = new Variable("angleX" + qCont);
                Variable angleY = new Variable("angleY" + qCont);
                Variable angleZ = new Variable("angleZ" + qCont);

                Variable forceX = new Variable("forceX" + qCont);
                Variable forceY = new Variable("forceY" + qCont);
                Variable forceZ = new Variable("forceZ" + qCont);
                Variable torqueX = new Variable("torqueX" + qCont);
                Variable torqueY = new Variable("torqueY" + qCont);
                Variable torqueZ = new Variable("torqueZ" + qCont);

                

                
                Variable comX = new Variable("comX" + qCont);
                Variable comY = new Variable("comY" + qCont);
                Variable comZ = new Variable("comZ" + qCont);
                



                pos.Add(temp.position.X);
                pos.Add(temp.position.Y);
                pos.Add(temp.position.Z);

                angles.Add(temp.Orientation.X * Math.PI / 180.0);
                angles.Add(temp.Orientation.Y * Math.PI / 180.0);
                angles.Add(temp.Orientation.Z * Math.PI / 180.0);

                mass.Add(temp.Mass);

                force.Add(temp.Forces.X);
                force.Add(temp.Forces.Y);
                force.Add(temp.Forces.Z);

                torque.Add(temp.Torque.X);
                torque.Add(temp.Torque.Y);
                torque.Add(temp.Torque.Z);


                lCenterOfMass.Add(temp.CenterOfMass.X);
                lCenterOfMass.Add(temp.CenterOfMass.Y);
                lCenterOfMass.Add(temp.CenterOfMass.Z);
                Function[,] centerOfMass = { { comX }, { comY }, { comZ }, { 1 } }; //point so need 1 in 4th position


                Function[,] A0 = { { 1, 0, 0, posX }, { 0, 1, 0,  posY }, { 0, 0, 1, posZ}, { 0, 0, 0, 1 } };
                Function[,] Aaux = { { 1, 0, 0, 0}, { 0, 1, 0, 0}, { 0, 0, 1, 0}, { 0, 0, 0, 1 } };
                            
                
                A0 = VM.mult(A0, VM.rotX(angleX, 0, 0, 0)); //conversion to radians
                A0 = VM.mult(A0, VM.rotY(angleY, 0, 0, 0)); //conversion to radians
                A0 = VM.mult(A0, VM.rotZ(angleZ, 0, 0, 0)); //conversion to radians
                
                if (temp.Cylinder.Use_Cylinder == true)
                {   
                    if (temp.Cylinder.Radius <= 0)
                    {
                        MessageBox.Show("Radius should be higher than zero","Error");
                        temp.Cylinder.Radius = 0.3;
                    }
                    if (temp.Cylinder.Initial_Angle > 360)
                        temp.Cylinder.Initial_Angle = 360;
                    else if (temp.Cylinder.Initial_Angle< 0)
                        temp.Cylinder.Initial_Angle = 0;

                    if (temp.Cylinder.Final_Angle > 360)
                        temp.Cylinder.Final_Angle = 360;
                    else if (temp.Cylinder.Final_Angle < 0)
                        temp.Cylinder.Final_Angle = 0;


                    if (temp.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotZAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }

                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    
                    //rb = VM.rotZ(temp.q0, 0, 0, 0);
                    
                    Function[,] I1 = inertiaMatrixCylinder(temp.vMass, temp.Cylinder.Radius, temp.Cylinder.Height);
                    Geometries.Cylinder cylinder = new Geometries.Cylinder(temp.Cylinder.Initial_Angle*Math.PI / 180.0, temp.Cylinder.Final_Angle * Math.PI /180.0, 0.0, temp.Cylinder.Height, temp.Cylinder.Radius);
                    temp.rigidBody = new RigidBody(Aaux, new Function[] { temp.q0 }, frictionForce(t, temp.q0), centerOfMass, I1, temp.vMass, temp.Name, t);
                    temp.rigidBody.CreateGeometry(GraphicsDevice, cylinder, textureMaps[10]);
                    temp.rigidBody.geometry = cylinder;
                }
                else if (temp.Sphere.Use_sphere == true)
                {
                    if (temp.Sphere.Radius <= 0)
                    {
                        MessageBox.Show("Radius should be higher than zero", "Error");
                        temp.Sphere.Radius = 0.5;
                    }
                    if (temp.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotZAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    //rb = VM.rotX(temp.q0, 0, 0, 0);

                    Function[,] I1 = inertiaMatrixSphere(temp.vMass, temp.Sphere.Radius);
                    Geometries.Sphere sphere = new Geometries.Sphere(temp.Sphere.Radius);
                    temp.rigidBody = new RigidBody(Aaux, new Function[] { temp.q0 }, frictionForce(t, temp.q0), centerOfMass, I1, temp.vMass, temp.Name, t);
                    temp.rigidBody.CreateGeometry(GraphicsDevice, sphere, textureMaps[10]);
                    temp.rigidBody.geometry = sphere;
                }
                else if (temp.Torus.Use_Torus == true)
                {
                    if (temp.Torus.Outer_Radius <= 0)
                    {
                        MessageBox.Show("Outer Radius should be higher than zero", "Error");
                        temp.Torus.Outer_Radius = 0.8;
                    }
                    if (temp.Torus.Inner_Radius <= 0)
                    {
                        MessageBox.Show("Inner Radius should be higher than zero", "Error");
                        temp.Torus.Inner_Radius = 0.2;
                    }

                    if (temp.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    //rb = VM.rotX(temp.q0, 0, 0, 0);
                    

                    Function[,] I1 = inertiaMatrixTorus (temp.vMass, temp.Torus.Outer_Radius, temp.Torus.Inner_Radius);
                    Geometries.Torus torus = new Geometries.Torus(temp.Torus.Outer_Radius, temp.Torus.Inner_Radius);
                    temp.rigidBody = new RigidBody(Aaux, new Function[] { temp.q0 }, frictionForce(t, temp.q0), centerOfMass, I1, temp.vMass, temp.Name, t);
                    temp.rigidBody.CreateGeometry(GraphicsDevice, torus, textureMaps[10]);
                    temp.rigidBody.geometry = torus;
                }
                else if (temp.ParametricModel.File != "")
                {
                    ParametricModelDStar3DVisualizer parametricModel = loadParametricModel(temp.ParametricModel.File, temp.ParametricModel.Width, temp.ParametricModel.Height);
                    // TRANSFORMATION CONSTRAINS (root Body)
                    if (temp.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(temp.q0, 0, 0, 0));
                    }
                    else if (temp.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(temp.q0, 0, 0, 0));
                    }
                    //CALCULATE INERTIA MATRIX HERE
                    // FOR NOW, USING THE SPHERE INERTIA MATRIX: 
                    Function[,] I1 = inertiaMatrixSphere(temp.vMass, temp.Sphere.Radius);
                    
                    temp.rigidBody = new RigidBody(Aaux, new Function[] { temp.q0 }, frictionForce(t, temp.q0),
                                    centerOfMass, I1, temp.vMass, temp.Name, t);
                    temp.rigidBody.parametricModel = parametricModel;
                    temp.rigidBody.isParametric = true;
                    // ***
                    // HEY ! DO NOT FORGET TO CHANGE THE CODE IN cRigidBodySystemIO class, functions TraverseTreeView() and visitChildNodes()
                    // ***
                }

                temp.rigidBody.orientation = new Variable[] { angleX, angleY, angleZ };
                temp.rigidBody.position = new Variable[] { posX, posY, posZ };

                temp.rigidBody.force = new Function[,] { { forceX }, { forceY }, { forceZ }, { 0 } }; //vector not point
                temp.rigidBody.torque = new Function[,] { { torqueX }, { torqueY }, { torqueZ }, { 0 } };

                //Display the Text of the Parent Node i.e. temp
                //MessageBox.Show(temp.Text);
                allJoints.Add(temp.rigidBody);
                qCont++;

                //Now Loop through each of the child nodes in this parent node i.e.temp
                for (int i = 0; i < temp.Nodes.Count; i++)
                {
                    temp2 = (cRigidBodyUI) temp.Nodes[i];
                    visitChildNodes(temp2, temp, t, allJoints, ref qCont, mass, pos, angles, force, torque, lCenterOfMass, qValues); //send every child to the function for further traversal   
                }
            }
            //Function.printCompilerSource = true;
            
            g = new Graph<JointNode>(temp.rigidBody, allJoints.ToArray());
            RigidBodySystem RBSTemp = new RigidBodySystem(g, t, GraphicsDevice, mass.ToArray(), pos.ToArray(), angles.ToArray(), force.ToArray(), torque.ToArray(), lCenterOfMass.ToArray());
           // Function.printCompilerSource = false;
            //RBSTemp.initialize();
            RBSTemp.initialize(qValues.ToArray());
            return RBSTemp;
            
            
        }

        private void visitChildNodes(cRigidBodyUI node, cRigidBodyUI father, Variable t, List<RigidBody> allJoints, ref long qCont, 
            List<double> mass, List<double> pos, List<double> angles, List<double> force, List<double> torque, List<double> lCenterOfMass, List<double> qValues)
        {
            Function[,] rb = null; // stores the constraints
            node.q0 = UnspecifiedFunction.functionOf("q"+qCont.ToString(),t);
            qValues.Add(node.genCoords.q0);
            node.vMass = new Variable("m" + qCont.ToString());

            Variable posX = new Variable("posX" + qCont);
            Variable posY = new Variable("posY" + qCont);
            Variable posZ = new Variable("posZ" + qCont);

            Variable angleX = new Variable("angleX" + qCont);
            Variable angleY = new Variable("angleY" + qCont);
            Variable angleZ = new Variable("angleZ" + qCont);

            Variable forceX = new Variable("forceX" + qCont);
            Variable forceY = new Variable("forceY" + qCont);
            Variable forceZ = new Variable("forceZ" + qCont);
            Variable torqueX = new Variable("torqueX" + qCont);
            Variable torqueY = new Variable("torqueY" + qCont);
            Variable torqueZ = new Variable("torqueZ" + qCont);

            Variable comX = new Variable("comX" + qCont);
            Variable comY = new Variable("comY" + qCont);
            Variable comZ = new Variable("comZ" + qCont);
            Function[,] centerOfMass = { { comX }, { comY }, { comZ }, { 1 } }; //point so need 1 in 4th position

            pos.Add(node.position.X);
            pos.Add(node.position.Y);
            pos.Add(node.position.Z);
            angles.Add(node.Orientation.X * Math.PI / 180.0);
            angles.Add(node.Orientation.Y * Math.PI / 180.0);
            angles.Add(node.Orientation.Z * Math.PI / 180.0);
            mass.Add(node.Mass);

            force.Add(node.Forces.X);
            force.Add(node.Forces.Y);
            force.Add(node.Forces.Z);

            torque.Add(node.Torque.X);
            torque.Add(node.Torque.Y);
            torque.Add(node.Torque.Z);


            lCenterOfMass.Add(node.CenterOfMass.X);
            lCenterOfMass.Add(node.CenterOfMass.Y);
            lCenterOfMass.Add(node.CenterOfMass.Z);


            Function[,] A0 = { { 1, 0, 0, posX }, { 0, 1, 0, posY }, { 0, 0, 1, posZ }, { 0, 0, 0, 1 } };
            Function[,] Aaux = { { 1, 0, 0, 0 }, { 0, 1, 0, 0 }, { 0, 0, 1, 0 }, { 0, 0, 0, 1 } };

            A0 = VM.mult(A0, VM.rotX(angleX, 0, 0, 0)); //conversion to radians
            A0 = VM.mult(A0, VM.rotY(angleY, 0, 0, 0)); //conversion to radians
            A0 = VM.mult(A0, VM.rotZ(angleZ, 0, 0, 0)); //conversion to radians
     
            if (node.Cylinder.Use_Cylinder == true)
            {
                if (node.Cylinder.Radius <= 0)
                {
                    MessageBox.Show("Radius should be higher than zero","Error");
                    node.Cylinder.Radius = 0.3;
                }

                if (node.Cylinder.Initial_Angle > 360)
                    node.Cylinder.Initial_Angle = 360;
                else if (node.Cylinder.Initial_Angle < 0)
                    node.Cylinder.Initial_Angle = 0;

                if (node.Cylinder.Final_Angle > 360)
                    node.Cylinder.Final_Angle = 360;
                else if (node.Cylinder.Final_Angle < 0)
                    node.Cylinder.Final_Angle = 0;

                Function[,] I1 = inertiaMatrixCylinder(node.vMass, node.Cylinder.Radius, node.Cylinder.Height);
                Geometries.Cylinder cylinder = new Geometries.Cylinder(node.Cylinder.Initial_Angle*Math.PI/180.0, node.Cylinder.Final_Angle*Math.PI/180.0, 0.0, node.Cylinder.Height, node.Cylinder.Radius);

                if (node.Surface == true)
                {
                    qCont++;
                    node.q1 = UnspecifiedFunction.functionOf("q" + qCont.ToString(), t);
                    qValues.Add(node.genCoords.q1);
                    qCont++;
                    if (!node.rigidBody.isParametric)
                    {
                        rb = father.rigidBody.geometry.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    else
                    {
                        //CHECK THIS FUNCTION
                        rb = father.rigidBody.parametricModel.SurfaceTangentFrame(node.q0, node.q1);
                    }

                    Aaux = VM.mult(A0, rb);
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0, node.q1 }, frictionForce(t, node.q0, node.q1), centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.CreateGeometry(GraphicsDevice, cylinder, textureMaps[11]);
                    node.rigidBody.force = new Function[,] { { forceX /** rb[0, 0] */}, { forceY }, { forceZ /** rb[2, 0] */}, { 0 } }; //vector not point
                }
                else
                {
                    if (node.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(node.q0, 0, 0, 0));
                    }
                    else if (node.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(node.q0, 0, 0, 0));
                    }
                    else if (node.RotZAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));
                    }
                       
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));   
                    }
                         
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0 }, frictionForce(t, node.q0), centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.CreateGeometry(GraphicsDevice, cylinder, textureMaps[11]);
                    node.rigidBody.force = new Function[,] { { forceX }, { forceY }, { forceZ }, { 0 } }; //vector not point                    
                    qCont++;

                }
                node.rigidBody.geometry = cylinder;
            }
            else if (node.Sphere.Use_sphere == true)
            {
                // TODO: validate the user input

                if (node.Sphere.Radius <= 0)
                {
                    MessageBox.Show("Radius should be higher than zero", "Error");
                    node.Sphere.Radius = 0.5;
                }

                Function[,] I1 = inertiaMatrixSphere(node.vMass, node.Sphere.Radius);
                Geometries.Sphere sphere = new Geometries.Sphere(node.Sphere.Radius);
                
                if (node.Surface == true)
                {
                    qCont++;
                    node.q1 = UnspecifiedFunction.functionOf("q" + qCont.ToString(), t);
                    qValues.Add(node.genCoords.q1);
                    qCont++;
                    if (!father.rigidBody.isParametric)
                    {
                        rb = father.rigidBody.geometry.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    else
                    {
                        //CHECK THIS FUNCTION
                        rb = father.rigidBody.parametricModel.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    Aaux = VM.mult(A0, rb);
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0, node.q1 }, frictionForce(t, node.q0, node.q1), centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.CreateGeometry(GraphicsDevice, sphere, textureMaps[11]);
                    node.rigidBody.force = new Function[,] { { forceX /** rb[0, 0]*/ }, { forceY }, { forceZ /** rb[2, 0] */}, { 0 } }; //vector not point
                    

                }
                else
                {
                    if (node.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(node.q0, 0, 0, 0));
                    }
                    else if (node.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(node.q0, 0, 0, 0));
                    }
                    else if (node.RotZAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));
                    }
                    
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));
                    }
                       
                    
                    //rb = VM.rotZ(node.q0, 0, 0, 0);
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0 }, frictionForce(t, node.q0), centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.CreateGeometry(GraphicsDevice, sphere, textureMaps[11]);

                    node.rigidBody.force = new Function[,] { { forceX }, { forceY }, { forceZ }, { 0 } }; //vector not point
                    qCont++;
                }
                node.rigidBody.geometry = sphere;

            }
            else if (node.Torus.Use_Torus == true)
            {
                // TODO: validate the user input

                if (node.Torus.Outer_Radius <= 0)
                {
                    MessageBox.Show("Outer Radius should be higher than zero","Error");
                    node.Torus.Outer_Radius = 0.8;
                }
                if (node.Torus.Inner_Radius <= 0)
                {
                    MessageBox.Show("Inner Radius should be higher than zero","Error");
                    node.Torus.Inner_Radius = 0.2;
                }
                //A0 = VM.mult (A0,VM.rotX(node.q0, node.Torus.Outer_Radius, 0, 0));

                Function[,] I1 = inertiaMatrixTorus(node.vMass, node.Torus.Outer_Radius, node.Torus.Inner_Radius);
                Geometries.Torus torus = new Geometries.Torus(node.Torus.Outer_Radius, node.Torus.Inner_Radius);

                if (node.Surface == true)
                {
                    qCont++;
                    node.q1 = UnspecifiedFunction.functionOf("q" + qCont.ToString(), t);
                    qValues.Add(node.genCoords.q1);
                    qCont++;
                    if (!father.rigidBody.isParametric)
                    {
                        rb = father.rigidBody.geometry.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    else
                    {
                        //CHECK THIS FUNCTION
                        rb = father.rigidBody.parametricModel.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    Aaux = VM.mult(A0, rb);
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0, node.q1 }, frictionForce(t, node.q0, node.q1), centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.CreateGeometry(GraphicsDevice, torus, textureMaps[11]);
                    node.rigidBody.force = new Function[,] { { forceX /** rb[0, 0] */}, { forceY  }, { forceZ /** rb[2, 0]*/ }, { 0 } }; //vector not point
                }
                else
                {
                    if (node.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(node.q0, 0, 0, 0));
                    }
                    else if (node.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(node.q0, 0, 0, 0));
                    }
                    else if (node.RotZAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));
                    }
                       
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));   
                    }
                         
                    node.rigidBody = new RigidBody(Aaux,   new Function[] { node.q0 }, frictionForce(t, node.q0), centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.CreateGeometry(GraphicsDevice, torus, textureMaps[11]);
                    node.rigidBody.force = new Function[,] { { forceX }, { forceY }, { forceZ }, { 0 } }; //vector not point
                    qCont++;
                }
                node.rigidBody.geometry = torus;
            }
            else if (node.ParametricModel.File != "")
            {
                ParametricModelDStar3DVisualizer parametricModel = loadParametricModel(node.ParametricModel.File, node.ParametricModel.Width, node.ParametricModel.Height);
                //CALCULATE INERTIA MATRIX HERE
                // FOR NOW, USING THE SPHERE INERTIA MATRIX: 
                Function[,] I1 = inertiaMatrixSphere(node.vMass, node.Sphere.Radius);
                
                if (node.Surface == true)
                {
                    qCont++;
                    node.q1 = UnspecifiedFunction.functionOf("q" + qCont.ToString(), t);
                    qValues.Add(node.genCoords.q1);
                    qCont++;
                    if (!father.rigidBody.isParametric)
                    {
                        rb = father.rigidBody.geometry.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    else
                    {
                        //CHECK THIS FUNCTION
                        rb = father.rigidBody.parametricModel.SurfaceTangentFrame(node.q0, node.q1);
                    }
                    Aaux = VM.mult(A0, rb);
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0 }, frictionForce(t, node.q0),
                                    centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.parametricModel = parametricModel;
                    node.rigidBody.isParametric = true;
                    node.rigidBody.force = new Function[,] { { forceX }, { forceY }, { forceZ }, { 0 } }; //vector not point
                }
                else
                {
                    // TRANSFORMATION CONSTRAINS (root Body)
                    if (node.RotXAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotX(node.q0, 0, 0, 0));
                    }
                    else if (node.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotY(node.q0, 0, 0, 0));
                    }
                    else if (node.RotYAxis == true)
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));
                    }
                    else
                    {
                        Aaux = VM.mult(A0, VM.rotZ(node.q0, 0, 0, 0));
                    }
                    node.rigidBody = new RigidBody(Aaux, new Function[] { node.q0 }, frictionForce(t, node.q0),
                                    centerOfMass, I1, node.vMass, node.Name, t);
                    node.rigidBody.parametricModel = parametricModel;
                    node.rigidBody.isParametric = true;
                    node.rigidBody.force = new Function[,] { { forceX }, { forceY }, { forceZ }, { 0 } }; //vector not point
                }
                // ***
                // HEY ! DO NOT FORGET TO CHANGE THE CODE IN cRigidBodySystemIO class, functions TraverseTreeView() and visitChildNodes()
                // ***
            }

            node.rigidBody.orientation = new Variable[] { angleX, angleY, angleZ };
            node.rigidBody.position = new Variable[] { posX, posY, posZ };

            node.rigidBody.torque = new Function[,] { { torqueX }, { torqueY }, { torqueZ}, { 0 } };
            allJoints.Add(node.rigidBody);
            father.rigidBody.c(node.rigidBody);
            
            //MessageBox.Show(node.Name);// + " Mass " + node.Mass.ToString());    
            //body.Mass = 2.0;
            //Loop Through this node and its childs recursively
            for (int j = 0; j < node.Nodes.Count; j++)
                visitChildNodes((cRigidBodyUI)node.Nodes[j], node, t, allJoints, ref qCont, mass, pos, angles, force, torque, lCenterOfMass, qValues);
        }

        /// <summary>
        /// This function will build a Generic Body System 
        /// specified by de UI
        /// </summary>
        /// <param name="rigidBodyHierarchy"></param>
        private RigidBodySystem genericRigidBodySystem(TreeView rigidBodyHierarchy)
        {
            return TraverseTreeView(rigidBodyHierarchy);
        }
        
        /// <summary>
        /// This function creates the geometries of the bodies.
        /// TreeView rigidBodyHierarchy, stores the hierarchy of the bodies
        /// </summary>
        /// <param name="rigidBodyHierarchy"></param>
        public void initializeRBSystem(TreeView rigidBodyHierarchy)
        {
            //if (rigidBodySystemUI == null)
            rigidBodySystemUI = genericRigidBodySystem(rigidBodyHierarchy);
            simulationOn = (int)simState.RESTART;
        }

        public void generateDynamics()
        {
            rigidBodySystemUI.generateDynamics();
            rigidBodySystemUI.updateInputGUIArray();
            rigidBodySystemUI.motion = true;
        }

        public void setFocusRigidBody(string bodyName)
        {
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    rb.colorWhenSelected.X = 1.0f;
                    rb.colorWhenSelected.Y = 0.0f;
                    rb.colorWhenSelected.Z = 0.0f;
                    rb.colorWhenSelected.W = 0.5f;
                }
                else
                {
                    rb.colorWhenSelected.X = 1.0f;
                    rb.colorWhenSelected.Y = 1.0f;
                    rb.colorWhenSelected.Z = 1.0f;
                    rb.colorWhenSelected.W = 1.0f;
                }
            }
        }

        public void displayAxis(string bodyName)
        {
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName && rb.displayAxis == false)
                {
                    rb.displayAxis = true; 
                }
                else if (rb.name == bodyName && rb.displayAxis == true)
                {
                    rb.displayAxis = false;
                }
                else
                {
                    rb.displayAxis = false;
                }
            }
        }

        public void updatePosition(string bodyName, Vector3 pos)
        {
            int index = 0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updatePos(index, pos);
        }

        public void updateOrientation(string bodyName, Vector3 orientation)
        {
            int index = 0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updateOrientation(index, orientation);
        }

        public void updateForces(string bodyName, Vector3 force)
        {
            int index = 0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updateForces(index, force);
        }

        public void updateTorque(string bodyName, Vector3 torque)
        {
            int index = 0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updateTorque(index, torque);
        }

        public void updateCenterOfMass(string bodyName, Vector3 com)
        {
            int index = 0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updateCenterOfMass(index, com);
        }

        public void updateMass(string bodyName, double mass)
        {
            int index=0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updateMass(index, mass);
        }

        public void updateCylinder(string bodyName, double initial_Angle, double final_Angle, double height, double radius)
        {

            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    Geometries.Cylinder cylinder = new Geometries.Cylinder(initial_Angle, final_Angle, 0.0, height, radius);
                    rb.CreateGeometry(this.GraphicsDevice, cylinder, textureMaps[10]);
                    rb.geometry = cylinder;
                    rb.isParametric = false;
                }

            }            
        }

        public void updateSphere(string bodyName, double radius)
        {
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    Geometries.Sphere sphere = new Geometries.Sphere (radius);
                    rb.CreateGeometry(this.GraphicsDevice, sphere, textureMaps[7]);
                    rb.geometry = sphere;
                    rb.isParametric = false;
                }
            }
        }

        public void updateTorus(string bodyName, double innerRadius, double outerRadius)
        {
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    Geometries.Torus torus = new Geometries.Torus(outerRadius, innerRadius);
                    rb.CreateGeometry(this.GraphicsDevice, torus, textureMaps[6]);
                    rb.geometry = torus;
                    rb.isParametric = false;
                }
            }
        }

        public void updateGenCoords(string bodyName, double q, string qName)
        {
            int index = 0;
            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    break;
                }
                index++;
            }
            rigidBodySystemUI.updateGenCoords(index, q, qName = qName.Substring(1));
        }


        private ParametricModelDStar3DVisualizer loadParametricModel(string fileName, double width, double height)
        {
            ParametricModelDStar3DVisualizer parametricModel = new ParametricModelDStar3DVisualizer(this.GraphicsDevice, 50, 50);
            XmlSerializer serializer = new XmlSerializer(typeof(ParametricModel3DData));
            ParametricModel3DData parametricModel3DData = new ParametricModel3DData();
            TextReader textReader = new StreamReader(fileName, Encoding.ASCII);
            parametricModel3DData = (ParametricModel3DData)serializer.Deserialize(textReader);
            BSplineCurve2D spline1 = new BSplineCurve2D();
            BSplineCurve2D spline2 = new BSplineCurve2D();

            if (parametricModel3DData.MaterialProperties.InternalColor != -Vector4.One)
            {
                parametricModel.MaterialColor = parametricModel3DData.MaterialProperties.InternalColor;
            }
            if (parametricModel3DData.MaterialProperties.AmbientFactor != -1.0f)
            {
                parametricModel.AmbientFactor = parametricModel3DData.MaterialProperties.AmbientFactor;
            }
            if (parametricModel3DData.MaterialProperties.DiffuseFactor != -1.0f)
            {
                parametricModel.DiffuseFactor = parametricModel3DData.MaterialProperties.DiffuseFactor;
            }
            if (parametricModel3DData.MaterialProperties.SpecularFactor != -1.0f)
            {
                parametricModel.SpecularFactor = parametricModel3DData.MaterialProperties.SpecularFactor;
            }
            int count = parametricModel3DData.ParametricCurves.Count;
            if (count >= 1)
            {
                spline1 = (BSplineCurve2D)parametricModel3DData.ParametricCurves[0];
                parametricModel._isClosedProfileCurve = spline1.IsClosedCurve;
            }
            if (count >= 2)
            {
                spline2 = (BSplineCurve2D)parametricModel3DData.ParametricCurves[1];
                parametricModel._isClosedCrossSectionCurve = spline2.IsClosedCurve;
            }
            switch (parametricModel3DData.ParametricEffectType)
            {
                case ParametricEffectType.SurfaceOfRevolution:
                    {
                        parametricModel.ApplySurfaceOfRevolutionEffect(
                                                    spline1, width, height);
                        break;
                    }
                case ParametricEffectType.ProfileProduct:
                    {
                        parametricModel.ApplyProfileProductEffect(
                                                    spline1,
                                                    spline2, width, height);
                        break;
                    }
                case ParametricEffectType.WireProduct:
                    {
                        parametricModel.ApplyWireProductEffect(
                                                        spline1,
                                                        spline2, width, height);
                        break;
                    }
                default:
                    {
                        parametricModel.ApplyNullEffect();
                        break;
                    }
            }
            return parametricModel;
        }

        public void updateShape(string bodyName, string fileName, double width, double height)
        {
            ParametricModelDStar3DVisualizer _parametricModelDStar3DVisualizer = loadParametricModel(fileName, width, height);

            foreach (RigidBody rb in rigidBodySystemUI.allNodes)
            {
                if (rb.name == bodyName)
                {
                    rb.isParametric = true;
                    rb.geometry = null;
                    rb.parametricModel = _parametricModelDStar3DVisualizer;
             
                }
            }
        }

        

        #region EXAMPLES
        private RigidBodySystem coilSystem(GraphicsDevice device, Texture2D[] texture)
        {
            Function.newContext();
            Variable t = new Variable("t");
            double _r0 = 3.0;
            double r1 = 0.5;
            double r2 = 0.09;
            int numRev = 6;
            double m = 1.0;
            double mSphere = 1.0;
            Function[,] r0 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            Function[,] R1 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            Geometries.CircularCoil cCoil = new Geometries.CircularCoil(_r0, r1, r2, numRev);
            Geometries.Sphere sphere = new Geometries.Sphere(r2 + 0.2);
            Function[,] I0 = inertiaMatrixTorus(m, r1, r2);
            Function[,] I1 = inertiaMatrixSphere(mSphere, r2 + 0.2);
            RigidBody rbParent = null, rbChild = null;
            List<RigidBody> allJoints = new List<RigidBody>();


            Function[,] A0 = { { 1, 0, 0, 0 }, { 0, 1, 0, 0 }, { 0, 0, 1, 0 }, { 0, 0, 0, 1 } };
            Function[,] flip = { { 0, 1, 0, 0 }, { 1, 0, 0, 0 }, { 0, 0, -1, 0 }, { 0, 0, 0, 1 } };
            int numBodies = 2;


            //coil
            A0 = VM.identity<Function>(4, 0, 1);
            A0 = VM.mult(A0, VM.rotZ(-Math.PI / 8, -5, 2 * _r0, 0));
            rbParent = new RigidBody(A0, new Function[] { }, new Function[] { }, r0, I0, m, "0", t);

            rbParent.CreateGeometry(device, cCoil, texture[0 % 4]);
            rbParent.force = new Function[,] { { 0 }, { -10 * m }, { 10 }, { 0 } }; //vector not point
            rbParent.torque = new Function[,] { { 0 }, { 00 }, { 0 }, { 0 } };
            allJoints.Add(rbParent);

            for (int i = 1; i <= numBodies; ++i)
            {
                UnspecifiedFunction qi_0 = UnspecifiedFunction.functionOf("q" + i + "_0", t);
                UnspecifiedFunction qi_1 = UnspecifiedFunction.functionOf("q" + i + "_1", t);
                Function[,] RBi0 = cCoil.SurfaceTangentFrame(qi_0, qi_1 + i * 5);
                /*
                UnspecifiedFunction qi_2 = UnspecifiedFunction.functionOf("q" + i + "_0", t);
                UnspecifiedFunction qi_3 = UnspecifiedFunction.functionOf("q" + i + "_1", t);
                Function[,] RBi1 = sphere.SurfaceTangentFrame(qi_0 , qi_1);
                */
                //A0 = VM.mult(
                A0 = VM.mult(RBi0, VM.rotX(qi_0 * qi_1, 0.0, 0.0, 0.0));
                //A0 = VM.mult(RBi0, VM.rigidBodyInverse(RBi1));
                rbChild = new RigidBody(A0, new Function[] { qi_0, qi_1 }, frictionForce(t, qi_0 / 2.0, qi_1 / 2.0), R1, I1, mSphere, i.ToString(), t);
                rbChild.CreateGeometry(device, sphere, texture[i % 4]);
                rbChild.force = new Function[,] { { 0 }, { -10 * mSphere }, { 0 }, { 0 } }; //vector not point
                rbChild.torque = new Function[,] { { 0 }, { 0 }, { 0 }, { 0 } };

                // ji.generalizedForce = new Function[] { -.1 *( Function.D(qi_0, t) + Function.D(qi_1, t) +Function.D(qi_2, t) +Function.D(qi_3, t)) };
                allJoints.Add(rbChild);
                rbParent.c(rbChild);
            }

            //allJoints.Reverse();

            Graph<JointNode> g = new Graph<JointNode>(rbParent, allJoints.ToArray());
            RigidBodySystem temp = new RigidBodySystem(g, t, device);
            temp.initialize();
            return temp;
        }

        private RigidBodySystem gyratory(GraphicsDevice device, Texture2D[] texture)
        {
            Function.newContext();
            Variable t = new Variable("t");
            double r1 = 1.5;
            double r2 = 0.1;
            double m = 1.0;
            double mSphere = 1.0;
            Function[,] r0 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            Function[,] R1 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            Geometries.Torus torus = new Geometries.Torus(r1, r2);
            Geometries.Sphere sphere = new Geometries.Sphere(r2 + 0.2);
            Function[,] I0 = inertiaMatrixTorus(m, r1, r2);
            Function[,] I1 = inertiaMatrixSphere(mSphere, r2 + 0.2);
            RigidBody rbParent = null, rbChild = null;
            List<RigidBody> allJoints = new List<RigidBody>();
            Function[,] A0 = { { 1, 0, 0, 0 }, { 0, 1, 0, 0 }, { 0, 0, 1, 0 }, { 0, 0, 0, 1 } };

            //sphere
            UnspecifiedFunction q0 = UnspecifiedFunction.functionOf("q0", t);
            A0 = VM.mult(A0, VM.rotX(Math.PI/4 * q0, 0, 2 * r1, -3));
            rbParent = new RigidBody(A0, new Function[] { q0 }, frictionForce(t, q0), r0, I0, m, "Body0", t);
            rbParent.CreateGeometry(device, sphere, texture[0 % 4]);
            rbParent.force = new Function[,] { { 0 }, { -10 * m }, { 0 }, { 0 } }; //vector not point
            rbParent.torque = new Function[,] { { 0 }, { 00 }, { 0 }, { 0 } };
            allJoints.Add(rbParent);
                        
            q0 = UnspecifiedFunction.functionOf("q1", t);
            A0 = VM.rotX(q0, 2 * (r2 + 0.2), 0, 0);
            rbChild = new RigidBody(A0, new Function[] { q0 }, frictionForce(t, q0), r0, I0, m, "1", t);
            rbChild.CreateGeometry(device, sphere, texture[1 % 4]);
            rbChild.force = new Function[,] { { 0 }, { -10 * m }, { 0 }, { 0 } }; //vector not point
            rbChild.torque = new Function[,] { { -5 }, { 00 }, { 0 }, { 0 } };
            allJoints.Add(rbChild);
            rbParent.c(rbChild);

            q0 = UnspecifiedFunction.functionOf("q2", t);
            A0 = VM.rotX(q0, -2 * (r2 + 0.2), 0, 0);
            rbChild = new RigidBody(A0, new Function[] { q0 }, frictionForce(t, q0), r0, I0, m, "1", t);
            rbChild.CreateGeometry(device, sphere, texture[1 % 4]);
            rbChild.force = new Function[,] { { 0 }, { -10 * m }, { 0 }, { 0 } }; //vector not point
            rbChild.torque = new Function[,] { { -5 }, { 00 }, { 0 }, { 0 } };
            allJoints.Add(rbChild);
            rbParent.c(rbChild);
            /*
            UnspecifiedFunction q1 = UnspecifiedFunction.functionOf("q2", t);
            A0 = VM.rotX(q1, (r2 + 0.2) + r1, 0, 0);
            rbChild1 = new RigidBody(A0, new Function[] { q1 }, frictionForce(t, q1), r0, I0, m, "2", t);
            rbChild1.CreateGeometry(device, torus, texture[2 % 4]);
            rbChild1.force = new Function[,] { { 0 }, { -10 * m }, { 0 }, { 0 } }; //vector not point
            rbChild1.torque = new Function[,] { { -5 }, { 00 }, { 0 }, { 0 } };
            allJoints.Add(rbChild1);
            rbChild.c(rbChild1);

            UnspecifiedFunction q2 = UnspecifiedFunction.functionOf("q3", t);
            UnspecifiedFunction q3 = UnspecifiedFunction.functionOf("q4", t);

            Function[,] RB0 = torus.SurfaceTangentFrame(q2 + 1, q3);
            rbChild2 = new RigidBody(RB0, new Function[] { q2, q3 }, frictionForce(t, q2 * 0.1, q3 * 0.1), R1, I1, mSphere, "3", t);
            rbChild2.CreateGeometry(device, sphere, texture[3 % 4]);
            rbChild2.force = new Function[,] { { RB0[0, 0] }, { -10 * mSphere }, { RB0[2, 0] }, { 0.0 } }; //vector not point
            rbChild2.torque = new Function[,] { { 0 }, { 0 }, { 0 }, { 0 } };
            allJoints.Add(rbChild2);
            rbChild1.c(rbChild2);

            q2 = UnspecifiedFunction.functionOf("q5", t);
            q3 = UnspecifiedFunction.functionOf("q6", t);
            RB0 = torus.SurfaceTangentFrame(q2 + 3, q3);
            rbChild2 = new RigidBody(RB0, new Function[] { q2, q3 }, frictionForce(t, q2 * 0.1, q3 * 0.1), R1, I1, mSphere, "3", t);
            rbChild2.CreateGeometry(device, sphere, texture[4 % 4]);
            rbChild2.force = new Function[,] { { RB0[0, 0] }, { -10 * mSphere }, { RB0[2, 0] }, { 0.0 } }; //vector not point
            rbChild2.torque = new Function[,] { { 0 }, { 0 }, { 0 }, { 0 } };
            allJoints.Add(rbChild2);
            rbChild1.c(rbChild2);
            
            */
            Graph<JointNode> g = new Graph<JointNode>(rbParent, allJoints.ToArray());
            //Function.printCompilerSource = true;
            RigidBodySystem temp = new RigidBodySystem(g, t, device);
            temp.initialize();
            return temp;

        }

        private RigidBodySystem ringSystem(GraphicsDevice device, Texture2D[] texture, int numSpheres)
        {
            Function.newContext();
            Variable t = new Variable("t");
            double r1 = 3.0;
            double r2 = 0.1;
            double m = 1.0;
            double mSphere = 1.0;
            Function[,] r0 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            Function[,] R1 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            Geometries.Torus torus = new Geometries.Torus(r1, r2);
            Geometries.Sphere sphere = new Geometries.Sphere(r2 + 0.2);
            Function[,] I0 = inertiaMatrixTorus(m, r1, r2);
            Function[,] I1 = inertiaMatrixSphere(mSphere, r2 + 0.2);
            RigidBody rbParent = null, rbChild = null;
            List<RigidBody> allJoints = new List<RigidBody>();

            Function[,] A0 = { { 1, 0, 0, 0 }, { 0, 1, 0, 0 }, { 0, 0, 1, 0 }, { 0, 0, 0, 1 } };
            Function[,] flip = { { 0, 1, 0, 0 }, { 1, 0, 0, 0 }, { 0, 0, -1, 0 }, { 0, 0, 0, 1 } };

            //ring
            A0 = VM.identity<Function>(4, 0, 1);
            A0 = VM.mult(A0, VM.rotX(-Math.PI / 1.7, 0, 2 * r1, 0));
            rbParent = new RigidBody(A0, new Function[] { }, new Function[] { }, r0, I0, m, "0", t);

            rbParent.CreateGeometry(device, torus, texture[0 % 4]);
            rbParent.force = new Function[,] { { 0 }, { -10 * m }, { 10 }, { 0 } }; //vector not point
            rbParent.torque = new Function[,] { { 0 }, { 00 }, { 0 }, { 0 } };
            allJoints.Add(rbParent);

            for (int i = 1; i <= numSpheres; ++i)
            {
                UnspecifiedFunction qi_0 = UnspecifiedFunction.functionOf("q" + i + "_0", t);
                UnspecifiedFunction qi_1 = UnspecifiedFunction.functionOf("q" + i + "_1", t);
                Function[,] RBi0 = torus.SurfaceTangentFrame(qi_0 + i / 2.5, qi_1);
                /*
                UnspecifiedFunction qi_2 = UnspecifiedFunction.functionOf("q" + i + "_0", t);
                UnspecifiedFunction qi_3 = UnspecifiedFunction.functionOf("q" + i + "_1", t);
                Function[,] RBi1 = sphere.SurfaceTangentFrame(qi_0 , qi_1);
                */
                //A0 = VM.mult(
                A0 = VM.mult(RBi0, VM.rotX(qi_0, 0.0, 0.0, 0.0));
                //A0 = VM.mult(RBi0, VM.rigidBodyInverse(RBi1));
                rbChild = new RigidBody(A0, new Function[] { qi_0, qi_1 }, frictionForce(t, qi_0 / 2.0, qi_1 / 2.0), R1, I1, mSphere, i.ToString(), t);
                rbChild.CreateGeometry(device, sphere, texture[i % 4]);
                rbChild.force = new Function[,] { { -RBi0[0, 0] }, { -10 * mSphere }, { -RBi0[2, 0] }, { 0.0 } }; //vector not point
                rbChild.torque = new Function[,] { { 0 }, { 0 }, { 0 }, { 0 } };


                allJoints.Add(rbChild);
                rbParent.c(rbChild);
            }
            //allJoints.Reverse();

            Graph<JointNode> g = new Graph<JointNode>(rbParent, allJoints.ToArray());
            RigidBodySystem temp = new RigidBodySystem(g, t, device);

            temp.initialize();
            return temp;


        }

        private RigidBodySystem torusLinkSystem(GraphicsDevice device, Texture2D[] texture, int numJoints)
        {
            Function.newContext();
            // Function.printCompilerSource = true;
            Variable t = new Variable("t");
            double m = 1.0;
            Function[,] r0 = { { 0 }, { 0 }, { 0 }, { 1 } }; //point so need 1 in 4th position
            double largeRad = 1, smallRad = .2;
            Function[,] I0 = inertiaMatrixTorus(m, largeRad, smallRad);
            Geometries.Torus tor = new Geometries.Torus(largeRad, smallRad);
            Function[,] flip = { { 0, 1, 0, 0 }, { 1, 0, 0, 0 }, { 0, 0, -1, 0 }, { 0, 0, 0, 1 } };
            RigidBody ji = null, jiPlus1 = null;
            List<RigidBody> allJoints = new List<RigidBody>();

            for (int i = numJoints - 1; i >= 0; i--)
            {
                Function[,] A0;

                UnspecifiedFunction qi_0 = UnspecifiedFunction.functionOf("q" + i + "_0", t), qi_1 = UnspecifiedFunction.functionOf("q" + i + "_1", t);
                Function[,] RBi0 = tor.SurfaceTangentFrame(qi_0, qi_1);
                UnspecifiedFunction qi_2 = UnspecifiedFunction.functionOf("q" + i + "_2", t), qi_3 = UnspecifiedFunction.functionOf("q" + i + "_3", t);
                Function[,] RBi1 = tor.SurfaceTangentFrame(qi_2, qi_3);
                //VM.print(VM.mult(RBi0,VM.rigidBodyInverse(RBi0)));
                if (i == 0)
                {
                    A0 = VM.mult(VM.translate(0, 2 * numJoints, 0), VM.rigidBodyInverse(RBi0));
                    ji = new RigidBody(A0, new Function[] { qi_0, qi_1 }, frictionForce(t, qi_0, qi_1), r0, I0, 1, i.ToString(), t);
                }
                else
                {
                    A0 = VM.mult(RBi0, VM.mult(flip, VM.rigidBodyInverse(RBi1)));
                    //A0 = VM.mult(RBi0, VM.rigidBodyInverse(RBi1));
                    ji = new RigidBody(A0, new Function[] { qi_0, qi_1, qi_2, qi_3 }, frictionForce(t, qi_0, qi_1, qi_2, qi_3), r0, I0, 1, i.ToString(), t);
                }

                ji.CreateGeometry(device, tor, texture[i % 4]);
                ji.force = new Function[,] { { 0 }, { -10 * m }, { 0 }, { 0 } }; //vector not point
                ji.torque = new Function[,] { { 0 }, { 0 }, { 0 }, { 0 } };
                // ji.generalizedForce = new Function[] { -.1 *( Function.D(qi_0, t) + Function.D(qi_1, t) +Function.D(qi_2, t) +Function.D(qi_3, t)) };
                allJoints.Add(ji);
                if (jiPlus1 != null) { ji.c(jiPlus1); }
                jiPlus1 = ji;
            }
            allJoints.Reverse();

            Graph<JointNode> g = new Graph<JointNode>(ji, allJoints.ToArray());
            RigidBodySystem temp = new RigidBodySystem(g, t, device);

            
            int dof = numJoints * 4 - 2;
            double[] qinit = new double[dof], qdinit = new double[dof];
            for (int i = 2; i < qinit.Length; i += 4)
            {
                qinit[i] = 1.4 * Math.PI;
                qinit[i + 1] = Math.PI; //small tube variable
                qinit[i + 2] = 0;
                qinit[i + 3] = Math.PI;  //small tube variable
            }

            //for (int i = 2; i < qinit.Length; i += 2) { qinit[i] = Math.PI / 2.0; }
            temp.initialize(qinit, qdinit);
            
            //temp.initialize();

            return temp;
        }
        #endregion

        /// <summary>
        /// FRICTION FORCE NEEDS TO BE DEFINED FOR EACH GENERALIZED COORDINATE
        /// </summary>
        /// <param name="t"></param>
        /// <param name="genCoords"></param>
        /// <returns></returns>
        public Function[] frictionForce(Variable t, params Function[] genCoords)
        {
            Function[] result = new Function[genCoords.Length];

            //for (int i = 0; i < genCoords.Length; i++) { result[i] = (- Math.PI/2.0 - genCoords[i] ) - .2 * Function.D(genCoords[i], t); } //this works only for one cylinder
            for (int i = 0; i < genCoords.Length; i++) { result[i] = -2 * Function.D(genCoords[i], t); }
            return result;
        }

        public Function[,] inertiaMatrixCylinder(double mass, double radius, double height)
        {
            // moment of inertia:
            return new Function[,]{{1/12.0*mass*(3*(radius*radius) + (height*height)), 0 ,0 ,0},
                         {0, .5*mass*(radius*radius), 0 ,0},
                         {0, 0, 1/12.0*mass*(3*(radius*radius) + (height*height)), 0},
                         {0,0,0,0}};
        }

        public Function[,] inertiaMatrixCylinder(Variable mass, double radius, double height)
        {
            // moment of inertia:
            return new Function[,]{{1/12.0*mass*(3*(radius*radius) + (height*height)), 0 ,0 ,0},
                         {0, .5*mass*(radius*radius), 0 ,0},
                         {0, 0, 1/12.0*mass*(3*(radius*radius) + (height*height)), 0},
                         {0,0,0,0}};
        }


        public Function[,] inertiaMatrixTorus(double mass, double largeRadius, double smallRadius)
        {
            return new Function[,]{{1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0 ,0 ,0},
                         {0, .25*mass*(4*largeRadius*largeRadius + 3*smallRadius*smallRadius), 0 ,0},
                         {0, 0, 1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0},
                         {0,0,0,0}};
        }

        public Function[,] inertiaMatrixTorus(Variable mass, double largeRadius, double smallRadius)
        {
            return new Function[,]{{1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0 ,0 ,0},
                         {0, .25*mass*(4*largeRadius*largeRadius + 3*smallRadius*smallRadius), 0 ,0},
                         {0, 0, 1/8.0*mass*(4*largeRadius*largeRadius + 5*smallRadius*smallRadius), 0},
                         {0,0,0,0}};
        }


        public Function[,] inertiaMatrixSphere(double mass, double radius)
        {
            return new Function[,]{ {0.4*mass*radius*radius, 0 ,0 ,0},
                                    {0, 0.4*mass*radius*radius, 0 ,0},
                                    {0, 0, 0.4*mass*radius*radius, 0},
                                    {0,0,0,0}};
        }
        public Function[,] inertiaMatrixSphere(Variable mass, double radius)
        {
            return new Function[,]{ {0.4*mass*radius*radius, 0 ,0 ,0},
                                    {0, 0.4*mass*radius*radius, 0 ,0},
                                    {0, 0, 0.4*mass*radius*radius, 0},
                                    {0,0,0,0}};
        }

        #region mouse keyboard imput
        /// <summary>
        /// Control camera, light, userSpring
        /// USAGE:
        /// 1. UserSpring: click on a link of a system while pressing "S" and drag the link with a mouse. Releasing "S" button will remove the userSpring.
        /// 2. Camera control: Left mouse button for viewing angle. Right mouse button for zoom. LeftShift + Left mouse button for translation of camera target.
        /// 3. Light control: LeftAlt + Left mouse button for light angle. LeftAlt + Right mouse button for light distance.
        /// 4. Start simulation by pressing SpaceBar
        /// 5. Start recording by pressing "R"
        /// 6. Exit by pressing "ESC"
        /// </summary>
        private void HandleInput()
        {
            KeyboardState newState = Keyboard.GetState();
            Mouse.WindowHandle = this.Handle;
            MouseState current_mouse = Mouse.GetState();
            /*
            //play/stop
            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Space))
            {
                simulationOn = 1;
                //IsFixedTimeStep = true;
            }
            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.P))
            {
                simulationOn = 0;
            }
             * */
/*
            //recording
            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.R))
            {
                simulationOn = 1;
                //IsFixedTimeStep = false; //save scene at every time step
                displayTimeStep = 1.0 / 30.0; //30frames per second
                saveScreenShots = true;
            }
*/
            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.H))
            {
                Console.WriteLine(
                @"USAGE:
 1. UserSpring: click on a link of a system while pressing S and drag the link with a mouse. 
                Releasing S button will remove the userSpring.
 2. Camera control: Left mouse button for viewing angle. 
                    Right mouse button for zoom. 
                    LeftShift + Left mouse button for translation of camera target.
 3. Light control: LeftAlt + Left mouse button for light angle. 
                   LeftAlt + Right mouse button for light distance.
 4. Start simulation by pressing SpaceBar
 5. Start recording by pressing R
 6. Exit by pressing ESC
         ");
            }

            //set control mode
            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.S))
            {
                userInputMode = UserInputMode.Spring;
            }
            else
            {
                userInputMode = UserInputMode.None;

                //disable userSpring
                if (userSpring.active) userSpring.active = false;
                //IsMouseVisible = true;
            }

            switch (userInputMode)
            {
                case UserInputMode.Spring:
                  /*
                  if (current_mouse.LeftButton == Microsoft.Xna.Framework.Input.ButtonState.Pressed)
                    {
                        if (!userSpring.active)
                        {
                            Vector3 intersection;
                            Vector2 textureCoord;
                            Link link;
                            RigidBodySystem system = RayIntersectionRigidBodies(current_mouse.X, current_mouse.Y, out link, out intersection, out textureCoord);
                            if (system != null)
                            {
                                //create spring
                                IsMouseVisible = false;
                                userSpring.active = true;
                                userSpring.position = intersection;
                                userSpring.attachedSystem = system;
                                userSpring.attachedLink = link;
                                userSpring.attachmentPoint = Vector3.Transform(intersection, Matrix.Invert(link.matWorld));
                            }
                        }
                        else
                        {
                            //update userSpring's2 position
                            Vector3 v = cameraTarget - cameraPos;
                            v.Normalize();
                            v = Vector3.Cross(v, Vector3.Up);

                            userSpring.position.X += 0.01f * v.X * (current_mouse.X - mouseX);
                            userSpring.position.Z += 0.01f * v.Z * (current_mouse.X - mouseX);
                            userSpring.position.Y -= 0.01f * (current_mouse.Y - mouseY);
                        }
                    }*/
                    break;

                default: // Camera and Light control
                    if (current_mouse.X > 0 && current_mouse.X < this.Width && current_mouse.Y > 0 && current_mouse.Y < this.Height)
                    {
                        if (current_mouse.LeftButton == Microsoft.Xna.Framework.Input.ButtonState.Pressed)
                        {
                            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.LeftAlt))
                            {
                                //control light position
                                lightTheta += 0.01f * (current_mouse.X - mouseX);
                                lightPhi -= 0.01f * (current_mouse.Y - mouseY);
                            }
                            else if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.LeftShift))
                            {
                                //control camera/light target
                                //update userSpring's2 position
                                Vector3 v = cameraTarget - cameraPos;
                                v.Normalize();
                                v = Vector3.Cross(v, Vector3.Up);

                                cameraTarget.X -= 0.01f * v.X * (current_mouse.X - mouseX);
                                cameraTarget.Z -= 0.01f * v.Z * (current_mouse.X - mouseX);
                                cameraTarget.Y += 0.01f * (current_mouse.Y - mouseY);
                            }
                            else
                            {
                                //control viewing direction
                                cameraTheta -= 0.01f * (current_mouse.X - mouseX);
                                cameraPhi += 0.01f * (current_mouse.Y - mouseY);
                                IContainerControl ctrl = this.GetContainerControl();
                            }
                        }

                        if (current_mouse.RightButton == Microsoft.Xna.Framework.Input.ButtonState.Pressed)
                        {
                            if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.LeftAlt))
                            {
                                //control light distance
                                lightDist += 0.1f * (current_mouse.X - mouseX);
                            }
                            else
                            {
                                //control zoom
                                cameraDist -= 0.1f * (current_mouse.X - mouseX);
                            }
                        }

                        //update viewMatrix
                        cameraPos.X = cameraTarget.X + cameraDist * (float)Math.Sin((float)cameraTheta) * (float)Math.Cos((float)cameraPhi);
                        cameraPos.Y = cameraTarget.Y + cameraDist * (float)Math.Sin((float)cameraPhi);
                        cameraPos.Z = cameraTarget.Z + cameraDist * (float)Math.Cos((float)cameraTheta) * (float)Math.Cos((float)cameraPhi);
                        pm.UpdateViewMatrix(cameraPos, cameraTarget);
                    }

                    break;
            }

            //exit on Escape
           // if (newState.IsKeyDown(Microsoft.Xna.Framework.Input.Keys.Escape)) Exit();

            oldKeyState = newState;
            mouseX = current_mouse.X;
            mouseY = current_mouse.Y;
        }
        #endregion

    }
}
